:- module(equational,[empty_rules/1,
		  delete_rule/3,
		  insert_rule/3,
		  match_rule/5,
		  term_type/3,
		  type_check_rule/3,
		  type_check_type/3,
		  unify/2,
		  unify_rule/3]).


:- use_module(library(change_arg),[change_arg/4]).
:- use_module(library(basics),[member/2]).

:- use_module(formulas,[left/2,left/3,
			mk_equation/5,
			number/2,
			orientation/2,
			orientation/3,
			right/2,right/3,
			type/2]).

:- use_module(satsrc(substitutiontree)).
:- use_module(global_vars, [cont/2,incr/1,store/2]).

:- use_module(st_terms).






unify_rule(L,STree,Ru):-
	st_us_unify(L,STree,Ru),
	prolog:'$acyclic'(L).

unify(S,S):-
	prolog:'$acyclic'(S).



/* ------------------------------------------------------------------------------

Matching rewrite rule


------------------------------------------------------------------------------ */

match_rule(Sig,STree,Term,R,Ru):-
	\+ \+ (match(Sig,STree,Term,R,Ru) ->
		  store('$found',(R,Ru))
	      ),
	(cont(irreducibility_check,true) ->
	    raise_exception(reducible)
	;
	    true
	),
	cont('$found',(R,Ru)),
				%			 proof(Ru,P),
	!.

match(Sig,STree,Term,R,Ru):-
	% no equations
	st_us_unify(Term,STree,eq(R,Ru)),
	type_check_rule(Sig,Term,Ru),
	orientation(Ru,Orient),
	(Orient=lr ->
	    true
	;%Orient=check
	    cont(ord,ord(OType,OParams)),
	    OType:greater(OParams,Term,R)
	).

/* ------------------------------------------------------------------------------

Insertion and Deletion

------------------------------------------------------------------------------ */

empty_rules(STree):-
	st_empty_tree(STree).

insert_rule(Ru,STree,STree1):-
	orientation(Ru,O),
	insert_rule(Ru,O,STree,STree1).

insert_rule(Ru,lr,STree,STree1):-
	left(Ru,L),
	right(Ru,R),
	term2prolog((L,R),(LP,RP)),
	st_insert_term(LP,eq(RP,Ru),STree,STree1).
insert_rule(Ru,rl,STree,STree1):-
	right(Ru,L),
	left(Ru,R),
	left(Ru,Ru1,L),
	right(Ru1,RuInv,R),
	term2prolog((L,R),(LP,RP)),
	st_insert_term(LP,eq(RP,RuInv),STree,STree1).
insert_rule(Ru,check,STree,STree1):-
	right(Ru,L),
	left(Ru,R),
	left(Ru,Ru1,L),
	right(Ru1,RuInv,R),
	term2prolog((L,R),(LP,RP)),
	st_insert_term(RP,eq(LP,Ru),STree,STree0),
	st_insert_term(LP,eq(RP,RuInv),STree0,STree1).









delete_rule(Ru,STree,STree2):-
	left(Ru,L),
	term2prolog(L,LP),	% can only delete variants of keys!
	number(Ru,N),           % hence need to disambiguate by number
	st_delete_term(LP,STree,STree1,equational:same_number(N)),
	orientation(Ru,O),
	(O=lr ->
	    STree2=STree1
	;			% need to delete the second orientation 
				% could already have been deleted during processing
				% of other orientation (only if first side
				% checked was reducible!)
	    right(Ru,R),
	    term2prolog(R,RP),
	    st_delete_term(RP,STree1,STree2,equational:same_number(N))
	).


same_number(N,eq(_,Ru)):-
	number(Ru,N).



/* ------------------------------------------------------------------------------

Type Checking

------------------------------------------------------------------------------ */
	

type_check_rule(Sig,Term,Ru):-
	left(Ru,L),
	(object_var_(L) ->
	    type(Ru,Ty),
	    term_type(Sig,Term,Ty_object_term),
	    type_match(Ty,Ty_object_term)
	;
	    % typing ok by matching of function symbols
	    true).

type_check_type(Sig,Term,Ty):-
	term_type(Sig,Term,Ty_object_term),
	type_match(Ty,Ty_object_term).




/* ------------------------------------------------------------------------------

  many-sorted type checking

------------------------------------------------------------------------------ */

term_type(_,'$neq'(_,_),prop):-
	!.
term_type(_,'$true',prop):-
	!.
term_type(_,'$false',prop):-
	!.
term_type(Sig,T,Ty):-
	nonvar_(T),
	functor_(T,F,N),
	member(F:(AT->Ty),Sig),
	length(AT,N).

	

type_match(T,T).


/* ------------------------------------------------------------------------------

Redundancy Detection


------------------------------------------------------------------------------ */


trivial(E):-
	left(E,L),
	right(E,R),
	L==R.



subsumed_by(E,Active):-
	incr(subsumption_ctr),
	left(E,L),
	right(E,R),
	type(E,Ty),
	subsumed_by(L,R,Ty,Active).

subsumed_by(L,R,Ty,Active):-
	path_arg_(P,L,Lsub),
	Lsub\=='$',
	path_arg_(P,R,Rsub),
	\+ \+ (   equational:unify_rule(Lsub,Active,eq(Rsub,Ru)),
		  type(Ru,Ty)
	      ),
	change_path_arg_(P,L,L1,'$'),
	change_path_arg_(P,R,R1,'$'),
	(
	  L1==R1
	;
	  subsumed_by(L1,R1,Ty,Active)
	),
	incr(success_sub_ctr),
	!.

