/* Property 3 of the Sheffer's 3-basis (for Boolean algebra).
   f is the Sheffer stroke. The property is derived from Veroff's 2-basis of f.
   (``Basis'' means that the set of axioms axiomatizes Boolean algebra.)

   The automatic introduction of new functions from splitting
   equations seems essential in solving the problem.
   Saturate ``invents'' f$2 (which is verum) and f$1 (negation).

   For Satellite we need to introduce these functions (n/1 and t/0)
   beforehand.

   This example was communicated to Thomas Hillenbrand by Dana Scott,
   also see the JAR 29,1 (2002) paper by McCune et al.

*/


:- use_module(dp(formulas),[mk_equation/5]).

theory([A1,A2,A3,A4,G]):-
	X='$VAR'(0),
	Y='$VAR'(1),
	Z='$VAR'(2),

				% Veroff's 2-basis:
	mk_equation(f(X,Y),      f(Y,X),          o, axiom(1), A1),
	mk_equation(f(f(X,Y),    f(X,f(Y,Z))), X, o, axiom(2), A2),

				% auxiliary functions (negation n, verum t)
	mk_equation(n(X),        f(X,X),          o, axiom(3), A3),
	mk_equation(f(X,f(X,X)), t,               o, axiom(4), A4),

				% (negated) property 3 of the Sheffer 3-basis
	mk_equation('$neq'(f(f(a,f(b,c)),f(a,f(b,c))),
			   f(f(f(b,b),a),f(f(c,c),a))),
		    '$true',
		    prop,
		    goal(1),
		    G).

prove:-
	theory([A1,A2,A3,A4,G]),
	hgstuff:time(saturation:sat(equational(sig([
						  t:([]->o),
						  n:([o]->o),
						  f:([o,o]->o),
						  a:([]->o),
						  b:([]->o),
						  c:([]->o)
						 ],
					     cons([])),
					     _),

				    ord(lpo,prec([],[f,n,t,a,b,c])),
				    [],
				    [A1,A2,A3,A4,G],
				    _A,_P
				   ),
		    T),
	format('~2nTime: ~p~2n',[T]).


/*
20021106 600 Mhz PIII by Saturate (without axiom(3) and axiom(4)).

Proof:
======

   2 : f(f(A,B),f(A,f(B,C)))=A                                          [input]
   3 + !A,B,C.(f(f(C,f(B,A)),f(C,f(B,A)))=f(f(f(B,B),C),f(f(A,A),C))) ->  [input]
   4 + !A,B.(f(f(B,f(A,s$3)),f(B,f(A,s$3)))=f(f(f(A,A),B),f(f(s$3,s$3),B))) ->  [expansion from 3]
   5 + !A.(f(f(A,f(s$2,s$3)),f(A,f(s$2,s$3)))=f(f(f(s$2,s$2),A),f(f(s$3,s$3),A))) ->  [expansion from 4]
   6 + f(f(s$1,f(s$2,s$3)),f(s$1,f(s$2,s$3)))=f(f(f(s$2,s$2),s$1),f(f(s$3,s$3),s$1)) ->  [expansion from 5]
   7 : f(f(f(A,B),A),A)=f(A,B)                           [chaining of 2 from 2]
  15 : f(f(f(A,B),f(B,C)),A)=f(A,B)                      [chaining of 2 from 2]
  27 : f(f(f(A,f(B,C)),B),A)=f(A,f(B,C))                 [chaining of 2 from 2]
  73 : f(f(A,A),f(A,B))=A                                [chaining of 7 from 2]
 105 : f(f(f(A,A),B),A)=f(A,A)                          [chaining of 73 from 2]
 800 : f(A,A)=f(A,f(f(A,B),B))                          [chaining of 2 from 27]
 990 : f(f(f(A,B),f(A,C)),f(f(A,B),f(A,C)))=f(f(f(A,B),f(A,C)),A) [chaining of 27 from 15]
1174 : f(f(A,B),f(A,f(A,A)))=f(f(A,B),f(A,B))         [chaining of 73 from 800]
1214 : f(A,f(f(A,B),f(f(A,B),f(B,C))))=f(A,A)         [chaining of 15 from 800]
1223 : f(f(A,A),B)=f(B,f(A,B))                        [chaining of 800 from 27]
1264 + f(f(f(s$2,s$2),s$1),f(f(s$3,s$3),s$1))=f(f(f(s$2,s$3),s$1),f(f(f(s$3,f(s$3,s$3)),f(s$3,f(s$3,s$3))),f(s$2,s$3))) ->  [reduction of 6 by [1174,1174,1223]]
1300 : f(f(A,f(B,C)),A)=f(f(f(A,B),f(A,B)),f(A,f(B,C))) [chaining of 2 from 1223]
1310 : f(f(f(A,A),B),B)=f(B,A)                        [chaining of 1223 from 7]
1358 : f(f(A,A),A)=f(f(f(A,B),f(A,B)),f(A,A))        [chaining of 73 from 1223]
1426 : f(f(f(A,B),f(B,A)),f(A,B))=f(f(A,A),f(f(A,B),f(B,A))) [chaining of 15 from 1223]
1485 : f(f(f(A,B),f(B,A)),f(A,B))=f(f(A,A),A)     [reduction of 1426 by [1358]]
1980 : f(f(f(A,A),B),f(f(f(A,A),B),f(B,C)))=f(f(f(f(A,A),B),f(B,C)),A) [chaining of 15 from 1310]
2201 : f(f(f(A,f(A,A)),B),f(f(A,B),f(A,B)))=f(A,f(A,A)) [chaining of 1174 from 2]
2258 : f(A,f(f(A,f(B,B)),f(f(A,f(B,B)),f(A,f(B,B)))))=f(f(f(A,f(B,B)),f(A,B)),f(f(A,f(B,B)),f(A,B))) [chaining of 2 from 1174]
2548 : f(f(A,B),f(f(A,B),f(A,B)))=f(f(f(A,f(A,A)),f(A,f(A,A))),f(A,B)) [chaining of 1174 from 1223]
2592 : f(f(A,f(B,f(B,B))),A)=f(B,f(B,B))          [reduction of 2201 by [1300]]
2593 : f(A,f(B,f(B,B)))=f(A,A)       [reduction of 2258 by [1980,1223,105,2,2]]
2597 : f(f(A,A),A)=f(f(f(A,f(A,A)),f(A,f(A,A))),f(A,B)) [reduction of 2548 by [1485]]
2598 + f(f(f(s$2,s$2),s$1),f(f(s$3,s$3),s$1))=f(f(f(s$2,s$3),s$1),f(f(s$3,s$3),s$3)) ->  [reduction of 1264 by [2597]]
2795 : f(f(A,f(B,f(B,B))),f(A,C))=A                  [chaining of 2593 from 73]
3003 : f(f(A,f(B,f(B,B))),C)=f(C,f(A,C))           [chaining of 2593 from 1223]
3007 : f(f(A,f(A,A)),f(B,f(A,f(A,A))))=f(f(B,B),f(B,B)) [chaining of 1223 from 2593]
3171 : f(A,f(B,f(B,B)))=f(A,f(C,f(C,C)))           [chaining of 2593 from 2593]
3179 : f(f(A,f(A,A)),f(B,f(A,f(A,A))))=B            [reduction of 3007 by [73]]
3182 : f(A,f(B,f(B,B)))=f$1(A)                 [variable elimination from 3171]
3183 : f(A,f(B,f(B,B)))=f$1(A)                 [variable elimination from 3171]
3187 : f(f$1(A),A)=f(B,f(B,B))                    [reduction of 2592 by [3182]]
3189 : f(A,f(A,A))=f$2                         [variable elimination from 3187]
3201 + f(f(f(s$2,s$2),s$1),f(f(s$3,s$3),s$1))=f(f(f(s$2,s$3),s$1),f$2) ->  [reduction of 2598 by [3189]]
3202 : f(A,f$2)=f(A,A)                            [reduction of 2593 by [3189]]
3207 : f(f$2,f(A,f$2))=A                     [reduction of 3179 by [3189,3189]]
3209 : f(f(A,f$2),B)=f(B,f(A,B))                  [reduction of 3003 by [3189]]
3213 : f(f(A,f$2),f(A,B))=A                       [reduction of 2795 by [3189]]
3215 : f(A,f$2)=f$1(A)                            [reduction of 3183 by [3189]]
3218 + f(f(f(s$2,s$2),s$1),f(f(s$3,s$3),s$1))=f$1(f(f(s$2,s$3),s$1)) ->  [reduction of 3201 by [3215]]
3233 : f(f$1(A),f(A,B))=A                         [reduction of 3213 by [3215]]
3237 : f(f$1(A),B)=f(B,f(A,B))                    [reduction of 3209 by [3215]]
3239 : f$1(f$1(A))=A                         [reduction of 3207 by [3215,3215]]
3241 : f$1(A)=f(A,A)                              [reduction of 3202 by [3215]]
3261 : f(A,f(f$1(f(B,C)),f(A,B)))=f$1(A)     [reduction of 1214 by [3237,3241]]
3267 + f(f(f$1(s$2),s$1),f(f$1(s$3),s$1))=f$1(f(f(s$2,s$3),s$1)) ->  [reduction of 3218 by [3241,3241]]
3268 : f(f$1(f(A,B)),f(A,f(B,C)))=f(f$1(f(B,C)),A) [reduction of 1300 by [3241]]
3269 : f$1(f(f(A,B),f(A,C)))=f(f(f(A,B),f(A,C)),A) [reduction of 990 by [3241]]
6293 : f(f$1(A),f$1(f(B,C)))=f(f$1(f(B,C)),f(A,B))   [chaining of 3261 from 27]
6719 : f(f$1(f(A,B)),f$1(f(C,A)))=f(f$1(f(A,B)),C) [reduction of 3268 by [6293]]
6960 : f(f$1(f(A,B)),f(f$1(C),f$1(f(A,B))))=f(f$1(f(C,A)),f$1(f(A,B))) [chaining of 6293 from 3237]
7242 : f(A,f$1(f(B,C)))=f(f$1(f(B,A)),C) [reduction of 6960 by [3237,3239,6719]]
7447 : f(A,f(f$1(f(B,A)),C))=f(f$1(f$1(f(B,C))),A) [chaining of 7242 from 3237]
7816 : f(A,f(f$1(f(B,A)),C))=f(f(B,C),A)          [reduction of 7447 by [3239]]
9976 : f(A,f(f$1(B),f$1(f(C,A))))=f(f(C,f(B,f$1(f(C,A)))),A) [chaining of 3237 from 7816]
10461 : f(f(A,f$1(B)),C)=f(f(f(C,B),A),C)    [reduction of 9976 by [7816,7816]]
10482 : f(f(f$1(A),f$1(B)),C)=f$1(f(f(C,A),f(C,B))) [reduction of 3269 by [10461,10461]]
10559 : f$1(f(f(f(A,B),C),A))=f(f(f$1(C),f$1(f$1(A))),f(A,B)) [chaining of 3233 from 10482]
10993 : f$1(f(f(A,f$1(B)),C))=f(f(f$1(A),C),f(C,B)) [reduction of 10559 by [10461,3239]]
11020 + false                             [reduction of 3267 by [10993,3239,L]]

Length = 65, Depth = 25, Search Depth = 0



Total time: 324754 milliseconds

Number of forward inferences: 10719
Number of tableau inferences: 3
Number of generated clauses: 11020
Number of kept clauses: 501


*/



/*  Satellite on mpino2202 20030904

Proofs 
28293:inference(reflexivity,[27943])
27943:inference(reduction,inference(reduction_max,eq(7 : (n(f(a,f(c,b))) \= f(f(a,n(c)),f(a,n(b)))) = $true),[27521]),[47,47])
27521:inference(cp,[2,27010])
27010:inference(cp,[26564,2])
26564:inference(cp,[26250,47])
26250:inference(reduction,inference(cp,[41,4163]),[47,47,2,7665,2,7543])
4163:inference(cp,[3982,58])
4566:inference(cp,[948,4036])
9352:inference(reduction,inference(cp,[1628,7543]),[47,7543,47])
8794:inference(reduction,inference(cp,[8195,7543]),[47,7543,47])
8640:inference(reduction,inference(cp,[8094,7543]),[47,7543,47])
24473:inference(cp,[2,24203])
24203:inference(reduction,inference(cp,[23928,7543]),[58])
23928:inference(cp,[2,23581])
23581:inference(reduction,inference(cp,[1029,7558]),[2,1559])
7558:inference(reduction,inference(cp,[169,7271]),[7271])
22527:inference(cp,[6909,21885])
21885:inference(cp,[21290,7305])
21290:inference(cp,[20608,15623])
20608:inference(cp,[19964,18776])
19964:inference(cp,[6323,19394])
19394:inference(cp,[18776,6955])
18776:inference(cp,[15623,6955])
15623:inference(reduction,inference(cp,[7790,15336]),[2,6909,6909,2,7665,7890,47,2,7790])
7305:inference(cp,[6909,6909])
6955:inference(reduction,inference(cp,[6537,6537]),[2])
8101:inference(reduction,inference(cp,[7545,7271]),[7271])
7557:inference(reduction,inference(cp,[7271,7271]),[7271])
16502:inference(cp,[2,5935])
14015:inference(reduction,inference(cp,[13867,1559]),[6])
16817:inference(reduction,inference(cp,[151,13780]),[47])
13780:inference(reduction,inference(cp,[13615,1559]),[2,6])
13628:inference(reduction,inference(cp,[2424,1559]),[6909,4610,6909,2,6,34])
6102:inference(reduction,inference(cp,[1406,6011]),[47,1559])
5935:inference(reduction,inference(cp,[4368,2299]),[2])
2425:inference(cp,[948,938])
1160:inference(reduction,inference(cp,[938,1078]),[938])
9351:inference(reduction,inference(cp,[254,1628]),[2,55,151,47,2,6909,2])
15903:inference(reduction,inference(cp,[15902,7552]),[52,2,2])
15902:inference(reduction,inference(reduction,inference(cp,[2419,84]),[47,2,1559]),[6909,2])
15336:inference(cp,[2,8099])
8099:inference(reduction,inference(cp,[7545,7665]),[5511,47])
7790:inference(reduction,inference(cp,[7271,7665]),[4610,47])
7556:inference(reduction,inference(cp,[3880,7271]),[7271])
6271:inference(reduction,inference(cp,[6190,1514]),[47,2,40])
6197:inference(reduction,inference(cp,[6098,1514]),[47,2,40])
5846:inference(reduction,inference(cp,[5,5369]),[2,2,1559])
6015:inference(reduction,inference(cp,[68,4157]),[2])
6949:inference(reduction,inference(cp,[1406,6537]),[2,47])
4788:inference(reduction,inference(cp,[1453,4610]),[47,2,40,47])
13867:inference(cp,[47,13759])
13759:inference(cp,[223,13615])
13615:inference(reduction,inference(cp,[41,2424]),[41])
2424:inference(reduction,inference(cp,[1406,948]),[1406,47,47,2])
13173:inference(cp,[12787,2])
12787:inference(cp,[2,12332])
12515:inference(cp,[12133,2])
12332:inference(cp,[11935,2])
12133:inference(cp,[6537,11264])
11935:inference(cp,[6537,10971])
11264:inference(cp,[10306,2])
10971:inference(cp,[8095,2])
10784:inference(cp,[7552,2])
10603:inference(cp,[1157,2])
10466:inference(cp,[1029,2])
10306:inference(cp,[2,995])
10002:inference(cp,[273,2])
8095:inference(reduction,inference(cp,[7545,1514]),[2,914])
7552:inference(reduction,inference(cp,[7271,1514]),[2,938])
1157:inference(reduction,inference(cp,[1078,259]),[2])
1029:inference(reduction,inference(cp,[985,259]),[2])
995:inference(reduction,inference(cp,[938,259]),[2])
952:inference(reduction,inference(cp,[914,259]),[2])
273:inference(reduction,inference(cp,[68,167]),[2])
7789:inference(reduction,inference(cp,[7665,7271]),[7271])
7678:inference(reduction,inference(cp,[7543,7271]),[7271])
1027:inference(reduction,inference(cp,[68,985]),[2,2])
1454:inference(reduction,inference(cp,[1406,1078]),[1406,47])
399:inference(reduction,inference(cp,[254,169]),[254,2,151])
1628:inference(reduction,inference(cp,[223,1559]),[1514])
9128:inference(cp,[9045,2])
9045:inference(cp,[223,8943])
8943:inference(cp,[8519,2])
8519:inference(cp,[6537,7277])
8195:inference(cp,[4368,1626])
8094:inference(reduction,inference(cp,[7545,3982]),[2,1514,6909])
7277:inference(cp,[6909,58])
6557:inference(cp,[6323,1087])
5207:inference(reduction,inference(cp,[1406,5072]),[1406,47,2,47])
2577:inference(cp,[1087,2])
1626:inference(reduction,inference(cp,[169,1559]),[1559])
7545:inference(reduction,inference(cp,[68,7271]),[7271])
7890:inference(cp,[2,7788])
7788:inference(reduction,inference(cp,[7665,1559]),[1559])
7665:inference(reduction,inference(cp,[68,7543]),[47,7543])
7543:inference(reduction,inference(cp,[1559,7271]),[7271])
7271:inference(reduction,inference(cp,[6909,1559]),[47])
6909:inference(cp,[6537,6323])
6537:inference(cp,[2,6323])
6323:inference(reduction,inference(cp,[4048,2461]),[4048,47,2,946])
4048:inference(reduction,inference(cp,[15,3981]),[2])
6190:inference(cp,[2,6098])
6098:inference(cp,[47,6011])
6011:inference(cp,[2,4157])
4157:inference(reduction,inference(cp,[3982,2299]),[2])
4368:inference(reduction,inference(cp,[3981,4233]),[47,47])
5369:inference(reduction,inference(cp,[71,5201]),[71,47,47])
5511:inference(cp,[5367,2])
5367:inference(cp,[2,5201])
5201:inference(cp,[5072,2])
5072:inference(reduction,inference(cp,[223,4917]),[223])
4917:inference(cp,[4776,2])
4776:inference(reduction,inference(cp,[68,4610]),[47,4610])
4610:inference(cp,[4235,2])
4235:inference(reduction,inference(cp,[15,4035]),[2])
4036:inference(reduction,inference(cp,[3981,3808]),[47,47])
3983:inference(reduction,inference(cp,[1048,3880]),[2])
4233:inference(cp,[2,4035])
4035:inference(reduction,inference(cp,[3981,1559]),[1559,47,2])
3982:inference(reduction,inference(cp,[3880,1559]),[1559,47])
3981:inference(cp,[47,3880])
3880:inference(reduction,inference(cp,[47,3808]),[47,2])
3808:inference(cp,[2881,307])
2881:inference(reduction,inference(cp,[2819,1559]),[47,40])
3580:inference(cp,[223,3452])
3452:inference(cp,[2,3300])
3300:inference(cp,[2,3187])
3187:inference(reduction,inference(cp,[3096,1559]),[40,47])
3096:inference(cp,[223,3030])
3030:inference(cp,[344,2934])
2934:inference(cp,[2,2880])
2880:inference(reduction,inference(cp,[2819,1014]),[47])
2819:inference(reduction,inference(cp,[217,2621]),[2,2])
2621:inference(reduction,inference(cp,[1574,191]),[2])
1574:inference(reduction,inference(cp,[217,1514]),[2,1514])
1087:inference(cp,[1048,894])
2461:inference(cp,[2,2419])
2419:inference(reduction,inference(cp,[216,948]),[216,47,47,2])
948:inference(reduction,inference(cp,[914,52]),[2])
916:inference(reduction,inference(cp,[894,731]),[2,2,68,2])
2299:inference(cp,[1184,2])
2262:inference(cp,[397,2])
1184:inference(cp,[1048,388])
397:inference(reduction,inference(cp,[201,169]),[201,2,151])
2136:inference(cp,[2,2050])
2050:inference(reduction,inference(cp,[1929,344]),[1929])
1929:inference(cp,[2,224])
1055:inference(reduction,inference(cp,[73,1014]),[73,47])
947:inference(reduction,inference(cp,[71,914]),[47])
224:inference(reduction,inference(cp,[41,15]),[2])
1085:inference(reduction,inference(cp,[216,1048]),[2,216,47,2])
1768:inference(cp,[1024,2])
1701:inference(cp,[946,2])
1024:inference(cp,[985,52])
946:inference(reduction,inference(cp,[58,914]),[47])
344:inference(cp,[259,217])
1559:inference(cp,[2,1514])
1514:inference(cp,[1482,307])
1482:inference(reduction,inference(cp,[13,307]),[2,1])
307:inference(cp,[217,217])
1453:inference(reduction,inference(cp,[1406,152]),[47,2])
1406:inference(cp,[985,57])
57:inference(reduction,inference(cp,[41,5]),[2])
1284:inference(cp,[84,991])
1148:inference(cp,[1078,152])
1023:inference(cp,[985,152])
991:inference(reduction,inference(cp,[938,152]),[2])
945:inference(reduction,inference(cp,[914,152]),[2])
434:inference(reduction,inference(cp,[133,223]),[133,2,151])
388:inference(reduction,inference(cp,[191,169]),[191,2,151])
1078:inference(reduction,inference(cp,[1048,42]),[201,40,47])
1048:inference(cp,[2,1014])
1014:inference(cp,[985,2])
985:inference(cp,[5,938])
938:inference(cp,[2,914])
914:inference(reduction,inference(cp,[47,894]),[47])
894:inference(cp,[199,47])
199:inference(reduction,inference(cp,[191,13]),[151])
73:inference(reduction,inference(cp,[68,68]),[2,68])
731:inference(cp,[2,658])
658:inference(cp,[2,531])
531:inference(cp,[263,340])
382:inference(cp,[2,169])
340:inference(cp,[5,259])
263:inference(reduction,inference(cp,[167,13]),[2])
223:inference(reduction,inference(cp,[15,13]),[2])
169:inference(reduction,inference(cp,[5,13]),[2])
259:inference(cp,[167,2])
217:inference(cp,[15,2])
167:inference(cp,[2,13])
254:inference(reduction,inference(cp,[216,152]),[47,2])
216:inference(reduction,inference(cp,[31,15]),[2,40])
15:inference(cp,[2,5])
201:inference(cp,[198,2])
198:inference(cp,[68,191])
191:inference(cp,[47,152])
152:inference(cp,[2,133])
13:inference(cp,[2,5])
162:inference(cp,[151,157])
157:inference(reduction,inference(cp,[151,133]),[47])
151:inference(reduction,inference(cp,[133,84]),[47,2,55])
133:inference(reduction,inference(cp,[41,126]),[2,40,2])
126:inference(cp,[47,35])
35:inference(reduction,inference(cp,[31,5]),[2])
71:inference(cp,[68,58])
96:inference(cp,[84,2])
84:inference(cp,[2,11])
11:inference(cp,[1,5])
42:inference(cp,[2,9])
68:inference(cp,[2,55])
55:inference(reduction,inference(cp,[5,41]),[2])
58:inference(cp,[47,52])
52:inference(cp,[2,41])
41:inference(reduction,inference(cp,[9,5]),[1])
47:inference(reduction_max,eq(19 : A = n(f(A,t))),[40])
40:inference(reduction,inference(cp,[1,9]),[1,6])
9:inference(reduction,inference(cp,[5,5]),[2,2])
36:inference(cp,[34,8])
34:inference(reduction,inference(cp,[31,8]),[1])
31:inference(reduction,inference(cp,[29,5]),[1,2])
29:inference(reduction,inference(cp,[19,6]),[2])
19:inference(cp,[8,1])
8:inference(cp,[6,5])
5:axiom(2)
7:inference(reduction,(goal 1),[1,2,1,2,2,2,2,1])
6:inference(reduction,axiom(4),[1])
2:axiom(1)
1:axiom(3)

Number of reductions = 74301
Number of subsumption checks [successful] = 32780 [4404]

Time: 59.685

*/
