:- module(saturation,[
		    active/2,
		    active/3,
		    ord/2,
		    ord/3,
		    sat/6
		   ]).



:- use_module(hgstuff,[
		       split/4,
		       unzip/3,
		       zip/3
		       
   ]).

:- use_module(rewrite,[
		       normalform/4,
		       normalform1/4
		      ]).

:- use_module(theory,[add_actives/3,
		      empty_actives/2,
		      delete_active/3,
		      inferences/3,
		      redundant/2,
		      unify_active/3]).

:- use_module(formulas).

:- use_module(library(change_arg),[change_arg/4]).

:- use_module(library(maplist),[maplist/2,maplist/3]).

:- use_module(library(basics), [
	member/2
   ]).

:- use_module(global_vars, [
		       cont/2,
		       initialize_vars/0,
		       restore_vars/1,
		       save_vars/1,
		       store/2,
		       timer/2
   ]).

:- use_module(log, [
		       log/2
   ]).



/* -----------------------------------------------------------------------------


State   ::= state(Active,Passive,Deleted,Order,EqCtr,Proofs)
Active  ::= Rules                     % active (processed) equations with theory
                                      % context

Passive ::= [Equation]                % ordered list (by Size)
Deleted ::= [nat]                     % list of previously deleted equations
EqCtr   ::= nat                       
Order   ::= ord(lpo,Precedence)
         |  ... more to come ...
Proofs  ::= [Proof]                   % proof for each active equation

EqCtr and Ord are not always uptodate.
They are stored globally (under cont), and accessed globally
during completion and formula construction.
Hence, upon recursive calls they need to be saved and restored explicitly.

----------------------------------------------------------------------------- */



active(S,A):-
	arg(1,S,A).

active(S,S1,A):-
	change_arg(1,S,S1,A).

passive(S,P):-
	arg(2,S,P).

passive(S,S1,P):-
	change_arg(2,S,S1,P).

deleted(S,P):-
	arg(3,S,P).

deleted(S,S1,P):-
	change_arg(3,S,S1,P).

ord(S,O):-
	arg(4,S,O).

ord(S,S1,O):-
	change_arg(4,S,S1,O).

eq_ctr(S,N):-
	arg(5,S,N).

eq_ctr(S,S1,N):-
	change_arg(5,S,S1,N).

proofs(S,P):-
	arg(6,S,P).

proofs(S,S1,P):-
	change_arg(6,S,S1,P).

/* -----------------------------------------------------------------------------


Initialization

----------------------------------------------------------------------------- */



init_empty(S,Thy):-
	S=state(Active,[],[],_Ord,_EqCtr,[]),
	empty_actives(Thy,Active).





/* -----------------------------------------------------------------------------


Saturation

----------------------------------------------------------------------------- */


sat(Thy,Ord,ActiveEqs,PassiveEqs,ActiveEqsOut,PassiveEqsOut):-
	save_vars(Saved),

	% initialize state
	init_empty(S,Thy),
	maplist(proof_line,ActiveEqs,InitialProofs),
	proofs(S,S0,InitialProofs),
	store(ord,Ord),
	length(ActiveEqs,C),
	store(eq_ctr,C),
	Ord=ord(OType,OParams),
	OType:init_ordering(OParams),

	% process initial active equations
	add_actives_to_thy(S0,S1,ActiveEqs),
	log(active,['$listnl'(ActiveEqs)]),
	
	% process initial passive equations
	insert_new_passive(PassiveEqs,[],Passive_with_weight),
	passive(S1,S2,Passive_with_weight),
	log(passive,['$listnl'(Passive_with_weight)]),
	
	% call the Waldmeister loop
	wm_loop(S2,SE),
	
	% (possibly) log result
	active(SE,ActiveEqsOut),
	passive(SE,PassiveEqsOut),
	findall(E,unify_active(_,ActiveEqsOut,eq(_,E)),Final),
	log(active,['$listnl'(Final)]),
%	log(passive,['$listnl'(PassiveEqsOut)]),
	proofs(SE,Proofs),
	log(proofs,['$listnl'(Proofs)]),
	cont(red_ctr,RC),
	cont(red_timer,RT),
	cont(match_ctr,MC),
	log(red_ctr,[RC,MC,RT]),
	cont(subsumption_ctr,SC),
	cont(success_sub_ctr,SS),
	log(subsumption_ctr,[SC,SS]),
%	deleted(SE,Del),
%	print('$listnl'(Del)),

	% restore global variables
	restore_vars(Saved).





/* -----------------------------------------------------------------------------


The Waldmeister loop.

However we do not explicitly interreduce old active
equations with newly generated ones.
Reduction is viewed as a simplifying inference, hence
is included in inference compuation.


----------------------------------------------------------------------------- */


:- dynamic spy_equation/1.

wm_loop(S,SE):-
	passive(S,[E|Passive]),   % E has minimal weight among Passive
	log(current_eqs,['$listnl'([E|Passive])]),
	!,
	% goals not refuted and at least one passive equation left
	log(select_active,[E]),
	deleted(S,Deleted),
	(orphan(E,Deleted) ->
	    passive(S,S1,Passive),
	    wm_loop(S1,SE)
	;
%	    (spy_equation(_),
%		number(E,N),
%		spy_equation(N) -> trace; true),
	    
	    normalize(S,E,E_normalized),
	    log(normalized_active,[E_normalized]),
	    active(S,Active),
	    
	    (redundant(E_normalized,Active) ->
		log(redundant,[E]),
		passive(S,S1,Passive),
		wm_loop(S1,SE)
	    ;inconsistency(E_normalized) ->
		passive(S,S0,Passive),
		add_actives_to_thy(S0,S1,[E_normalized]),
		proof(E_normalized,P),
		number(E_normalized,N),
		proofs(S,Proofs),
		proofs(S1,SE,[N:P|Proofs]),
		log(inconsistent,[E_normalized])
	    ;
		wm_loop_body(S,S2,E_normalized,Deleted,Passive),
		% continue
		wm_loop(S2,SE)
	    )
	).
wm_loop(S,S).


wm_loop_body(S1,SE,E_normalized,Deleted,Passive) :-
	complete_eq(E_normalized),
	log(new_active,[E_normalized]),
%	(number(E_normalized,223) -> trace; true),

	        % add new active equation
	add_actives_to_thy(S1,S2,[E_normalized]),
	active(S2,Active_with_E),

		% compute inferences from new active eq and all active eqs
	inferences(Active_with_E,E_normalized,NewEqs),
	log(new,['$listnl'(NewEqs)]),

	        % identify deletable actives (those which are made
	        % redundant by the new equaitons) from the proofs
	        % of the new equations
	redundant_active(Active_with_E,NonRedundantActive,NewEqs,Del1/Deleted),
	active(S2,S2E,NonRedundantActive),

	        % reduce new equations
	timer(maplist:maplist(saturation:normalize(S2E),NewEqs,P2),red_timer),
	split(redundant(Active_with_E),P2,_,New_nonredundant),
	log(nonredundant_new,['$listnl'(New_nonredundant)]),

		% insert new passive equations and number them
	insert_new_passive(New_nonredundant,Passive,NewPassive),


		% update state
	passive(S2E,S2EP,NewPassive),
	deleted(S2EP,S2EPD,Del1),
	proofs(S2EPD,Proofs),
	proof(E_normalized,P),
	number(E_normalized,N),
	proofs(S2EPD,SE,[N:P|Proofs]),
		
	!.

/* -----------------------------------------------------------------------------

Redundant Rule Deletion

----------------------------------------------------------------------------- */

redundant_active(Active,NewActive,[E|Eqs],Deleted/End):-
	(proof(E,inference(reduction_max,eq(Ru),_)) ->
	    delete_active(Ru,Active,Active1),
	    number(Ru,N),
	    End1=[N|End]
	
	;
	    Active1=Active,
	    End1=End
	),
	redundant_active(Active1,NewActive,Eqs,Deleted/End1).
redundant_active(Active,Active,[],Del/Del).




/* -----------------------------------------------------------------------------

Normalization of equations

We return the original equation (same number) if irreducible.
We leave the number open (uninstantiated), if equation was reduced.
Normalization is independent of the formula structure by first
converting the formula to a term.

----------------------------------------------------------------------------- */

normalize(S,E,E_normalized):-
	formula2term(E,T,Ty),
	active(S,Rs),
	normalform(T,TN,Used/[],Rs),
	(Used=[] ->
	    % irreducible equation
	    E_normalized=E
	;
	    proof(E,P),
	    Proof=inference(reduction,P,Used),
	    mk_equation(eq(TN,Ty,Proof),E_normalized) % unnumbered eq
	).


/* -----------------------------------------------------------------------------

Orphan Detection

orphan(+E:Equation,+Deleted:[nat])
  is true if one equation in the proof of E is among the
  deleted equations in Deleted.
  At this point there's no deep checking.


----------------------------------------------------------------------------- */

orphan(E,Deleted):-
	proof(E,P),
	orphan_by_proof(P,Deleted),
	log(orphan,[E,P]).

orphan_by_proof(inference(_Type,Proof_of_main_premise,SidePremises),Deleted):-
	(
	  member(N,SidePremises),
	  member(N,Deleted)
	;
	  orphan_by_proof(Proof_of_main_premise,Deleted)
	),
	!.
orphan_by_proof(inference(_Type,SidePremises),Deleted):-
	member(N,SidePremises),
	member(N,Deleted),
	!.


/* -----------------------------------------------------------------------------

New passive equatons

----------------------------------------------------------------------------- */
	
% first sort eqs by weight and then insert them simultaneously
% brings a big performance improvement for large eq sets

insert_new_passive(Eqs,Passive,Passive1):-
	maplist(weight,Eqs,Ws),
	zip(WeightedEqs,Ws,Eqs),
	sort(WeightedEqs,SortedWeightedEqs),
	insert_by_weight(SortedWeightedEqs,Passive,Passive1).

insert_new_passive(Eqs,Passive,Passive1):-
	user:append(Eqs,Passive,Passive1).

insert_by_weight([],Es,Es):-
	!.
% we number the new equations
insert_by_weight(WEs,[],Es):-
	unzip(WEs,_,Es),
	maplist(number,Es,_).
insert_by_weight([W-E|WEs],[E1|Es],New):-
	weight(E1,W1),
	(W<W1 ->
	    number(E,_),
	    New=[E,E1|Es1],
	    insert_by_weight(WEs,Es,Es1)
	;
	% if W=W1 we insert E after E1 (E must have a larger number than E1)
	    insert_by_weight([W-E|WEs],Es,EEs),
	    New=[E1|EEs]
	),
	!.






/* -----------------------------------------------------------------------------

Auxiliary

----------------------------------------------------------------------------- */
	


proof_line(E,N:P):-
	number(E,N),
	proof(E,P).



add_actives_to_thy(S0,S1,A):-
	active(S0,Active0),
	add_actives(Active0,Active1,A),
	active(S0,S1,Active1).


