% -*- Module: formulas; -*-

:- module(formulas,[
		    complete/1,
		    equation_template/1,
		    formula2term/3,
		    left/2,
		    left/3,
		    inconsistency/1,
		    equation2clause/2,
		    equation2clause/5,
		    number/2,
		    number/3,
		    proof/2,
		    proof/3,
		    right/2,
		    right/3,
%		    skolemized/3,
		    type/2,
		    type/3,
		    weight/2,
		    weight/3
		   ]).

:- use_module(library(change_arg),[change_arg/4]).


:- use_module(global_vars, [
        cont/2,
	store/2
   ]).

:- use_module(st_terms).

:- use_module(order,[orient/3]).



/* ----------------------------------------------------------------------------

Clause      ::= cl(Kind,Literals,Number,Weight,Proof)

Kind        ::= unit | negative | positive | clause
Literals    ::= [Literal]
Literal     ::= lit(Sign,Left,Right,Ty)
Sign        ::= + | -
Left        ::= Term                             % left side
Right       ::= Term                             % right side
Ty          ::= Type
Number      ::= nat
Weight      ::= number
Proof       ::= inference(Inference,Premise,[Premise])
             |  inference(Inference,[Premise])


Upon creation of a clause, the Literals and Proof have to be provided.
Weight and Number will be computed lazily.

---------------------------------------------------------------------------- */


equation_template(cl(_,_,_,_,_)).


/* ----------------------------------------------------------------------------

  Creation

---------------------------------------------------------------------------- */

equation2clause(S\=T,'$true',_Ty,P,cl(unit,[lit(-,S,T,prop)],_N,_W,P)):-
	!.
equation2clause(L,R,Ty,P,          cl(unit,[lit(+,L,R,Ty)]  ,_N,_W,P)).



equation2clause(eq(L,R,T,P),C):-
	equation2clause(L,R,T,P,C).

equation2clause(eq(L=R,T,P),C):-
	equation2clause(L,R,T,P,C).



/* ----------------------------------------------------------------------------

  Conversion

---------------------------------------------------------------------------- */

formula2term(cl(unit,[lit(-,S,T,prop)]),S\=T,prop).
formula2term(cl(unit,[lit(+,S,T,Ty)]),S=T,Ty).



/* ----------------------------------------------------------------------------

  Access

---------------------------------------------------------------------------- */


% literal level

sign(E,S):-
	arg(1,E,S).

sign(E,E1,S):-
	change_arg(1,E,E1,S).


left(E,L):-
	arg(2,E,L).

left(E,E1,L):-
	change_arg(2,E,E1,L).

right(E,R):-
	arg(3,E,R).

right(E,E1,R):-
	change_arg(3,E,E1,R).


type(E,T):-
	arg(4,E,T).

type(E,E1,T):-
	change_arg(4,E,E1,T).






% clause level

kind(E,S):-
	arg(1,E,S).

kind(E,E1,S):-
	change_arg(1,E,E1,S).

literals(E,L):-
	arg(2,E,L).

literals(E,E1,L):-
	change_arg(2,E,E1,L).

number(E,N):-
	arg(3,E,N),
	(var(N) ->
	    cont(eq_ctr,M),
	    N is M+1,
%	    format('~Nnew number:~p',[E]),
	    store(eq_ctr,N)
	;
	    true).

number(E,E1,N):-
	change_arg(3,E,E1,N).

weight(E,S):-
	% special handling of unit equations
	arg(E,1,unit),
	!,
	arg(E,2,lit(Sign,L,R,_),
	arg(4,E,S),
	(var(S) ->
	    proof(E,P),
	    left(E,L),
	    right(E,R),
	    orient(L,R,O),
	    (O\==check ->
		(O=lr ->
		    L1=L,
		    R1=R
		; %O=rl 
		    L1=R,
		    R1=L
		),
		term_weight(L1,SL),
		term_weight(R1,SR),
		S is SL*(SL+SR)	% Waldmeister weight function
	    ;
		term_weight(L,SL),
		term_weight(R,SR),
		S1 is max(SL,SR)*(SL+SR),
		((P=axiom(_);P=goal(_)) ->
		    S is S1/2
		;
		    S1=S
		)
	    )
	;
	    true
	).


weight(E,E1,N):-
	change_arg(4,E,E1,N).

proof(E,N):-
	arg(5,E,N).

proof(E,E1,N):-
	change_arg(5,E,E1,N).



	
complete(E):-
	number(E,_),
	weight(E,_).
	



/* ----------------------------------------------------------------------------

                Term Size Computation


---------------------------------------------------------------------------- */

term_weight( T, 0.7 ):-
        var_(T), !.
%/*
term_weight(T,0.7):-
        object_var_(T), !.
%*/
term_weight('$neq'(L,R),W):-
	term_weight(L,W1),
	term_weight(R,W2),
	W is max(W1,W2)/2.
term_weight( T, S ):-
	functor_(T,_O,N),
	term_weight_from_to(0,N,T,S1),
	S is S1 + 1.


term_weight_from_to(I,N,T,S):-
	(I=N ->
	     S=0
	;
	     I1 is I+1,
	     arg_(I1,T,T1),
	     term_weight(T1,S1),
	     term_weight_from_to(I1,N,T,S2),
	     S is S1+S2
	).



term_size( T, 1 ):-
        var_(T), !.
term_size(T,1):-
        object_var_(T), !.
term_size( T, S ):-
	functor_(T,_O,N),
	term_size_from_to(0,N,T,S1),
	S is S1 + 1.


term_size_from_to(I,N,T,S):-
	(I=N ->
	     S=0
	;
	     I1 is I+1,
	     arg_(I1,T,T1),
	     term_size(T1,S1),
	     term_size_from_to(I1,N,T,S2),
	     S is S1+S2
	).






/* ----------------------------------------------------------------------------

                Testing

---------------------------------------------------------------------------- */

inconsistency(eq('$true','$false',_,_,_,_)).
inconsistency(eq('$false','$true',_,_,_,_)).


/* ----------------------------------------------------------------------------

                Printing

---------------------------------------------------------------------------- */


:- multifile user:portray/1.

user:portray(cl(unit,[lit(+,Left,Right,Type)],Number,Weight,Proof)):-
	portray(eq(Left,Right,Number,Weight,Type,Proof)).
user:portray(eq(Left,Right,Number,_Weight,_Type,_Proof)):-
	Orientation=lr,  
	(Orientation==rl ->
	    L=Right,
	    R=Left
	;
	    L=Left,
	    R=Right
	),
	(var(Number) ->
	    \+ \+ (   numbervars((L,R),0,_),
%		      format('~t  ~7| ~p = ~p [~p,~p]',[L,R,_LS,_RS]))
		      format('~t  ~7| ~p = ~p',[L,R]))
	;
	    \+ \+ (   numbervars((L,R),0,_),
%		      format('~t~d :~7| ~p = ~p [~p,~p]',[Number,L,R,_LS,_RS]))
		      format('~t~d :~7| ~p = ~p',[Number,L,R]))
	).
