% -*- Module: st_terms; -*-

:- module(st_terms,[
		    arg_/3,
		    contains_var/2,
		    functor_/3,
		    genarg_/3,
		    is_function_symbol/1,
		    is_term/1,
		    is_type/1,
		    nonvar_/1,
		    prolog2term/2,
		    sample_terms/1,
		    univ_/2,
		    var_/1
		    ]).

:- use_module(satsrc(mylists),
	      [
	       is_list/1,
	       nth1/3
	      ]).

:- use_module(library(maplist),[maplist/2,maplist/3]).
:- use_module(library(terms),[subsumes_chk/2]).


/* ---------------------------------------------------------------------------

term data structure:


% oriented equations 

equation  ::= eq(orientation,lhs_term,rhs_term)
orientation ::= uniform                       % from left to right
             |  non_uniform                   % both orientations possible


% simply typed terms
	     
term      ::= function_symbol @ [terms]
           |  variable
variable  ::= v(prolog_variable, type)
function_symbol ::= prolog_atom / type


% simple types (can be higher-order and can have type variables for
% implementing polymorphic types

type      ::= tc(type_cons,[type])            % type constructor
           |  tv(type_var)                    % type variable
	   |  [type]                          % product type
	   |  type -> type                    % function type
	   
	   |  o                               % abbrev. for tc(o,[])
	   |  arity                           % abbrev. for
		              % [o,...,o] -> o, with arity many arguments;
	                      % to be used in untyped applications
	   
type_cons ::= prolog_atom
type_var  ::= prolog_var
arity     ::= prolog_nat



------------------------------------------------------------------------------ */


/* ----------------------------------------------------------------------------

Type checking

------------------------------------------------------------------------------ */

is_term(T):-
	nonvar(T),
	T=(F@Args),
	is_function_symbol(F),
	is_list(Args),
	maplist(is_term,Args).
is_term(T):-
	nonvar(T),
	T=v(X,Ty),
	var(X),
	is_type(Ty).

is_function_symbol(F):-
	nonvar(F),
	F=FS/Ty,
	atomic(FS),
	is_type(Ty).

is_type(Ty):-
	nonvar(Ty),
	Ty=tc(TC,TyArgs),
	atomic(TC),
	is_list(TyArgs),
	maplist(is_type,TyArgs),
	(TC=o -> TyArgs=[]; true).
is_type(Ty):-
	nonvar(Ty),
	Ty=tv(TV),
	var(TV).
is_type(Ty):-
	nonvar(Ty),
	is_list(Ty),
	maplist(is_type,Ty).
is_type(Ty):-
	Ty==o.
is_type(Ty):-
	integer(Ty),
	Ty>=0.



/* ----------------------------------------------------------------------------

    sample terms for testing

------------------------------------------------------------------------------ */

sample_terms(T):-
	Nat = tc(nat,[]),
	M    = '*'/([Nat,Nat]->Nat),
	M1   = '*'/3,
	I1   = i/1,
	Zero = 0/Nat,
	Zero1 = 0/0,
	
	X=v(_X,Nat),
	Y=v(_Y,Nat),
	X1=v(_X,o),
	Y1=v(_Y,o),
	
	T=[X,
	   M@[Y,T2],
	   I@[M@[X,Y]],
	   M@[I@[X],I@[Y]],
	   I1@[M1@[X1,Y1]],
	   M1@[I1@[X1],I1@[Y1]],
	   Zero1@[]
	  ],
	
	T2=(M@[Y,T3]),
	T3=(Zero@[]).


% :- sample_terms([_T1,_T2,_T3,_T4,T5|_]),genarg_(I,T5,T5I),arg_(2,T5I,T5I2).

/* ----------------------------------------------------------------------------

Primitive operations on terms as we have them for Prolog terms

------------------------------------------------------------------------------ */


nonvar_(_@_).

var_(v(_,_)).

functor_(F@Args,F,N):-
	length(Args,N).


arg_(I,F@Args,ArgI):-
	integer(I),
	!,
	(nth1(I,Args,ArgI) ->
	    true
	;
	    raise_exception(
	     user_error(
                 '~N! Index too large in argument ~w of ~q/~w~n! Goal: ~p~n',
		 [1,arg_,3,arg_(I,F@Args,ArgI)]
		       ))
	).
arg_(I,T,TI):-
	(nonvar(I) ->
	    raise_exception(
	     user_error(
                 '~N! Type error in argument ~w of ~q/~w~n! Goal: ~p~n',
		 [1,arg_,3,arg_(I,T,TI)]
		       ))
	;
	    genarg_(I,T,TI)).




% code adapted from arg.pl in the Quintus library

genarg_(N, Term, Arg) :-
	integer(N),
	nonvar(Term),
	!,
	arg_(N, Term, Arg).
genarg_(N, Term, Arg) :-
	var(N),
	nonvar(Term),
	!,
	functor_(Term, _, Arity),
	genarg_(Arity, Term, Arg, N).
genarg_(N, Term, Arg) :-
	var(Term),
	!,
	raise_exception(
	 user_error(
            '~N! Instantiation fault in argument ~w of ~q/~w~n! Goal: ~p~n',
            [2,genarg_,3,genarg_(N,Term,Arg)]
		   )).
genarg_(N, Term, Arg) :-
	raise_exception(
	 user_error(
	    '~N! Type failure in argument ~w of ~q/~w~n! Goal: ~p~n',
	    [1,genarg_,3,genarg_(N,Term,Arg)])
		       ).


genarg_(1, Term, Arg, 1) :- !,
	arg_(1, Term, Arg).
genarg_(N, Term, Arg, N) :-
	arg_(N, Term, Arg).
genarg_(K, Term, Arg, N) :-
	K > 1, J is K-1,
	genarg_(J, Term, Arg, N).


univ_(T,[F|Ts]):-
	(nonvar_(T) ->
	    T=(F@Ts)
	;
	    raise_exception(
			    user_error(
	    '~N! Type failure in argument ~w of ~q/~w~n! Goal: ~p~n',
	    [1,univ_,2,univ_(T,[F|Ts])])
			   )
	).




/* ----------------------------------------------------------------------------

                Nonprimitive operations

Some code is adapted from Quintus Prolog library occurs.pl


------------------------------------------------------------------------------ */



%   contains_var(+Variable, +Term)
%   is true when the given Term contains at least one sub-term which
%   is identical to the given Variable.  We use '=='/2 to check for
%   the variable so it can be used to check
%   for arbitrary terms, not just variables.

contains_var(Variable, Term) :-
	\+ free_of_var(Variable, Term).


%   free_of_var(+Variable, +Term)
%   is true when the given Term contains no sub-term identical to the
%   given Variable (which may actually be any term, not just a var).
%   For variables, this is precisely the "occurs check" which is
%   needed for sound unification.

free_of_var(Variable, Term) :-
	Term == Variable,
	!,
	fail.
free_of_var(Variable, Term) :-
	nonvar_(Term),
	!,
	functor_(Term, _, Arity),
	free_of_var(Arity, Term, Variable).
free_of_var(_, _).

free_of_var(1, Term, Variable) :- !,
	arg_(1, Term, Argument),
	free_of_var(Variable, Argument).
free_of_var(N, Term, Variable) :-
	arg_(N, Term, Argument),
	free_of_var(Variable, Argument),
	M is N-1, !,
	free_of_var(M, Term, Variable).




/* ----------------------------------------------------------------------------

                Conversions


------------------------------------------------------------------------------ */

%  prolog2term(+PT, ?STT)
%  converts a Prolog term PT into a simply typed term STT
%  by using shorthand notation for types for the untyped case

prolog2term(X,v(X,o)):-
	var(X),
	!.
prolog2term(T,(F/N)@TTArgs):-
	T=..[F|TArgs],
	length(TArgs,N),
	maplist(prolog2term,TArgs,TTArgs).
	


term2prolog(v(X,_),X).
term2prolog((F/_N)@Args,T):-
	T=..[F|TArgs],
	maplist(term2prolog,Args,TArgs).
