%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Path orderings
%
% Uwe Waldmann <uwe@mpi-sb.mpg.de> 1996/02/09
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The module "pathorderings" implements:
% 
% * Rubio/Nieuwenhuis ordering (RTA'93, pp. 374-388):
%   - rnGreater(T1,T2)
%   - rnGreaterEqual(T1,T2)
%       Succeeds if T1 is greater than (or equal to) T2 w.r.t. the
%       Rubio/Nieuwenhuis ordering, fails otherwise.
%   - gndRnGreater(T1,T2)
%   - gndRnGreaterEqual(T1,T2)
%       Same as rnGreater and rnGreaterEqual, except that variables
%       are taken as maximal constants below "first_predicates".
% * Recursive path ordering:
%   - rpoGreater(T1,T2)
%   - rpoGreaterEqual(T1,T2)
%   - gndRpoGreater(T1,T2)
%   - gndRpoGreaterEqual(T1,T2)
%       Similar for the recursive path ordering.
% * Lexicographic path ordering:
%   - lpoGreater(T1,T2)
%   - lpoGreaterEqual(T1,T2)
%   - gndLpoGreater(T1,T2)
%   - gndLpoGreaterEqual(T1,T2)
%       Similar for the lexicographic path ordering.
% * Equality modulo AC:
%   - acEqual(T1,T2)
%       Succeeds if T1 and T2 are equal modulo AC.
% * Equality modulo C.
%   - cEqual(T1,T2)
%       Succeeds if T1 and T2 are equal modulo C.
% 
% For the Rubio/Nieuwenhuis ordering and acEqual, every operator f for
% which multisetStatus(f/_) succeeds is considered as an AC-operator.
% 
% For the recursive path ordering and cEqual, every operator f for
% which multisetStatus(f/_) succeeds is considered as a C-operator.
% 
% 
% Caveat 1:
% For a variable X, both rnGreater(X,'$min') and acEqual(X,'$min') fail,
% but rnGreaterEqual(X,'$min') succeeds. (Analogously for the other
% orderings.)
% 
% Caveat 2:
% In the presence of a minimal constant $min, axioms X >=rpo $min, and
% operators with multiset status, deciding the recursive path ordering
% is NP-complete. To keep the runtime polynomoial, this implementation
% of the RPO is incomplete: If the terms t1 and t2 contain variables and
% $min' below an operator with multiset status, then rpoGreater(t1,t2)
% may fail although t1 >rpo t2 is in fact provable. As the interpretation
% function that is used in the Rubio/Nieuwenhuis ordering introduces
% occurrences of '$min', the same problem may occur for rnGreater
% whenever the arguments contain variables below two multiset operators.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



:- module(aclpo, [ 
	rnGreater/2,
	rnGreaterEqual/2,
        gndRnGreater/2,
        gndRnGreaterEqual/2,
	rpoGreater/2,
	rpoGreaterEqual/2,
        gndRpoGreater/2,
        gndRpoGreaterEqual/2,
	lpoGreater/2,
	lpoGreaterEqual/2,
        gndLpoGreater/2,
        gndLpoGreaterEqual/2,
        cEqual/2,
	acEqual/2
   ]).


:- use_module(adt_precedence, [
	greater_op/2,
	is_first_predicate/1
   ]).

:- use_module(achelp, [
	ac_ist_AC/1
   ]).

:- use_module(adt_term, [
	var1/1
   ]).

:- use_module(props, [
	ac_ist_C/1
   ]).

:- use_module(library(basics), [member/2]).
:- use_module(library(samsort), [samsort/2]).
:- use_module(library(arg), [genarg/3]).
:- use_module(library(occurs), [contains_var/2]).
:- use_module(library(basics), [member/2]).

% FROM WHERE?
% should be the AC-operators for aclpo
% should be the C-operators for rpo

multisetStatus(O/2):-
	(termOrder:currentOrdering(aclpo) ->
	     ac_ist_AC(O)
	;
	     ac_ist_C(O)
	).



%--------- Interface Predicates -----------------------------------------

% C-Equality

cEqual(T1,T2) :-
  order(T1,T1Ordered),
  order(T2,T2Ordered),
  T1Ordered == T2Ordered.


% AC-Equality

acEqual(T1,T2) :-
  flattenAndOrder(T1,T1FlatOrdered),
  flattenAndOrder(T2,T2FlatOrdered),
  T1FlatOrdered == T2FlatOrdered.


% Rubio/Nieuwenhuis Ordering

rnGreater(T1,T2) :-
  flattenAndOrder(T1,T1FlatOrdered),
  flattenAndOrder(T2,T2FlatOrdered),
  rnGrEq(T1FlatOrdered,T2FlatOrdered,gt).


rnGreaterEqual(T1,T2) :-
  flattenAndOrder(T1,T1FlatOrdered),
  flattenAndOrder(T2,T2FlatOrdered),
  rnGrEq(T1FlatOrdered,T2FlatOrdered,_).


% Rubio/Nieuwenhuis Ordering - ground version

gndRnGreater(T1,T2) :-
  flattenAndOrder(T1,T1FlatOrdered),
  flattenAndOrder(T2,T2FlatOrdered),
  gndRnGrEq(T1FlatOrdered,T2FlatOrdered,gt).


gndRnGreaterEqual(T1,T2) :-
  flattenAndOrder(T1,T1FlatOrdered),
  flattenAndOrder(T2,T2FlatOrdered),
  gndRnGrEq(T1FlatOrdered,T2FlatOrdered,_).


% RPO

rpoGreater(T1,T2) :-
  order(T1,T1Ordered),
  order(T2,T2Ordered),
  rpoGrEq(T1Ordered,T2Ordered,gt).


rpoGreaterEqual(T1,T2) :-
  order(T1,T1Ordered),
  order(T2,T2Ordered),
  rpoGrEq(T1Ordered,T2Ordered,_).


% RPO - ground version

gndRpoGreater(T1,T2) :-
  order(T1,T1Ordered),
  order(T2,T2Ordered),
  gndRpoGrEq(T1Ordered,T2Ordered,gt).


gndRpoGreaterEqual(T1,T2) :-
  order(T1,T1Ordered),
  order(T2,T2Ordered),
  gndRpoGrEq(T1Ordered,T2Ordered,_).


% LPO

lpoGreater(T1,T2) :-
  lpoGrEq(T1,T2,gt).


lpoGreaterEqual(T1,T2) :-
  rpoGrEq(T1,T2,_).


% LPO - ground version

gndLpoGreater(T1,T2) :-
  gndLpoGrEq(T1,T2,gt).


gndLpoGreaterEqual(T1,T2) :-
  gndRpoGrEq(T1,T2,_).


%--------- Normalize Terms (flatten + order subterms w.r.t. @<) ---------

flattenAndOrder(T,T) :-
  var1(T),
  !.
flattenAndOrder(T,'$varyadic'(Func,TFlatOrderedList)) :-
  functor(T,Func,Arity),
  multisetStatus(Func/_),
  !,
  flattenAndOrderAC(Func,T,Arity,[],TFlatList),
  samsort(TFlatList,TFlatOrderedList).
flattenAndOrder(T,TFlat) :-
  functor(T,Func,Arity),
  functor(TFlat,Func,Arity),
  flattenAndOrderSubt(T,Arity,TFlat).


flattenAndOrderSubt(_T,0,_TFlat) :-
  !.
flattenAndOrderSubt(T,N,TFlat) :-
  arg(N,T,TN),
  N1 is N - 1,
  flattenAndOrder(TN,TNFlat),
  arg(N,TFlat,TNFlat),
  flattenAndOrderSubt(T,N1,TFlat).


flattenAndOrderAC(_Func,_T,0,TFlatList,TFlatList) :-
  !.
flattenAndOrderAC(Func,T,N,TFlatListRest,TFlatList) :-
  arg(N,T,TN),
  var1(TN),
  N1 is N - 1,
  !,
  flattenAndOrderAC(Func,T,N1,[TN|TFlatListRest],TFlatList).
flattenAndOrderAC(Func,T,N,TFlatListRest,TFlatList) :-
  arg(N,T,TN),
  functor(TN,Func,Arity),
  N1 is N - 1,
  !,
  flattenAndOrderAC(Func,TN,Arity,TFlatListRest,TFlatList1),
  flattenAndOrderAC(Func,T,N1,TFlatList1,TFlatList).
flattenAndOrderAC(Func,T,N,TFlatListRest,TFlatList) :-
  arg(N,T,TN),
  N1 is N - 1,
  flattenAndOrder(TN,TNFlat),
  flattenAndOrderAC(Func,T,N1,[TNFlat|TFlatListRest],TFlatList).


%--------- Normalize Terms (order subterms w.r.t. @<) -------------------

order(T,T) :-
  var1(T),
  !.
order(T,'$varyadic'(Func,TOrderedList)) :-
  functor(T,Func,Arity),
  multisetStatus(Func/_),
  !,
  functor(T1,Func,Arity),
  orderSubt(T,Arity,T1),
  T1 =.. [_|T1List],
  samsort(T1List,TOrderedList).
order(T,TOrdered) :-
  functor(T,Func,Arity),
  functor(TOrdered,Func,Arity),
  orderSubt(T,Arity,TOrdered).


orderSubt(_T,0,_TOrdered) :-
  !.
orderSubt(T,N,TOrdered) :-
  arg(N,T,TN),
  N1 is N - 1,
  order(TN,TNOrdered),
  arg(N,TOrdered,TNOrdered),
  orderSubt(T,N1,TOrdered).



%--------- Rubio/Nieuwenhuis Ordering -----------------------------------

rnGrEq(S,T,Rel) :-
  var1(S),
  !,
  Rel = geq,
  ( T == S
  ; T == '$min'
  ),
  !.
rnGrEq(S,T,Rel) :-
  S == T,
  !,
  Rel = geq.
rnGrEq(S,T,gt) :-
  var1(T),
  !,
  operator(S,F,_),
  ( is_first_predicate(F) ->
      true
  ; contains_var(T,S)
  ).
rnGrEq(S,T,Rel) :-
  operator(S,F,1),
  operator(T,F,1),
  !,
  arg(1,S,S1),
  arg(1,T,T1),
  rnGrEq(S1,T1,Rel).
rnGrEq(S,T,Rel) :-
  setof(X,interpr(S,X),SInterpr),
  setof(X,interpr(T,X),TInterpr),
  rnGrEq1(SInterpr,TInterpr,S,T,Rel).


rnGrEq1(_,[],_,_,gt) :-
  !.
rnGrEq1(SInterpr,[X|TInterpr],S,T,Rel) :-
  existsRpoGrEq(SInterpr,X,Rel1),
  !,
  ( Rel1 = gt ->
      rnGrEq1(SInterpr,TInterpr,S,T,Rel)
  ; rnGrEq2(SInterpr,TInterpr,S,T,Rel)
  ).


rnGrEq2(_,[],'$varyadic'(F,SList),'$varyadic'(F,TList),Rel) :-
  !,
  rnGrEqMs(SList,TList,Rel).
rnGrEq2(_,[],S,T,Rel) :-
  !,
  functor(S,_,N),
  functor(T,_,M),
  rnGrEqLex(0,S,N,T,M,Rel).
rnGrEq2(SInterpr,[X|TInterpr],S,T,Rel) :-
  existsRpoGrEq(SInterpr,X,_),
  !,
  rnGrEq2(SInterpr,TInterpr,S,T,Rel).


rnGrEqMs(K,L,Rel) :-
  mutualDifference(K,L,KMinusL,LMinusK),
  delVarsAndMin(KMinusL,LMinusK,KL1,LK1),
  rnGrEqMs1(KL1,[],LK1,Rel).


rnGrEqMs1([],[],[],Rel) :-
  !,
  Rel = geq.
rnGrEqMs1(_K1,_K2,[],gt) :-
  !.
rnGrEqMs1(K1,K2,[T|L],gt) :-
  rnGrEqMs3(K2,T,_,_),
  !,
  rnGrEqMs1(K1,K2,L,_Rel).
rnGrEqMs1(K1,K2,[T|L],Rel) :-
  rnGrEqMs2(K1,T,T1,K1MinusT1,Rel1),
  ( Rel1 = gt ->
      Rel = gt,
      rnGrEqMs1(K1MinusT1,[T1|K2],L,_Rel)
  ; rnGrEqMs1(K1MinusT1,K2,L,Rel)
  ).


rnGrEqMs2([T0|L],T,T1,Remainder,Rel) :-
  rnGrEq(T0,T,Rel1),
  !,
  ( Rel1 = gt ->
      Rel = gt,
      T1 = T0,
      Remainder = L
  ; ( rnGrEqMs3(L,T,T1,Remainder2) ->
        Rel = gt,
        Remainder = [T0|Remainder2]
    ; Rel = geq,
      T1 = T0,
      Remainder = L
    )
  ).
rnGrEqMs2([T0|L],T,T1,[T0|Remainder],Rel) :-
  rnGrEqMs2(L,T,T1,Remainder,Rel).


rnGrEqMs3([T0|L],T,T0,L) :-
  rnGrEq(T0,T,gt),
  !.
rnGrEqMs3([T0|L],T,T1,[T0|Remainder]) :-
  rnGrEqMs3(L,T,T1,Remainder).


rnGrEqLex(N,_,N,_,M,Rel) :-
  !,
  N = M,
  Rel = geq.
rnGrEqLex(M,_,_N,_,M,gt) :-
  !.
rnGrEqLex(I,S,N,T,M,Rel) :-
  I1 is I + 1,
  arg(I1,S,SI1),
  arg(I1,T,TI1),
  rnGrEq(SI1,TI1,Rel1),
  ( Rel1 = geq ->
      rnGrEqLex(I1,S,N,T,M,Rel)
  ; Rel = gt
  ).


%--------- Interpretation Function (non-deterministic!) -----------------

interpr(T,T) :-
  var1(T),
  !.
interpr('$varyadic'(Func,L),'$varyadic'(Func,LInterprOrdered)) :-
  !,
  interprAC(Func,L,LInterpr),
  samsort(LInterpr,LInterprOrdered).
interpr(T,TInterpr) :-
  functor(T,Func,Arity),
  functor(TInterpr,Func,Arity),
  interprSubt(T,Arity,TInterpr).


interprSubt(_T,0,_TInterpr) :-
  !.
interprSubt(T,N,TInterpr) :-
  arg(N,T,TN),
  N1 is N - 1,
  interpr(TN,TNInterpr),
  arg(N,TInterpr,TNInterpr),
  interprSubt(T,N1,TInterpr).


interprAC(_Func,[],[]) :-
  !.
interprAC(Func,[T|L],[T|LInterpr]) :-
  var1(T),
  !,
  interprAC(Func,L,LInterpr).
interprAC(Func,[T|L],TLInterpr) :-
  interpr(T,TInterpr),
  setof(X,maxSubtBelow(Func,TInterpr,X),S),
  member(TMax,S),
  ( \+ var1(TMax),
    TMax = '$varyadic'(Func,Args) ->
      interprAC(Func,L,LInterpr),
      append(Args,LInterpr,TLInterpr)
  ; TLInterpr = [TMax|LInterpr],
    interprAC(Func,L,LInterpr)
  ).


maxSubtBelow(_Func,T,T) :-
  var1(T),
  !.
maxSubtBelow(Func,T,TMax) :-
  operator(T,G,N),
  greater_op(Func,G),
  !,
  ( N = 0 ->
      TMax = '$min'
  ; maxSubt(T,TM),
    maxSubtBelow(Func,TM,TMax)
  ).
maxSubtBelow(_Func,T,T).


maxSubt('$varyadic'(_,L),TN) :-
  !,
  delDuplicates(L,L1),
  maxElem(L1,TN).
maxSubt(T,TN) :-
  genarg(N,T,TN),
  \+ ( genarg(M,T,TM),
       ( M < N,
         rpoGrEq(TM,TN,gt)
       ; M > N,
         rpoGrEq(TM,TN,_)
       )
     ).


maxElem(L,T) :-
  member(T,L),
  \+ ( member(T1,L),
       rpoGrEq(T1,T,gt)
     ).


%--------- RPO ----------------------------------------------------------

rpoGrEq(S,T,Rel) :-
  var1(S),
  !,
  Rel = geq,
  ( T == S
  ; T == '$min'
  ),
  !.
rpoGrEq(S,T,gt) :-
  var1(T),
  !,
  operator(S,F,_),
  ( is_first_predicate(F) ->
      true
  ; contains_var(T,S)
  ).
rpoGrEq(S,T,Rel) :-
  operator(S,F,N),
  operator(T,G,M),
  rpoGrEqSubt(F,G,N,M,S,T,Rel).


rpoGrEqSubt(F,F,_N,_M,'$varyadic'(_F,SList),'$varyadic'(_F,TList),Rel) :-
  !,
  rpoGrEqMs(SList,TList,Rel).
rpoGrEqSubt(F,F,N,M,S,T,Rel) :-
  !,
  ( is_first_predicate(F) ->
      rpoGrEqLexFP(0,S,N,T,M,Rel)
  ; rpoGrEqLex(0,S,N,T,M,Rel)
  ).
rpoGrEqSubt(F,G,_N,M,S,T,gt) :-
  ( greater_op(F,G) ->
      ( T = '$varyadic'(_G,TList) ->
          rpoGreaterList(S,TList)
      ; rpoGreaterArgs(S,0,T,M)
      )
  ; subterm(S,SK),
    rpoGrEq(SK,T,_)
  ),
  !.


rpoGrEqMs(K,L,Rel) :-
  mutualDifference(K,L,KMinusL,LMinusK),
  delVarsAndMin(KMinusL,LMinusK,KL1,LK1),
  rpoGrEqMs1(KL1,[],LK1,Rel).


rpoGrEqMs1([],[],[],Rel) :-
  !,
  Rel = geq.
rpoGrEqMs1(_K1,_K2,[],gt) :-
  !.
rpoGrEqMs1(K1,K2,[T|L],gt) :-
  rpoGrEqMs3(K2,T,_,_),
  !,
  rpoGrEqMs1(K1,K2,L,_Rel).
rpoGrEqMs1(K1,K2,[T|L],Rel) :-
  rpoGrEqMs2(K1,T,T1,K1MinusT1,Rel1),
  ( Rel1 = gt ->
      Rel = gt,
      rpoGrEqMs1(K1MinusT1,[T1|K2],L,_Rel)
  ; rpoGrEqMs1(K1MinusT1,K2,L,Rel)
  ).


rpoGrEqMs2([T0|L],T,T1,Remainder,Rel) :-
  rpoGrEq(T0,T,Rel1),
  !,
  ( Rel1 = gt ->
      Rel = gt,
      T1 = T0,
      Remainder = L
  ; ( rpoGrEqMs3(L,T,T1,Remainder2) ->
        Rel = gt,
        Remainder = [T0|Remainder2]
    ; Rel = geq,
      T1 = T0,
      Remainder = L
    )
  ).
rpoGrEqMs2([T0|L],T,T1,[T0|Remainder],Rel) :-
  rpoGrEqMs2(L,T,T1,Remainder,Rel).


rpoGrEqMs3([T0|L],T,T0,L) :-
  rpoGrEq(T0,T,gt),
  !.
rpoGrEqMs3([T0|L],T,T1,[T0|Remainder]) :-
  rpoGrEqMs3(L,T,T1,Remainder).


rpoGrEqLexFP(N,_,N,_,M,Rel) :-
  !,
  N = M,
  Rel = geq.
rpoGrEqLexFP(M,_,_N,_,M,gt) :-
  !.
rpoGrEqLexFP(I,S,N,T,M,Rel) :-
  I1 is I + 1,
  arg(I1,S,SI1),
  arg(I1,T,TI1),
  rpoGrEq(SI1,TI1,Rel1),
  ( Rel1 = geq ->
      rpoGrEqLexFP(I1,S,N,T,M,Rel)
  ; Rel = gt
  ).


rpoGrEqLex(N,_,N,_,M,Rel) :-
  !,
  N = M,
  Rel = geq.
rpoGrEqLex(M,_,_N,_,M,gt) :-
  !.
rpoGrEqLex(I,S,N,T,M,Rel) :-
  I1 is I + 1,
  arg(I1,S,SI1),
  arg(I1,T,TI1),
  ( rpoGrEq(SI1,TI1,Rel1) ->
      ( Rel1 = geq ->
          rpoGrEqLex(I1,S,N,T,M,Rel)
      ; Rel = gt,
        rpoGreaterArgs(S,I1,T,M)
      )
  ; Rel = gt,
    genarg(K,S,SK),
    K > I1,
    rpoGrEq(SK,T,_),
    !
  ).


rpoGreaterList(_,[]) :-
  !.
rpoGreaterList(S,[T|L]) :-
  rpoGrEq(S,T,gt),
  rpoGreaterList(S,L).


rpoGreaterArgs(_,N,_,N) :-
  !.
rpoGreaterArgs(S,I,T,M) :-
  I1 is I + 1,
  arg(I1,T,TI1),
  rpoGrEq(S,TI1,gt),
  rpoGreaterArgs(S,I1,T,M).


existsRpoGrEq([Y|L],X,Rel) :-
  rpoGrEq(Y,X,Rel1),
  !,
  ( Rel1 = gt ->
      Rel = gt
  ; existsRpoGrEq1(L,X,Rel)
  ).
existsRpoGrEq([_|L],X,Rel) :-
  existsRpoGrEq(L,X,Rel).


existsRpoGrEq1([],_,geq).
existsRpoGrEq1([Y|_],X,gt) :-
  rpoGrEq(Y,X,gt),
  !.
existsRpoGrEq1([_|L],X,Rel) :-
  existsRpoGrEq1(L,X,Rel).


%--------- LPO ----------------------------------------------------------

lpoGrEq(S,T,Rel) :-
  var1(S),
  !,
  Rel = geq,
  ( T == S
  ; T == '$min'
  ),
  !.
lpoGrEq(S,T,gt) :-
  var1(T),
  !,
  functor(S,F,_),
  ( is_first_predicate(F) ->
      true
  ; contains_var(T,S)
  ).
lpoGrEq(S,T,Rel) :-
  functor(S,F,N),
  functor(T,G,M),
  lpoGrEqSubt(F,G,N,M,S,T,Rel).


lpoGrEqSubt(F,F,N,M,S,T,Rel) :-
  !,
  ( is_first_predicate(F) ->
      lpoGrEqLexFP(0,S,N,T,M,Rel)
  ; lpoGrEqLex(0,S,N,T,M,Rel)
  ).
lpoGrEqSubt(F,G,_N,M,S,T,gt) :-
  ( greater_op(F,G) ->
      lpoGreaterArgs(S,0,T,M)
  ; genarg(_,S,SK),
    lpoGrEq(SK,T,_)
  ),
  !.


lpoGrEqLexFP(N,_,N,_,M,Rel) :-
  !,
  N = M,
  Rel = geq.
lpoGrEqLexFP(M,_,_N,_,M,gt) :-
  !.
lpoGrEqLexFP(I,S,N,T,M,Rel) :-
  I1 is I + 1,
  arg(I1,S,SI1),
  arg(I1,T,TI1),
  lpoGrEq(SI1,TI1,Rel1),
  ( Rel1 = geq ->
      lpoGrEqLexFP(I1,S,N,T,M,Rel)
  ; Rel = gt
  ).


lpoGrEqLex(N,_,N,_,M,Rel) :-
  !,
  N = M,
  Rel = geq.
lpoGrEqLex(M,_,_N,_,M,gt) :-
  !.
lpoGrEqLex(I,S,N,T,M,Rel) :-
  I1 is I + 1,
  arg(I1,S,SI1),
  arg(I1,T,TI1),
  ( lpoGrEq(SI1,TI1,Rel1) ->
      ( Rel1 = geq ->
          lpoGrEqLex(I1,S,N,T,M,Rel)
      ; Rel = gt,
        lpoGreaterArgs(S,I1,T,M)
      )
  ; Rel = gt,
    genarg(K,S,SK),
    K > I1,
    lpoGrEq(SK,T,_),
    !
  ).


lpoGreaterArgs(_,N,_,N) :-
  !.
lpoGreaterArgs(S,I,T,M) :-
  I1 is I + 1,
  arg(I1,T,TI1),
  lpoGrEq(S,TI1,gt),
  lpoGreaterArgs(S,I1,T,M).


%--------- Rubio/Nieuwenhuis Ordering - ground --------------------------

gndRnGrEq(S,T,Rel) :-
  S == T,
  !,
  Rel = geq.
gndRnGrEq(S,T,Rel) :-
  operator(S,F,1),
  operator(T,F,1),
  !,
  arg(1,S,S1),
  arg(1,T,T1),
  gndRnGrEq(S1,T1,Rel).
gndRnGrEq(S,T,Rel) :-
  setof(X,gndInterpr(S,X),SInterpr),
  setof(X,gndInterpr(T,X),TInterpr),
  gndRnGrEq1(SInterpr,TInterpr,S,T,Rel).


gndRnGrEq1(_,[],_,_,gt) :-
  !.
gndRnGrEq1(SInterpr,[X|TInterpr],S,T,Rel) :-
  gndExistsRpoGrEq(SInterpr,X,Rel1),
  !,
  ( Rel1 = gt ->
      gndRnGrEq1(SInterpr,TInterpr,S,T,Rel)
  ; gndRnGrEq2(SInterpr,TInterpr,S,T,Rel)
  ).


gndRnGrEq2(_,[],'$varyadic'(F,SList),'$varyadic'(F,TList),Rel) :-
  !,
  gndRnGrEqMs(SList,TList,Rel).
gndRnGrEq2(_,[],S,T,Rel) :-
  !,
  operator(S,_,N),
  operator(T,_,M),
  gndRnGrEqLex(0,S,N,T,M,Rel).
gndRnGrEq2(SInterpr,[X|TInterpr],S,T,Rel) :-
  gndExistsRpoGrEq(SInterpr,X,_),
  !,
  gndRnGrEq2(SInterpr,TInterpr,S,T,Rel).


gndRnGrEqMs(K,L,Rel) :-
  mutualDifference(K,L,KMinusL,LMinusK),
  gndRnGrEqMs1(KMinusL,[],LMinusK,Rel).


gndRnGrEqMs1([],[],[],Rel) :-
  !,
  Rel = geq.
gndRnGrEqMs1(_K1,_K2,[],gt) :-
  !.
gndRnGrEqMs1(K1,K2,[T|L],gt) :-
  gndRnGrEqMs3(K2,T,_,_),
  !,
  gndRnGrEqMs1(K1,K2,L,_Rel).
gndRnGrEqMs1(K1,K2,[T|L],Rel) :-
  gndRnGrEqMs2(K1,T,T1,K1MinusT1,Rel1),
  ( Rel1 = gt ->
      Rel = gt,
      gndRnGrEqMs1(K1MinusT1,[T1|K2],L,_Rel)
  ; gndRnGrEqMs1(K1MinusT1,K2,L,Rel)
  ).


gndRnGrEqMs2([T0|L],T,T1,Remainder,Rel) :-
  gndRnGrEq(T0,T,Rel1),
  !,
  ( Rel1 = gt ->
      Rel = gt,
      T1 = T0,
      Remainder = L
  ; ( gndRnGrEqMs3(L,T,T1,Remainder2) ->
        Rel = gt,
        Remainder = [T0|Remainder2]
    ; Rel = geq,
      T1 = T0,
      Remainder = L
    )
  ).
gndRnGrEqMs2([T0|L],T,T1,[T0|Remainder],Rel) :-
  gndRnGrEqMs2(L,T,T1,Remainder,Rel).


gndRnGrEqMs3([T0|L],T,T0,L) :-
  gndRnGrEq(T0,T,gt),
  !.
gndRnGrEqMs3([T0|L],T,T1,[T0|Remainder]) :-
  gndRnGrEqMs3(L,T,T1,Remainder).


gndRnGrEqLex(N,_,N,_,M,Rel) :-
  !,
  N = M,
  Rel = geq.
gndRnGrEqLex(M,_,_N,_,M,gt) :-
  !.
gndRnGrEqLex(I,S,N,T,M,Rel) :-
  I1 is I + 1,
  arg(I1,S,SI1),
  arg(I1,T,TI1),
  gndRnGrEq(SI1,TI1,Rel1),
  ( Rel1 = geq ->
      gndRnGrEqLex(I1,S,N,T,M,Rel)
  ; Rel = gt
  ).


%--------- Interpretation Function (non-deterministic!) - ground --------

gndInterpr(T,T) :-
  var1(T),
  !.
gndInterpr('$varyadic'(Func,L),'$varyadic'(Func,LInterprOrdered)) :-
  !,
  gndInterprAC(Func,L,LInterpr),
  samsort(LInterpr,LInterprOrdered).
gndInterpr(T,TInterpr) :-
  functor(T,Func,Arity),
  functor(TInterpr,Func,Arity),
  gndInterprSubt(T,Arity,TInterpr).


gndInterprSubt(_T,0,_TInterpr) :-
  !.
gndInterprSubt(T,N,TInterpr) :-
  arg(N,T,TN),
  N1 is N - 1,
  gndInterpr(TN,TNInterpr),
  arg(N,TInterpr,TNInterpr),
  gndInterprSubt(T,N1,TInterpr).


gndInterprAC(_Func,[],[]) :-
  !.
gndInterprAC(Func,[T|L],[T|LInterpr]) :-
  var1(T),
  !,
  gndInterprAC(Func,L,LInterpr).
gndInterprAC(Func,[T|L],TLInterpr) :-
  gndInterpr(T,TInterpr),
  setof(X,gndMaxSubtBelow(Func,TInterpr,X),S),
  member(TMax,S),
  ( \+ var1(TMax),
    TMax = '$varyadic'(Func,Args) ->
      gndInterprAC(Func,L,LInterpr),
      append(Args,LInterpr,TLInterpr)
  ; TLInterpr = [TMax|LInterpr],
    gndInterprAC(Func,L,LInterpr)
  ).


gndMaxSubtBelow(_Func,T,T) :-
  var1(T),
  !.
gndMaxSubtBelow(Func,T,TMax) :-
  operator(T,G,N),
  gndGreaterOp(Func,G),
  !,
  ( N = 0 ->
      TMax = '$min'
  ; gndMaxSubt(T,TM),
    gndMaxSubtBelow(Func,TM,TMax)
  ).
gndMaxSubtBelow(_Func,T,T).


gndMaxSubt('$varyadic'(_,L),TN) :-
  !,
  delDuplicates(L,L1),
  gndMaxElem(L1,TN).
gndMaxSubt(T,TN) :-
  genarg(N,T,TN),
  \+ ( genarg(M,T,TM),
       ( M < N,
         gndRpoGrEq(TM,TN,gt)
       ; M > N,
         gndRpoGrEq(TM,TN,_)
       )
     ).


gndMaxElem(L,T) :-
  member(T,L),
  \+ ( member(T1,L),
       gndRpoGrEq(T1,T,gt)
     ).


%--------- RPO - ground -------------------------------------------------

gndRpoGrEq(S,T,Rel) :-
  operator(S,F,N),
  operator(T,G,M),
  gndRpoGrEqSubt(F,G,N,M,S,T,Rel).


gndRpoGrEqSubt(F,G,_N,_M,S,T,Rel) :-
  F == G,
  \+ var1(F),
  S = '$varyadic'(_F,SList),
  T = '$varyadic'(_F,TList),
  !,
  gndRpoGrEqMs(SList,TList,Rel).
gndRpoGrEqSubt(F,G,N,M,S,T,Rel) :-
  F == G,
  !,
  ( gndIsFirstPredicate(F) ->
      gndRpoGrEqLexFP(0,S,N,T,M,Rel)
  ; gndRpoGrEqLex(0,S,N,T,M,Rel)
  ).
gndRpoGrEqSubt(F,G,_N,M,S,T,gt) :-
  ( gndGreaterOp(F,G) ->
      ( \+ var1(T),
        T = '$varyadic'(_G,TList) ->
          gndRpoGreaterList(S,TList)
      ; gndRpoGreaterArgs(S,0,T,M)
      )
  ; \+ var1(S),
    subterm(S,SK),
    gndRpoGrEq(SK,T,_)
  ),
  !.


gndRpoGrEqMs(K,L,Rel) :-
  mutualDifference(K,L,KMinusL,LMinusK),
  gndRpoGrEqMs1(KMinusL,[],LMinusK,Rel).


gndRpoGrEqMs1([],[],[],Rel) :-
  !,
  Rel = geq.
gndRpoGrEqMs1(_K1,_K2,[],gt) :-
  !.
gndRpoGrEqMs1(K1,K2,[T|L],gt) :-
  gndRpoGrEqMs3(K2,T,_,_),
  !,
  gndRpoGrEqMs1(K1,K2,L,_Rel).
gndRpoGrEqMs1(K1,K2,[T|L],Rel) :-
  gndRpoGrEqMs2(K1,T,T1,K1MinusT1,Rel1),
  ( Rel1 = gt ->
      Rel = gt,
      gndRpoGrEqMs1(K1MinusT1,[T1|K2],L,_Rel)
  ; gndRpoGrEqMs1(K1MinusT1,K2,L,Rel)
  ).


gndRpoGrEqMs2([T0|L],T,T1,Remainder,Rel) :-
  gndRpoGrEq(T0,T,Rel1),
  !,
  ( Rel1 = gt ->
      Rel = gt,
      T1 = T0,
      Remainder = L
  ; ( gndRpoGrEqMs3(L,T,T1,Remainder2) ->
        Rel = gt,
        Remainder = [T0|Remainder2]
    ; Rel = geq,
      T1 = T0,
      Remainder = L
    )
  ).
gndRpoGrEqMs2([T0|L],T,T1,[T0|Remainder],Rel) :-
  gndRpoGrEqMs2(L,T,T1,Remainder,Rel).


gndRpoGrEqMs3([T0|L],T,T0,L) :-
  gndRpoGrEq(T0,T,gt),
  !.
gndRpoGrEqMs3([T0|L],T,T1,[T0|Remainder]) :-
  gndRpoGrEqMs3(L,T,T1,Remainder).


gndRpoGrEqLexFP(N,_,N,_,M,Rel) :-
  !,
  N = M,
  Rel = geq.
gndRpoGrEqLexFP(M,_,_N,_,M,gt) :-
  !.
gndRpoGrEqLexFP(I,S,N,T,M,Rel) :-
  I1 is I + 1,
  arg(I1,S,SI1),
  arg(I1,T,TI1),
  gndRpoGrEq(SI1,TI1,Rel1),
  ( Rel1 = geq ->
      gndRpoGrEqLexFP(I1,S,N,T,M,Rel)
  ; Rel = gt
  ).


gndRpoGrEqLex(N,_,N,_,M,Rel) :-
  !,
  N = M,
  Rel = geq.
gndRpoGrEqLex(M,_,_N,_,M,gt) :-
  !.
gndRpoGrEqLex(I,S,N,T,M,Rel) :-
  I1 is I + 1,
  arg(I1,S,SI1),
  arg(I1,T,TI1),
  ( gndRpoGrEq(SI1,TI1,Rel1) ->
      ( Rel1 = geq ->
          gndRpoGrEqLex(I1,S,N,T,M,Rel)
      ; Rel = gt,
        gndRpoGreaterArgs(S,I1,T,M)
      )
  ; Rel = gt,
    genarg(K,S,SK),
    K > I1,
    gndRpoGrEq(SK,T,_),
    !
  ).


gndRpoGreaterList(_,[]) :-
  !.
gndRpoGreaterList(S,[T|L]) :-
  gndRpoGrEq(S,T,gt),
  gndRpoGreaterList(S,L).


gndRpoGreaterArgs(_,N,_,N) :-
  !.
gndRpoGreaterArgs(S,I,T,M) :-
  I1 is I + 1,
  arg(I1,T,TI1),
  gndRpoGrEq(S,TI1,gt),
  gndRpoGreaterArgs(S,I1,T,M).


gndExistsRpoGrEq([Y|L],X,Rel) :-
  gndRpoGrEq(Y,X,Rel1),
  !,
  ( Rel1 = gt ->
      Rel = gt
  ; gndExistsRpoGrEq1(L,X,Rel)
  ).
gndExistsRpoGrEq([_|L],X,Rel) :-
  gndExistsRpoGrEq(L,X,Rel).


gndExistsRpoGrEq1([],_,geq).
gndExistsRpoGrEq1([Y|_],X,gt) :-
  gndRpoGrEq(Y,X,gt),
  !.
gndExistsRpoGrEq1([_|L],X,Rel) :-
  gndExistsRpoGrEq1(L,X,Rel).


%--------- LPO - ground -------------------------------------------------

gndLpoGrEq(S,T,Rel) :-
  operator(S,F,N),
  operator(T,G,M),
  gndLpoGrEqSubt(F,G,N,M,S,T,Rel).


gndLpoGrEqSubt(F,G,N,M,S,T,Rel) :-
  F == G,
  !,
  ( gndIsFirstPredicate(F) ->
      gndLpoGrEqLexFP(0,S,N,T,M,Rel)
  ; gndLpoGrEqLex(0,S,N,T,M,Rel)
  ).
gndLpoGrEqSubt(F,G,_N,M,S,T,gt) :-
  ( gndGreaterOp(F,G) ->
      gndLpoGreaterArgs(S,0,T,M)
  ; \+ var1(S),
    genarg(_,S,SK),
    gndLpoGrEq(SK,T,_)
  ),
  !.


gndLpoGrEqLexFP(N,_,N,_,M,Rel) :-
  !,
  N = M,
  Rel = geq.
gndLpoGrEqLexFP(M,_,_N,_,M,gt) :-
  !.
gndLpoGrEqLexFP(I,S,N,T,M,Rel) :-
  I1 is I + 1,
  arg(I1,S,SI1),
  arg(I1,T,TI1),
  gndLpoGrEq(SI1,TI1,Rel1),
  ( Rel1 = geq ->
      gndLpoGrEqLexFP(I1,S,N,T,M,Rel)
  ; Rel = gt
  ).


gndLpoGrEqLex(N,_,N,_,M,Rel) :-
  !,
  N = M,
  Rel = geq.
gndLpoGrEqLex(M,_,_N,_,M,gt) :-
  !.
gndLpoGrEqLex(I,S,N,T,M,Rel) :-
  I1 is I + 1,
  arg(I1,S,SI1),
  arg(I1,T,TI1),
  ( gndLpoGrEq(SI1,TI1,Rel1) ->
      ( Rel1 = geq ->
          gndLpoGrEqLex(I1,S,N,T,M,Rel)
      ; Rel = gt,
        gndLpoGreaterArgs(S,I1,T,M)
      )
  ; Rel = gt,
    genarg(K,S,SK),
    K > I1,
    gndLpoGrEq(SK,T,_),
    !
  ).


gndLpoGreaterArgs(_,N,_,N) :-
  !.
gndLpoGreaterArgs(S,I,T,M) :-
  I1 is I + 1,
  arg(I1,T,TI1),
  gndLpoGrEq(S,TI1,gt),
  gndLpoGreaterArgs(S,I1,T,M).


%--------- Auxiliary Functions ------------------------------------------

mutualDifference([],V,[],V) :-
  !.
mutualDifference(U,[],U,[]) :-
  !.
mutualDifference([X|U],[Y|V],UMinusV,VMinusU) :-
  X == Y,
  !,
  mutualDifference(U,V,UMinusV,VMinusU).
mutualDifference([X|U],[Y|V],[X|UMinusYV],YVMinusU) :-
  X @< Y,
  !,
  mutualDifference(U,[Y|V],UMinusYV,YVMinusU).
mutualDifference([X|U],[Y|V],XUMinusV,[Y|VMinusXU]) :-
  mutualDifference([X|U],V,XUMinusV,VMinusXU).


delVarsAndMin([X|U],[Y|V],L1,[Y|L2]) :-
  var1(X),
  Y @< '$min',
  !,
  delVarsAndMin([X|U],V,L1,L2).
delVarsAndMin([X|U],[Y|V],L1,L2) :-
  var1(X),
  Y == '$min',
  !,
  delVarsAndMin(U,V,L1,L2).
delVarsAndMin(U,V,U,V).


delDuplicates([X,Y|R],L) :-
  X == Y,
  !,
  delDuplicates([Y|R],L).
delDuplicates([X|R],[X|L]) :-
  !,
  delDuplicates(R,L).
delDuplicates([],[]).


%--------- Generalizations of functor, genarg, var, ... -----------------

operator(F,Op,Arity) :-
  var1(F),
  !,
  Op = F,
  Arity = 0.
operator('$varyadic'(F,_),Op,Arity) :-
  !,
  Op = F,
  Arity = '$varyadic'.
operator(T,F,N) :-
  functor(T,F,N).


subterm('$varyadic'(_,L),S) :-
  !,
  member(S,L).
subterm(T,S) :-
  genarg(_,T,S).



gndIsFirstPredicate(F) :-
  \+ var1(F),
  is_first_predicate(F).


gndGreaterOp(F,G) :-
  ( var1(F) ->
      ( var1(G) ->
          F @> G
      ; \+ is_first_predicate(G)
      )
  ; var1(G) ->
      is_first_predicate(F)
  ; greater_op(F,G)
  ).


