%declarations%

:- module(adt_proof, [ 
	create_history/2,
	depth/2,
	get_first_history/2,
	get_second_history/2,
	get_key_history/2,
	first_premise/2,
	initial_history/2,
	recover_proof/0,
	recover_proof_for/1,
	second_premise/2,
	parents/2,
	proof_depth/2,
	proof_file_entry/2,
	s_depth/2,
	theoryIndicator/2
   ]).

%:- use_module(user,_,[append/3]).

:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(maplist), [
	maplist/3
   ]).

:- use_module(mylists, [
	is_list/1
   ]).

:- use_module(classes, [
        constraint/3,
	history/2,
	is_object/2,
	number/2,
	number/3,
	empty/1
   ]).

:- use_module(adt_declarations, [
	binary_inference/1,
	hi/3,
	simplification_inference/1,
	unary_inference/1
   ]).

:- use_module(auxiliary, [
	delete_repeats/2
   ]).

:- use_module(proofState, [
	proof_file/2,
	rule/2
   ]).

:- use_module(output, [
	print_terms/1
   ]).

:- use_module(orderConstraints, [
	true/1
   ]).



%end%



%							ADT_PROOF

%------------------------------------------------------------------------------
% history ::- [ key, premiseNumbers, Theory, proofDepth, searchDepth ]
%
%  no ... don't store the compete info in the premises as this will
%  result in a very big structure: each clause will have all its
%  predecessors in its proof and each unification will have to
%  instantiate a possibly long proof . Only store the ancestor literals.
%  Need to be done sometime.
% premiseNumbers is either
%  - a one-element or two-element list of (premise) clause numbers, or
%  - a two-element list [N,NL], consisting of a clause number and a 
%    list NL of clause
%    numbers. 
% The latter format occurs when a clause is obtained by reduction from
% clause N with the reducers in NL.

% Theory = [theory|goal] indicates as to whether or not the clause depends 
%           solely on theory clauses.
%------------------------------------------------------------------------------

initial_history( T, [init,[],T,1,0] ).

create_history( [Key,NL,Premises], [Key,NL,T,D,S] ):-
	maxDepth(Premises,NL,D1,T),
	D is D1+1,
	maxSearchDepth(Premises,S1),
	search_inference(Key,Inc),
	S is S1+Inc.





get_first_history( [_,[N,_]|_], N ).
get_first_history( [_,[N]|_], N ).
get_first_history( [_,[]|_],[] ).
get_second_history( [_,[_,N]|_], N ) :- !.
get_second_history( _, [] ).
get_key_history( [Key|_], Key ).

first_premise([_,_,[C|_]|_],C).

second_premise([_,_,[_,C]|_],C).

depth([_,_,_,D|_],D).

s_depth([_,_,_,_,D|_],D).

theoryIndicator([_,_,T|_],T).



search_inference(Key,1):-
	binary_inference(Key),
	!.
search_inference(_Key,0).




maxSearchDepth([],0).
maxSearchDepth([Cl],D):-
	history(Cl,H1),
	s_depth(H1,D).
maxSearchDepth([Cl1,Cl2],D):-
	history(Cl1,H1),
	s_depth(H1,D1),

	(\+is_object(Cl2,_) ->
	     D2=0
	;
	     history(Cl2,H2),
	     s_depth(H2,D2)
	),
	D is max(D1,D2).



maxDepth([],_,0,theory).
maxDepth([Cl],[_|NL],D,T):-
	history(Cl,H1),
	depth(H1,D1),
	theoryIndicator(H1,T1),
	maxDepth(NL,D2,T2),
	D is max(D1,D2),
	maxTh(T1,T2,T).
maxDepth([Cl1,Cl2],_,D,T):-
	history(Cl1,H1),
	depth(H1,D1),
	theoryIndicator(H1,T1),

	(\+is_object(Cl2,_) ->
	     D2=0,
	     T2=theory
	;
	     history(Cl2,H2),
	     depth(H2,D2),
	     theoryIndicator(H2,T2)
	),
	D is max(D1,D2),
	maxTh(T1,T2,T).



maxDepth([],0,theory).
maxDepth([NL],D,T):-
	is_list(NL),
	!,
	maxDepth(NL,D,T).
maxDepth([N|NL],D,T):-
	maxDepth(NL,D1,T1),
	proof_depth(N,D2),
	theoryIndicatorN(N,T2),
	D is max(D1,D2),
	maxTh(T1,T2,T).



%------------------------------------------------------------------------------
%	recover the proof for a given clause number
%------------------------------------------------------------------------------

recover_proof_for( N ):-
	get_proof_clauses( N, P ),
        display_proof( P ),
	length(P,PL),
	proof_depth(N,D),
	search_depth(N,SD),
	format('~nLength = ~d, Depth = ~d, Search Depth = ~d~n',[PL,D,SD]),
	!.        

%------------------------------------------------------------------------------
%	recover the proof (for the empty clause)
%------------------------------------------------------------------------------

recover_proof:-
	rule( N, Ru ),
	empty( Ru ),
	format('~2nProof:~n======~n',[]),
	recover_proof_for( N ),!.

get_proof_clauses( N, P2 ):- 
	rec_p( N, P ),
%        delete_repeats( P, P1 ),
        sort( P, P2 ), !.

display_proof( [] ):- nl,!.
display_proof( [N|R] ):-
	(N=0 ->
	     true
	;
             proof_file_entry(Cl,N),
	     display_code( Cl ),
	     display_proof( R )
	).
%------------------------------------------------------------------------------
% the hi predicate is defined in module adt_declarations
%------------------------------------------------------------------------------
proof_file_entry(Cl,N):-
	proof_file(N, Cl ),
	!.

display_code( Cl ):-
	history( Cl, H ),
	get_first_history( H, N1 ),
	get_second_history( H, N2 ),
	get_key_history( H, Key ),
%	true(Ct),
%	constraint(Ct,Cl,Cl2),
	print(proofClause(Cl)),
	hi( Key, _, Text ),  
	parents(N1,N2,P),
%	on_exception(consistency_error(_,_,_,_),format(Text, [N1,N2]),true). % until 3#5
	on_exception(Ex,format(Text, P),
		     (Ex=consistency_error(_,_,_,_);Ex=domain_error(_,_,_,_))).
display_code( _ ):- 
	raise_exception(saturate_error("~nerror in display_code.",[])).


parents([],[],[]).
parents(N,[],[N]).
parents([],N,[N]).
parents(N1,N2,[N1,N2]).


proved_from(N,M):-
	inferred_from(N,M).
proved_from(N,M):-
	inferred_from(N,M1),
	proved_from(M1,M).

inferred_from(N,M):-
	proof_file(N,Cl),
	history([N|Cl],H),
	get_first_history(H,N1),
	get_second_history(H,N2),
	get_key_history(H,Key),
	(binary_inference(Key) -> (M=N1;M=N2)
	; unary_inference(Key) -> M=N1
	; simplification_inference(Key) -> M=N1).
	




print_info( NL, [nl|Text]):-
	nl, print_info( NL, Text ), !.
print_info( [N|L], [arg|Text]):-
	print_terms( [N] ),
	print_info( L, Text ), !. 
print_info( NL, [Text|R]):-
	print_terms( [Text] ),
	print_info( NL, R ), !.
print_info( _, []):- !.
	 
rec_p( N, P):-
	rec_p(N,[],_,P).


rec_p( N, V, V, [] ):-
        var(N),!.
rec_p( 0, V, V, [] ).
rec_p( 'L', V, V, [] ).
rec_p( [], V, V, [] ):- !.
rec_p( [N|R], Visited,V,P ):-
	rec_p( N, Visited, V1, P1 ),
	rec_p( R, [N|V1], V, P2 ),
	append( P1, P2, P ), !.
rec_p( N, Visited, Visited, [] ):-
	member(N,Visited),
	!.
rec_p( N, Visited, V, [N|P] ):-
	proof_file_entry(Cl,N),
	history( Cl, F ),
	get_first_history( F, N1 ),
	get_second_history( F, N2 ),
        rec_p( N1, [N|Visited], V1, P1 ),
        (is_list(N1) ->
	     append(V1,[N|N1],V2)
	;
	     V2=[N|V1]
	),
	rec_p( N2, V2, V, P2 ),
        append( P1, P2, P ),!.
rec_p( N, _Visited, _V, _ ):-
	raise_exception(saturate_error('~nrec_p: don''t know the proof of ~p',[N])).


parents(C,Cs):-
	history(C,H),
	get_first_history( H, N1 ),
	get_second_history( H, N2 ),
	(is_list(N1) ->
	     (is_list(N2) ->
		  append(N1,N2,N)
	     ;
                  N=[N2|N1]
	     )
	;is_list(N2) ->
	     N=[N1|N2]
	;
	     N=[N2,N1]
	),
	maplist(proof_file_entry,Cs,N).


	

proof_depth(0,1):-
	!.
proof_depth(N,D):-
	number(N),
	proof_file_entry(Cl,N),
	!,
	history(Cl,H),
	depth(H,D).
proof_depth(_,0).


search_depth(0,0):-
	!.
search_depth(N,D):-
	number(N),
	proof_file_entry(Cl,N),
	!,
	history(Cl,H),
	s_depth(H,D).
search_depth(_,0).


theoryIndicatorN(0,theory):-
	!.
theoryIndicatorN(N,T):-
	number(N),
	proof_file_entry(Cl,N),
	!,
	history(Cl,H),
	theoryIndicator(H,T).
theoryIndicatorN(_,theory).


maxTh(theory,theory,theory):-
	!.
maxTh(_,_,goal).



% eof ---------------------------------------------------------------------------



