
%							ADT_TERM  
%declarations%

:- module(adt_term, [ 
	abstract_from_terms/3,
        beta_reduction/4,
	binder/4,
	binder/5,
        binding_operator/1,
	builtInOp/1,
	containsAllVarsOf/2,
	collect_symbols/3,
	deBruijn/2,
	deBruijn_index/1,
	disjointVars/2,
	get_term_operators/2,
	get_term_vars/2,
	external_binder/4,
	external_quantifiers/2,
	extraVars/2,
	free_vars/2,
	f_substring/2,
	find_value_of_extra_vars/3,
	insert_shifts/2,
	instantiate/1,
        is_closed/1,
	is_constant/1,
	logical_symbol/1,
	logical_symbol_complexity/2,
	logical_operator/1,
	normal_term_weight/2,
	path_arg1/3,
	path_arg2/3,
	pi_constraints/4,
	pi_constraints_geq/4,
	predicate_in_formula/3,
	replacesubterm_eq/4,
	ruleInstanceOf_in_/2,
	s_match/2,
	select_skolem_term/2,
	subst/3,
	subst/4,
	subterm/4,
	standard_atom/1,
	term_info/4,
	term_size/2,
	term_size_weight/2,
	term_var_weight/2,
	term_depth_weight/2,
	term_var_weight_gt_small/2,
	unique_bound_vars/2,
        unDeBruijn/2,
        var1/1,
	variant/2
   ]).


% :-meta_predicate collect_symbols(:,+,-).

/* ----------------------------------------------------------------------------


Operations on First-Order Terms

Certain assumptions about the representation of first-order terms
are fixed throughout the system. In particular, free and bound 
first-order variables are Prolog variables.
Some things are abstract such as the distinction between 
free and bound variables.


---------------------------------------------------------------------------- */



%  flags, gauges (z.B. normal_increment) -> adt_flags
%  pi_*  -> AC terms

:-use_module(global,[cont/2]).

:- use_module(proofState, [
	goal_symbol/1,
	theory_symbol/1,
	irreducible_by_current_rules/1
   ]).

:- use_module(props, [
	max_deBruijn/1,
	max_skolem/1,
	partial_congruence_relation_symbol/1,
	opOccursInBuiltInClauses/1,
	search_depth/1,
	special_relation/1
   ]).

:- use_module(constraintSolver, [
	simplify/2
   ]).

%:- use_module(library(arg), [
%	genarg/3
%   ]).
:- use_module(library(basics), [
	member/2
   ]).

:- use_module(auxiliary, [
	disjoint_vars/2,
	get_prolog_data_operators/2,
	member212/2
   ]).

:- use_module(achelp, [
	ac_del_assoc/3,
	ac_ist_AC/1,
	rreduce/3
   ]).


:- use_module(equationalTheory, [
	match/2
   ]).

:- use_module(match, [
	ac_equal/2,
	ac_subsumes_chk/2
   ]).

:- use_module(orderConstraints, [
	and_eq/3,
	and_ge/3,
	and_gt/3,
	false/1
   ]).

:- use_module(myterms, [
	list_of_vars/2,
	mapargs/3
   ]).

:- use_module(mylists, [
	is_list/1
   ]).

:- use_module(adt_precedence, [
	abbreviating_proposition/1,
	data_complexity/2,
	is_predicate/1,
	is_skolem_function/1,
	new_skolem_function/2,
	special_symbol_conv/2
   ]).

:- use_module(adt_atom, [
	atom_as_term/2,
	create_atom/2,
	is_atom/1
   ]).

:- use_module(library(change_arg), [
	change_arg/4
   ]).

:- use_module(library(occurs), [
	contains_var/2
   ]).

:- use_module(library(maplist),[maplist/3]).

:- use_module(library(flatten), [
	flatten/2
   ]).


%end%

%--------------------------------------------------------------------------------
% term ::- prolog term
%--------------------------------------------------------------------------------


%------------------------------------------------------------------------------
% gets all the operators of a given term
%------------------------------------------------------------------------------

get_term_operators( T, L ):-
	get_prolog_data_operators( T, L ), !.

%------------------------------------------------------------------------------
% term is a constant?
%------------------------------------------------------------------------------

is_constant( T ):-
	nonvar(T),
	functor(T,_,0).

%------------------------------------------------------------------------------
% collect symbols of a given kind
% special symbols might have been converted to DFG-notation -> need to be de-converted
%------------------------------------------------------------------------------
collect_symbols(_,X,[]):-
	var(X),
	!.
collect_symbols(M:P,C,[(F/N)|Fs]):-
	functor(C,F,N),
	special_symbol_conv(FD/N,F/N),  % de-convert
	Call=..[P,FD/N],
	M:Call,
	C=..[F|Args],
	!,
	maplist(collect_symbols(M:P),Args,FsL),
	flatten(FsL,Fs).
collect_symbols(P,C,Fs):-
	C=..[_F|Args],
	maplist(collect_symbols(P),Args,FsL),
	flatten(FsL,Fs).


%------------------------------------------------------------------------------
% gets all the vars of the given term
%------------------------------------------------------------------------------

get_term_vars( T, L ):-
	list_of_vars( T, L ), !.




%------------------------------------------------------------------------------
% free variables of a first-order term
%------------------------------------------------------------------------------


free_vars(T,L1):-
	free_vars(T,[],L/[]),
	sort(L,L1).

free_vars(X,BV,V):-
	var(X),
	!,
	(member212(X,BV) ->
	     V= L/L
	;
	     V=[X|L]/L
	).
free_vars(T1,BV,V):-
	(  binder(T1,T,X,_T,_)
	;
	   external_binder(T1,X,T,_)
	),
	!,
	free_vars(T,[X|BV],V).
free_vars(T,BV,V):-
	functor(T,_,N),
	free_vars_args(N,N,T,BV,V).


free_vars_args(0,_,_,_,V/V):-
	!.
free_vars_args(I,N,T,BV,VA/VE):-
	arg(I,T,TI),
	free_vars(TI,BV,VA/V),
	I1 is I-1,
	free_vars_args(I1,N,T,BV,V/VE).


is_closed(T):-
	free_vars(T,V),
	!,
	V=[].


bound_vars(T,L1):-
	bound_vars1(T,L/[]),
	sort(L,L1).

bound_vars1(X,V/V):-
	var(X),
	!.
bound_vars1(X,[X|V]/V):-
	atom(X),
	deBruijn_index(X),
	!.
bound_vars1(T,V):-
	functor(T,_,N),
	bound_vars_args(N,N,T,V).


bound_vars_args(0,_,_,V/V):-
	!.
bound_vars_args(I,N,T,VA/VE):-
	arg(I,T,TI),
	bound_vars1(TI,VA/V),
	I1 is I-1,
	bound_vars_args(I1,N,T,V/VE).




external_binder(forall(X,T),X,T,all).
external_binder(all(X,T),X,T,all).
external_binder(exists(X,T),X,T,exists).
external_binder(lambda(X,T),X,T,lambda).


external_binder_segment(S,[X|Xs],T,B):-
	external_binder(S,X,S1,B),  % same binder
	!,
	external_binder_segment(S1,Xs,T,B).
external_binder_segment(S,[],S,_).



quantifier(all).
quantifier(exists).

binder('$all'(Sk,X,T),Sk,X,T,all).
binder('$exists'(Sk,X,T),Sk,X,T,exists).
binder('$lambda'(Sk,X,T),Sk,X,T,lambda).

binder_segment(S,[Skt|Skts],[X|Xs],T,B):-
	binder(T1,Skt,X,T,B),
	binder_segment(S,Skts,Xs,T1,B).
binder_segment(T,[],_,T,_B).
	       

binder('$all'(Sk,T),Sk,T,all).
binder('$exists'(Sk,T),Sk,T,exists).
binder('$lambda'(Sk,T),Sk,T,lambda).

binding_operator(all/2).
binding_operator(exists/2).
binding_operator(lambda/2).
binding_operator('$all'/2).
binding_operator('$exists'/2).
binding_operator('$lambda'/2).


logical_symbol(= /2).
logical_symbol(== /2).
logical_symbol(and/_).
logical_symbol(or/_).
logical_symbol(=> /2).
logical_symbol(<=> /2).
logical_symbol((~)/1).
logical_symbol((@)/2).
logical_symbol('$true$'/0).
logical_symbol('$min'/0).


logical_operator(O):-
	binding_operator(O/_).
logical_operator(V):-
	deBruijn_index(V).
logical_operator(O):-
	logical_symbol(O/_).



var1(X):-
	var(X),
	!.
var1('$VAR'(_)):-
	!.
var1(I):-
%	% we risk non-termination at small dearch depth
%	search_depth(D),
%	D=<2,
	deBruijn_index(I).



/* ---------------------------------------------------------------------------


  predicate_in_formula(+F,?P/N,?S)

is true when predicate symbol P/N occurs in the formula F
in a subformula matching P. No bindings are produced by the match.

--------------------------------------------------------------------------- */



predicate_in_formula(X,_,_):-
	var(X),
	!,
	fail.
predicate_in_formula(T,R,S):-
	functor(T,P,N),
	(logical_symbol(P/N),
	 \+ (P/N = (=)/2)	
	->
           genarg(_I,T,TI),
	   predicate_in_formula(TI,R,S)
	;binder(T,_,TI,_) ->
	   predicate_in_formula(TI,R,S)
	;
	   \+ \+ match(S,T),
	   R=P/N
	).


abstract_from_terms(X,Y,[Y=X]):-
	var(X),
	!.
abstract_from_terms(A,AA,S):-
	is_atom(A),
	!,
	atom_as_term(A,T),
	abstract_from_terms(T,AT,S),
	create_atom(AT,AA).
abstract_from_terms(T,T,[]):-
	binder(T,_,_,_),
	!.
abstract_from_terms(T,A,S):-
	functor(T,P,N),
	logical_symbol(P/N),
	\+ (P/N = (=)/2),
	!,
	functor(A,P,N),
	abstract_from_terms(N,N,T,A,S).
abstract_from_terms(T,A,S):-
	T=..[P|Ts],   % must be a predicate
	!,
	abstract_list_from_terms(Ts,As,S),
	A=..[P|As].



abstract_list_from_terms([],[],[]).
abstract_list_from_terms([T|Ts],[A|As],S):-
	(bound_vars(T,[]) ->
	     A=X,
	     S=[X=T|S1]
	;
	     A=T,
	     S=S1
	),
	abstract_list_from_terms(Ts,As,S1).





abstract_from_terms(0,_N,_T,_A,[]):-
	!.
abstract_from_terms(I,N,T,A,S):-
	arg(I,T,TI),
	arg(I,A,AI),
	abstract_from_terms(TI,AI,SI),
	I1 is I-1,
	abstract_from_terms(I1,N,T,A,S1),
	append(SI,S1,S).




/* ---------------------------------------------------------------------------


   logical symbols and their ordering complexities

--------------------------------------------------------------------------- */



logical_symbol_complexity('$all',1817).
logical_symbol_complexity('$exists',1816).

logical_symbol_complexity(<=>,1815).
logical_symbol_complexity(=>,1814).
logical_symbol_complexity(or,1813).
logical_symbol_complexity((~),1812).
logical_symbol_complexity(and,1811).
logical_symbol_complexity(all,1810).
logical_symbol_complexity(exists,1809).

logical_symbol_complexity(==,1806).
logical_symbol_complexity(= ,1805).

logical_symbol_complexity((@),1804).
logical_symbol_complexity('$lambda',1803).


/* ---------------------------------------------------------------------------


   Introduction of deBruijn Indexes


--------------------------------------------------------------------------- */

% V is an Index
deBruijn_index(V):-
	atom(V),
	atom_concat('i$',_,V).
%	name(V,[105,36|_]). % i$ ...


% it is important that deBruijn indexes be greater than Skolem functions
% to make Skolemization monotonic


% bidirectional conversion between numbers I and indexes V
deBruijn_index(V,I):-
	atom(V),
	!,
	name(V,[105,36|NI]), 
	name(I,NI).
deBruijn_index(V,I):-
	number(I),
	!,
	name(I,NI),
	name(V,[105,36|NI]).


firstIndex('i$1').

nextIndex(I,I1):-
	name(I,[105,36|N]),
	name(Nu,N),
	Nu1 is Nu+1,
	max_deBruijn(M),
	Nu1=<M,
	name(Nu1,N1),
	name(I1,[105,36|N1]),
	!.
nextIndex(_I,_I1):-
	raise_exception(saturate_error("max number deBruijn indices exceeded",[])).

/* ---------------------------------------------------------------------------


deBruijn(?A,?B)

is true when B is A with bound variables replaced by deBruijn-Indexes.
The predicate is bidirectional. When called with input B, 
the result A will have pairwise different bound variables,
but internal binders (cf. deBruijn1). unDeBrjuin is like deBruijn1
with second parameter 
except that external binders are re-introduced and Skolem functions
are forgotten.


--------------------------------------------------------------------------- */
deBruijn(S,T):-
	var(T),
	!,
	quantifiers_with_dependencies(S,T1),
	deBruijn1(T1,T).
deBruijn(S,T):-
	deBruijn1(S,T).



deBruijn1(X,X):-
	var(X),
	!.
deBruijn1(S,T):-
	binder(S,SkT,X,A,Op),
	binder(T,SkT1,A1,Op),
	!,
	firstIndex(I0),
	(var(SkT)  ->
	     % replacement of indexes by variables
	     subst(A1,I0,X,AI),
	     subst(SkT1,I0,X,SkT),
	     deBruijn1(A,AI)
	;
	     subst(A,X,I0,AI),
	     subst(SkT,X,I0,SkT1),
	     deBruijn1(AI,A1)
	).
deBruijn1(A,B):-
	mapargs(deBruijn1,A,B).

unDeBruijn(X,X):-
	var(X),
	!.
unDeBruijn(S,T):-
	external_binder(S,X,A,Op),
	binder(T,_SkT1,A1,Op),
	!,
	firstIndex(I0),
	subst(A1,I0,X,AI),
	unDeBruijn(A,AI).
unDeBruijn(A,B):-
	mapargs(unDeBruijn,A,B).




quantifiers_with_dependencies(S,S):-
	var(S),
	!.
quantifiers_with_dependencies(S,T):-
	external_binder_segment(S,Xs,A,Op),
	nonvar(Op),
	!,
	free_vars(S,V),
	maplist(new_skolem_term(Op,V),Xs,SkTs),
	quantifiers_with_dependencies(A,A1),
	binder_segment(T,SkTs,Xs,A1,Op).
quantifiers_with_dependencies(S,T):-
	mapargs(quantifiers_with_dependencies,S,T).

external_quantifiers(S,S):-
	var(S),
	!.
external_quantifiers(S,T):-
	binder(S,_Skt,X,A,Op),
	!,
	external_quantifiers(A,A1),
	external_binder(T,X,A1,Op).
external_quantifiers(S,T):-
	mapargs(external_quantifiers,S,T).


new_skolem_term(Op,DepOn,_ForThisVar,SkT):-
	( quantifier(Op)->
	    new_skolem_function(_,F),
	    SkT=..[F|DepOn]
	;
	    SkT='$bound'
	).
	

select_skolem_term(T,T1):-
	% this is an optimization and requires that
	% new skolem constants be smaller than older skolem constants
	\+atom(T),
	\+is_constant(T),

	free_vars(T,TV),
	length(TV,M),
	M=0,
%	functor(T,_,N),M<N,
        cont(skolemCtr,SC),
	max_skolem(Max),
	SC<Max/10,
	!,
	new_skolem_function(M,F),
	T1=..[F|TV].
select_skolem_term(T,T).



/* ----------------------------------------------------------------------------

subst(+S,+X,+E,?T)
is true when T is the result of substituting free occurrences of
X in S by E.

subst works both for representations of S with explicit variables
or with deBruijn indices. In the first case, X is assumed to be a 
Prolog variable, whereas in the second case X is a deBruijn index I,
referring to a binder I levels above S.

---------------------------------------------------------------------------- */

subst(Old,Old1,New,New):-
	Old==Old1,
	!.
subst(X,_,_,X):-
	var(X),
	!.
subst(S,X,I,T):-
	binder(S,SkT,Y,A,Op),
	binder(T,SkT1,Y,A1,Op),
	X\==Y,
	!,
	nextIndex(I,I1),
	subst(A,X,I1,A1),
	subst(SkT,X,I1,SkT1).
subst(S,_X,_I,S):-
	binder(S,_SkT,_Y,_Body,_Op),
	!.
subst(S,I,X,T):-
	binder(S,SkT,A,Op),
	binder(T,SkT1,A1,Op),
	!,
	nextIndex(I,I1),
	shift(0,X,X0),
%	X=X0,
	subst(A,I1,X0,A1),
	subst(SkT,I1,X0,SkT1).
subst(A,Old,New,B):-
	mapargs(subst1(Old,New),A,B).





subst1(Old,New,A,B):-
	subst(A,Old,New,B).


subst(A,New,B):-
	firstIndex(I),
	subst(A,I,New,B).

/* ----------------------------------------------------------------------------


insert_shifts(+T,?S)

   is true whenever S is a variant of T in which all free Prolog
   variables X are shifted (by replacing X by $shift(I,X)) by
   the binding depth I, if I>0, of the respective occurrence of X.


---------------------------------------------------------------------------- */



insert_shifts(S,T):-
	insert_shifts(0,Changed,S,T1),
	!,
	(var(Changed) ->
	     S=T
	;
	     (is_list(S) ->
		 % internal data structure such as atoms
		 maplist(insert_shifts,S,T)
	     ;
		 T='$shift'(T1)
	     )
	).

insert_shifts(0,_,X,X):-
	var(X),
	!.
insert_shifts(I,changed,X,'$shift'(I,X)):-
	var(X),
	!.
insert_shifts(I,C,S,T):-
	binder(S,Sk,S1,B),
	binder(T,Sk1,T1,B),
	!,
	I1 is I+1,
	insert_shifts(I1,C,S1,T1),
	insert_shifts(I1,C,Sk,Sk1).
insert_shifts(I,C,S,T):-
	mapargs(insert_shifts(I,C),S,T).




/* ----------------------------------------------------------------------------


beta_reduction(+S,+SR,?T,?TR)

   normalizes S with respect to shift and 1 step of root beta reduction
   to yield T. SR and TR are irreducibility indicators.

---------------------------------------------------------------------------- */

beta_reduction('$subst'(F,A),_,T,T):-
	% after 1 step of beta reduction more reductions are possible
	% hence TR = T
	!,
	subst(F,A,T).
beta_reduction('$shift'(S),SR,T,SR):-
	!,
	shift(S,T).
beta_reduction(S,SR,S,SR).


/* ----------------------------------------------------------------------------


shift(+S,?T)

   replaces '$shift'(I,'$$J') in  by '$$(I+J)' to yield T,
   for every free variable '$$J' in S.
   We assume that S contains no free Prolog variables and
   that S contains no '$shift(_,_) subterms if it is not the form
   S = '$shift'(S1).


---------------------------------------------------------------------------- */




shift(S,T):-
	!,
	shift(0,0,S,T).



shift(_I,_D,'$shift'(J,S),T):-
	!,
	shift(J,0,S,T).
shift(I,D,V,V1):-
	deBruijn_index(V,J),
	!,
	(J>=D ->
	     J1 is I+J,
	     deBruijn_index(V1,J1)
	;
	     V1=V
	).
shift(I,D,S,T):-
	binder(S,Sk,S1,B),
	binder(T,Sk1,T1,B),
	!,
	D1 is D+1,
	shift(I,D1,S1,T1),
	shift(I,D1,Sk,Sk1).
shift(I,D,S,T):-
	mapargs(shift(I,D),S,T).

	     
	     
/* ----------------------------------------------------------------------------


shift(+K,+S,?T)
   succeeds, if T is S in which all variables I>=K+binder nesting depth
   are lifted by 1


---------------------------------------------------------------------------- */
	

shift(_K,S,T):-
	var(S),
	!,
	T=S.
shift(K,V,V1):-
	deBruijn_index(V,J),
	!,
	(J>=K ->
	    J1 is J+1,
	    deBruijn_index(V1,J1)
	;
	    V1=V).
shift(K,S,T):-
	binder(S,Sk,S1,B),
	binder(T,Sk1,T1,B),
	!,
	K1 is K+1,
	shift(K1,S1,T1),
	shift(K1,Sk,Sk1).
shift(K,S,T):-
	mapargs(shift(K),S,T).










	


unique_bound_vars(A,B):-
	deBruijn(A,A1),
	deBruijn(B,A1).







disjointVars(S,T):-
	list_of_vars(T,VT),
	list_of_vars(S,VS),
	disjoint_vars(VS,VT).





containsAllVarsOf(S,T):-
	list_of_vars(T,VT),
	list_of_vars(S,VS),
	\+ (member(X,VT),
	    \+ member212(X,VS)).


extraVars(V1,V2):-
	member(X,V1),
	\+ member212(X,V2),
	!.

find_value_of_extra_vars( ToChooseFrom, Vars, NotFromHere ):-
	find_value_of_extra_vars1( ToChooseFrom, Vars, NotFromHere ).
find_value_of_extra_vars( ToChooseFrom, Vars, NotFromHere ):-
	\+find_value_of_extra_vars1( ToChooseFrom, Vars, NotFromHere ),
	find_value_of_extra_vars1( ToChooseFrom, Vars ).



find_value_of_extra_vars1( _, [], _ ).
find_value_of_extra_vars1( LVC, [V|R], LVR ):-
	member212( V, LVC ), !,
	find_value_of_extra_vars1( LVC, R, LVR ).
find_value_of_extra_vars1( LVC, [V|R], LVR ):-
	member( V1, LVC ),
	\+ member212( V1, LVR ),
	V = V1,
	find_value_of_extra_vars1( LVC, R, LVR ).

find_value_of_extra_vars1( _, []).
find_value_of_extra_vars1( LVC, [V|R]):-
	member212( V, LVC ), !,
	find_value_of_extra_vars1( LVC, R ).
find_value_of_extra_vars1( LVC, [V|R] ):-
	member( V, LVC ),
	find_value_of_extra_vars1( LVC, R ).





%------------------------------------------------------------------------------
% term size
%------------------------------------------------------------------------------
normal_term_weight((S,D,_),SD):-
	SD is integer(S*D).   % standard until 20020820
%	SD is integer(S*D/2).   % standard until 20020820
%	SD is integer(sqrt(S*D)). % for operator depth
%	SD is integer(S*D).  
%	SD is integer(S+D).  
%	SD is integer(S*max(D-4,1)).

term_size_weight((S,_,_),S).

term_var_weight((S1,D,NVars),S):-
	S is integer((NVars*NVars+3)*S1*(D/4)),
%	S is integer((NVars+1)*S1*(D/2)),
	!.

term_var_weight_gt_small((S1,_D,NVars),S):-
	S is NVars*S1,
%	S is integer((NVars+0.01)*S1*(D/2)),
	!.


term_depth_weight((S1,D,NVars),S):-
	S is integer((NVars*NVars+3)*S1*(D*D)/16),
	!.



term_size( T, 1 ):-
        var(T), !.
term_size( [], 0 ):- !.
term_size( [H|R], S ):-
	!,
        term_size( H, S1 ),
        term_size( R, S2 ),
        S is S1 + S2.
term_size( apply(H,R), S ):-
	!,
        term_size( H, S1 ),
        term_size( R, S2 ),
        S is S1 + S2.
term_size( comp(H,R), S ):-
	!,
        term_size( H, S1 ),
        term_size( R, S2 ),
        S is S1 + S2.
term_size( T, S ):-
	functor(T,O,N),
	term_size_from_to(0,N,T,S1),
	(
%	  (goal_symbol(O/N);is_skolem_function(O/N)) ->
%	  N=0,is_skolem_function(O/N) ->
%	    S is S1
%	;
	  goal_symbol(O/N) ->
	     S is S1+1/2
	;
	     S is S1 + 1
	).


term_size_from_to(I,N,T,S):-
	(I=N ->
	     S=0
	;
	     I1 is I+1,
	     arg(I1,T,T1),
	     term_size(T1,S1),
	     term_size_from_to(I1,N,T,S2),
	     S is S1+S2
	).


term_info(T,Size,Depth,Vars):-
%	term_size(T,Size,Depth,V,[]),
	term_sizeD(T,Size,OpDepths,DD,V,[]),
	max_depth(OpDepths,D),
	user:sqrt(DD*D,D1), % don't have sqrt in Quintus
	Depth is D1*D,  % most terms have D=1 
%	Depth is DD*D*D,  % most terms have D=1
	sort(V,Vars),
	!.

term_size( T, 1, 1, [T|End], End ):-
        var(T), !.
term_size( [], 0, 0, E, E ):- !.
term_size( [H|R], S, D, L1, E):-
        term_size( H, S1, D1, L1, L2 ),
        term_size( R, S2, D2, L2, E ),
        S is S1 + S2, 
	D is max(D1,D2), !.
term_size( T, S, D, L, E ):-
	functor(T,O,N),
	term_size_from_to(0,N,T,S1,D1,L,E),
	(
	  (O/N=apply/2;O/N=comp/2) ->
	     S is S1
	;
%	  (goal_symbol(O/N);is_skolem_function(O/N)) ->
	  N=0,is_skolem_function(O/N) ->
	    S is S1
	;
	  goal_symbol(O/N) ->
	     S is S1+1/2
	;
	  abbreviating_proposition(O/N) ->
	    S is S1 + 5
	)
	,
%	data_complexity(O,C1),
	D is D1 + 1,
	!.

term_size_from_to(N,N,_,0,0,End,End):-
	!.
term_size_from_to(I,N,T,S,D,L1,E):-
	I1 is I+1,
	arg(I1,T,T1),
	term_size(T1,S1,D1,L1,L2),
	term_size_from_to(I1,N,T,S2,D2,L2,E),
        S is S1 + S2, 
	D is max(D1,D2), !.
	




% new version where we collect nesting information about each function symbol


term_sizeD( T, 1, [], 1, [T|End], End ):-
        var(T), !.
term_sizeD( [], 0, [], 0, E, E ):- !.
term_sizeD( [H|R], S, D, DD, L1, E):-
        term_sizeD( H, S1, D1, DD1, L1, L2 ),
        term_sizeD( R, S2, D2, DD2, L2, E ),
        S is S1 + S2, 
	max_depth(D1,D2,D),
	DD is max(DD1,DD2),
	!.
term_sizeD( T, S, D, DD, L, E ):-
	functor(T,O,N),
	(binding_operator(O/N)->
	    arg(2,T,T2),
	    term_sizeD( T2, S, D, DD1, L, E ),
	    DD is DD1+1
	;
	    term_size_from_toD(0,N,T,S1,D1,DD1,L,E),
	    (	
		( O/N=apply/2;O/N=comp/2) ->
		S is S1
	    ;	
%	        (goal_symbol(O/N);is_skolem_function(O/N)) ->
		N=0,is_skolem_function(O/N) ->
		S is S1+1/2
	    ;	
		goal_symbol(O/N) ->
		S is S1+1/2
	    ;	
		abbreviating_proposition(O/N) ->
		S is S1 + 5
	    ;	
%	        data_complexity(O,C1),
		S is S1 + 1
	    ),
	    incr_depth(O/N,D1,D),
	    DD is DD1+1
	),
	!.

term_size_from_toD(N,N,_,0,[],0,End,End):-
	!.
term_size_from_toD(I,N,T,S,D,DD,L1,E):-
	I1 is I+1,
	arg(I1,T,T1),
	term_sizeD(T1,S1,D1,DD1,L1,L2),
	term_size_from_toD(I1,N,T,S2,D2,DD2,L2,E),
        S is S1 + S2, 
	max_depth(D1,D2,D),
	DD is max(DD1,DD2),
	!.
	



incr_depth(O/N,[],[(O/N,1)]).
incr_depth(O/N,[(O1/N1,D)|Os],[(O1/N1,D1)|Os]):-
	(   O=O1,N=N1
%	;   N>0,N1>0,
%	    is_skolem_function(O/N),
%	    is_skolem_function(O1/N1)
	),
	!,
	D1 is D+1.
incr_depth(O/N,[(O1/N1,D1)|Os],[(O1/N1,D1)|Os1]):-
	incr_depth(O/N,Os,Os1).


max_depth([(O/N,D)|Os],Os1,Os2):-
	max_depth_for((O/N,D),Os1,Os3),
	max_depth(Os,Os3,Os2).
max_depth([],Os,Os).


max_depth_for((O/N,D),[],[(O/N,D)]).
max_depth_for((O/N,D),[(O1/N1,D1)|Os],[(O1/N1,DM)|Os]):-
	(
	    O=O1,N=N1
%	;   N>0,N1>0,
%	    is_skolem_function(O/N),
%	    is_skolem_function(O1/N1)
	),
	!,
	DM is max(D,D1).
max_depth_for((O/N,D),[(O1/N1,D1)|Os],[(O1/N1,D1)|Os1]):-
	max_depth_for((O/N,D),Os,Os1).



	
max_depth([],0).
max_depth([('.'/2,_D)|Os],D1):-  % internal data structures
	!,
	max_depth(Os,D1).
max_depth([(comp/2,_D)|Os],D1):-  % HOL encodings
	!,
	max_depth(Os,D1).
max_depth([(apply/2,_D)|Os],D1):-  % HOL encodings
	!,
	max_depth(Os,D1).
max_depth([(F/N,_D)|Os],D1):-  % logical symbol
	binding_operator(F/N),
%	(logical_symbol(F/N);binding_operator(F/N)),
	!,
	max_depth(Os,D1).
max_depth([(_ON,D)|Os],DM):-
	max_depth(Os,D1),
	DM is max(D,D1).



% end new version








/*-----------------------------------------------------------------------------


 ruleInstanceOf_atpos_( (?S1,?PolS,+RSS,?S2), (+T1,+PolT,?RST,?T2), ?Pos ): 
  iff  S1=T1/p  and
       S1 RSS^PolS S2 implies T1[S1]_p RTS^PolT T1[S2]_p = T2
       is a consequence of the monotonicity laws.



 position ::- { lambda, notlambda }
-----------------------------------------------------------------------------*/
/* old


ruleInstanceOf_in_atpos_( (T,Pol,RS,S2), (T,Pol,RS,S2), lambda ).
ruleInstanceOf_in_atpos_( (S1,PolS,RSS,S2), (T,PolT,RST,NT), notlambda ):-
        nonvar( T ),
	functor(T,F,_N),
	monotone_operator(F,I,RSSub,PolI,RST),
	genarg(I,T,Sub),
	PolSub is PolT*PolI,
	ruleInstanceOf_in_( (S1,PolS,RSS,S2), (Sub,PolSub,RSSub,NTI) ),
	change_arg(I,T,NT,NTI).

*/
	


ruleInstanceOf_in_( Ru, Inst ):-
	ruleInstanceOf_in_atpos_( Ru, Inst, _ ).

ruleInstanceOf_in_atpos_( (T,Pol,RS,S2), (T,Pol,RS,S2), [] ).
ruleInstanceOf_in_atpos_( (S1,Pol,RS,S2), (T,Pol,RS1,NT), [_|_] ):-
        nonvar( T ),
	partial_congruence_relation_symbol(RS),
	(RS1=RS;RS1= ==),
	replacesubterm_eq_atpos_(S1,T,S2,NT,[_|_]).
ruleInstanceOf_in_atpos_( (T1,Pol,==,S2), (T1==T2,Pol,==,S2==T2), [1] ).
ruleInstanceOf_in_atpos_( (T2,Pol,==,S2), (T1==T2,Pol,==,T1==S2), [2] ).


	

% ------------- old version -------------------
%      for the case of (one) equality relation;
%      to be kept for efficiency reasons
% --------------------------------------------	
replacesubterm_eq( T, T, S2, S2 ).
replacesubterm_eq( S1, T, S2, NT ):-
        nonvar( T ),
	genarg(I,T,TI),
	replacesubterm_eq(S1,TI,S2,NTI),
	change_arg(I,T,NT,NTI).




replacesubterm_eq_atpos_( S1, T, S2, NT, [_|_] ):-
        nonvar( T ),
	genarg(I,T,TI),
	replacesubterm_eq_atpos_(S1,TI,S2,NTI,_),
	change_arg(I,T,NT,NTI).
replacesubterm_eq_atpos_( T, T, S2, S2, [] ).

%-------------------------------------------------
% subterm replacement
%-------------------------------------------------

genarg(N, Term, Arg) :-
%	var(N),       % assumption
%	nonvar(Term), % assumption
%	!,
	functor(Term, _, Arity),
	genarg(Arity, Term, Arg, N).



genarg(1, Term, Arg, 1) :- !,
	arg(1, Term, Arg).
genarg(N, Term, Arg, N) :-
	arg(N, Term, Arg).
genarg(K, Term, Arg, N) :-
	K > 1, J is K-1,
	genarg(J, Term, Arg, N).



path_arg1( [], T, T ).
path_arg1([I|Path], S, Sub ):-
        nonvar( S ),
	\+ skolemTerm(S),
	genarg(I,S,SI),
	path_arg1(Path,SI,Sub).


path_arg2( [], T, T ).
path_arg2([I|Path], S, Sub ):-
        nonvar( S ),
	genarg(I,S,SI),
	\+irreducible_by_current_rules(SI),
	path_arg2(Path,SI,Sub).




% Skolem functions are those that do not occur in input clauses
% with nonvariable arguments. That is a poor man's basic strategy.
% What are its consequences with respect to simplification?

skolemTerm(T):-
	functor(T,F,N),
	is_skolem_function(F/N).


% skolemTerm(i(_)).


subterm(_,0,_,_):-
	!,
	fail.
subterm(N,N,T,S):-
	arg(N,T,S).
subterm(I,N,T,S):-
	K is N-1,
	subterm(I,K,T,S).



f_substring(X,S):-
	path_arg1(_,X,Sub),
	f_sub(S,Sub).


f_sub([],_).
f_sub([F],X):-
	nonvar(X),
	functor(X,F,_).
f_sub([F,G|S],X):-
	nonvar(X),
	functor(X,F,_),
	genarg(_,X,XI),
	f_sub([G|S],XI).





standard_atom(A):-
	var(A),
	!,
	fail.
standard_atom('$true$').
standard_atom(A):-
	functor(A,P,N),
	is_predicate(P/N),
	\+ (N=2,special_relation(P)).  % 2002/08/02



%------------------------------------------------------------------------------
% unification
%------------------------------------------------------------------------------




s_match(L,T):-
	match(L,T).

instantiate([]):-!.
instantiate([ [X,X]|Rest]):-
                instantiate(Rest),!.
instantiate([ X-X|Rest]):-
                instantiate(Rest),!.
instantiate([ X=X|Rest]):-
                instantiate(Rest),!.


/*-------------------------------------------------------------------
matching up to permutativity with computation of corresponding
substitution and constraints

pi_match(
        Pattern,
	Term,
	InConstraints,  (triple of 2 constraints and 1 equality constraint
	  first: accumulated constraint from permutating subterms
	  second: corresponding complement (inequality part)
	  third: equality part of the constraint)

	OutConstraints,
	InSubst,
	OutSubst
)
*/
pi_match(P,T,C,C,S,S):-
	P==T,
	!.
pi_match( Left, T, _, _, _, _ ):-
	\+ac_subsumes_chk( Left, T ), !, fail.
pi_match(Left,T,Ci,Co,S,S):-
	var(Left),
	in_subst([Left,B],S),!,
	ac_equal(B,T),
	pi_constraints_geq([T],[B],Ci,Co).
pi_match(Left,T,C,C,S,[[Left,T]|S]):-
	var(Left),
	!.
pi_match(Left,T,Ci,Co,Si,So):-
	nonvar(T),
	functor(Left,F,N),
	\+ac_ist_AC(F),
	!,
	functor(T,F,N),
	Left =.. [_|ARGLeft],
	T =.. [_|ARGT],
	pi_match_list(ARGLeft,ARGT,Ci,Co,Si,So).
pi_match(Left,T,Ci,Co,Si,So):-
	nonvar(T),
	Left =.. [F,P1,P2],
	T =.. [F,T1,T2],
	pi_match(P1,T1,Ci,C1,Si,S1),
	pi_match(P2,T2,C1,Co,S1,So).
pi_match(Left,T,Ci,Co,Si,So):-
	nonvar(T),
	Left =.. [F,P1,P2],
	T =.. [F|Ts],
	ac_del_assoc(F,Ts,Ts1),
%	!,
	append([T1|TL],[T2|TR],Ts1),
	append([T1|TL],TR,Ts2),
	rreduce(F,Ts2,T3),
	pi_match(P1,T2,Ci,C1,Si,S2),
	pi_match(P2,T3,C1,C2,S2,So),
	pi_constraints_geq([T1|TL],[T2],C2,Co).
pi_match(Left,T,Ci,Co,Si,So):-
	nonvar(T),
	Left =.. [F,P1,P2],
	T =.. [F|Ts],
	ac_del_assoc(F,Ts,Ts1),
	!,
	append([T1,T2|TL],TR,Ts1),
	append(TL,[T1|TR],Ts2),
	rreduce(F,Ts2,T3),
	pi_match(P1,T2,Ci,C1,Si,S2),
	pi_match(P2,T3,C1,C2,S2,So),
	pi_constraints_geq([T1],[T2|TL],C2,Co).

pi_constraints([],_,C,C).
pi_constraints([T|Ts],Rs,Ci,Co):-
	pi_constraints1(T,Rs,Ci,C1),
	pi_constraints(Ts,Rs,C1,Co).

pi_constraints1(T1,[T2|Ts],[Ci,CCi,Ei], Co):-
	and_eq(Ci,[T1,T2],C11),
	simplify(C11,C1),
	\+ false(C1),
	(and_gt(Ci,[T2,T1],C3),
         simplify(C3,C3S),
	 false(C3S) ->
            CC1=CCi
	;   CC1=[[T1,T2]|CCi]
	),
	!,
	pi_constraints1(T1,Ts,[C1,CC1,[T1=T2|Ei]],Co).
pi_constraints1(_T,[],C,C).




pi_constraints_geq([],_,C,C).
pi_constraints_geq([T|Ts],Rs,Ci,Co):-
	pi_constraints1_geq(T,Rs,Ci,C1),
	pi_constraints_geq(Ts,Rs,C1,Co).

pi_constraints1_geq(T1,[T2|Ts],[Ci,CCi,Ei], Co):-
	and_ge(Ci,[T1,T2],C11),
	simplify(C11,C1),
	\+ false(C1),
	(and_gt(Ci,[T2,T1],C3),
         simplify(C3,C3S),
	 false(C3S) ->
            CC1=CCi
	;   CC1=[[T1,T2]|CCi]
	),
	!,
	pi_constraints1_geq(T1,Ts,[C1,CC1,Ei],Co).
pi_constraints1_geq(_T,[],C,C).







/*
pi_match(_8084+ (_8084+_8088),_7340+ (_7343+ (_7346+_7340)),yes,[[[[],[]]],[],[]],_8205,[],_8207).
*/

/*
match_list( [],[],L,L ):- !.
match_list([A1|ARGS1],[A2|ARGS2],L1,L2 ):-
                match(A1,A2,L1,L3),
                match_list(ARGS1,ARGS2,L3,L2),
		!.
*/

pi_match_list( [],[],L,L,S,S ).
pi_match_list([A1|ARGS1],[A2|ARGS2],Ci,Co,Si,So):-
	pi_match(A1,A2,Ci,C1,Si,S1),
	pi_match_list(ARGS1,ARGS2,C1,Co,S1,So).


in_subst( [L,T], [[X,T]|_] ):- L == X,!.
in_subst( X, [_|R] ):- in_subst( X,R ).

%--------------------------------------------------------------------------------
/*
subterm_match( Left, T, Right, Right ):-
                match( Left, T ), !.

subterm_match(Left,T,Right, NT) :-
                nonvar(T),
                T =.. [F|Args1],
                subterm_match_list(Left,Args1,Right, Args2),
                NT =.. [F|Args2].

subterm_match_list(Left,[T|Rest],Right, [NT|Rest]):-
                nonvar(T),
                subterm_match(Left,T, Right, NT).
subterm_match_list(Left,[T|Rest1],Right, [T|Rest2]):-
                subterm_match_list(Left,Rest1,Right, Rest2).
*/



builtInOp('$true$').
builtInOp('$min').
builtInOp(=).
builtInOp(==).
builtInOp(O) :-
	opOccursInBuiltInClauses(O).

%------------------------------------------------------------------------------
% checks whether two terms are variants one of the other, that is, if they are
% equal up to variable renaming
%------------------------------------------------------------------------------

variant( T1, T2 ):-
%	copy_term(T1,T),
        subsumes_chk( T1, T2 ),
        subsumes_chk( T2, T1 ),!.


% eof -------------------------------------------------------------------------

