%declarations%

:- module(clause_processing, [ 
	process_clause/3
   ]).

:- use_module(props, [
	flag_on/2
   ]).

:- use_module(adt_atom, [
	atom_for_demodulation/5,
	true_literal/1
   ]).

:- use_module(classes, [
	dont_simplify/1,
	antecedent/2,
	empty/1,
	number/2,
	literals/3,
	createRuleWithSelection/5,
	createRule/4,
	maximal_succedent/2,
	is_atom/1,
	small_rule/1
   ]).

:- use_module(adt_declarations, [
	message/2
   ]).

:- use_module(equationalTheory, [
	match/2
   ]).

:- use_module(adt_term, [
	path_arg1/3
   ]).

:- use_module(clause_redundancy_prover, [
	condensement/3,
	demodulate_clause/4,
	extensionality_resolution/2,
	redundant_clause_by_case_analysis/1,
	simplify_by_crw/2,
	remove_repetitions_and_identities/2,
	simplify_by_var_elim/2,
	tautology/3,
	totality_resolution/3
   ]).

:- use_module(termOrder,[greater/2]).

:- use_module(global, [
        cont/2,
	store/2
   ]).

:- use_module(timing, [
	push_timing/1,
	pop_timing/1
   ]).

:- use_module(proofState, [
           deleteAllRules/0,
	   recordRule/2,
	   recordProofOfClause/1,
	   deleteIrreducibility/2,
	   deleteNormalform/2

   ]).


:- use_module(propLogic,[tellPropClause/4]).

%end%

%							CLAUSE PROCESSING


%------------------------------------------------------------------------------
%
%	PROCESS THE CLAUSES
%------------------------------------------------------------------------------

process_clause( Cl, ST, STO ):-
	message( trcl, [Cl] ),
	simplify_clause_into_rules( Cl, ST, STO ),
	!.
process_clause(Cl,ST,STO):-
%	trace,
	raise_exception(saturate_error("~nprocess_clause should not be failing",[])),
	process_clause(Cl,ST,STO).



% the following might produce 0, 1 or more than one new rule
% upon backtracking

simplify_clause_into_rules(Cl, ST, STO):-
	push_timing(csim),
	findall(Dl,simplify_clause( Cl, Dl, ST),Dls),
	pop_timing(csim),
	createRules(Dls,ST,STO).
	

createRules([Dl|Dls],ST,STO):-
	push_timing(clpp),
	tellPropClause(Dl,Dl1,ST,STT), 
	convert_into_rule( Dl1, Ru, STT, ST1 ),
	pop_timing(clpp),
	(empty(Ru) -> deleteAllRules; true), 
	number(Ru,N),
	(small_rule(Ru) ->
	    store(new_small_rule,[N])
	;   true),
	recordRule(N,Ru),
	cont(new_rules,Rs),
	store(new_rules,[N|Rs]),
	createRules(Dls,ST1,STO).
createRules([],ST,ST).

convert_into_rule( Dl, Ru, ST, STO ):-
	\+ flag_on( ma, _ ), !,
	createRule( Dl, Ru, ST, STO ),
	message( nwru, [Ru] ), !.
convert_into_rule( Dl, Ru, ST, STO ):-
	antecedent( Dl, [] ),
	createRule( Dl, Ru, ST, STO ),
	message( nwru, [Ru] ), !.
convert_into_rule( Dl, Ru, ST, STO ):-
	flag_on( ma,user),
	nl, write( 'Marking:' ), nl,
	print( Dl ),
	nl, write( '    Introduce:' ), nl,
	write( '       the number of the equation to be marked' ), nl,
	write( '       0 if none' ), nl, nl,
	read( MKN ),
	( ( MKN=0, createRule( Dl, Ru, ST, STO ) )
	;
	  createRuleWithSelection( MKN, Dl, Ru, ST, STO )
	),
	message( nwru, [Ru] ), !.
convert_into_rule( Dl, Ru, ST, STO ):-
	flag_on(ma,M),
	createRuleWithSelection( M, Dl, Ru, ST, STO ),
	message( nwru, [Ru] ), !.













% ---------------------------------------------------------
% A clause may be simplified into a conjunction
% of clauses. These are obtained through backtracking.
% At present only variable elimination may produce
% more than one result.
% ---------------------------------------------------------

simplify_clause(C,C,_ST):-
	dont_simplify(C),
	!,
	recordProofOfClause(C),
	!.
simplify_clause( C1, C7, ST ):-
	number(C1,N1),
	(user:spy_clause(N1)-> trace; true),
	extensionality_resolution(C1,C11),    
	recordProofOfClause(C11),
	totality_resolution(C11,C2,AC2),
	% we should reduce first before we check for tautology
	% so that definitions are first expanded
	on_exception(reduced_to_true(RN),
	             iterateReduction(maybeReducible,ST,C2,C5),
		     tell_tautology(C2,RN)),
	tautology( C5, C51, no ),
	recordProofOfClause( C1 ),
	record_proof(var(AC2),C2), 

	  % the following simplification may produce more than one result
	simplify_by_var_elim(C51,C6),
	\+ redundant_clause_by_case_analysis( C6 ),
	simplify_by_crw(C6,C61),
	number(C6,N6),
	number(C51,N51),
	( N51 \== N6 ->
	      recordProofOfClause(C51),
	      simplify_clause(C61,C7,ST)
	; 
              C61=C7
	).


iterateReduction(wasIrreducible,_ST,C,C):- !.
iterateReduction(RFlag,ST,C,D):-
%	remove_repetitions_and_identities( C, C2 ),
        C2=C,
	condensement( C2, Cond, ACond ),
%	record_proof(var(ACond),Cond),
	((var(ACond);RFlag=maybeReducible) ->
%	write('.'),ttyflush,
             demodulate_clause( Cond, C3, AC3, ST ),
%	write('.'),ttyflush,
             record_proof(var(AC3),C3),
	     (var(AC3) ->
		  NF=irreducible
	     ;
	          NF=wasIrreducible
	     ),
	     iterateReduction(NF,ST,C3,D)
	;D=C2
	),
	!.
	

tell_tautology(C1,RN):-
	true_literal(T),
	literals([T],C1,C11),
%	redundancy_info((tautology,[]),C11,C2),

	(RN\==[] ->
	     message( cdem, [C1,RN,C11] )
	;
   	     true
	),
	!,
	fail.

record_proof(P,C):-
	(P -> recordProofOfClause(C); true).


% eof -------------------------------------------------------------------------




