%%% -*- Mode: Prolog; Module: dpll; -*-
% Author: Harald Ganzinger, MPI Informatik

:- module(dpll, [test_case/3,
		 dpll/3,
		 sat/2
		]).





:- use_module(library(assoc),
	      [
	       del_min_assoc/4,
	       gen_assoc/3,
	       get_assoc/3,
	       max_assoc/3,
	       empty_assoc/1,
	       put_assoc/4
	      ]).

:- use_module(library(lists),
	      [
	       append/3,
	       nth/3,
	       member/2,
	       reverse/2,
	       select/3
	      ]).
:- use_module(library(ordsets),
	      [
	       ord_add_element/3
	      ]).

/*
:- use_module(library(maplist),
	      [
	       maplist/3
	      ]).
*/



%===============================================================================

/*
Data structures with backtrackable assignments go into
parameters (set of clauses, current variable assigment).
Non-backtrackable data (positive and negatively watched
occurrences of a variable) go into the data base of
blackboard.
*/





/*

Global state represented by blackboard variables
================================================

Clause  index  = [0      ..  99999]
Variable index = [100000 .. 109999]
Literal  index = small integer
Value          = 0|1|undef     -- truth values

For i in Clause index :
            bb[i] : clause(W1,W2,Lits,Prod_Ctr)
             Lits : [Sign(Var)]
	     Sign : [+|-]
             Var  : Variable index
	     W1,W2: Literal index
	     Prod_ctr : small int
				% counts how many units the clause produces
	     

		     
For j in Variable index :
            bb[j]    = var(Pos_Watched,Neg_Watched,Pos_Occ_Count,Neg_Occ_Count)
            Pos_Watched, Neg_Watched : [occ(Clause_ length,Clause index,W)],
	    Pos_Occ_Count, Neg_Occ_Count : small int
	    W : 1|2   % which of the 2 pointers is it?


global variables:
            bb[log] : true | false   % logging?
	    bb[learning] : true | all | false
				% learning sometimes, always, or  off
	    
	    bb[learned]  : Lits
				% learned clause
	    
	    bb[freq_selection] : true | false
				% selection based on frequency of
				% occurrence
	    
	    bb[max_freq_pos], bb[max_freq_neg] : (Var,Nmb_occ)
                      % Var is a variable with maximal number
	              % Nmb_occ number of posive [negative]
	              % occurrences
	    
	    bb[nmb_clauses] : small int
				% current number of clause
				% including learned ones

	    
Local state
===========

Assigment, representing the current variable assigment,
  is a transient parameter that gets destructively
  updated by new variable assigments.
  Those updates are backtrackable.

Assignment       : [Var - Var_Info]
Var_Info         : var(Val,Source)
Val              : Value
Source           : unit(Clause,Depth) | branch(Depth)
Clause           : Clause index
Depth            : small int
		 
Source tells how the variable assignment was generated:
either from branching at depth Depth, or else from unit propagation
via clause Clause at banching depth Depth.
Source is used for learning from conflicts.


Formula stores information about the clause set.
*** obsolete:
[True_Count counts how many clauses are true at the moment,
and is to be compared with the total number of clauses
stored in Clause_Count, the current number of clauses not including
learned clauses. (These are derived and, hence, true in any model
of the other clauses).]

Formula: form(Search_Depth,Clause_Count,True_Count)
Clause_Count : small int
Search_depth, True_Count : *small int

		 


current variable assignment
Newly_assigned_var:  newly assigned var that has not yet been
  processed with regard to unit propagation and watching
  
*/


dpll(Formula,Assignment):-
	choose_undef_variable(Assignment,Var,Val,Freq),
	% variable assignment not yet complete
	% and no conflicting clause found,
	% hence the status is undetermined
	!,

	(
	  inc_search_depth(Formula,NextDepth),
	  log_choice(Var,NextDepth,Freq),
	  Source=branch(NextDepth),
	  Try=Val
	
	;
	  
				% we are back to previous search depth
	  search_depth(Formula,Depth),
	  log_backtrack(Var,Depth),
	  (bb_get(learned,C),C\==none ->
	      watchable_literals(C,Assignment,I,LI,K,LK),
	      (I\==0,I=K,LI=_-Var ->
				% learned clause forces propagation of
				% Var=Try which should be the complement of Val
		  signed_value(-,Val,Try),
		  new_clause_watching(C,I,LI,K,LK,CIndex),
		  bb_put(learned,none), % we've used it
		  Source=unit(CIndex,Depth)
	      ;I=0 ->
				% otherwise we need to reduce c.e. more
				% at the current depth
		  learning_from_conflict(Assignment,Depth,no_restrict,C),
				% get back to previous selection level
		  fail
	      ;% clause is not conflicting 
				% this can happen if learning is only partial
		  new_clause_watching(C,I,LI,K,LK,CIndex),
		  bb_put(learned,none), % we've used it
		  inc_search_depth(Formula,NextDepth),
				% try other value of Var
		  signed_value(-,Val,Try),
		  Source=branch(NextDepth)
	      )
	  ;
				% continue search with other value for Var
	      inc_search_depth(Formula,NextDepth),
	      signed_value(-,Val,Try),
	      Source=branch(NextDepth)
	  
	  )
	),
	extension(Var=Try,Source,Formula,Assignment),
	deducible_units([Var=Try],Formula,Assignment),
	dpll(Formula,Assignment).
dpll(_Formula,_Assignment). % success







sat(F,A):-
	statistics(runtime,[T,_]),
	sat_check(F,A),
	statistics(runtime,[T1,_]),
	Time is T1-T,
	format('[Time=~w msec]',[Time]).
	    
sat_check(F,A):-
	on_exception(restart,
		     on_exception(inconsistent,dpll(F,A),Status=fail),
		     Status=restart
		    ),

	(var(Status) -> 
	    format('~Nsatisfiable. ',[])
	;Status=fail ->
	    format('~Nunsatisfiable. ',[])
	;%Status=restart
	    format('Nrestart~n',[]),
	    inc(nmb_restarts,_R),
	    reset_assignment(A,A1),
	    F=form(Depth_ptr,_,_),
	    update_mutable(0,Depth_ptr),
	    bb_put(learning,all),
%	    trace,
	    sat_check(F,A1)
	).
	

/* =============================================================================

extension(+Var=+Val,+Source,+Formula,+Assignment)

assigns Val to Var in Assigment. Assigment contains an association
Var-var(V,Source), with V its (old) value.
V should normally be undef, that is, uninstantiated. V gets
assigned Val.

Source is information about how the unit was produced, either
from branching or from unit propagation..
*/


extension(Var=Val,Source,form(Depth_ptr,_,_True_Count_ptr),Assignment):-
	get_assoc(Var,Assignment,var(Val,Source)),
	get_mutable(Depth,Depth_ptr),
	log_extension(Var=Val,Source,Depth),
	!.


extensions([],_S,_F,_A).
extensions([Unit|Units],Source,F,A):-
	extension(Unit,Source,F,A),
	extensions(Units,Source,F,A).




/* =============================================================================

deducible_units(+Units,+Formula,+Assigment)
  Units: [Var=Val]

precondition:
all watched literal pointers, except the ones
pointing to literals Var xor Val, for Var=Val in Units.
These outdated pointers point 
to positive occurrences of Var when Val=0, and to
negative occurrences of Var, respectively, when Var=1,
point to literals that are true or undef in the Assigment.
The assigments in Units are already reflected in Assigment,
that is, if Var=Val is in Units, then Var-var(Val,_) is in
Assigment.

postcondition (if the predicate succeeds):
all watched literal pointers
point to literals that are true or undef in the Assigment
as we have it upon return.

side effect:
deducible_units moves those
watched pointers pointing to a false literal
when making the literals in Units true.
From this one may infer new units, which are
handled by calling deducible_units recursively.
Assigment gets updated by the newly inferred units.

deducible_unit, therefore,
fails if a watched literal pointer pointing
to a false literal cannot be moved to a true or undefined


*/



deducible_units([Var=Val|Units],F,A):-
	bb_get(Var,var(Pos_Watched,Neg_Watched,PO,NO)),
	(Val=1 ->
	    % move pointers in small clauses first
	    % finds smaller conflicting clauses
	    % hence Pos_Watched and Neg_Watched are ordered by clause length
	    move_ptr(F,A,Neg_Watched,Not_Moved,NewUnits),
	    bb_put(Var,var(Pos_Watched,Not_Moved,PO,NO))
	;
	    move_ptr(F,A,Pos_Watched,Not_Moved,NewUnits),
	    bb_put(Var,var(Not_Moved,Neg_Watched,PO,NO))
	),
	% if one watched literal pointer to be moved cannot be moved,
	% the assigment falsifies one of the clauses, hence we fail.
	Not_Moved=[],
	append(Units,NewUnits,U),
	deducible_units(U,F,A),
	!.
deducible_units([],_F,_A).



propagated_unit(A,C,Var,Val):-
	select(S-Var,C,C1),
	literal_value(A,S,Var,undef),
	!,
	signed_value(S,1,Val),
	false(A,C1).

false(_A,[]).
false(A,[S-V|Ls]):-
	literal_value(A,S,V,0),
	false(A,Ls).


/* =============================================================================

moving watched literal pointers

move_ptr(+Formula,+Assignment,+Occurrences,?Not_Moved,?New_Units)

expects in Occurrences the list of watched occurrences of a literal
L that has just become false by assigning the variable V
of L. Occurences are of the form

    occ(Clause length,Clause Index,I)
    
with I : 1|2 identifying which of the 2 wl pointers WI
it is that points to this occurrence.

move_ptr attempts to move WI in C to a true or undefined
literal. If the only such choice is the literal K to which
W(I+1) is pointing to, and if K is undefined, K is a new deducible
unit to be returned in New_Units.

If no K can be found, we have found a conflict, and then
move_ptr stops, returning, in Not_Moved, the subset
of occurrences of L that have not been moved. When finding
a conflict clause, learning is invoked, and if a clause
is learned it is returned in the blackboard variable
learned.

If Not_moved is not empty upon returing from move_ptr,
this indicates that a conflict clause was found. 

*/

move_ptr(F,A,[occ(_N,C,WhichW)|Cs],Not_Moved,NewUnits):-
	search_depth(F,Depth),
	bb_get(C,clause(W1,W2,Lits,Prod_Ctr)),
	(WhichW=1 ->
	    Move=W1, Other=W2, NewW1=NewIndex, NewW2=W2
	;
	    Move=W2, Other=W1, NewW1=W1, NewW2=NewIndex
	),
	(true_or_unit(A,Lits,Other,NewIndex,NewSign,NewVar,NewVal) ->
	    (Other=NewIndex, NewVal=undef ->
				% found a new deducible unit
		signed_value(NewSign,1,ForcedVal),
		extension(NewVar=ForcedVal,unit(C,Depth),F,A),
		New=[NewVar=ForcedVal],
				% we indcrease the productivity counter for C
		PC is Prod_Ctr+1
	    ;
		New=[],
		PC = Prod_Ctr
	    )
	;
				% C is a conflicting clause
	    log_conflict(Lits,Depth),
	    learning_from_conflict(A,Depth,restrict,Lits),
	    fail
	),
	bb_put(C,clause(NewW1,NewW2,Lits,PC)),
	length(Lits,N),
	add_to_watched_occurrences(NewVar,NewSign,occ(N,C,WhichW)),
	append(New,Units,NewUnits),
	move_ptr(F,A,Cs,Not_Moved,Units),
	!.
move_ptr(_F,_A,Cs,Cs,[]).


add_to_watched_occurrences(Var,Sign,Occ):-
	bb_get(Var,var(Pos_Watched,Neg_Watched,Pos_Occ,Neg_Occ)),
	(Sign= + ->
	    (member(Occ,Pos_Watched) ->
		PW=Pos_Watched
	    ;
		ord_add_element(Pos_Watched,Occ,PW)
	    ),
	    NW=Neg_Watched
	;
	    PW=Pos_Watched,
	    (member(Occ,Neg_Watched) ->
		NW=Neg_Watched
	    ;
		ord_add_element(Neg_Watched,Occ,NW)
	    )
	),
	bb_put(Var,var(PW,NW,Pos_Occ,Neg_Occ)),
	!.
add_to_watched_occurrences(Var,Sign,Occ):-
	% may happen if data base is corrupt and cont(Var,var(_,_,_,_)) fails
	raise_exception(add_to_watched_occurrences_failed(Var,Sign,Occ)).

add_to_watched_occurrences(Sign-Var,Occ):-
	add_to_watched_occurrences(Var,Sign,Occ).

inc_occ_counters([]).
inc_occ_counters([Sign-Var|Ls]):-
	inc_occ_counter(Var,Sign),
	inc_occ_counters(Ls).


inc_occ_counter(Var,Sign):-
	bb_get(Var,var(Pos_Watched,Neg_Watched,Pos_Occ,Neg_Occ)),
	(Sign= + ->
	    P is Pos_Occ+1,
	    N is Neg_Occ
	;
	    P is Pos_Occ,
	    N is Neg_Occ+1
	),
	bb_put(Var,var(Pos_Watched,Neg_Watched,P,N)),
	!.
	
/* 
true_or_unit(Assigment,Lits,Other,NewIndex,NewSign,NewVar,NewVal)

  succeeds if a NewIndex exists such that Lits[NewIndex] is undefined
  or true. If NewIndex can be chosen different from Other, such a choice
  is made. NewSign, NewVar, NewVal are, respectively, sign, variable, and
  value of the literal Lits[NewIndex]. 

  Failure means than a conflict was found.
*/

true_or_unit(A,Lits,Other,NewIndex,NewSign,NewVar,NewVal):-
	nth(NewIndex,Lits,NewSign-NewVar),
	(NewIndex=Other ->
	    bb_put(other_watched,NewSign-NewVar),
	    fail
	;
	    true
	),
	literal_value(A,NewSign,NewVar,NewVal),
	NewVal\==0,
	!.
true_or_unit(A,_Lits,Other,Other,NewSign,NewVar,NewVal):-
	% can only choose other watched literal, if at all
	bb_get(other_watched,NewSign-NewVar),
	literal_value(A,NewSign,NewVar,NewVal),
	NewVal\==0.


literal_value(Assigment,Sign,Var,Val):-
	get_assoc(Var,Assigment,var(VarVal,_)),
	(var(VarVal) ->
	    Val=undef
	;
	    signed_value(Sign,VarVal,Val)
	),
	!.


signed_value(+,V,V).
signed_value(-,0,1).
signed_value(-,1,0).
signed_value(-,undef,undef).




/* ============================================================================

Learning

C is a conflict clause. C contains a deduced literal
produced at the current search depth Depth.
Otherwise C would have been a propagator, hence not a conflict.
The learned clause is the one resulting from repeatedly
resolving all propagated units by ther producing clauses
at the current Depth.

*/

learned_clause(A,Depth,C,Learned):-
	select(Lit,C,C1),
	Lit=_-V,
	variable_source(V,A,unit(D,Depth)),
	% D is the producer of Lit
%	format('~Nresolving on ~w source ~w',[Lit,unit(D,Depth)]),
	resolve(Lit,C1,D,Resolvent),
	% proceed with next literal
	learned_clause(A,Depth,Resolvent,Learned),
	!.
learned_clause(_A,_D,C,C). % all units of current Depth resolved


resolve(Lit,C,DIndex,CDFactored):-
	complement(Lit,NLit),
	bb_get(DIndex,clause(_,_,D,_)),
%	format('~Nresolving ~w with ~w=~w',[Lit,DIndex,D]),
	select(NLit,D,D1),
	append(D1,C,CD),
	sort(CD,CDFactored).

variable_source(V,A,S):-
	get_assoc(V,A,var(_Val,S)).


max_depth_lit(_A,[],-1).
max_depth_lit(A,[_-V|Ls],M):-
	variable_source(V,A,S),
	(S=unit(_,Depth);S=branch(Depth)),
	max_depth_lit(A,Ls,M1),
	M is max(M1,Depth),
	!.

complement((+)-V,(-)-V).
complement((-)-V,(+)-V).


% interface
learning_from_conflict(A,Depth,Restrict,Lits):-
	(learn_more ->
	    learned_clause(A,Depth,Lits,Learned),
	    (Learned=[] ->
		log_learned(Learned,Depth),
		raise_exception(inconsistent)
	    ;
		length(Learned,N),
		(restricted_learn(Restrict,N) ->
		    log_learned(Learned,Depth),
		    bb_put(learned,Learned)
		;
		    bb_put(learned,none)
		)
	    )
	;
	    bb_put(learned,none)
	),
	!.


restricted_learn(Restrict,N):-
	bb_get(learning,all)
	;    Restrict=no_restrict
	;    N=<4.

learn_more:-
	\+bb_get(learning,false),
	bb_get(nmb_clauses,M),
	M=<15000.





/* ============================================================================

variable selection for branching

*/

% frequency-based selection
choose_undef_variable(Assigment,Var,Val,Freq):-
	bb_get(freq_selection,true),
	!,
	bb_put(max_freq_pos,(0,-1000000)),
	bb_put(max_freq_neg,(0,-1000000)),
	find_max_undef(Assigment),
	bb_get(max_freq_pos,(VarP,ValP)),
	bb_get(max_freq_neg,(VarN,ValN)),
	(ValP>ValN ->
	    Freq=ValP,
	    Var=VarP,
	    Val=0
	;
	    Freq=ValN,
	    Var=VarN,
	    Val=1
	),
	!,
	Var>0.
% select the first undefined variable in Assigment
choose_undef_variable(Assigment,Var,1,0):-
	gen_assoc(Var,Assigment,var(Val,_)),
	var(Val),
	!.

	
find_max_undef(Assigment):-
	gen_assoc(Var,Assigment,var(Val,_)),
	var(Val),
	bb_get(Var,var(_P,_N,A2,B2)),
%	length(_P,A1),
	A is A2,
%	length(_N,B1),
	B is B2,
	bb_get(max_freq_pos,(_,MaxP)),
	bb_get(max_freq_neg,(_,MaxN)),
	(A>MaxP ->
	    bb_put(max_freq_pos,(Var,A))
	;
	    true
	),
	(B>MaxN ->
	    bb_put(max_freq_neg,(Var,B))
	;
	    true
	),
	fail.
find_max_undef(_Assigment).




/* =============================================================================

Status determination

obsolete

*/

status(F,satisfiable):-
	F=form(_,NmbClauses,True_Count_ptr),
	get_mutable(NmbClauses,True_Count_ptr), % NmbClauses=True_Count
	!.
status(_,undetermined).


/* =============================================================================

Logging

*/



/* =============================================================================

Auxiliary

*/


store(K,V):-
	bb_put(K,V).

cont(K,V):-
	bb_get(K,V).


log_extension(Var=Val,Source,Depth):-
	(bb_get(log,true) ->
	    (Source=unit(CIndex,_) ->
		bb_get(CIndex,clause(_,_,C,_)),
		format('~N~*c|~w=~w produced by ~w~n',[Depth,0'-,Var,Val,C])
	    ;
		format('~N~*c|~w=~w~n',[Depth,0'-,Var,Val])
	    )
		
	;
	    true
	).

log_choice(Var,Depth,Freq):-
	(bb_get(log,true) ->
	    format('~N~*c|branching on ~w d=~w freq=~w~n',[Depth,0'-,Var,Depth,Freq])
	;
	    true
	).
	
log_backtrack(Var,Depth):-
	(bb_get(log,true) ->
	    format('~N~*c|backtracking on ~w~n',[Depth,0'-,Var])
	;
	    true
	).

log_learned(C,Depth):-
	(bb_get(log,true) ->
	    format('~N~*c|learned ~w~n',[Depth,0'-,C])
	;
	    true
	).

log_conflict(C,Depth):-
	(bb_get(log,true) ->
	    format('~N~*c|conflicting ~w~n',[Depth,0'-,C])
	;
	    true
	).

log_out_of_choices(Depth):-
	(bb_get(log,true) ->
	    format('~N~*c|out of choices~n',[Depth,0'-])
	;
	    true
	).


nat(1).
nat(I):-
	I is J+1,
	nat(J).

clause_number(I):-
	nat(I),
	bb_get(nmb_clauses,M),
	(I=<M ->
	    true
	;I>M ->
	    !,
	    fail
	).

nthrev(N,[_X|Xs],Y):-
	nthrev(M,Xs,Y),
	N is M+1.
nthrev(1,[X|_],X).
	
	





inc_search_depth(form(Depth_ptr,_,_),D):-
	get_mutable(Depth,Depth_ptr),
	D is Depth+1,
	update_mutable(D,Depth_ptr).

search_depth(form(Depth_ptr,_,_),Depth):-
	get_mutable(Depth,Depth_ptr).


inc(Key,NewVal):-
	bb_get(Key,Val),
	NewVal is Val+1,
	bb_put(Key,NewVal).




/* =============================================================================

Initialization

*/

dpll_init(F1,A1):-
	bb_put(var_offset,100000),
	bb_put(nmb_clauses,0),
	bb_put(learning,all),
	bb_put(freq_selection,true),
	bb_put(nmb_restarts,1),
	bb_put(max_restarts,2),
	empty_assoc(A1),
	F1=form(Depth,0,Ptr2zero),
	create_mutable(0,Ptr2zero),
	create_mutable(0,Depth).


reset_assignment(A,A1):-
	empty_assoc(A0),
	reset(A,A0,A1).

reset(A,A0,A1):-
	del_min_assoc(A,V,_,B),
	!,
	put_assoc(V,A0,var(_,_),A2),
	reset(B,A2,A1).
reset(_,A,A).

	







init_variable(Var,A,A1):-
	(get_assoc(Var,A,var(_Val,_Source)) ->
	    A=A1
	;
	    bb_put(Var,var([],[],0,0)),
	    put_assoc(Var,A,var(_UndefVal,_UndefSource),A1)
	).

init_variables_in_clause([],A,A).
init_variables_in_clause([_-Var|Ls],A,A1):-
	init_variable(Var,A,A2),
	init_variables_in_clause(Ls,A2,A1).



/* =============================================================================

Clause generation

*/
	
new_input_clause(C,F,A,F1,A1):-
	% does increase the clause count in the Formula data structure
	% and puts clause on occurrence lists of variables
	% also generates variables if they are new
	F=form(Depth,NClauses,TC),
	Clause_Index is NClauses+1,
	F1=form(Depth,Clause_Index,TC),
	normalized_vars(C,D),
	init_variables_in_clause(D,A,A1),
	new_clause(D,A1,_).


new_clause(D,A,Clause_Index):-
	% does not increase the clause count in the Formula data structure
	% does not enter clause on occurrence lists for variables
	% if needed this has to be done in addition
	% we don't need it for derived clauses (from learning)
	% we assume all variables in the clause exist
	% the main purpose is to find suitable watched literals
	watchable_literals(D,A,I,LI,K,LK),
	new_clause_watching(D,I,LI,K,LK,Clause_Index),
	!.
new_clause(D,A,Clause_Index):-
	raise_exception(new_clause_fails(D,A,Clause_Index)).


new_clause_watching(D,I,LI,K,LK,Clause_Index):-
	inc(nmb_clauses,Clause_Index),
	bb_get(var_offset,Offset),
	Clause_Index<Offset,  % otherwise clauses flow into variables
	length(D,N),
	bb_put(Clause_Index,clause(I,K,D,0)),
	process_variables(D,Clause_Index),
	add_to_watched_occurrences(LI,occ(N,Clause_Index,1)),
	add_to_watched_occurrences(LK,occ(N,Clause_Index,2)),
	inc_occ_counters(D),
	bb_get(nmb_restarts,R),
	bb_get(max_restarts,M),
	(R=<M,Clause_Index > R*3000 ->
	    raise_exception(restart)
	;
	    true
	).
	

watchable_literals(D,A,I,SI-VI,K,SK-VK):-
	% I and K will be different, if possible
	nth(I,D,SI-VI),
	literal_value(A,SI,VI,XI),
	XI\==0,
	nthrev(K,D,SK-VK),
	literal_value(A,SK,VK,XK),
	XK\==0,
	!.
% otherwise clause is conflicting; we return 0 for the pointer.
watchable_literals(_D,_A,0,_LI,0,_LK).
% 


new_input_clauses([],F,A,F,A).
new_input_clauses([C|Cs],F,A,F1,A1):-
	new_input_clause(C,F,A,F2,A2),
	new_input_clauses(Cs,F2,A2,F1,A1).
	

normalized_vars([],[]).
normalized_vars([V|Vs],[VN|VsN]):-
	normalized_var(V,VN),
	normalized_vars(Vs,VsN).

normalized_var(V,Sign-VIndex):-
	bb_get(var_offset,Offset),
	(V= -A ->
	    Sign= -
	;V=A ->
	    Sign= +
	),
	atom_chars(A,[_|Number]), % strip leading letter
	number_chars(N,Number),
	VIndex is N+Offset.

	

process_variables([],_).
process_variables([Sign-Var|Lits],Clause_index):-
	bb_get(Var,var(PW,NW,Pos_Occ,Neg_Occ)),
	(Sign= + ->
	    Pos_Occ1 is Pos_Occ+1,
	    Neg_Occ1=Neg_Occ
	;
	    Neg_Occ1 is Neg_Occ+1,
	    Pos_Occ1=Pos_Occ
	),
	bb_put(Var,var(PW,NW,Pos_Occ1,Neg_Occ1)),
	process_variables(Lits,Clause_index).





add_clause_on_occ_list(Clause_index,Occ,Occ1):-
	(member(Clause_index,Occ) ->
	    Occ1=Occ
	;
	    Occ1=[Clause_index|Occ]
	).





/* =============================================================================

test cases

*/


test_case(1,F,A):-
	dpll_init(F1,A1),
	Clauses=[[-p1,p2],[-p2,p3],[-p2,p1],[p1,p2,p3],[-p1,-p2,-p3]],
	new_input_clauses(Clauses,F1,A1,F,A).
test_case(2,F,A):-
	dpll_init(F1,A1),
	Clauses=[[-p1,p2],[-p2,p3],[-p3,p1],[p1,p2,p3],[-p1,-p2,-p3]],
	new_input_clauses(Clauses,F1,A1,F,A).
test_case(3,F,A):-
	dpll_init(F1,A1),
	ensure_loaded('../examples/tests/prop140-3'),
	problem(Clauses),
	new_input_clauses(Clauses,F1,A1,F,A).

