:-module(genericC,
	 [ simplify/2,
	   simplify_light/2,
	   simplify_med/2,
	   simplify_constraint_vars/3
	 ]).

:-use_module(library(sets),[select/3]).

:-use_module(termOrder,[greater/2]).

:-use_module(equationalTheory,[equal/2,unifiable/2]).

:-use_module(myterms,[apply_substitution/3]).

:-use_module(library(basics)).

:- use_module(timing, [
	timing/2
   ]).

:-use_module(constraintSolverBasics, 
	     [
	     satisfiable_ineq/1,
%	     simplify_constraint_vars/3,
	     true_ineq/1,
	     impliesA/2
   ]).

:- use_module(equationalTheory,[mgu/2]).
:- use_module(orderConstraints,[delete_implying_ands/2]).
:- use_module(library(change_arg)).

:- use_module(adt_term, [
	path_arg1/3,
	term_size/2
   ]).
:- use_module(thisProlog,[timeOut/3]).
:- use_module(hgstuff, [
	setof0/3
   ]).





%------------------------------------------------------------------------------
% CONSTRAINT SIMPLIFICATION
%------------------------------------------------------------------------------
simplify_constraint_vars(A,B,C):-
        constraintSolverBasics:simplify_constraint_vars(A,B,C).


% zu "andern: Konjunktionen als (saturierte) Module
% Identit"at modulo Kongruenz


simplify([conj([],[],[])],[conj([],[],[])]):-
	!.
simplify(C,C1):-
%	format('~n(',[]),ttyflush,
	timing(cons,simplify(2,C,C1)),
%	format(')',[]),ttyflush,
	!.


simplify_med(I,I1):-
	simplify(1,I,I1).

simplify_light(I,I1):-
	simplify(0,I,I1).






%------------------------------------------------------------------------------
% simplify( +Level, +C1, ?C2 ): 
% is true when C2 is a simplified form of C1. Level gives an indication
% of how hard should be worked on the simplification
%------------------------------------------------------------------------------
simplify(L,C,C2):-
	simplify1(L,C,C1),
	(member( conj([],[],[]), C1 ) ->
	    C2 = [conj([],[],[])]
	;   
	    delete_implying_ands( C1, C2 )
        ).


simplify1( L, [And|C], C1 ):-
	(simplified_conjunction(L,And,And1) ->
	     C1=[And1|CS]
	;    C1=CS),
	simplify1( L, C, CS ).
simplify1( _, [], [] ).


/*
simplified_conjunction(+Level,+C,?C1)
is true when C1 is an equivalent but somehow simplified form
of C and when C is potentially satisfiable wrt. some
chaining-based, incomplete method, where L is the
depth of saturation.
*/
simplified_conjunction( L, conj(I,E,O), conj(I2,U,O1) ):-
	(mgu( E, U ) ->
	     apply_substitution( U, I, I1 ),
	     apply_substitution( U, O, O1 )
	;
	     I1=I,
	     O1=O
	),
	partially_solved_form( L, I1, I2 ).


/*
partially_solved_form(+L,+I,?SF)
is true when SF is an approximation to a solved form of I;
is false when unsatisfiability of I is detected.
The solved form is produced by partial saturation (up to depth L) of I by
some form of chaining modulo the given ordering.
*/
partially_solved_form(_,[],[]).
partially_solved_form(L,I,SF):-
	depth_count(I,I1),   % initialize I for saturation
	on_exception(unsatisfiable,
		     (process_new_ineqs([],I1,I3),
		      depth_restricted_chaining(L,I3,I2)
		     ),fail),
	depth_count(SF1,I2),
	sort(SF1,SF).

depth_count([],[]).
depth_count([E|Es],[(D,_,E)|Fs]):-
	(var(D) -> D=0;true),
	depth_count(Es,Fs).


%pr(I1,I2):- 
%	format('~n~p->~p',[I1,I2]).
%	!.
%	length(I1,N1),length(I2,N2),
%	format('~p->~p;',[N1,N2]).


/*
depth_restricted_chaining(+Depth,+I,?I1)
is true when I1 is the saturation (up to depth Depth) of I.
Raises the exception unsatisfiable, when unsatisfiablity of
I is detected.
I and I1 are sets of triples (D,M,E), where D is the proof depth of
E, M is a mark that indicates, when uninstantiated, that the
inequation E is new. Chaining inferences are computed between
a new and all old inequations, after which the former inequation
becomes old.
*/
depth_restricted_chaining(_Depth,[X],[X]):-
!.
depth_restricted_chaining(Depth,I,I1):-
	member((D,New,E),I),
	var(New),  % really is new
	New=old,
	(D>=Depth ->
	     I2=I
	;    setof0(E2,genericC:
		         inferenceWithSome(Depth,(D,_,E),I,E2),NewIneqs),
	     process_new_ineqs(I,NewIneqs,I2)
	),
	depth_restricted_chaining(Depth,I2,I1).
depth_restricted_chaining(_Depth,I,I).




inferenceWithSome(Depth,(D,_,E),I,(D2,_,E2)):-
	member((D1,Old,E1),I),
%	E1\==E,  % useful?
	Old==old,
	D1<Depth,
	D2 is max(D,D1)+1,
	(chain(E,E1,E2);chain(E1,E,E2)).

	


process_new_ineqs(_,[(_,_,E1)|_],_):-
	\+satisfiable_ineq(E1),
	raise_exception(unsatisfiable).
process_new_ineqs(I,[(_,_,E1)|I1],I2):-
	true_ineq(E1),
	process_new_ineqs(I,I1,I2).
process_new_ineqs(I,[(_,_,E1)|I1],I2):-
	member((_,_,E),I),
	impliesA(E,E1),
	process_new_ineqs(I,I1,I2).
process_new_ineqs(I,[(D1,New1,E1)|I1],I2):-
	select((_,_,E),I,II),
	impliesA(E1,E),
	process_new_ineqs(II,[(D1,New1,E1)|I1],I2).
process_new_ineqs(I,[(D1,New1,E1)|I1],I2):-
	process_new_ineqs([(D1,New1,E1)|I],I1,I2).
process_new_ineqs(I,[],I).



chain(E,E1,E2):-
	functor(E,O,_),
	arg(1,E,L),
	arg(2,E,R),
	functor(E1,O1,_),
	arg(1,E1,L1),
	arg(2,E1,R1),
	!,
	(   %chain  L2 = k[L] >= k[R] >= k[SL1] = L1 >= R1 = R2
	    Middle=L1,
	    path_arg1(Pos,L1,SL1),
%	    SL1=L1,Pos=[],       % restricted case
	    (greater(R,SL1) -> O2= >
	    ;equal(R,SL1) -> compo(O,O1,O2)
	    ),
	    change_path_arg(Pos,L1,L2,L),
	    R2=R1
	;   % chain L2 = L >= R = k[SR] >= k[L1] >= k[R1] = R2
	    Middle=R,
	    path_arg1(Pos,R,SR),
%	    fail,                % restricted case 
	    Pos\==[],
	    (greater(SR,L1) -> O2= >
	    ;equal(L1,SR) -> compo(O,O1,O2)
	    ),
	    change_path_arg(Pos,R,R2,R1),
	    L2=L
	),
	(O2 = >=, equal(L2,R2),\+ (unifiable(Middle,L2);unifiable(Middle,R2)) ->
	     raise_exception(unsatisfiable)
	;    true),
	E2=..[O2,L2,R2],
	(satisfiable_ineq(E2) ->
	     true
	;    raise_exception(unsatisfiable)).
	    
	    

compo(>,>,>).
compo(>,>=,>).
compo(>=,>,>).
compo(>=,>=,>=).

	


