%declarations%
:- module(lpo, [ 
	greater/2,
	greater_ground/2,
	greaterEqual/2
   ]).

:- use_module(adt_precedence, [
	greater_op/2,
	is_first_predicate/1
   ]).

:- use_module(library(occurs), [contains_var/2]).

:- use_module(adt_term, [deBruijn_index/1,var1/1,term_size/2]).




%end%

%------------------------------------------------------------------------------
% greater( S, T ):
%	S is greater than T with respect to the lexicographic path ordering
%------------------------------------------------------------------------------


% new implementation

greater( S, _ ):-
	var1(S),
	!,
	fail.
greater( S, T ):-
	\+var1(T),
	!,
	functor(S,F,N),
	functor(T,G,M),
	greater(F,G,N,M,S,T).
greater( S, T ):-
	% T is a variable (of type var1)
	functor(S,F,_),
	(is_first_predicate(F) -> 
	     true
	;   \+deBruijn_index(T),
	    contains_var(T,S)),
	!.


greater(F,F,N,M,S,T):-
	!,
        (is_first_predicate(F) ->
             lexicographic1(0,S,N,T,M)
        ;
             lexicographic(0,S,N,T,M)
	).
greater(F,G,N,M,S,T):-
        (greater_op(F,G)->
             greater_than_list(S,0,T,M)

        ; % greater_op(G,F)
             some_subterm_greater_equal(0,S,N,T)
        ),
        !.



some_subterm_greater_equal(N,_S,N,_T):-
	!,
	fail.
some_subterm_greater_equal(I,S,N,T):-
	I1 is I+1,
	arg(I1,S,SI1),
	(greaterEqual( SI1, T) ->
	     true
	;
	     some_subterm_greater_equal(I1,S,N,T)
	).

	




lexicographic1(N,_,N,_,_M):-
	!,  % here : S[j]==T[j], for 1=<j=<N and N=<M
	fail.
lexicographic1(I,S,N,T,M):-
	% here : S[j]==T[j], for 1=<j=<I,  and I=<M and I<N
	I1 is I+1,
	arg(I1,S,SI),   % otherwise: N=<M
	(arg(I1,T,TI) ->
	     (SI==TI ->
		  lexicographic1(I1,S,N,T,M)
	     ;
	          greater(SI,TI)
	     )
	;% M<N
	     true
	).
	

lexicographic(N,_,N,_,_M):-
	!,  % here : S[j]==T[j], for 1=<j=<N and N=<M
	fail.
lexicographic(I,S,N,T,M):-
	% here : S[j]==T[j], for 1=<j=<I,  and I=<M and I<N
	I1 is I+1,
	arg(I1,S,SI),   % otherwise: N=<M
	(arg(I1,T,TI) ->
	     (SI==TI ->
		  lexicographic(I1,S,N,T,M)
	     ;greater(SI,TI) ->
		  greater_than_list(S,I1,T,M)
	     ;
                  some_subterm_greater_equal(I1,S,N,T)
	     )
	;% M<N     
	     true
	).

greater_than_list(_,N,_,N):-
	!.
greater_than_list(S,I,T,M):-
	I1 is I+1,
	arg(I1,T,TI),
	greater(S,TI),
	greater_than_list(S,I1,T,M).




%------------------------------------------------------------------------------
% greaterEqual( S, T ):
%	S is greater than T with respect lpo or S is equal to T
%------------------------------------------------------------------------------

greaterEqual(T1,T2):- T1 == T2,!.
greaterEqual(_T1,T2):- '$min' == T2,!.
greaterEqual(T1,T2):- 
	'$true$' == T2,
	T1\=='$min'.
greaterEqual(T1,T2):- greater(T1,T2),!.

%------------------------------------------------------------------------------
% maximal_term( S, T, M ):
%------------------------------------------------------------------------------

maximal_term( S, T, S ):- greaterEqual( S, T ), !.
maximal_term( S, T, T ):- greater( T, S ), !.

%------------------------------------------------------------------------------
% ground lpo: for ground completion in tautology detection
% it considers variables in the ordering as smallest Skolem constants
%
% hg: changed to considering variables biggest; reason: if we also want
% to apply unit clauses of the database for convergence proofs
% of succedents, more such units may become applicable if antecedents
% of the form  X=t, X variable, rewrite X into t, if X does not occur in t.
%
% hg: in order to avoid divergence of ground chaining for inequations
%     we need an ordering that has no ascending chains. Such an ordering
%     is obtained by combining the lpo lexicographically with the
%     term size.
%------------------------------------------------------------------------------



greater_ground(S,T):-
%       if we don't do subterm chaining for non-symmetric relations
%       we can do ground completion with arbitrary orderings
%	term_size(S,N),
%	term_size(T,M),
%	(N>M ->
%	     true
%	;M=N ->
	     gt_ground(S,T)
%	)
	.


gt_ground( S, T ):-
	ground_functor(S,F,N),
	ground_functor(T,G,M),
	gt_ground(F,G,N,M,S,T).

gt_ground(F,G,N,M,S,T):-
	F==G,
	!,
	(ground_is_first_predicate(F) ->
	     ground_lexicographic1(0,S,N,T,M)
	;    
	     ground_lexicographic(0,S,N,T,M)
	),
	!.
gt_ground(F,G,N,M,S,T):-
	(gt_ground_op(F,G)->
	     ground_greater_than_list(S,0,T,M)
	; % gt_ground_op(G,F)
             some_subterm_greater_equal_ground(0,S,N,T)
	),
	!.

	

some_subterm_greater_equal_ground(_,S,_,_):-
	var1(S),
	!,
	fail.
some_subterm_greater_equal_ground(N,_S,N,_T):-
	!,
	fail.
some_subterm_greater_equal_ground(I,S,N,T):-
	I1 is I+1,
	arg(I1,S,SI1),
	(geq_ground( SI1, T) ->
	     true
	;
	     some_subterm_greater_equal_ground(I1,S,N,T)
	).

	

ground_functor(T,T,0):-
	var(T),
	!.
ground_functor('$VAR'(N),'$VAR'(N),0):-
	!.
ground_functor(T,F,N):-
	functor(T,F,N).

ground_is_first_predicate(F):-
	nonvar(F),
	is_first_predicate(F).




ground_lexicographic1(N,_,N,_,_M):-
	!,
	fail.
ground_lexicographic1(M,_,N,_,M):-  % arity of first term greater  than second
	N>M.                        % for the case of variadic symbols
ground_lexicographic1(I,S,N,T,M):-
	I1 is I+1,
	arg(I1,S,SI),
	arg(I1,T,TI),
	(SI==TI ->
	     ground_lexicographic1(I1,S,N,T,M)
	;gt_ground(SI,TI) ->
	     true
	;
             some_subterm_greater_equal_ground(I1,S,N,T)
	).
	

ground_lexicographic(N,_,N,_,_M):-
	!,
	fail.
ground_lexicographic(M,_,N,_,M):- % arity of first term greater  than second
	N>M.                      % for the case of variadic symbols
ground_lexicographic(I,S,N,T,M):-
	I1 is I+1,
	arg(I1,S,SI),
	arg(I1,T,TI),
	(SI==TI ->
	     ground_lexicographic1(I1,S,N,T,M)
	;gt_ground(SI,TI) ->
	     ground_greater_than_list(S,I1,T,M)
	;
	     some_subterm_greater_equal_ground(I1,S,N,T)
	).

ground_greater_than_list(_,N,_,N):-
	!.
ground_greater_than_list(S,I,T,M):-
	I1 is I+1,
	arg(I1,T,TI),
	gt_ground(S,TI),
	ground_greater_than_list(S,I1,T,M).

geq_ground( T1, T2 ):- T1 == T2,!.
geq_ground( T1, T2 ):- gt_ground( T1, T2 ).

% ordering: first_predicates > variables > other symbols
gt_ground_op(F,G):-
	(var1(F) ->
	     (var1(G) -> F @> G; \+is_first_predicate(G))
	;var1(G) -> 
	     is_first_predicate(F)
	; 
	     greater_op(F,G)).

