:- module(lpoC, [ 
	simplify/2,
	simplify_light/2,
	simplify_med/2,
	simplify_constraint_vars/3
   ]).




:- use_module(lpo, [
	greater/2,
	greaterEqual/2
   ]).

:- use_module(proofState, [
        s_con/2,
	record_s_con/2
   ]).


:- use_module(props, [
        memo_constraints/1
   ]).

:- use_module(timing, [
	timing/2
   ]).



:- use_module(adt_precedence, [
	greater_op/2,
	is_first_predicate/1,
	is_predicate/1
   ]).

:- use_module(constraintSolverBasics, [
	implies/2,
	satisfiable_ineq/1
   ]).

:- use_module(equationalTheory, [
        mgu/2,
	unify/2
     ]).

:- use_module(orderConstraints, [
	true/1,
	delete_implying_ands/2

   ]).

:- use_module(myterms, [
	apply_substitution/3,
	list_of_vars/2,
	occurs_in/2,
	skolemized/3
   ]).


:- use_module(library(maplist), [
	maplist/3
   ]).

:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(sets), [
	select/3
   ]).


:- use_module(thisProlog, [tryAShortTime/1]).





%------------------------------------------------------------------------------
% CONSTRAINT SIMPLIFICATION
%------------------------------------------------------------------------------
simplify_constraint_vars(A,B,C):-
	constraintSolverBasics:simplify_constraint_vars(A,B,C).

%------------------------------------------------------------------------------
% simplify_constraint_light( C1, C2 ): 
% 	efficient method that simplifies constraints, although not totally, i.e.
% 	if C1 is insatisfiable, then C2 will not always be syntactically bottom
%------------------------------------------------------------------------------

simplify_light( [And|C], C1 ):-
	\+ can_be_satisfiable_and( And ),
	simplify_light( C, C1 ), !.
simplify_light( C, C ):- !.

can_be_satisfiable_and( conj(I,E,[]) ):-
	unif_list( E ),
	can_be_satisfiable_list_ineqs( I ).
can_be_satisfiable_and( conj(_I,_E,[_|_]) ).


can_be_satisfiable_list_ineqs( [] ).
can_be_satisfiable_list_ineqs( [E|R] ):-
	satisfiable_ineq(E),
	can_be_satisfiable_list_ineqs( R ).

unif_list( [] ).
unif_list( [S=T|E] ):-
	unify( S, T ),
	unif_list( E ).

%------------------------------------------------------------------------------
% simplify_constraint_med( C1, C2 ): 
% 	simplifies a constraint at medium level
% 	if C1 is insatisfiable, then C2 will not necessarily 
%	be syntactically bottom
%------------------------------------------------------------------------------

simplify_med( C, C ):- true( C ), !.
simplify_med( [], [] ).
simplify_med( C, [conj([],[],[])] ):-
	member( conj([],[],[]), C ), !.	
simplify_med( C, C2 ):- 
	simplify_med1( C, C1 ),
	(member( conj([],[],[]), C1 ) ->
	     C2 = [conj([],[],[])]
	;    delete_implying_ands( C1, C2 ) ), !.

simplify_med1( [], [] ).
simplify_med1( [A|C], C1 ):-
	simplified_and( A, A1 ), !,
	( A1 = conj([],[],[]) ->
	      C1 = [conj([],[],[])]
	;     simplify_med1( C, C2 ), 
	      C1 = [A1|C2]  ), !. 
simplify_med1( [_|C], C1 ):-
	simplify_med1( C, C1 ), !.

simplified_and( conj(I,E,O), conj(I3,U,O) ):-
	mgu( E, U ), 
	apply_substitution( U, I, I1 ),
%	maplist(transform_fp,I1,I11),
	partially_solved_form( I1, I1, I2 ),
	eliminate_implied_ineqs( I2, I3 ), !.

%------------------------------------------------------------------------------
% simplify( C1, C2 ): 
% 	simplifies a constraint totally
% 	e.g. if C1 is insatisfiable, then C2 will be syntactically bottom
%------------------------------------------------------------------------------


simplify( [conj([],[],[])], [conj([],[],[])] ):-
	!.
simplify( [], [] ):-
	!.
simplify( C, [conj([],[],[])]):-
	member( conj([],[],[]), C ), !.	
simplify( C, C2):-
	(memo_constraints(on) ->
	     normalize_constraint(C,CN), % to normalize a little before caching
	     skolemized(CN,CS,Subst),
	     hash_term(CS,Key),
	     (s_con(Key,(CS, CS1)) ->
%	          write('|'),ttyflush,
		  skolemized(C2,CS1,Subst)
	     ;
	          timing(cons,simplify_t(CN,C2)),
		  record_s_con(Key,(CN,C2))
	     )
	;
	     timing(cons,simplify_t(C,C2))
	),
	!.





simplify_nt(CN,C2):-
%	format('~N{working on constraint ~p',[CN]),ttyflush,
	simplify1(CN,C1),   
	(member( conj([],[],[]), C1 ) ->
	     C2 = [conj([],[],[])]
	;	 
	     delete_implying_ands( C1, C2 )
	).
%	write('}'),ttyflush,




simplify_t(C,C3):-
%       format('~N{working on constraint ~p',[C]),ttyflush,
%	simplify_med(C,CMS),

	CMS=C,
	(tryAShortTime(simplify1(CMS,C1)) ->
	     (member( conj([],[],[]), C1 ) ->
		  C3 = [conj([],[],[])]
	     ;	 
	          delete_implying_ands( C1, C3 )
	     )
	; 
%	     format('~Ntime-out on constraint ~p',[CMS]),
	     C3=CMS
	).
%	write('}'),ttyflush,
%	write(q),ttyflush,







normalize_constraint(C,C1):-
	maplist(normalize_c,C,C11),
	sort(C11,C1).

normalize_c(conj(Is,Es,Os),conj(IsS,EsS,OsS)):-
	sort(Is,IsS),
	sort(Os,OsS),
	sort(Es,EsS).







simplify1( [], [] ).
simplify1( [A|C], C1 ):-
	simplified_satisfiable_and( A, A1 ),
	( A1 = conj([],[],[]) ->
	      C1 = [conj([],[],[])] 
	; 
	      simplify1( C, C2 ), 
	      C1 = [A1|C2] ),
	!. 
simplify1( [_|C], C1 ):-
	simplify1( C, C1 ),
	!.
simplify1(C,_):-
	format('~NSimplify1 failed on ~p',[C]).




simplified_satisfiable_and( conj(I,E,O), conj(I3,U,O1) ):-
	mgu( E, U ),
	apply_substitution( U, I, I1 ),
	apply_substitution( U, O, O1 ),
%	maplist(transform_fp,I1,I11),
	partially_solved_form( I1,I1, I2 ),	
	eliminate_implied_ineqs( I2, I3 ),
	\+ (\+ satisfiable( 1,[],[],I3 )  ), !.


transform_fp(S>=T,S1>=T1):-
	transform_fp(S,T,S1,T1).
transform_fp(S>T,S1>T1):-
	transform_fp(S,T,S1,T1).

transform_fp(S,T,S1,T1):-
	nonvar(S),
	nonvar(T),
	functor(S,F,N),
	functor(T,G,M),
	is_first_predicate(F),
	N>0,
	is_first_predicate(G),
	M>0,
	!,
	S=..[_|Si],
	T=..[_|Ti],
	append(Si,[F],SiF),
	append(Ti,[G],TiF),
	S1=..['$c'|SiF],
	T1=..['$c'|TiF].
transform_fp(S,T,S,T).

% this simplification is dependent only on the properties of a partial
% ordering and hence can be used for any term ordering besides lpo
eliminate_implied_ineqs( I1, I2 ):- 
	sort( I1, I3 ),
	eliminate_implied_ineqs1( I3, I2 ), !.

eliminate_implied_ineqs1( I1, I2 ):-
	select( E, I1, I3 ),
	member( E1, I3 ),
	implies( E1, E ), 
%	new_implies( E1, E ), 
	eliminate_implied_ineqs1( I3, I2 ), !.
eliminate_implied_ineqs1( I, I ).



new_implies(E1,E):-
	delete_implied_ones([E1],[E],I),
	!,
	I=[].

%------------------------------------------------------------------------------
% satisfiable_and: checks whether an and of > and >= is satisfiable or not
%------------------------------------------------------------------------------
% old

satisfiable_and( [] ):- !. 
satisfiable_and( I ):- 
	one_solved_and( I, S ),
	list_of_vars( S, LV ),
	delete_vars_only_big_side( LV, S, LV1, S1 ),
	delete_vars_only_small_side( LV1, S1, LV2, S2 ),
	sat_solved_and( LV2, S2 ).

sat_solved_and( LV, S ):-
	\+ (( member( X, LV ), right_term_var( X, S ))), !.
sat_solved_and( LV, S1 ):-
	select( X, LV, LV1 ),
	right_term_var( X, S1 ),
	is_maximal_var( X, S1, S2, SMA, BIG, SEQ, BEQ ),
	combine_smaller_and_bigger( >,  SMA, BIG, S2, L1 ),
	combine_smaller_and_bigger( >,  SMA, BEQ, L1, L2 ),
	combine_smaller_and_bigger( >,  SEQ, BIG, L2, L3 ),
	combine_smaller_and_bigger( >=, SEQ, BEQ, L3, L4 ),
	combine_smaller_and_bigger( >,  LV1, BIG, L4, L5 ),
%	combine_smaller_and_bigger( >=, LV1, BEQ, L5, S3 ),
	combine_smaller_and_bigger( >=, LV1, BEQ, L5, L6 ),
%	sort(L6,S3),
	eliminate_implied_ineqs(L6,S3),
	satisfiable_and( S3 ), !.

right_term_var( X, S ):-
	member( I, S ),
	( I = (Y>_) ; I = (Y>=_) ),
	Y == X, !.

%------------------------------------------------------------------------------
% is_maximal_var( X, S1, S2, SMA, BIG, SEQ, BEQ ):
% 
%  Optimized method NR92 extended to include >= in solved forms.
%
% X is possibly maximal var if 
% 		it does not appear in Y>T[X] or Y>=T[X]_p  (p not lambda). 
%
% T[X]_p > Y  and  T[X]_p >= Y  disappear for all p including lambda 
% SMA contains terms T with X>T
% BIG contains terms T with T>X
% SEQ contains terms T with X>=T
% BEQ contains terms T with T>=X
% S2 is the set of literals in which X does not appear
%------------------------------------------------------------------------------

is_maximal_var( _, [], [], [], [], [], [] ).
is_maximal_var( X, [I|_], _, _, _, _, _ ):-		% X not maximal
	( I = (Y>T) ; (I = (Y>=T), X\==T) ),
	var( Y ),
	occurs_in( X, T ), !, fail.
is_maximal_var( X, [I|S], S1, SM, BI, SE, BE ):-	% T[X] >>= Y disappears 
	( I = (T>Y) ; I = (T>=Y) ),
	nonvar( T ),
	var( Y ),
	occurs_in( X, T ), !,
	is_maximal_var( X, S, S1, SM, BI, SE, BE ), !.
is_maximal_var( X, [Y>T |S], S1, [T|SM], BI, SE, BE ):-	%if X>T  keep T in SM
	X == Y, !,
	is_maximal_var( X, S, S1, SM, BI, SE, BE ), !.
is_maximal_var( X, [Y>=T|S], S1, SM, BI, [T|SE], BE ):-	%if X>=T keep T in SE
	X == Y, !,
	is_maximal_var( X, S, S1, SM, BI, SE, BE ), !.
is_maximal_var( X, [T>Y |S], S1, SM, [T|BI], SE, BE ):-	%if T>X  keep T in BI
	X == Y, !,
	is_maximal_var( X, S, S1, SM, BI, SE, BE ), !.
is_maximal_var( X, [T>=Y|S], S1, SM, BI, SE, [T|BE] ):-	%if T>=X keep T in BE
	X == Y, !,
	is_maximal_var( X, S, S1, SM, BI, SE, BE ), !.
is_maximal_var( X, [I|S], [I|S1], SM, BI, SE, BE ):-	%other I's  with no X
	is_maximal_var( X, S, S1, SM, BI, SE, BE ), !.




%------------------------------------------------------------------------------
% delete_vars_only_big_side:
% detects variables that appear only in big sides and deletes these inequalities.
% Attention: after deleting one such variable, it must try again previous ones:
% that's how it is done here.
% It is not possible that after deleting vars_only_small, new vars_only_big
% can be deleted, but the other way round it can happen.
%------------------------------------------------------------------------------

delete_vars_only_big_side( LV, S, LV1, S1 ):-
	select( X, LV, LV2 ),
	delete_big_side( X, S, S2 ), !,
	delete_vars_only_big_side( LV2, S2, LV1, S1 ), !.
delete_vars_only_big_side( LV, S, LV, S ):- !.

delete_big_side( _, [], [] ).
delete_big_side( X, [I|_], _ ):-
	( I = (_>T) ; I = (_>=T) ),
	occurs_in( X, T ), !, fail. 
delete_big_side( X, [I|S], S1 ):-
	( I = (T>_) ; I = (T>=_) ),
	occurs_in( X, T ), !, 
	delete_big_side( X, S, S1 ),!.
delete_big_side( X, [I|S], [I|S1] ):-
	delete_big_side( X, S, S1 ),!.

%------------------------------------------------------------------------------
% delete_vars_only_small_side:
% detects variables appearing only in small sides and replaces them by the
% smallest constant.
% After deleting one such variable, it must try again previous ones.
%------------------------------------------------------------------------------

delete_vars_only_small_side( LV, S, LV1, S1 ):-
	select( X, LV, LV2 ),
	delete_small_side( X, S, S2 ), !,
	X = '$min', !,
	delete_vars_only_small_side( LV2, S2, LV1, S1 ), !.
delete_vars_only_small_side( LV, S, LV, S ):- !.

delete_small_side( _, [], [] ).
delete_small_side( X, [I|_], _ ):-
	( I = (T>_) ; I = (T>=_) ),
	occurs_in( X, T ), !, fail. 
delete_small_side( X, [I|S], S1 ):-
	( I = (_>T) ; I = (_>=T) ),
	X == T, !, 
	delete_small_side( X, S, S1 ),!.
delete_small_side( X, [I|S], [I|S1] ):-
	delete_small_side( X, S, S1 ),!.


combine_smaller_and_bigger( _, _, [], L, L ):- !.
combine_smaller_and_bigger( Rel, SM, [T|BI], L1, L2 ):-
	relate_term_list( Rel, T, SM, L1, L3 ),
	combine_smaller_and_bigger( Rel, SM, BI, L3, L2 ), !.

relate_term_list( _, _, [], L, L ).
relate_term_list( >,  T, [S|L], L1, [T>S |L2] ):-
	relate_term_list( >, T, L, L1, L2 ), !.
relate_term_list( >=, T, [S|L], L1, [T>=S|L2] ):-
	relate_term_list( >=, T, L, L1, L2 ).


%------------------------------------------------------------------------------
% partially_solved_form:
%------------------------------------------------------------------------------

partially_solved_form( _, [], [] ):-!.
partially_solved_form( I, [S>T|R], R1 ):- 
	greater( S, T ), !, 
	partially_solved_form( I, R, R1 ), !.
partially_solved_form( I, [S>=T|R], R1 ):- 
	greaterEqual( S, T ), !, 
	partially_solved_form( I, R, R1 ), !.
partially_solved_form( I, [S>T|_], _ ):-
	implies(I,T>=S),
	!,
%	greaterEqual( T, S ), !, 
	fail.
partially_solved_form( I, [S>=T|_], _ ):-
	implies(I,T>S),
	!,
%	greater( T, S ), !, 
	fail.
partially_solved_form( I, [T1>T2|R], R1 ):-
        nonvar(T1),
        nonvar(T2),
        T1 =.. [F1|Args1],
        T2 =.. [F2|Args2],
        project_to_elem( >, I, T1, F1, Args1,  T2, F2, Args2,  L),
	!,
        partially_solved_form( I, L, L2 ), 
        partially_solved_form( I, R, R2 ),
        user:append( L2, R2, R1 ), !.
partially_solved_form( I, [T1>=T2|R], R1 ):-
        nonvar(T1),
        nonvar(T2),
        T1 =.. [F1|Args1],
        T2 =.. [F2|Args2],
        project_to_elem( >=, I, T1, F1, Args1,    T2, F2, Args2,  L),
	!,
        partially_solved_form( I, R, R2 ), 
        partially_solved_form( I, L, L2 ), 
        user:append( L2, R2, R1 ), !.
partially_solved_form( I, [D|R], [D|R1] ):-
        partially_solved_form( I, R, R1 ), !.


project_to_elem( _, _I, T1,F1,_,    _,F2,Args2,  L ):-
        greater_op( F1, F2 ),
	!,
        greater_term_list( T1, Args2, L ).
project_to_elem( _, _I, _,F1,Args1,    T2,F2,_, [A >= T2] ):-
        greater_op( F2, F1 ),
	!,
	delete_smaller_ones( T2, Args1, [A] ).
project_to_elem( >, I, T1, F, L1,    T2, F, L2,   L ):-
	create_lex_list( I, T1, T2, L1, L2, L ), !.
project_to_elem( >=, I, T1, F, L1,    T2, F, L2,   L ):-
	\+ unify( T1, T2 ), !,
	create_lex_list( I, T1, T2, L1, L2, L ),
	!.
project_to_elem( >=, _I, T1, F, L1,    _, F, L2,   L ):-
	create_lex_list_geq( T1, L1, L2, L ), !.

					% Let a > b >= c
create_lex_list( I ,_, T, [S1,S2], [T1,T2], [S1>=T1] ):-   	
        % f(x,a) > f(y,b) => x>=y  [no!]
        % f(x,a) > f(y,b) & f(y,b)> a => x>=y
	implies(I,S2>T2),
	implies(I,T>S2),
	!.
%	greater( S2, T2 ), !.
create_lex_list( _I, _, _, [S1,S2], [T1,T2], [S1>T1] ):-   	
	% f(x,z) > f(y,z) => x>y   [ z>=f(y,z) is impossible
        %                            x> f(y,z) implies x>y]
	S2 == T2, 
	!.
create_lex_list( _, _, _, [S1,S2], [T1,T2], [S1>T1] ):-   	
        % f(x,y) > f(y,x) => x>y        [ok] 
	(S1>S2) == (T2>T1), !.
create_lex_list( I, S, _,[S1,S2], [T1,T2], [S1>T1, S>T2] ):-	
        % f(x,c)>f(y,b) => x>y & f(x,c)>b
	% [in particular x>=f(y,b) implies x>y and f(x,c)>b]
	implies(I,T2>=S2),
	!.
%	greaterEqual( T2, S2 ), !.				
create_lex_list( I, _, T, [S1,S2], [T1,_T2], [S2>=T] ):-	
        % f(b,x)>f(a,y) -> x>=f(a,y)
	implies(I,T1>S1),
	!.
create_lex_list( _, _, _, [S], [T], [S>T] ):- !.		% f(x)>f(y) => x>y 
create_lex_list( I, T1, T2, [S|RS], [T|RT], L ):-		% f(x,Y)>f(x,Z) => 
	S == T, !,						% f(Y)>f(Z)
	create_lex_list( I, T1, T2, RS, RT, L ), !.


create_lex_list_geq( _, [], [], [] ):- !.
create_lex_list_geq( _, AS, AT, [S>=T] ):-
	delete_synt_equals( AS, AT, [S], [T] ), !.

delete_synt_equals( [], [], [], [] ):- !.
delete_synt_equals( [S|RS], [T|RT], S1, T1 ):-
	S == T, !,
	delete_synt_equals( RS, RT, S1, T1 ), !.
delete_synt_equals( [S|RS], [T|RT], [S|S1], [T|T1] ):-
	delete_synt_equals( RS, RT, S1, T1 ), !.

greater_term_list( _, [], [] ).
greater_term_list( T, [S|L], [T>S|L1] ):-
	greater_term_list( T, L, L1 ).

greatereq_list_term( [], _, [] ).
greatereq_list_term( [S|L], T, [S>=T|L1] ):-
	greatereq_list_term( L, T, L1 ).

delete_smaller_ones(  _, [], [] ).
delete_smaller_ones( T, [S|R], R1  ):-
	greater( T, S ), !,
	delete_smaller_ones( T, R, R1 ), !.
delete_smaller_ones( T, [S|R], [S|R1] ):-
	delete_smaller_ones( T, R, R1 ), !.



%------------------------------------------------------------------------------
% one_solved_and: computes one possible solved_form of an and
% unifying for elimination of equalities.
%------------------------------------------------------------------------------


one_solved_and( [], [] ):-!.
one_solved_and( [ T1>T2 |_], _ ):-
	greaterEqual( T2, T1 ), !, fail.
one_solved_and( [ T1>=T2 |_], _ ):-
	greater( T2, T1 ), !, fail.
one_solved_and( [ T1>T2 |R], I ):-
	greater( T1, T2 ), !,
	one_solved_and( R, I ).
one_solved_and( [ T1>=T2 |R], I ):-
	greaterEqual( T1, T2 ), !,
	one_solved_and( R, I ).
one_solved_and( [T1>T2|R], I ):-
	( var( T1 ) ; var( T2 ) ),!,
	one_solved_and( R, I1 ),
	reconsider( T1>T2, I1, I ).
one_solved_and( [T1>=T2|R], I ):-
	( var( T1 ) ; var( T2 ) ),!,
	one_solved_and( R, I1 ),
	reconsider( T1>=T2, I1, I ).
one_solved_and( [T1>T2|R], I ):-
	T1 =.. [F1|Args1],
	T2 =.. [F2|Args2], !,
	solved_elem( >,  T1, F1, Args1, T2, F2, Args2, I1 ),
	one_solved_and( R, I2 ),
	user:append( I1, I2, I3 ),
	one_solved_and( I3, I ).
one_solved_and( [T1>=T2|R], I ):-
	T1 =.. [F1|Args1],
	T2 =.. [F2|Args2],
	solved_elem( >=, T1, F1, Args1, T2, F2, Args2, I1 ),
	one_solved_and( R, I2 ),
	user:append( I1, I2, I3 ),
	one_solved_and( I3, I ).

reconsider( T1 >  T2, _, _ ):-
	greaterEqual( T2, T1 ), !, fail.
reconsider( T1 >= T2, _, _ ):-
	greater( T2, T1 ), !, fail.
reconsider( T1 >  T2, I, I ):-
	greater( T1, T2 ), !.
reconsider( T1 >= T2, I, I ):-
	greaterEqual( T1, T2 ), !.
reconsider( T1 >  T2, I, [T1>T2|I] ):-
	(var(T1);var(T2)),!.
reconsider( T1 >= T2, I, [T1>=T2|I] ):-
	(var(T1);var(T2)),!.
reconsider( T1 >  T2, I1, I  ):-
	one_solved_and( [T1>T2|I1],  I ).
reconsider( T1 >= T2, I1, I  ):-
	one_solved_and( [T1>=T2|I1], I ).


	% New definition for solved_elem:

solved_elem( _, T1,F1,_,    _, F2,Args2, L):-
	greater_op( F1, F2 ),
	greater_term_list( T1, Args2, L1 ),!,
	one_solved_and( L1, L ).
solved_elem( _, _,F1,Args1, T2,F2,_,    L):-
	greater_op( F2, F1 ), !,
	lors_d3( Args1, T2, L1 ),
	one_solved_and( L1, L ).
solved_elem( >, T1, F, Args1, T2, F, Args2, L):-!,
	(lors_d4( T1, Args1, Args2, L1 ) ; lors_d3( Args1, T2, L1 ) ),
	one_solved_and( L1, L ).
solved_elem( >=, T1, F, Args1, T2, F, Args2, L):-!,
	(lors_d4_geq( T1, Args1, Args2, L1 ); lors_d3( Args1, T2, L1 )), % hg
%	lors_d4_geq( T1, Args1, Args2, L1 ),
	one_solved_and( L1, L ).










lors_d3( Args, T, []    ):- 
	member( E, Args ), 
	unify_constraint( E, T ).
lors_d3( Args, T, [E>T] ):- 
	member( E, Args ).

lors_d4( T1, [E1|_], [E2|Args2], [E1>E2|L] ):-
	greater_term_list( T1, Args2, L ).
lors_d4( T1, [E1|Args1], [E2|Args2], [E1>=E2|L] ):-
	lors_d4( T1, Args1, Args2, L ).

lors_d4_geq( _, [], [], [] ).
lors_d4_geq( T1, [E1|_], [E2|Args2], [E1>E2|L] ):-
	greater_term_list( T1, Args2, L ).
lors_d4_geq( T1, [E1|Args1], [E2|Args2], [E1>=E2|L] ):-
	lors_d4_geq( T1, Args1, Args2, L ).

%------------------------------------------------------------------------------
% Is like unify, but does not unify variables with p(...) if p is a predicate.
%------------------------------------------------------------------------------

unify_constraint( T1, T2 ):- T1 == T2, !.
unify_constraint( X, Y ):- 
	var( X ),
	var( Y ),
	!,
	X = Y.
unify_constraint( X, T ):- 
	var( X ), !, 
	functor( T, F, N ), 
	\+ is_predicate( F/N ),  
	unify( X, T ),
	\+ occurs_in( X, T ),
	!.
unify_constraint( T, X ):- 
	var( X ), !, 
	functor( T, F, N ), 
	\+ is_predicate( F/N ),  
	unify( X, T ),
	!.
unify_constraint( T1, T2 ):- 
	T1 =.. [F|Args1],  
	T2 =.. [F|Args2], !,
	unify_constraintlist( Args1, Args2 ), !.
unify_constraintlist( [], [] ):-!.
unify_constraintlist( [T1|R1], [T2|R2] ):-
	unify_constraint( T1, T2 ), 
	unify_constraintlist( R1, R2 ), !.




% ======= new: satisfiable( Old, New, Tra )


satisfiable( _, Old, New, _ ):- New \== [], append(Old,New,ON), cycle(ON), !,fail.
satisfiable( _, Old, [], _ ):- cycle(Old), !,fail.
satisfiable( _, _, [], [] ):- !.
satisfiable( N, Old, New, [I|Tra] ):- !,
	solved_form_atom( [], I, L ),
	treat_eqs( Old, New, Tra, L, L1, Old1, New1, Tra1 ),
	add_non_implied( Old1, L1, New1, New2, Tra1, Tra2 ),
	bottom_test(Tra2,Tra3,Old1,New2),
	N1 is N+1,
	satisfiable( N1, Old1, New2, Tra3 ).
satisfiable( N, Old, [I|New], [] ):- !,
	delete_implied_ones(I,Old,Old1),
	trans(I,Old1,Tra),!,
	sort(Tra,Tra1),!,
	bottom_test(Tra1,Tra2,[I|Old1],New),
	N1 is N+1,
	satisfiable( N1, [I|Old1], New, Tra2 ).


bottom_test([],[],_,_):-!.
bottom_test([S>T|_],_,Old,New):-
	greq_lpo_implied(Old,New,T,S),!,fail.
bottom_test([S>T|Tra1],Tra2,Old,New):-
	greater_lpo_implied(Old,New,S,T),!,
	bottom_test(Tra1,Tra2,Old,New),!.
bottom_test([S>=T|_],_,Old,New):-
	greater_lpo_implied(Old,New,T,S),!,fail.
bottom_test([S>=T|Tra1],Tra2,Old,New):-
	greq_lpo_implied(Old,New,S,T),!,
	bottom_test(Tra1,Tra2,Old,New),!.
bottom_test([I|Tra1],[I|Tra2],Old,New):-
	bottom_test(Tra1,Tra2,Old,New),!.


bottom([S>T|_]):-
	greaterEqual(T,S),!.
bottom([S>=T|_]):-
	greater(T,S),!.
bottom([_|L]):-
	bottom(L),!.


add_non_implied( _, [], L, L, T, T).
add_non_implied( _, [S>T|_], _,_,_,_ ):-
	greaterEqual(T,S),!,fail.
add_non_implied( _, [S>=T|_], _,_,_,_ ):-
	greater(T,S),!,fail.
add_non_implied( Old, [S>T|L], New1, New2, Tra1, Tra2 ):-
	greater_lpo_implied(Old,New1,S,T),!,	
	add_non_implied( Old, L, New1, New2, Tra1, Tra2 ),!.
add_non_implied( Old, [S>=T|L], New1, New2, Tra1, Tra2 ):-
	greq_lpo_implied(Old,New1,S,T),!,	
	add_non_implied( Old, L, New1, New2, Tra1, Tra2 ),!.
add_non_implied( Old, [S>T|L], New1, [S>T|New2], Tra1, Tra2 ):-
	(var(S);var(T)),!,
	add_non_implied( Old, L, New1, New2, Tra1, Tra2 ),!.
add_non_implied( Old, [S>=T|L], New1, [S>=T|New2], Tra1, Tra2 ):-
	(var(S);var(T)),!,
	add_non_implied( Old, L, New1, New2, Tra1, Tra2 ),!.
add_non_implied( Old, [I|L], New1, New2, Tra1, [I|Tra2] ):-
	add_non_implied( Old, L, New1, New2, Tra1, Tra2 ),!.


treat_eqs( Old, New, Tra, [], [], Old, New, Tra ).
treat_eqs( Old, New, Tra, [X=T|L], L1, Old1, New1, Tra1 ):-
	var(X),!,
	remove_x(X,Old,Old2,Tra,Tra2),
	remove_x(X,New,New2,Tra2,Tra3),
	unify(X,T),
	treat_eqs( Old2, New2, Tra3, L, L1, Old1, New1, Tra1 ),!.
treat_eqs( Old, New, Tra, [T=X|L], L1, Old1, New1, Tra1 ):-
	remove_x(X,Old,Old2,Tra,Tra2),
	remove_x(X,New,New2,Tra2,Tra3),
	unify(X,T),
	treat_eqs( Old2, New2, Tra3, L, L1, Old1, New1, Tra1 ),!.
treat_eqs( Old, New, Tra, [I|L], [I|L1], Old1, New1, Tra1 ):-
	treat_eqs( Old, New, Tra, L, L1, Old1, New1, Tra1 ),!.


remove_x(_,[],[],Tra,Tra).
remove_x(X,[I|L1],L2,Tra1,[I|Tra2]):-
	is_side(X,I),!,
	remove_x(X,L1,L2,Tra1,Tra2),!.
remove_x(X,[I|L1],[I|L2],Tra1,Tra2):-
	remove_x(X,L1,L2,Tra1,Tra2),!.

is_side( X, Y >  _ ):-X==Y,!.
is_side( X, Y >= _ ):-X==Y,!.
is_side( X, _ >  Y ):-X==Y,!.
is_side( X, _ >= Y ):-X==Y,!.


delete_implied_ones(_,[],[]):-!.
delete_implied_ones(I,[S>T|Old],Old1):-
	greater_lpo_implied([I],[],S,T),!,
	delete_implied_ones(I,Old,Old1),!.
delete_implied_ones(I,[S>=T|Old],Old1):-
	greq_lpo_implied([I],[],S,T),!,
	delete_implied_ones(I,Old,Old1),!.
delete_implied_ones(I,[I1|Old],[I1|Old1]):-
	delete_implied_ones(I,Old,Old1),!.


trans( _, [], [] ).
trans( I1, [I2|A], [I3|L] ):-
	( comb(I1,I2,I3); comb(I2,I1,I3) ), !,
	trans( I1, A, L ),!.
trans( I, [_|A], L ):-
	trans( I, A, L ),!.


comb( A>B,  C>D,  A>D  ):- !, B==C, var(B),!.
comb( A>B,  C>=D, A>D  ):- !, B==C, var(B),!.
comb( A>B,  C=D,  A>D  ):- !, B==C, var(B),!.

%comb( A>B,  C>D,  A>D  ):- !, nonvar(A), B==C, var(B),!.
%comb( A>B,  C>=D, A>D  ):- !, nonvar(A), B==C, var(B),!.
%comb( A>B,  C=D,  A>D  ):- !, nonvar(A), B==C, var(B),!.

comb( A>=B, C>D,  A>D  ):- !, B==C, var(B),!.
comb( A>=B, C>=D, A=B  ):- B==C, A==D, !.
comb( A>=B, C>=D, A>=D ):- !, B==C, var(B),!.
comb( A>=B, C=D,  A>=D ):- !, B==C, var(B),!.

comb( A=B,  C>D,  A>D  ):- !, B==C, var(B),!.
comb( A=B,  C>=D, A>=D ):- !, B==C, var(B),!.

%=================================================================

cycle(A):- 
	strict_elim(A,A1),
	sort(A1,A2),
	\+ unif_vars(A2),!.


% elimina ineqs con vars a la dcha y reemplaza >= por >.

strict_elim([],[]).
strict_elim([_>X|A],A1):-
	var(X),!,
	strict_elim(A,A1),!.
strict_elim([_>=X|A],A1):-
	var(X),!,
	strict_elim(A,A1),!.
strict_elim([S>=T|A],[S>T|A1]):-
	strict_elim(A,A1),!.
strict_elim([I|A],[I|A1]):-
	strict_elim(A,A1),!.


x_part( X, [Y>T|A], [T|L], Rest ):-
	X==Y,!,
	x_part( X, A, L, Rest ),!.
x_part( _, A, [], A ):-!.


unif_vars([]).
unif_vars([X>T|A]):-
	x_part(X,A,P,Rest),
	list_of_vars( [T|P], L ),
	unify(X,L),
	unif_vars(Rest),!.


invars( X, Y ):- X==Y, !.
invars( X, T ):-
	nonvar(T),
	functor(T,_,N),
	invars_list( X, T, N).

invars_list( _, _, 0):-!,fail.
invars_list( X, T, N):-
	arg(N,T,Tn),
	invars(X,Tn),!.
invars_list( X, T, N):-
	N1 is N-1,
	invars_list( X, T, N1),!.


%unify(S,T):- unify_with_occurs_check(S,T).


%=================================================================

solved_form_atom( _, S>T,  _ ):-	  greaterEqual( T, S ), !, fail.
solved_form_atom( L, S>T,  L  ):-	  greater( S, T ), !.
solved_form_atom( L, S>T,  [S>T|L] ):-    (var(S);var(T)),!.

solved_form_atom( _, S>=T, _ ):-	  greater( T, S ), !, fail.
solved_form_atom( L, S>=T, L   ):-	  greaterEqual( S, T ), !.
solved_form_atom( L, S>=T, [S>=T|L] ):-   (var(S);var(T)),!.

solved_form_atom( _, S=T,  _ ):-	  \+ unify(S,T), !, fail.
solved_form_atom( L, S=T,  L   ):-	  S==T,!.
solved_form_atom( L, S=T,  [S=T|L] ):-    (var(S);var(T)),!.


solved_form_atom( L1, S=T,  L2 ):-
	functor(S,F,N),
	functor(T,F,_),
        solved_eq( S,N,T,L1, L2 ),!.
solved_form_atom( L1, S>T,  L2 ):-
	functor(S,FS,NS),
	functor(T,FT,NT),
        one_solved_gr( S,FS,NS, T,FT,NT, L1, L2 ).
solved_form_atom( L1, S>=T, L2 ):-
	functor(S,FS,NS),
	functor(T,FT,NT),
        one_solved_greq( S,FS,NS, T,FT,NT, L1, L2 ).


solved_eq( _,0,_,L, L ):-!.
solved_eq( S,N,T,L1, L2 ):-
	arg( N, S, Sn ),
	arg( N, T, Tn ),
	solved_form_atom( L1, Sn=Tn, L3 ),
	N1 is N-1,
	solved_eq( S,N1,T,L3, L2 ),!.


one_solved_gr( S,FS,_,  T,FT,N,  L1, L2 ):-
	greater_op( FS, FT ),!,
	solved_gr_than_all_args( S, T, 1, N, L1, L2 ).
one_solved_gr( S,F,N,   T,F,_,   L1, L2 ):-
	solved_lexicographic_gr( S, T, N, L1, L2 ).
one_solved_gr( S,_,N,   T,_,_,   L1, L2 ):-
	solved_one_arg_greq( S, T, N, L1, L2 ).


one_solved_greq( S,FS,_,  T,FT,N,  L1, L2 ):-
	greater_op( FS, FT ),!,
	solved_gr_than_all_args(   S, T, 1, N, L1, L2 ).
one_solved_greq( S,F,N,   T,F,_,   L1, L2 ):-
	solved_lexicographic_greq( S, T, N, L1, L2 ).
one_solved_greq( S,_,N,   T,_,_,   L1, L2 ):-
	solved_one_arg_greq( S, T, N, L1, L2 ).


solved_lexicographic_gr( S, T, 1, L1, L2 ):-!,
	arg( 1, S, S1 ),
	arg( 1, T, T1 ),
	solved_form_atom( L1, S1>T1, L2 ).
solved_lexicographic_gr( S, T, 2, L1, L2 ):-!,
        arg( 1, S ,S1 ),
        arg( 1, T ,T1 ),
        arg( 2, S ,S2 ),
        arg( 2, T ,T2 ),
        args_gr_lex_two( S,T, S1,S2,T1,T2,L1, L2 ).
solved_lexicographic_gr( S, T, N, L1, L2 ):-
	solved_lexicographic_gr1( S, T, 1, N, L1, L2 ).


solved_lexicographic_gr1( S, T, N1,  N2, L1, L2 ):-
	N1 =< N2,
	arg( N1, S, Sn ),
	arg( N1, T, Tn ),
	solved_form_atom( L1, Sn>Tn, L3 ),
	N3 is N1+1,
	solved_gr_than_all_args( S, T, N3, N2, L3, L2 ).
solved_lexicographic_gr1( S, T, N1,  N2, L1, L2 ):-
	N1 < N2,
	arg( N1, S, Sn ),
	arg( N1, T, Tn ),
	solved_form_atom( L1, Sn=Tn, L3 ),
	N3 is N1+1,
	solved_lexicographic_gr1( S, T, N3, N2, L3, L2 ).


args_gr_lex_two( _, _, S1,S2,T1,T2,L1,L2 ):-
        S1 == T1, !, 
        solved_form_atom( L1, S2>T2, L2 ).
args_gr_lex_two( _, _, S1,S2,T1,T2,L1, L2 ):-
        S2 == T2, !, 
        solved_form_atom( L1, S1>T1, L2 ).
args_gr_lex_two( S, _, S1,_,T1,T2,L1, L2 ):-
        greater( S1, T1 ), !,
        solved_form_atom( L1, S>T2, L2 ).
args_gr_lex_two( _, T, S1,S2,T1,_,L1, L2 ):-
        greater( T1, S1 ), !,
        solved_form_atom( L1, S2>=T, L2 ).
args_gr_lex_two( _, _, S1,S2,T1,T2,L1,L2 ):-
        S1 == S2,
        T1 == T2,!,
        solved_form_atom( L1, S1>T1, L2 ).
args_gr_lex_two( _, T, S1,S2,T1,T2,L1, L2 ):-
        greater( T, S2 ),
        greater( S2, T2 ), !,
        solved_form_atom( L1, S1>=T1, L2 ).
args_gr_lex_two( S, _, S1,S2,T1,T2,L1, L2 ):-
        greater( S, T2 ),
        greater( T2, S2 ), !,
        solved_form_atom( L1, S1>T1, L2 ).
args_gr_lex_two( S, T, _,_,_,_, L1, L2 ):-
	solved_lexicographic_gr1( S, T, 1, 2, L1, L2 ).


solved_lexicographic_greq( S, T, 1, L1, L2 ):-!,
	arg( 1, S, S1 ),
	arg( 1, T, T1 ),
	solved_form_atom( L1, S1>=T1, L2 ).
solved_lexicographic_greq( S, T, 2, L1, L2 ):-!,
        arg( 1, S ,S1 ),
        arg( 1, T ,T1 ),
        arg( 2, S ,S2 ),
        arg( 2, T ,T2 ),
        args_greq_lex_two( S,T, S1,S2,T1,T2,L1, L2 ).
solved_lexicographic_greq( S, T, N, L1, L2 ):-
	solved_lexicographic_greq1( S, T, 1, N, L1, L2 ).


args_greq_lex_two( _, _, S1,S2,T1,T2,L1,L2 ):-
        S1 == T1, !, 
        solved_form_atom( L1, S2>=T2, L2 ).
args_greq_lex_two( _, _, S1,S2,T1,T2,L1, L2 ):-
        S2 == T2, !, 
        solved_form_atom( L1, S1>=T1, L2 ).
args_greq_lex_two( S, _, S1,_,T1,T2,L1, L2 ):-
        greater( S1, T1 ), !,
        solved_form_atom( L1, S>T2, L2 ).
args_greq_lex_two( _, T, S1,S2,T1,_,L1, L2 ):-
        greater( T1, S1 ), !,
        solved_form_atom( L1, S2>=T, L2 ).
args_greq_lex_two( _, T, S1,S2,T1,T2,L1, L2 ):-
        greater( T, S2 ),
        greater( S2, T2 ), !,
        solved_form_atom( L1, S1>=T1, L2 ).
args_greq_lex_two( S, _, S1,S2,T1,T2,L1, L2 ):-
        greater( S, T2 ),
        greater( T2, S2 ), !,
        solved_form_atom( L1, S1>T1, L2 ).
args_greq_lex_two( S, T, _,_,_,_, L1, L2 ):-
	solved_lexicographic_greq1( S, T, 1, 2, L1, L2 ).


solved_lexicographic_greq1( S, T, N1,  N2, L1, L2 ):-
        N1==N2, !,
	arg( N1, S, Sn ),
	arg( N1, T, Tn ),
	solved_form_atom( L1, Sn>=Tn, L2 ).
solved_lexicographic_greq1( S, T, N1,  N2, L1, L2 ):-
	arg( N1, S, Sn ),
	arg( N1, T, Tn ),
	solved_form_atom( L1, Sn>Tn, L3 ),
	N3 is N1+1,
	solved_gr_than_all_args( S, T, N3, N2, L3, L2 ).
solved_lexicographic_greq1( S, T, N1,  N2, L1, L2 ):-
	arg( N1, S, Sn ),
	arg( N1, T, Tn ),
	solved_form_atom( L1, Sn=Tn, L3 ),
	N3 is N1+1,
	solved_lexicographic_greq1( S, T, N3, N2, L3, L2 ).


solved_one_arg_greq( _, _, 0, _,  _  ):- !, fail.
solved_one_arg_greq( S, T, 1, _,  _  ):- 
	functor(S,F,_),
	functor(T,F,_),
	!, fail.
solved_one_arg_greq( S, T, N, L1, L2 ):-
	arg( N, S, Sn ),
	solved_form_atom( L1, Sn>=T, L2 ).
solved_one_arg_greq( S, T, N, L1, L2 ):-
	N1 is N-1,
	solved_one_arg_greq( S, T, N1, L1, L2 ).



solved_gr_than_all_args( _, _, N1, N2, L,  L  ):- N1>N2, !.
solved_gr_than_all_args( S, T, N1, N2, L1, L2 ):-
	arg( N1, T, Tn ),
	solved_form_atom( L1, S>Tn, L3 ),
	N3 is N1+1,
	solved_gr_than_all_args( S, T, N3, N2, L3, L2 ).


% =============== CONSTRAINT IMPLIES LPO INEQ ===============================


greater_lpo_implied( C1, _, S, T ):- member212(S>T,C1),!.
greater_lpo_implied( _, C2, S, T ):- member212(S>T,C2),!.
greater_lpo_implied( C1, _, S, T ):-
		var(S),
		member22(I,C1),
		(I=(S1>T1);I=(S1>=T1)),
		S==S1,
		is_proper_subterm(T,T1),!.
greater_lpo_implied( _, C2, S, T ):-
		var(S),
		member22(I,C2),
		(I=(S1>T1);I=(S1>=T1)),
		S==S1,
		is_proper_subterm(T,T1),!.
greater_lpo_implied( _,  _, S, _ ):- var(S),!,fail.
greater_lpo_implied( C1, C2,  S, T ):-
                nonvar(T),
                S =.. [F|Args1],
                T =.. [F|Args2],
                gr_lexicographic_implied( C1, C2, S, Args1, Args2 ),!.
greater_lpo_implied( C1, C2, S, T ):-
                nonvar(T),
                S =.. [F|_],
                T =.. [G|Args2],
		F \== G,
                greater_op(F, G), !,
                greater_than_list_implied( C1, C2, S, Args2), !.
greater_lpo_implied( C1, C2, S, T ):-
                S =.. [_|Args1],
                member22( S1, Args1 ),
                greq_lpo_implied( C1, C2, S1, T), !.

greq_lpo_implied( _,  _, S, T ):- S==T,!.
greq_lpo_implied( C1, _, S, T ):- member22(I,C1), (I==(S>T);I==(S>=T)),!.
greq_lpo_implied( _, C2, S, T ):- member22(I,C2), (I==(S>T);I==(S>=T)),!.
greq_lpo_implied( C1, _, S, T ):-
		var(S),
		member22(I,C1),
		(I=(S1>T1);I=(S1>=T1)),
		S==S1,
		is_proper_subterm(T,T1),!.
greq_lpo_implied( _, C2, S, T ):-
		var(S),
		member22(I,C2),
		(I=(S1>T1);I=(S1>=T1)),
		S==S1,
		is_proper_subterm(T,T1),!.
greq_lpo_implied( _, _, S, _ ):- var(S),!,fail.
greq_lpo_implied( C1, C2, S, T ):-
                nonvar(T),
                S =.. [F|_],
                T =.. [G|Args2],
		F \== G,
                greater_op(F, G), !,
                greater_than_list_implied( C1, C2, S, Args2), !.
greq_lpo_implied( C1, C2, S, T ):-
                nonvar(T),
                S =.. [F|Args1],
                T =.. [F|Args2],
                greq_lexicographic_implied( C1, C2, S, Args1, Args2 ),!.
greq_lpo_implied( C1, C2, S, T ):-
                S =.. [_|Args1],
                member22( S1, Args1 ),
                greq_lpo_implied( C1, C2, S1, T), !.


gr_lexicographic_implied( C1, C2, S, [S1|_], [T1|R2] ):-
                greater_lpo_implied( C1, C2, S1, T1 ),
                greater_than_list_implied( C1, C2, S, R2 ),!.
gr_lexicographic_implied( C1, C2, S, [S1|R1], [T1|R2]):-
                S1 == T1,
                gr_lexicographic_implied( C1, C2, S, R1, R2 ), !.


greq_lexicographic_implied( _,  _, _, [], _ ):- !.
greq_lexicographic_implied( C1, C2, S, [S1|_], [T1|R2] ):-
                greater_lpo_implied( C1, C2, S1, T1 ),
                greater_than_list_implied( C1, C2, S, R2 ),!.
greq_lexicographic_implied( C1, C2, S, [S1|R1], [T1|R2]):-
                S1 == T1, !,
                greq_lexicographic_implied( C1, C2, S, R1, R2 ), !.
greq_lexicographic_implied( C1, C2, S, [S1|R1], [T1|R2]):-
                greq_lpo_implied(C1,C2,S1,T1),
                greq_lexicographic_implied( C1, C2, S, R1, R2 ), !.


greater_than_list_implied( _,  _, _, [] ):- !.
greater_than_list_implied( C1, C2, S, [T|R] ):-
                greater_lpo_implied( C1, C2, S, T ),
                greater_than_list_implied( C1, C2, S, R ),!.

is_proper_subterm(S,T):-
	nonvar(T),
	T=..[_|Args],
	member22(T1,Args),
	is_subterm(S,T1),!.

is_subterm(S,T):-S==T,!.
is_subterm(S,T):-
	nonvar(T),
	T=..[_|Args],
	member22(T1,Args),
	is_subterm(S,T1),!.


guapo([]).
guapo(I):- \+ (\+ guapo1(I) ),!.

guapo1(I):- numbervars(I,0,_), write_term(I,[numbervars(true)]),nl,!.


guapos( L ):- \+ (\+ guapos1(L)),!.
guapos1( L ):- numbervars(L,0,_), guapos2(L),!.

guapos2([]).
guapos2( [nl|L] ):- !, nl,guapos1(L),!.
guapos2( [X|L] ):- write_term(X,[numbervars(true)]),
	guapos2(L),!.


% =============== LPO =====================================

/*
greater_lpo( S, _ ):-var(S),!,fail.
greater_lpo( S, T ):-
                nonvar(T),
                S =.. [F|Args1],
                T =.. [F|Args2],
                lexicographic( S, Args1, Args2 ),!.
greater_lpo( S, T ):-
                nonvar(T),
                S =.. [F|_],
                T =.. [G|Args2],
		F \== G,
                greater_op(F, G), !,
                greater_than_list(S, Args2), !.
greater_lpo( S, T ):-
                S =.. [_|Args1],
                member22( S1, Args1 ),
                grequal_lpo( S1, T), !.

lexicographic( S, [S1|_], [T1|R2] ):-
                greater_lpo( S1, T1 ),
                greater_than_list( S, R2 ),!.
lexicographic( S, [S1|R1], [T1|R2]):-
                S1 == T1,
                lexicographic( S, R1, R2 ), !.

greater_than_list( _, [] ):- !.
greater_than_list( S, [T|R] ):-
                greater_lpo( S, T ),
                greater_than_list( S, R ),!.

grequal_lpo(T1,T2):- T1 == T2,!.
grequal_lpo(T1,T2):- greater_lpo(T1,T2),!.

*/



%_________________________________________________________________________
%
%	AUXILIARY PREDICATES
%_________________________________________________________________________

member22(X,[X|_]).
member22(X,[_|L]):- member22(X,L).

member212(X,[Y|_]):- X == Y,!.
member212(X,[_|L]):- member212(X,L),!.

delete32(X,[X|L],L).
delete32(X,[Y|L],[Y|L1]):- delete32(X,L,L1).

concat312( [],L,L):-!.
concat312( [X|L1],L2,[X|L3]):- concat312(L1,L2,L3),!.

concat33( [],L,L):-!.
concat33( [X|L1],L2,[X|L3]):- concat33(L1,L2,L3),!.

delete_all( _, [], [] ):- !.
delete_all( X, [Y|TL], L ):- X == Y,!, delete_all( X, TL, L ).
delete_all( X, [Y|TL], [Y|L]):- delete_all( X, TL, L ).

delete_one32( X, [X|L], L ).
delete_one32( X, [Y|L], [Y|L1] ):- delete_one32( X, L, L1 ).

delete_one312( X, [Y|L], L ):- X == Y, !.
delete_one312( X, [Y|L], [Y|L1] ):- delete_one312( X, L, L1 ).

delete_repeats( [], [] ):- !.
delete_repeats( [X|R], R1 ):-
	member22( X1, R ),
	X == X1, !,
	delete_repeats( R, R1 ).
delete_repeats( [X|R], [X|R1] ):-
	delete_repeats( R, R1 ), !.

/*
%-------------------------------------------------
% list of variables of a prolog structure
%-------------------------------------------------

list_of_vars( T, L ):-
        list_vars( T, L1 ),
        sort( L1, L ),!.

list_vars( X, [X] ):- var(X), !.
list_vars( X, [] ):- atomic(X), !.
list_vars( [X|R], L ):-
        list_vars( X, L1 ),
        list_vars( R, L2 ),
        append( L1, L2, L ), !.
list_vars( T, L ):-
        T =.. [_|Args],
        list_vars( Args, L ),!.

*/

%--------------------------------------------------------------------------------
% gets all the operators of a given prolog data expression
%--------------------------------------------------------------------------------

get_prolog_data_operators( T, L2 ):-
	get_prolog_data_operators1( T, L1 ),
	sort( L1, L2 ), !.

get_prolog_data_operators1( T, [] ):-
	var( T ), !.
get_prolog_data_operators1( T, [] ):-
	T == [], !.
get_prolog_data_operators1( T, [T] ):-
	atomic( T ), !.
get_prolog_data_operators1( [C|R], L3 ):-
	get_prolog_data_operators1( C, L1 ),
	get_prolog_data_operators1( R, L2 ),
	append( L1, L2, L3 ), !.
get_prolog_data_operators1( T, [Op|L] ):-
	T =.. [Op|R],
	get_prolog_data_operators1( R, L ), !.


% eof --------------------------------------------------------------------



