/* OO 


This file contains the predicates required for compiling
class definitions.


Objects are records represented by lists.
The order of instance variables is such that instance variables of 
superclasses come first.

Objects are represented in clauses within a class definition C
by terms of the form
       self(Context,IV) or myself(Context,IV).
Hereby IV is the list of instance variable values according to the number
of instance variables newly declared in C.
Context is a pair (Super,Sub), where Super and Sub hold the instance variables
from super and subclasses, respectively.
These terms are at compile-time translated into the list 
       Super@IV@Sub  or Super@IV, respectively,
consisting all instance variables of an object. The myself variant
refers to objects of C excluding objects of subclasses. It is mainly
used in methods that create objects.

In order to make code in class C applicable to objects of subclasses C1
of C it is mandatory that patterns of the form self((Super,Sub),IV)
do not constrain Sub. Then Super@IV@Sub becomes an open-ended list
matching more specific objects as well.


Methodological considerations imply that
Super should not be constrained either.
That means that both Super and Sub should usually be variables,
and then they simply name object parts without constraining them.
If self(X,IV1) and self(X,IV2) occur in one clause, then the second
term denotes an object that is just like the first except for the values
of the local instance variables IV for C-objects.

The declarations for classes with the number of instance variables
for their objects together with the subclass relationship are relevant for
compiling [my]self(Context,IV) terms into lists of appropriate length
in which the IV section occurs at the right position.
This is achieved by term_expansion clauses in oo.pl.

Class membership of an object is determined at runtime by inspecting the
length of this list. This is implicit in that clauses
for a predicate p written in class C, hence mentioning
only objects names or object representations of the form
[my]self(Context,IV), necessarily fail if they are applied to objects
of superclasses of C as these have fewer instance variables.
(They don't necessarily fail for objects in subclasses if
Context is unconstrained.)
This implies that each class must introduce at least one
new instance variable.

If a predicate p has clauses in more than one class,
one wants clauses in subclasses to be tried first.
This implies that subclasses must textually preceed their superclasses
which is automatically achievde by term_expansion.
Cuts may be used to avoid application of clauses in superclasses
when this is not appropriate. Otherwise, a superclass may
contain the more general clauses while subclasses contain more specific
cases of a predicate definition.


*/

:-module(classExpansion, [ex/0,objects_version/1]).


:- use_module(library(maplist),[maplist/2,maplist/3]).
:- use_module(library(sets), [
	select/3
   ]).
:- use_module(library(basics), [
	member/2
   ]).

:- use_module(mylists, [
	is_list/1,
	rev/2
   ]).

%:- use_module(user,_,[append/3]).

:- use_module(auxiliary,[appendAtoms/3]).

:- use_module(thisProlog,[prolog_version/1]).


objects_version(o).  % old

ex:- 
	Classes=[
		% definition of the class hiearchy


	object,
%	  literal,
	  simpleClause,
	    clause,
	      rule

		],



	rev(Classes,RClasses),

	(retract(user:dontExpand);true),
	expandHeader(rule),

	open('classesCode.pl',write,Str),
	expandClasses(Str,RClasses),
	close(Str),
	format('~n classesCode written.~n',[]).




expandClasses(_,[]).
expandClasses(Str,[C|Cs]):-
	expandClass(Str,C),
	expandClasses(Str,Cs).


expandHeader(File):-
	appendAtoms([File,'pl'],'.',F),
	see(F),
	repeat,
	read(T),
	(   evalClassDef(T)
	;
	    T=end_of_file
	),
	seen,
%	close(F),
	!.

	     

evalClassDef((:- class(ClassName,SuperClass,IV,Methods))):-
	class(ClassName,SuperClass,IV,Methods).

expandClass(Str,File):-
	appendAtoms([File,'pl'],'.',F),
%	format(Str,'~n Clauses for ~p~3n',[File]),
	see(F),
	repeat,
	read(T),
	(classExpansion(T,E) ->
	     true
	;
	     trace,
	     classExpansion(T,E)
	),
	portray_clause_for_module(E,Str),
	T=end_of_file,
	seen,
	!.





classExpansion(X,X):-
	var(X),
	!.
classExpansion(end_of_file,(:-true)).
classExpansion(self((Super,Sub),Local),All):-
	!,
	currentClass(Class),
	classDefinition(Class,SuperClass,IVC,_Methods),
	length(IVC,M),
	length(Local,M),
	CVars=..[Class,Local],
	instanceVariables(SuperClass,All-[CVars|Sub],Super-[]).
classExpansion(self(Local),All):-
	!,
	currentClass(Class),
	classDefinition(Class,SuperClass,IVC,_Methods),
	length(IVC,M),
	length(Local,M),
	CVars=..[Class,Local],
	instanceVariables(SuperClass,All-[CVars|_Sub],_Super-[]).
classExpansion(myself(Super,Local),All):-
	!,
	currentClass(Class),
	classDefinition(Class,SuperClass,IVC,_Methods),
	length(IVC,M),
	length(Local,M),
	CVars=..[Class,Local],
	instanceVariables(SuperClass,All-[CVars],Super-[]).
classExpansion((:- class(ClassName,_SuperClass,_IV,_Methods)),(:- true)):-
	!,
	(retract(currentClass(_));true),
	asserta(currentClass(ClassName)).
classExpansion(T,T1):-
	T=..[F|Ts],
	maplist(classExpansion,Ts,Ts1),
	T1=..[F|Ts1].






% instanceVariables(+C,?IV1, ?IV2)
% is true when IV1 and IV2 are two copies of the instance variable structure of
% objects of class C. These are represented by difference lists
instanceVariables(C,IV-Last,IV1-Last1):-
	classDefinition(C,S,_IVC,_Methods),
%	length(_IVC,M),
%	length(CopyIVC,M),
%	CVars=..[C|CopyIVC],
	CVars=..[C,_],
	(nonvar(S) ->
	     instanceVariables(S,IV-[CVars|Last],IV1-[CVars|Last1])
	;    
	IV=[CVars|Last],
	IV1=[CVars|Last1]
	).




% Actions to declare a class and load its superclasses
% before reading the class definition itself
class(ClassName,SuperClass,IV,Methods):-
	addClass(ClassName,SuperClass,IV,Methods),
	(var(SuperClass) ->
	     true
	;atom(SuperClass) ->
	     seen, % close previous input and read super class header
	     expandHeader(SuperClass)
	;true).


addClass(ClassName,SuperClass,IV,Methods):-
	(atom(ClassName) ->
	     true
	;raise_exception(error(className,ClassName))),
	((atom(SuperClass);var(SuperClass)) ->
	     true
	;raise_exception(error(className,SuperClass))),
	(is_list(IV), maplist(atom,IV) -> 
	     true
	;raise_exception(error(ivars,IV))),
	(is_list(Methods), maplist(predSpec,Methods) -> 
	     true
	;raise_exception(error(methods,Methods))),

	(retract(classDefinition(ClassName,_,_,_));true),
	assert(classDefinition(ClassName,SuperClass,IV,Methods)).




% auxiliary



predSpec(P):-
	nonvar(P),
	P=..[/,F,A],
	atom(F),
	integer(A),
	A>=0.
/*
pair(X,P,X1):-
	var(P),
	(prolog_version(sicstus),exportedPredicate(X) ->
	    X1=P
	;
	    X1=X,
	    assert(exportedPredicate(X))
	)
	,
	!.
pair(X,P,X1):-
	(prolog_version(sicstus),exportedPredicate(X) ->
	    X1=P
	;
	    X1=(X,P),
	    assert(exportedPredicate(X))
	)
	,
	!.
*/


% error Message generation through message hook
:- multifile user:generate_message_hook/3.       % two implicit arguments in DCGs

user:generate_message_hook(error(className,Y)) -->
	['Class names must be atoms: ~p'-[Y],nl].
user:generate_message_hook(error(ivars,Y)) -->
	['Instance variables must be lists of atoms: ~p'-[Y],nl].
user:generate_message_hook(error(methods,Y)) -->
	['Method specifications must be lists of predicate declarations: ~p'-[Y],nl].



portray_clause_for_module((:-mymultifile B),Str):-
	!,   % SICStus 3#5 bug
	format(Str,'~N~q.~n',(:-mymultifile B)).
portray_clause_for_module((Module:H:-B),Str):-
	!,
	write(Str,Module),
	write(Str,':'),
	(prolog_version(sicstus3) ->
	     portray_clause(Str,(H:-B))
	;
	     format(Str,'~N~q.~n',(H:-B))
	).
portray_clause_for_module(Module:H,Str):-
	!,
	write(Str,Module),
	write(Str,':'),
	(prolog_version(sicstus3) ->
	     portray_clause(Str,H)
	;
	     format(Str,'~N~q.~n',H)
	),
	portray_clause(Str,H).
portray_clause_for_module(C,Str):-
	(prolog_version(sicstus3) ->
	     portray_clause(Str,C)
	;
	     format(Str,'~N~q.~n',C)
	).




