
:- module(orderConstraints, [ 
	and/3,
	and_eq/3,
	and_ge/3,
	and_gt/3,
	and_list/2,
	and_not_eq/3,
	and_other/3,
	constraint_operators/2,
	delete_implying_ands/2,
	false/1,
	implies/2,
	implies_s/2,
	implies_light/2,
	nonstrict_inequalities_as_disjunction/2,
	not/2,
	or/3,
	or_eq/3,
	or_ge/3,
	or_gt/3,
	or_list/2,
	or_not_eq/3,
	or_other/3,
	print_constraint/1,
	project/3,
	true/1,
	valid_s/1
   ]).

:- use_module(constraintSolver, [
	simplify/2
   ]).

:- use_module(termOrder, [
        greater/2,
        greaterEqual/2
   ]).

:- use_module(props, [
        flag_on/2
   ]).

:- use_module(auxiliary, [
	get_prolog_data_operators/2,
	member212/2
   ]).

:- use_module(mylists, [
	delete/3
   ]).

:- use_module(library(maplist), [
	maplist/2,
	maplist/3
   ]).

:- use_module(library(basics), [
	member/2
   ]).


:- use_module(library(sets), [
	select/3
   ]).


:- use_module(myterms, [
	list_of_vars/2,
	deskolemize/2
   ]).

:- use_module(adt_term, [
	extraVars/2
   ]).

:- use_module(hgstuff, [
	split/4
   ]).







%	constraint handling

%------------------------------------------------------------------------------
%  constraint ::- [and]
%  and ::- conj( [ineq_geq], [eq], [other] )
%  ineq_geq ::- ineq | geq
%  ineq ::- term>term
%  geq ::- term>=term
%  eq ::- term=term
%  other ::- atom
%
%------------------------------------------------------------------------------

false( [] ).

true( [ conj([],[],[]) ] ).

or([], C, C ):- 
	!.
or(C, [], C):-
	!.
or(C1, C2, C4):- 
	user:append( C1, C2, C3 ),
	delete_implying_ands(C3,C4).

or_gt( 	C, [S,T],  [ conj([S>T],[],[])  |C] ).
or_ge( 	C, [S,T],  [ conj([S>=T],[],[]) |C] ).
or_eq( 	C, [S,T],  [ conj([],[S=T],[])  |C] ).
or_not_eq( C, [S,T],  [ conj([S>T],[],[]), conj([T>S],[],[])  |C] ).
or_other(C, Other,  [ conj([],[],[Other])|C] ).

	
and_gt( [], _, [] ).
and_gt( [conj(D,E,O)|Rest], [S,T], [conj([S>T|D],E,O)|C1] ):-
	and_gt( Rest, [S,T], C1 ).


and_ge( [], _, [] ).
and_ge( [conj(D,E,O)|Rest], [S,T], [conj([S>=T|D],E,O)|C1] ):-
		and_ge( Rest, [S,T], C1 ).


and_eq( [], _, [] ).
and_eq( [conj(D,E,O)|Rest], [S,T], [conj(D,[S=T|E],O)|C1] ):-	
	and_eq( Rest, [S,T], C1 ).

and_not_eq( C1, [S,T], C2 ):-		% indep of the repr.
	true( C3 ),
	and_gt( C3, [S,T], C4 ),
	or_gt(  C4, [T,S], C5 ),
	and1( C1, C5, C2 ).             % no simplification

and_other([],_,[]).
and_other([conj(D,E,O)|Rest],Other,[conj(D,E,[Other|O])|C1]):-
	and_other( Rest, Other, C1 ).



and( [ conj([],[],[]) ], C, C1 ):-
	simplify( C, C1 ),
	!.
and( C, [ conj([],[],[]) ], C ):-
	!.
and( C1, C2, Conj ):-
	and1( C1, C2, Conj1 ),
	simplify( Conj1, Conj ),
	!.


and_list([],Top):-
	true(Top).
and_list([C|Cs],K):-
	and_list(Cs,K1),
	and(C,K1,K),
	!.


or_list([],[]).
or_list([C|Cs],Ct):-
	or_list(Cs,Ct1),
	or(C,Ct1,Ct).


and1( [], _, [] ).
and1( [A|Rest], C, Conj ):-
	and_conjunction_constraint( A, C, C1 ),
	and1( Rest, C, C2 ),
	user:append( C1, C2, Conj ).

and_conjunction_constraint( _, [], [] ).
and_conjunction_constraint( conj(D,E,O), [conj(D1,E1,O1)|Rest], [conj(D2,E2,O2)|C1] ):-
	user:append( E, E1, E2 ),
	user:append( D, D1, D2 ),
	user:append( O, O1, O2 ),
	and_conjunction_constraint( conj(D,E,O), Rest, C1 ), !.


not( [], [conj([],[],[])] ):- !.
not( [A|R], C ):-
	not_conjunction( A, NA ),
	not( R, NR ),
	and( NR, NA, C ).

not_conjunction( conj([],[],[]), [] ).
not_conjunction( conj([S>T|R],E,O), C ):-
	not_conjunction( conj(R,E,O), NR ),
	or_ge( NR, [T,S], C ).
not_conjunction( conj([S>=T|R],E,O), C ):-
	not_conjunction( conj(R,E,O), NR ),
	or_gt( NR, [T,S], C ).
not_conjunction( conj([],[S=T|R],O), C ):-
	not_conjunction( conj([],R,O), NR ),
	or_not_eq( NR, [S,T], C ).
not_conjunction( conj([],[],[O|R]), C ):-
	not_conjunction( conj([],[],R), NR ),
	or_other( NR, ~O, C ).





nonstrict_inequalities_as_disjunction([],[]).
nonstrict_inequalities_as_disjunction([Conj|Conjs],C):-
	nsi_as_disj_conj(Conj,C1),
	nonstrict_inequalities_as_disjunction(Conjs,C2),
	append(C1,C2,C).


nsi_as_disj_conj(conj([],E,O),[conj([],E,O)]).
nsi_as_disj_conj(conj([S>T|I],E,O),C):-
	nsi_as_disj_conj(conj(I,E,O),C1),
	and_gt(C1,[S,T],C).
nsi_as_disj_conj(conj([S>=T|I],E,O),C):-
	nsi_as_disj_conj(conj(I,E,O),C1),
	and_gt(C1,[S,T],C2),
	and_eq(C1,[S,T],C3),
	or(C2,C3,C).




/*
------------------------------------------------------------------------------
 projection


project(+V,+C,?CP)
is true when CP is the projection of the constraint C onto the
variables in V.
Logically that means that  CP => (exists V) C.
Note that projections need not be precise;
we allow for a constraint weakening.
In the implementation  below we simply drop all atomic constraints
that contain variables not in C.
We assume that C is already simplified.

------------------------------------------------------------------------------
*/


project(V,C,CP):-
	maplist(projectDisjunct(V),C,CP).
%	simplify(CP,C1).



projectDisjunct(V,conj(I,E,O),conj(IP,EP,OP)):-
	noextraVarAtoms(V,I,IP),
	noextraVarAtoms(V,O,OP),
	noextraVarAtoms(V,E,EP).
	

noextraVarAtoms(_,[],[]).
noextraVarAtoms(V,[X|Xs],Ys):-
	(extraVarAtom(V,X) ->
	     Ys=Xs1
	;
	     Ys=[X|Xs1]
	),
	noextraVarAtoms(V,Xs,Xs1).


extraVarAtom(V,S=T):-
	list_of_vars(S=T,V1),
	extraVars(V1,V),
	!.
extraVarAtom(V,S>T):-
	list_of_vars(S=T,V1),
	extraVars(V1,V),
	!.
extraVarAtom(V,S>=T):-
	list_of_vars(S=T,V1),
	extraVars(V1,V),
	!.
extraVarAtom(V,O):-
	list_of_vars(O,V1),
	extraVars(V1,V),
	!.

	

%------------------------------------------------------------------------------
% constraint printing
%------------------------------------------------------------------------------

print_constraint( Ct ):-
	true( Ct ),
	!.
print_constraint( Ct ):-
	format(' {~p}', [Ct] ).


%------------------------------------------------------------------------------
% constraint entailment
%------------------------------------------------------------------------------





implies( _, [conj([],[],[])] ):-
	!.
implies( C1, C2 ):-
	not( C2, NC2 ),
	and( C1, NC2, Im ),
	simplify( Im, Im1 ),
	!,
	false( Im1 ).


% the same for skolemized constraints

valid_s(C):-
	implies_s([conj([],[],[])],C).




implies_s( _, [conj([],[],[])] ):-
	!.
implies_s( _, [conj(InEqs,[],[])] ):-
	maplist(greaterC,InEqs),
	!.
implies_s(C1,C2):-
	(flag_on(ci,orde) ->
	     (flag_on(rp,ccrw) ->
		  implies_hard_s(C1,C2)
	     ;
	          implies_light_s(C1,C2)   % we try harder
	     )
	;
	     implies_s_simple(C1,C2)
	).

greaterC(S>T):-
	greater(S,T).
greaterC(S>=T):-
	greaterEqual(S,T).


implies_light_s(C1,C2):-
	deskolemize((C1,C2),(D1,D2)),
	!,
	implies_light(D1,D2).

implies_hard_s(C1,C2):-
	deskolemize((C1,C2),(D1,D2)),
	!,
	implies(D1,D2).



implies_s_simple(Ct,[conj(J,[],[])]):-
	\+ (
	   member(conj(I,_,_),Ct), 
	   member(InEq,J),
	   \+ constraintSolverBasics: implies(I,InEq)
	   ).



implies_light( _, [conj([],[],[])]):-
	!.
implies_light( [], _ ):- !.
implies_light( [A|R], Ct ):-
	and_implies_constraint_light( A, Ct ),
	implies_light( R, Ct ), !.

delete_implying_ands( C1, C2 ):-
	select( A, C1, C3 ),
	member( A1, C3 ),
	implies_and( A, A1 ), 
	delete_implying_ands( C3, C2 ), !.
delete_implying_ands( C, C ).

implies_and( conj(I1,E1,O1), conj(I,E,O) ):-
	\+ (member(Eq,E),\+member212(Eq,E1)),
	\+ (member(Ot,O),\+member212(Ot,O1)),
	\+ (member(IEq,I),\+ (member(IEq1,I1),constraintSolverBasics:implies(IEq1,IEq))).



%------------------------------------------------------------------------------
% and_implies_constraint_light is incomplete: to compute 	
% 	(1) 	forall sigma  (a --> b1  or...or  a --> bn)
% we do 
%	(2)	(forall sigma a --> b1)  or...or  (forall sigma a --> bn)
% where (2) implies (1) but not vice versa.
%------------------------------------------------------------------------------

and_implies_constraint_light( A, Ct ):-
	member( A1, Ct ),
	implies( [A], [A1] ), !.

constraint_operators( Ct, L ):-
	get_prolog_data_operators( Ct, L1 ),
	delete( L1,=, L2 ),
	delete( L2, >, L3 ),
	delete( L3, >=, L ), !.

