% st_retrieval.pl
% Retrieval in Substitution trees
% Author: pb

:- module(st_retrieval,[
	st_unify/3,
	st_unify/2,
	st_retrieve/4,
	st_retrieve/3,
	st_match/3,
	st_match/2,
	st_us_unify/3,
	st_us_unify/2
		       
]).

:- use_module(substitution).
:- use_module(normalization).
:- use_module(library(basics),[member/2]).

% st_unify(Term,ST).
% unifies Term and a term t inserted in ST such that
% Term and t unfies.
% Note, variables in Term and t are bound.
% ATTENTION: use only with findall (otherwise ST is damaged)
st_unify(Term,ST) :-
	st_unify(Term,ST,_).

st_unify(Term,substitutiontree(U,ST,_),Info) :-
	U = Term,
	st_unify2(ST,Info).

st_unify2(st(Subst,Nodes),Info) :-
	subst_self_apply(Subst),
	(Nodes = leaf(Infos) ->
 	    member(FinalSubst-Info,Infos),
	    subst_self_apply(FinalSubst)
	;
	    member(Node,Nodes),
	    st_unify2(Node,Info)
	).

% st_us_unify(Term,ST).
% unifies Term and a term t inserted in ST such that
% Term and t unfies.
% Note, variables in Term and t are bound.
% ATTENTION: use only with findall (otherwise ST is damaged)
st_us_unify(Term,ST) :-
	st_us_unify(Term,ST,_).

st_us_unify(Term,substitutiontree(U,ST,_),Info) :-
	U = Term,
	st_us_unify2(ST,Info).

st_us_unify2(st(Subst,Nodes),Info) :-
	us_subst_self_apply(Subst),
	(Nodes = leaf(Infos) ->
 	    member(FinalSubst-Info,Infos),
	    us_subst_self_apply(FinalSubst)
	;
	    member(Node,Nodes),
	    st_us_unify2(Node,Info)
	).

us_subst_self_apply([]).
us_subst_self_apply([V-V|S]) :-
	us_subst_self_apply(S).

% st_retrieve(Term,ST,FinalSubst,Info).
% st_retrieve(Term,ST,FinalSubst).
% retrieves unifiable Term. In order to get unifiable Term apply
% FinalSubst. 
% Useful if variants (variable-renamed terms) need to be retrieved.
% FinalSubst represents the variablerenaming
% ATTENTION: use only with findall (otherwise ST is damaged)
st_retrieve(Term,substitutiontree(U,ST,_),FinalSubst) :-
	st_retrieve(Term,substitutiontree(U,ST,_),FinalSubst,_).
	
st_retrieve(Term,substitutiontree(U,ST,_),FinalSubst,Info) :-
	U = Term,
	st_retrieve2(ST,FinalSubst,Info).

st_retrieve2(st(Subst,Nodes),FinalSubst,Info) :-
	subst_self_apply(Subst),
	(Nodes = leaf(Infos) ->
	    member(FinalSubst-Info,Infos)
	;
	    member(Node,Nodes),
	    st_retrieve2(Node,FinalSubst,Info)
	).


% matching; I don't like it.
% Not yet very efficient; just the functionality
st_match(Term,substitutiontree(U,ST,_)) :-
	st_match(Term,substitutiontree(U,ST,_),_).

st_match(Term,substitutiontree(U,ST,_),Info) :-
	normalization(Term,NormalTerm,StarSubst,[],NewVars),
	st_match_number_vars(NewVars,1),
	U = NormalTerm,
	st_match2(ST,StarSubst,Info).

st_match2(st(Subst,Nodes),QueryStarSubst,Info) :-
	subst_match(Subst),
	(Nodes = leaf(Infos) ->
	    member(FinalSubst-Info,Infos),
	    st_match_merge_subst(FinalSubst,QueryStarSubst)
	;
	    member(Node,Nodes),
	    st_match2(Node,QueryStarSubst,Info)
	).

st_match_merge_subst([],_).
st_match_merge_subst([X-Var|FinalSubst],QueryStarSubst) :-
	atomic(X),!,
	(dom_member(QueryStarSubst,X,OriginalVar) ->
	    Var = OriginalVar % a numbered variable; take the variable
	;
	    Var = X % no numbered variable
	),
	st_match_merge_subst(FinalSubst,QueryStarSubst).
st_match_merge_subst([X-Var|FinalSubst],QueryStarSubst) :-
	st_match_merge_term(X,Var,QueryStarSubst),
	st_match_merge_subst(FinalSubst,QueryStarSubst).

st_match_merge_term(X,Y,QueryStarSubst) :-
	atomic(X),!,
	(dom_member(QueryStarSubst,X,OriginalVar) ->
	    Y = OriginalVar % a numbered variable; take the variable
	;
	    Y = X % no numbered variable
	).
st_match_merge_term(X,Y,QueryStarSubst) :-
	% nonvar(X), % is assured, since term to match on is ground
	functor(X,F,A),
	functor(Y,F,A),
	st_match_merge_term(A,X,Y,QueryStarSubst).

st_match_merge_term(0,_X,_Y,_QueryStarSubst).
st_match_merge_term(A,X,Y,QueryStarSubst) :-
	arg(A,X,SubX),
	arg(A,Y,SubY),
	st_match_merge_term(SubX,SubY,QueryStarSubst),
	A1 is A-1,
	st_match_merge_term(A1,X,Y,QueryStarSubst).



st_match_number_vars([],_).
st_match_number_vars([NewVar|NewVars],N) :-
	st_match_gen_name(NewVar,N),
	N1 is N+1,
	st_match_number_vars(NewVars,N1).
	
st_match_gen_name(NewVar,N) :-
	(N =< 254 ->
	    Id = "$st_match_gen_name_ident_",
	    name(NewVar,[N|Id])
	;
	    N1 is N // 254,
	    N2 is N - (N1*254) + 1,
	    Id = "$st_match_gen_name_ident_",
	    name(NewVar,[N1,N2|Id])
	).


% assume that query term was ground!
subst_match([]).
subst_match([X-X|Subst]) :-
	subst_match(Subst).

/*
% copied member from SICStus lib since it is not in quintus library(lists) :-(
member(Element, [Head|Tail]) :-
        member_(Tail, Head, Element).

% auxiliary to avoid choicepoint for last element
member_(_, Element, Element).
member_([Head|Tail], _, Element) :-
        member_(Tail, Head, Element).

*/

%
%
%% Example
%% insert_all_info([f(a,X)-g(X),f(Y,b)-h(Y)],ST),st_match(f(a,Z),ST,R).
