


 :- module(translate,[make_file/4, logic_list/1, translate/6, translate/0]).


/**************************************************************************
*
*		         OPERATORS
*
* EXPLANATION:
* Operators [and,or,=>,<=>,~,dia,box] are the logical
* symbols of the input logic
* Operators [::,^^] are function symbols of the output language
*
**************************************************************************/

	:- op(920,yfx,[and]).
	:- op(930,yfx,[or]).
	:- op(950,xfx,[<=>]).
	:- op(950,xfy,[=>]).
%	:- op(900,fy,[~,dia,box]).
	:- op(900,fy,[dia,box]).
	:- op(500,fy,[~]).
        :- op(910,yfx,[::]).
        :- op(910,xfx,[^^]).

 :- use_module(library(basics)).

 :- use_module(adt_precedence,[new_skolem_function/2]).

/************************************************************
*
*		        MISCELLANEOUS
*
************************************************************/

new_symbol(Prefix,Symbol) :-
		( retract(new_symbol_counter(Prefix,Value)) ; Value = 0 ),
		NewValue is Value+1,
		asserta(new_symbol_counter(Prefix,NewValue)),
		name(Prefix,P),
		name(NewValue,N),
		user:append(P,N,SymbolList),
		name(Symbol,SymbolList), !.

new_variable_number(N) :-
	( retract(new_variable_counter(N)) ; N = 0 ),
	M is N+1,
	asserta(new_variable_counter(M)).

my_delete(_,[],[]) :- !.
my_delete(X,[X|List],NewList) :- !,
	my_delete(X,List,NewList).
my_delete(X,[A|List],[A|NewList]) :- !,
	my_delete(X,List,NewList).

empty_intersection([],_).
empty_intersection(_,[]).
empty_intersection([A|_],B) :-
	member(A,B),!,fail.
empty_intersection([_|Rest],B) :-
	empty_intersection(Rest,B).

read_from_stream(Stream,List) :- read(Stream,Term),
		((Term = end_of_file, List = []) ;
		 (read_from_stream(Stream,MoreList),
		  List = [Term | MoreList])).

/************************************************************
*
*		NEGATION NORMAL FORM
*
* EXPLANATION
* norm_pos(F,.) produces the negation normal form of F
* norm_neg(F,.) produces the negation normal form of ~F
*
************************************************************/

norm_pos(A and B,A1 and B1) :- !,
	norm_pos(A,A1),
	norm_pos(B,B1).
norm_pos(A or B,A1 or B1) :- !,
	norm_pos(A,A1),
	norm_pos(B,B1).
norm_pos(all(X,A),all(X,A1)) :- !,
	norm_pos(A,A1).
norm_pos(exists(X,A),exists(X,A1)) :- !,
	norm_pos(A,A1).
norm_pos(box A,box A1) :- !,
	norm_pos(A,A1).
norm_pos(dia A,dia A1) :- !,
	norm_pos(A,A1).
norm_pos(~A,A1) :- !,
	norm_neg(A,A1).
norm_pos(A => B,A1 or B1) :- !,
	norm_neg(A,A1),
	norm_pos(B,B1).
norm_pos(A <=> B,(A1 or B1) and (B2 or A2)) :- !,
	norm_neg(A,A1),
	norm_pos(B,B1),
	norm_pos(A,A2),
	norm_neg(B,B2).
norm_pos(A,A).
norm_neg(A and B,A1 or B1) :- !,
	norm_neg(A,A1),
	norm_neg(B,B1).
norm_neg(A or B,A1 and B1) :- !,
	norm_neg(A,A1),
	norm_neg(B,B1).
norm_neg(all(X,A),exists(X,A1)) :- !,
	norm_neg(A,A1).
norm_neg(exists(X,A),all(X,A1)) :- !,
	norm_neg(A,A1).
norm_neg(box A,dia A1) :- !,
	norm_neg(A,A1).
norm_neg(dia A,box A1) :- !,
	norm_neg(A,A1).
norm_neg(~A,A1) :- !,
	norm_pos(A,A1).
norm_neg(A => B,A1 and B1) :- !,
	norm_pos(A,A1),
	norm_neg(B,B1).
norm_neg(A <=> B,(A1 or B1) and (A2 or B2)) :- !,
	norm_pos(A,A1),
	norm_pos(B,B1),
	norm_neg(A,A2),
	norm_neg(B,B2).
norm_neg(A,~A).

/************************************************************
*
*		TRANSLATION INTUITIONISTIC -> S4
*
************************************************************/

trans(intui,A and B,NewA and NewB) :-
	trans(intui,A,NewA),
	trans(intui,B,NewB).
trans(intui,A or B,NewA or NewB) :-
	trans(intui,A,NewA),
	trans(intui,B,NewB).
trans(intui,A => B,box(NewA => NewB)) :-
	trans(intui,A,NewA),
	trans(intui,B,NewB).
trans(intui,A <=> B,box(NewA <=> NewB)) :-
	trans(intui,A,NewA),
	trans(intui,B,NewB).
trans(intui,~A,box ~ NewA) :-
	trans(intui,A,NewA).
trans(intui,exists(Var,F),exists(Var,NewF)) :-
	trans(intui,F,NewF).
trans(intui,all(Var,F),all(Var,NewF)) :-
	trans(intui,F,NewF).
trans(intui,A,box A).

/************************************************************
*
*		CLASSICAL LOGIC TRANSLATION
*
* EXPLANATION
* This is how classical formulae are to be translated.
* Since the input formula is in fact already in the desired
* form the only thing to be done is to rename the variables
* such that later program parts can deal with them accordingly.
*
************************************************************/

trans(none,A and B,_,NewA and NewB) :-
	trans(none,A,_,NewA),
	trans(none,B,_,NewB).
trans(none,A or B,_,NewA or NewB) :-
	trans(none,A,_,NewA),
	trans(none,B,_,NewB).
trans(none,exists(Var,P),_,exists('$VAR'(X),NewP)) :-
	new_variable_number(X),
	replace(P,Var,'$VAR'(X),PVarX),
	trans(none,PVarX,_,NewP).
trans(none,all(Var,P),_,all('$VAR'(X),NewP)) :-
	new_variable_number(X),
	replace(P,Var,'$VAR'(X),PVarX),
	trans(none,PVarX,_,NewP).
trans(none,Literal,_,Literal).

/************************************************************
*
*		SEMI-FUNCTIONAL TRANSLATION
*
* EXPLANATION
* See my paper "First-Order Theorem Proving and Functional
* Simulation" IJCAI'93
*
************************************************************/

trans(semi_functional,P and Q,U,P1 and Q1) :- !,
	trans(semi_functional,P,U,P1),
	trans(semi_functional,Q,U,Q1).
trans(semi_functional,P or Q,U,P1 or Q1) :- !,
	trans(semi_functional,P,U,P1),
	trans(semi_functional,Q,U,Q1).
trans(semi_functional,all(Var,P),U,all('$VAR'(X),~'$E'(U,'$VAR'(X)) or NewP)) :- !,
	new_variable_number(X),
	replace(P,Var,'$VAR'(X),PVarX),
	trans(semi_functional,PVarX,U,NewP).
trans(semi_functional,exists(Var,P),U,exists('$VAR'(X),NewP)) :- !,
	new_variable_number(X),
	replace(P,Var,'Ex$'(U)^^'$VAR'(X),PVarUX),
	trans(semi_functional,PVarUX,U,NewP).
trans(semi_functional,box P,U,all('$VAR'(V),~'R$'(U,'$VAR'(V)) or NewP)) :- !,
	new_variable_number(V),
	trans(semi_functional,P,'$VAR'(V),NewP).
trans(semi_functional,dia P,U,'Normal$'(U) and exists('$VAR'(X),NewP)) :- 
	new_variable_number(X),
	trans(semi_functional,P,U::'$VAR'(X),NewP).
trans(semi_functional,~P,U,~X) :- !, P =.. [Y|Z], X =.. [Y,U|Z].
trans(semi_functional,P,U,X) :- !, P =.. [Y|Z], X =.. [Y,U|Z].

/************************************************************
*
*		FUNCTIONAL TRANSLATION
*
* EXPLANATION
* Actually the functional translation method differs from the
* semi_functional one. A direct translation however, is difficult
* to implement because of the strange kind of skolemization
* which is required. I avoided this problem by translating
* semi_functionally and adding a suitable background theory (see below).
*
************************************************************/

trans(functional,Formula,World,TranslatedFormula) :-
	trans(semi_functional,Formula,World,TranslatedFormula).

/************************************************************
*
*		RELATIONAL TRANSLATION
*
* This is a direct transformation according to the Kripke
* model semantics. Nothing special about it.
*
************************************************************/

trans(relational,P and Q,U,P1 and Q1) :- !,
	trans(relational,P,U,P1),
	trans(relational,Q,U,Q1).
trans(relational,P or Q,U,P1 or Q1) :- !,
	trans(relational,P,U,P1),
	trans(relational,Q,U,Q1).
trans(relational,all(Var,P),U,all('$VAR'(X),~'$E'(U,'$VAR'(X)) or NewP)) :- !,
	new_variable_number(X),
	replace(P,Var,'$VAR'(X),PVarX),
	trans(relational,PVarX,U,NewP).
trans(relational,exists(Var,P),U,exists('$VAR'(X),'$E'(U,'$VAR'(X)) and NewP)) :- !,
	new_variable_number(X),
	replace(P,Var,'$VAR'(X),PVarX),
	trans(relational,PVarX,U,NewP).
trans(relational,box P,U,all('$VAR'(V),~'R$'(U,'$VAR'(V)) or NewP)) :- !,
	new_variable_number(V),
	trans(relational,P,'$VAR'(V),NewP).
trans(relational,dia P, U, exists('$VAR'(V),'R$'(U,'$VAR'(V)) and NewP)) :- !,
	new_variable_number(V),
	trans(relational,P,'$VAR'(V),NewP).
trans(relational,~P,U,~NewP) :- !, P =.. [Y | Z], NewP =.. [Y,U|Z].
trans(relational,P,U,NewP) :- !, P =.. [Y | Z], NewP =.. [Y,U|Z].
	

/************************************************************
*
*	        ORDERING AND WRITING
*
* saturate expects sequent clauses as input.
* order_clauses accepts a clause and produces a pair
* of atom sets, one element for the negative literals
* and one for the positive ones respectively.
* write_sequent_clauses writes the respective sequent clauses
* to STREAM which has been opened for output (the outputfile).
*
************************************************************/

order_clauses([],[]) :- !.
order_clauses([Clause|Rest],[NewClause|NewRest]) :- !,
	order_clause(Clause,NewClause),
	order_clauses(Rest,NewRest).

order_clause([],[[],[]]) :- !.
order_clause([~A|Rest],[[A|C],D]) :- !,
	order_clause(Rest,[C,D]).
order_clause([A|Rest],[C,[A|D]]) :- !,
	order_clause(Rest,[C,D]).

write_sequent_clauses(_,[]) :- !.
write_sequent_clauses(Stream,[First|Rest]) :- !,
	write_sequent_clause(Stream,First),
	write_sequent_clauses(Stream,Rest).

write_sequent_clause(Stream,[[],Positives]) :- !,
	write_atoms(Stream,Positives),
	format(Stream,".~n",[]).
write_sequent_clause(Stream,[Negatives,[]]) :- !,
	write_atoms(Stream,Negatives),
	format(Stream," -> [].~n",[]).
write_sequent_clause(Stream,[Negatives,Positives]) :- !,
	write_atoms(Stream,Negatives),
	format(Stream," -> ",[]),
	write_atoms(Stream,Positives),
	format(Stream,".~n",[]).

write_atom(Stream,A) :-
	A =.. ['$VAR',_], write(Stream,A).
write_atom(Stream,A) :-
	A =.. [Top|Rest],
	format(Stream,"~q",[Top]),
	((Rest == [],true);
	 (write(Stream,'('),write_atoms(Stream,Rest),write(Stream,')'))).

write_atoms(_,[]).
write_atoms(Stream,[OnlyOne]) :- write_atom(Stream,OnlyOne).
write_atoms(Stream,[First|Rest]) :-
	write_atom(Stream,First),write(Stream,','),write_atoms(Stream,Rest).
	

/************************************************************
*
*	        CLAUSIFICATION
*
* EXPLANATION
* skolemize expects a formula given in negation normal form
* and performs the skolemization (also collects the skolem
* function symbols).
* clausify expects a formula in negation normal form AFTER
* skolemization. For conjunctions there is nothing to be done.
* Disjunctions are treated as the scalar product of the respective
* disjunctive elements (merge).
*
************************************************************/

skolemize(A and B,Vars,C and D,SkSymbols) :- !,
	skolemize(A,Vars,C,SkSymbols1),skolemize(B,Vars,D,SkSymbols2),
	user:append(SkSymbols1,SkSymbols2,SkSymbols).
skolemize(A or B,Vars,C or D,SkSymbols) :- !,
	skolemize(A,Vars,C,SkSymbols1),skolemize(B,Vars,D,SkSymbols2),
	user:append(SkSymbols1,SkSymbols2,SkSymbols).
skolemize(all(X,P),Vars,Q,SkSymbols) :- !,
	skolemize(P,[X|Vars],Q,SkSymbols).
skolemize(exists(X,P),Vars,Q,[U|SkSymbols]) :- !,
	length(Vars,N),
        new_skolem_function(N,U),
%	new_symbol('$sk',U),
	SkolemTerm =.. [U|Vars],
	replace(P,X,SkolemTerm,NewP),
	skolemize(NewP,Vars,Q,SkSymbols).
skolemize(A,_,A,[]) :- !.

clausify(A and B,C) :- !,
	clausify(A,A1),
	clausify(B,B1),
	user:append(A1,B1,C).
clausify(A or B,C) :- !,
	clausify(A,A1),
	clausify(B,B1),
	merge(A1,B1,C).
clausify(A,[[A]]) :- !.

merge([],_,[]) :- !.
merge([Clause|MoreClauses],OtherClauses,Result) :- !,
	merge_single(Clause,OtherClauses,List),
	merge(MoreClauses,OtherClauses,RestList),
	user:append(List,RestList,Result).

merge_single(_,[],[]) :- !.
merge_single(Clause,[First|Rest],[NewFirst|NewRest]) :- !,
	user:append(Clause,First,NewFirst),
	merge_single(Clause,Rest,NewRest).

/************************************************************
*
*	        SUBSTITUTION
*
* EXPLANATION
* replace(A,B,C,D) replaces each occurrence of B in expression A by C.
* replace_list is similar, but A is a list of expressions.
*
************************************************************/

replace_list(Var,_,_,Var) :- var(Var), !.
replace_list([],_,_,[]) :- !.
replace_list([Expr|Rest],ToBeReplaced,Replacer,[NewExpr|NewRest]) :- !,
	replace(Expr,ToBeReplaced,Replacer,NewExpr),
	replace_list(Rest,ToBeReplaced,Replacer,NewRest).

replace(Expr,ToBeReplaced,Replacer,Replacer) :- Expr == ToBeReplaced, !.
replace(Expr,_,_,Expr) :- var(Expr), !.
replace(Expr,ToBeReplaced,Replacer,NewExpr) :- !,
	Expr =.. [Head|Body],
	replace_list(Body,ToBeReplaced,Replacer,NewBody),
	NewExpr =.. [Head|NewBody].

/************************************************************
*
*	   LIST OF PREDICATE AND FUNCTION SYMBOLS
*
* EXPLANATION
* symbolList computes the list of all the predicate and function
* symbols used in the input formula (essentially by a recursive descent).
* symbolSet then makes sets out of these lists.
*
************************************************************/

symbolSet(Formula,PSet,FSet) :-
	symbolList(Formula,PList,FList),!,
	sort(PList,PSet),
	sort(FList,FSet).

symbolList(Formula,PList,FList) :-
	Formula =.. [Symb|Args],
	((member(Symb,[and,or,'=>','<=>']),
	   Args = [First,Second],
	   symbolList(First,PFirst,FFirst),
	   symbolList(Second,PSecond,FSecond),
	   user:append(PFirst,PSecond,PList),
	   user:append(FFirst,FSecond,FList));
	 (member(Symb,[all,exists]),
	   Args = [Var,Second],
	   symbolList(Second,PList,FList1),
	   my_delete(Var,FList1,FList));
	 (member(Symb,['~',box,dia]),
	   Args = [TheArg],
	   symbolList(TheArg,PList,FList));
	 (PList = [Symb], fsymblist(Args,FList))).

fsymblist([],[]).
fsymblist([Term | Rest],FSymbs) :-
	fsymbs(Term,List),
	fsymblist(Rest,Restlist),
	user:append(List,Restlist,FSymbs).

fsymbs(Term,[Head|List]) :-
	Term =.. [Head|Args],
	fsymblist(Args,List).

write_final_form(Stream,Form) :-
	write_sequent_clauses(Stream,Form).

/****************************************************************
*
*                   THE KIND OF LOGIC USED
*
* EXPLANATION
* logic_type finds out (solely by syntactic means) whether the
* input theorem is of classical logic (propositional or first-order)
* or of modal logic (again propositional or first-order)
*
*****************************************************************/

logic_type_list([],[]) :- !.
logic_type_list([Formula | Formulae],[Type | Types]) :- !,
	logic_type(Formula,Type),
	logic_type_list(Formulae,Types).

logic_type(Formula,Type) :-
	Formula =.. [Head | Args],
	((member(Head,[and,or,'=>','<=>','~']), !,
	  logic_type_list(Args,Types),
	  maximum(Type,Types));
	 (member(Head,[exists,all]), !,
	  Args = [_|RelevantArgs],
	  logic_type_list(RelevantArgs,Types),
	  maximum(Type,[fol | Types]));
	 (member(Head,[box,dia]), !,
	  logic_type_list(Args,Types),
	  maximum(Type,[propmod | Types]));
	 (Args = [], !, Type = prop);
	 (Type = fol), !).

maximum(Type,ListOfTypes) :-
	((member(folmod,ListOfTypes), !, Type = folmod);
	 (member(fol,ListOfTypes),member(propmod,ListOfTypes),!,Type = folmod);
	 (member(fol,ListOfTypes), !, Type = fol);
	 (member(propmod,ListOfTypes), !, Type = propmod);
	 (Type = prop)).

/**************************************************************
*
*                       Starting    
*
***************************************************************/

logic_list(List) :-
	findall(Logics,theory(_,Logics,_,_,_,_),L),
	sort(L,LL),
	remove_vars(LL,List).

make_file(File,skolems(Skolems),TheoryClauses,TheoremClauses):-
        retractall(new_symbol_counter(_,_)),
        open(File,write,Stream),
        format(Stream,"~n%   Skolemsymbols:  ~q.~n~n",[Skolems]),
        ((TheoryClauses == [],true);
	 (format(Stream,"%   Theory clauses:~n~n",[]),
          write_sequent_clauses(Stream,TheoryClauses),
	  format(Stream,"~n",[]))),
	((TheoremClauses == [],true);
         (format(Stream,"%   Theorem clauses:~n~n",[]),
          write_sequent_clauses(Stream,TheoremClauses))),
        close(Stream).

remove_vars([],[]).
remove_vars([Var|Rest],Result) :-
	var(Var),remove_vars(Rest,Result).
remove_vars([Symbol|Rest],[Symbol|Result]) :-
	remove_vars(Rest,Result).

/************************************************************
*
* The predicate "translate/6" has arguments
*      i) the theorem to be proved
*     ii) the logic (classic, intui  versus modal logics)
*    iii) a list of options as they should be given in an input file
*     iv) --> a list of skolemsymbols
*      v) --> the list of theory clauses
*     vi) --> the list of (negated) theorem clauses
* the first three arguments are mandatory.
*
************************************************************/

translate :-
	format("~n>translate/6< is a 6-ary predicate.~n~n",[]),
	format("1. Theorem to be proved (mandatory).~n",[]),
	format("2. Logic within which it is to be proved (mandatory).~n",[]),
	format("3. List of options (see help-file HELP) (mandatory).~n",[]),
	format("4. Skolemsymbols (result).~n",[]),
	format("5. Theory clauses (result).~n",[]),
	format("6. Theorem clauses (result).~n~n",[]),
	format("The predicate >make_file/4< can then be used with the~n",[]),
	format("final three arguments from above to obtain an output-file.~n",[]).

translate(Theorem,_,_,_,_,_) :-
	var(Theorem),
	format("INPUT ERROR: No theorem is given.~n",[]),
	format("The query is ignored!",[]), !, fail.

translate(none,Logic,OptionList,Skolemsymbols,TheoryClauses,[]) :-
	((member(domain(_),OptionList),
	  translate(none,folmod,Logic,OptionList,Skolemsymbols,TheoryClauses,_));
	 (translate(none,propmod,Logic,OptionList,Skolemsymbols,TheoryClauses,_))).

translate(Theorem,Logic,OptionList,Skolemsymbols,TheoryClauses,NegTheoremClauses) :-
	retractall(new_symbol_counter(_,_)), !, 
	logic_type(Theorem,Kind),
	translate(Theorem,Kind,Logic,OptionList,Skolemsymbols,TheoryClauses,
	              NegTheoremClauses).

translate(none,Kind,Logic,Options,skolems(['o$'|TheorySkolems]),TheoryClauses,[]) :-
	retractall(new_variable_counter(_)),
	determine_translation(Kind,Logic,Options,Translation),
	determine_fora(Kind,Logic,Translation,Options,ForA),
	determine_domain(Kind,Logic,Translation,ForA,Options,Domain),
	theory(Kind,Logic,Translation,ForA,Domain,Theory),
	prepare_theory(Theory,TheorySkolems,TheoryClauses).

translate(Theorem,Kind,GivenLogic,Options,skolems(Skolems),TheoryClauses,
	            NegTheoremClauses) :-
	retractall(new_variable_counter(_)),
	determine_logic(Kind,GivenLogic,Logic),
	determine_translation(Kind,Logic,Options,Translation),
	determine_fora(Kind,Logic,Translation,Options,ForA),
	determine_domain(Kind,Logic,Translation,ForA,Options,Domain),
	theory(Kind,Logic,Translation,ForA,Domain,Theory),
	prepare_theory(Theory,TheorySkolems,TheoryClauses),
	((Logic == classic,AllTheorySkolems = TheorySkolems);
	 (AllTheorySkolems = ['o$'|TheorySkolems])),
	((Logic == intuitionistic,trans(intui,Theorem,RealTheorem));
	 (RealTheorem = Theorem)),
	norm_neg(RealTheorem,NegTheorem),
	trans(Translation,NegTheorem,'o$',NewNegTheorem),
	skolemize(NewNegTheorem,[],SkNegTheorem,TheoremSkolems),
	clausify(SkNegTheorem,NegTheoremClauses_simple),
	order_clauses(NegTheoremClauses_simple,NegTheoremClauses), !,
	user:append(AllTheorySkolems,TheoremSkolems,Skolems).

determine_logic(Kind,Given,Logic) :-
	findall(Logics,theory(Kind,Logics,_,_,_,_),LLL),
	sort(LLL,LL),
	remove_vars(LL,LogicList),
	((nonvar(Given),
	  member(Given,LogicList),
	  Logic = Given);
	 (nonvar(Given),
	  format("Sorry, the logic ~q makes no sense here.~n",[Given]),
	  format("Please choose one from the list ~q.~n",[LogicList]),
	  read(Input),
	  determine_logic(Kind,Input,Logic));
	 (format("You didn't specify a logic within which the theorem is to be proved.~n",[]),
	  format("Please choose one from the list ~q.~n",[LogicList]),
	  read(Input),
	  determine_logic(Kind,Input,Logic))).
	
determine_translation(Kind,Logic,Options,Translation) :-
	findall(Translations,theory(Kind,Logic,Translations,_,_,_),T),
	sort(T,TT),
	remove_vars(TT,PossTrans),
	determine_translation_1(PossTrans,Options,Translation).
determine_translation_1([],_,_).
determine_translation_1([T],_,T) :-
	format("~q is the only possible translation here.~n",[T]).
determine_translation_1(PossTrans,Options,Trans) :-
	member(translation(GivenTrans),Options),
	((member(GivenTrans,PossTrans),Trans = GivenTrans);
	 (member(semi_functional,PossTrans),Trans = semi_functional);
	 (PossTrans = [Trans|_])).
determine_translation_1(PossTrans,_,semi_functional) :-
	member(semi_functional,PossTrans).
determine_translation_1([FirstPossTrans|_],_,FirstPossTrans).

determine_fora(Kind,Logic,Translation,Options,ForA) :-
	findall(ForAs,theory(Kind,Logic,Translation,ForAs,_,_),F),
	sort(F,FF),
	remove_vars(FF,PossForAs),
	determine_fora_1(PossForAs,Options,ForA).
determine_fora_1([],_,_).
determine_fora_1([F],_,F).
determine_fora_1(PossForAs,Options,ForA) :-
	member(theory(GivenForA),Options),
	((member(GivenForA,PossForAs),ForA = GivenForA);
	 (member(abbreviated,PossForAs),ForA = abbreviated);
	 (PossForAs = [ForA|_])).
determine_fora_1(PossForAs,_,abbreviated) :-
	member(abbreviated,PossForAs).
determine_fora_1([FirstPossForA|_],_,FirstPossForA).

determine_domain(Kind,Logic,Translation,ForA,Options,Domain) :-
	findall(Domains,theory(Kind,Logic,Translation,ForA,Domains,_),D),
	sort(D,DD),
	remove_vars(DD,PossDomains),
	determine_domain_1(PossDomains,Options,Domain).
determine_domain_1([],_,_).
determine_domain_1([D],_,D).
determine_domain_1(PossDomains,Options,Domain) :-
	member(domain(GivenDomain),Options),
	((member(GivenDomain,PossDomains),Domain = GivenDomain);
	 (member(constant,PossDomains),Domain = constant);
	 (PossDomains = [Domain|_])).
determine_domain_1(PossDomains,_,constant) :-
	member(constant,PossDomains).
determine_domain_1([FirstPossDomain|_],_,FirstPossDomain).

prepare_theory(none,[],[]).
prepare_theory(Theory,TheorySkolems,TheoryClauses) :-
	norm_pos(Theory,NormTheory),
	trans(none,NormTheory,_,NewTheory),
	skolemize(NewTheory,[],SkTheory,TheorySkolems),
	clausify(SkTheory,TheoryClauses_simple),
	order_clauses(TheoryClauses_simple,TheoryClauses).


/**************************************************************
*
*                       BACKGROUND THEORIES
*
* EXPLANATION
* The respective background theory depends on several parameters.
* Each theory literal from below describes (in its final argument)
* this background theory.
*
***************************************************************/

%theoryFormula(Formula) :-
%	logic_kind(Kind),
%	logic(Logic),
%	translation(Translation),
%	theory(Theory),
%	domain(Domain),
%	formula(Kind,Logic,Translation,Theory,Domain,Formula).

theory(prop,classic,_,_,_,none).
theory(fol,classic,_,_,_,none).

theory(prop,intuitionistic,Trans,ForA,_,Res) :-
	theory(propmod,s4,Trans,ForA,_,Res).
theory(fol,intuitionistic,Trans,ForA,increasing,Res) :-
	theory(folmod,s4,Trans,ForA,increasing,Res).
theory(propmod,intuitionistic,Trans,ForA,_,Res) :-
	theory(propmod,s4,Trans,ForA,_,Res).
theory(folmod,intuitionistic,Trans,ForA,increasing,Res) :-
	theory(folmod,s4,Trans,ForA,increasing,Res).

theory(propmod,k,relational,_,_,none).
theory(propmod,k,semi_functional,_,_,
	all(u,'Normal$'(u) => all(x,'R$'(u,u::x)))).
theory(propmod,k,functional,_,_,
	all(u,'Normal$'(u) => all(x,'R$'(u,u::x)))).
theory(folmod,k,relational,_,varying,
	all(u,exists(x,'$E'(u,x)))).
theory(folmod,k,semi_functional,_,varying,
	all(u,'Normal$'(u) => all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x)))).
theory(folmod,k,functional,_,varying,
	all(u,'Normal$'(u) => all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x)))).
theory(folmod,k,relational,_,increasing,
	all(u,exists(x,'$E'(u,x))) and
      all(u,all(v,all(x,'$E'(u,x) and 'R$'(u,v) => '$E'(v,x))))).
theory(folmod,k,semi_functional,_,increasing,
	all(u,'Normal$'(u) => all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(y,all(x,'$E'(u,y) and 'Normal$'(u) => '$E'(u::x,y))))).
theory(folmod,k,functional,_,increasing,
	all(u,'Normal$'(u) => all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(x,all(y,'Normal$'(u) => exists(z,('Ex$'(u::x)^^z) = ('Ex$'(u)^^y)))))).
theory(folmod,k,relational,_,decreasing,
	all(u,exists(x,'$E'(u,x))) and
      all(u,all(v,all(x,'$E'(v,x) and 'R$'(u,v) => '$E'(u,x))))).
theory(folmod,k,semi_functional,_,decreasing,
	all(u,'Normal$'(u) => all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(y,all(x,'Normal$'(u) and '$E'(u::x,y) => '$E'(u,y))))).
theory(folmod,k,functional,_,decreasing,	
	all(u,'Normal$'(u) => all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(x,all(y,'Normal$'(u) => exists(z,('Ex$'(u)^^z) = ('Ex$'(u::y)^^x)))))).
theory(folmod,k,relational,_,constant,
	all(u,all(x,'$E'(u,x)))).
theory(folmod,k,semi_functional,_,constant,
	all(u,'Normal$'(u) => all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,x)))).
theory(folmod,k,functional,_,constant,	
	all(u,'Normal$'(u) => all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,x)))).

theory(propmod,kd,relational,_,_,
	all(u,exists(v,'R$'(u,v)))).
theory(propmod,kd,semi_functional,_,_,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,'Normal$'(u))).
theory(propmod,kd,functional,_,_,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,relational,_,varying,
	all(u,exists(x,'R$'(u,x))) and
      all(u,exists(x,'$E'(u,x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,semi_functional,_,varying,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,functional,_,varying,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,relational,_,increasing,
	all(u,exists(x,'R$'(u,x))) and
      all(u,exists(x,'$E'(u,x))) and
      all(u,all(v,all(x,'$E'(u,x) and 'R$'(u,v) => '$E'(v,x)))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,semi_functional,_,increasing,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(y,all(x,'$E'(u,y) => '$E'(u::x,y)))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,functional,_,increasing,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(x,all(y,exists(z,('Ex$'(u::x)^^z) = ('Ex$'(u)^^y))))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,relational,_,decreasing,
	all(u,exists(x,'R$'(u,x))) and
      all(u,exists(x,'$E'(u,x))) and
      all(u,all(v,all(x,'$E'(v,x) and 'R$'(u,v) => '$E'(u,x)))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,semi_functional,_,decreasing,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(y,all(x,'$E'(u::x,y) => '$E'(u,y)))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,functional,_,decreasing,	
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(x,all(y,exists(z,('Ex$'(u)^^z) = ('Ex$'(u::y)^^x))))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,relational,_,constant,
	all(u,exists(x,'R$'(u,x))) and
      all(u,all(x,'$E'(u,x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,semi_functional,_,constant,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd,functional,_,constant,	
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,x))) and
      all(u,'Normal$'(u))).

theory(propmod,s4,relational,_,_,
	all(u,'R$'(u,u)) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w))))).
theory(propmod,s4,semi_functional,abbreviated,_,
	all(u,'R$'('o$',u)) and
      all(u,'R$'(u,u)) and
      all(u,all(v,all(x,'R$'(u,v) => 'R$'(u,v::x)))) and
      all(u,'Normal$'(u))).
theory(propmod,s4,semi_functional,full,_,
	all(u,'R$'(u,u)) and
      all(u,all(x,'R$'(u,u::x))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,'Normal$'(u))).
theory(propmod,s4,functional,_,_,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,exists(x,u::x = u)) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(x,y) => exists(z,(u::z) = y)))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,relational,_,varying,
	all(u,exists(x,'$E'(u,x))) and
      all(u,'R$'(u,u)) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w))))).
theory(folmod,s4,semi_functional,abbreviated,varying,
	all(u,'R$'('o$',u)) and
      all(u,'R$'(u,u)) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(v,all(x,'R$'(u,v) => 'R$'(u,v::x)))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,semi_functional,full,varying,
	all(u,'R$'(u,u)) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(x,'R$'(u,u::x))) and
      all(u,all(v,all(x,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,functional,_,varying,
	all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(x,'R$'(u,u::x))) and
      all(u,exists(x,u::x = u)) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(x,y) => exists(z,(u::z) = y)))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,relational,_,increasing,
	all(u,'R$'(u,u)) and
      all(u,exists(x,'$E'(u,x))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(x,'$E'(u,x) and 'R$'(u,v) => '$E'(v,x))))).
theory(folmod,s4,semi_functional,abbreviated,increasing,
	all(u,'R$'('o$',u)) and
      all(u,'R$'(u,u)) and
      all(u,all(v,all(x,'R$'(u,v) => 'R$'(u,v::x)))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(y,all(x,'$E'(u,y) => '$E'(u::x,y)))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,semi_functional,full,increasing,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,'R$'(u,u)) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(x,'$E'(u,x) and 'R$'(u,v) => '$E'(v,x)))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,functional,_,increasing,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,exists(x,u::x = u)) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(x,y) => exists(z,(u::z) = y)))) and
      all(u,all(x,all(y,'R$'(u,x) and '$E'(u,y) => exists(z,('Ex$'(x)^^z) = y)))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,relational,_,decreasing,
	all(u,'R$'(u,u)) and
      all(u,exists(x,'$E'(u,x))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(x,'$E'(v,x) and 'R$'(u,v) => '$E'(u,x))))).
theory(folmod,s4,semi_functional,abbreviated,decreasing,
	all(u,'R$'('o$',u)) and
      all(u,'R$'(u,u)) and
      all(u,all(v,all(x,'R$'(u,v) => 'R$'(u,v::x)))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(y,all(x,'$E'(u::x,y) => '$E'(u,y)))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,semi_functional,full,decreasing,
	all(u,'R$'(u,u)) and
      all(u,all(x,'R$'(u,u::x))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,all(v,all(x,'$E'(v,x) and 'R$'(u,v) => '$E'(u,x)))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,functional,_,decreasing,	
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,exists(x,u::x = u)) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(x,y) => exists(z,(u::z) = y)))) and
      all(u,all(x,all(y,'R$'(u,y) and '$E'(y,x) => exists(z,('Ex$'(u)^^z) = x)))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,relational,_,constant,
	all(u,'R$'(u,u)) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(x,'$E'(u,x)))).
theory(folmod,s4,semi_functional,abbreviated,constant,
	all(u,'R$'(u,u)) and
      all(u,'R$'('o$',u)) and
      all(u,all(v,all(x,'R$'(u,v) => 'R$'(u,v::x)))) and
      all(u,all(x,'$E'(u,x))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,semi_functional,full,constant,
	all(u,'R$'(u,u)) and
      all(u,all(x,'R$'(u,u::x))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(x,'$E'(u,x))) and
      all(u,'Normal$'(u))).
theory(folmod,s4,functional,_,constant,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,exists(x,u::x = u)) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(x,y) => exists(z,(u::z) = y)))) and
      all(u,all(x,'$E'(u,x))) and
      all(u,'Normal$'(u))).

theory(propmod,kd45,relational,_,_,
	all(u,exists(v,'R$'(u,v))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w))))).
theory(propmod,kd45,semi_functional,abbreviated,_,
	all(u,all(v,all(x,'R$'(u,v::x)))) and
      all(u,'Normal$'(u))).
theory(propmod,kd45,semi_functional,full,_,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w)))) and
      all(u,'Normal$'(u))).
theory(propmod,kd45,functional,_,_,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(x,y) => exists(z,(u::z) = y)))) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(u,y) => exists(z,(x::z) = y)))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,relational,_,increasing,
	all(u,exists(v,'R$'(u,v))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w)))) and
      all(u,exists(x,'$E'(u,x))) and
      all(u,all(v,all(x,'R$'(u,v) and '$E'(u,x) => '$E'(v,x))))).
theory(folmod,kd45,relational,_,decreasing,
	all(u,exists(v,'R$'(u,v))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w)))) and
      all(u,exists(x,'$E'(u,x))) and
      all(u,all(v,all(x,'R$'(u,v) and '$E'(v,x) => '$E'(u,x))))).
theory(folmod,kd45,relational,_,constant,
	all(u,exists(v,'R$'(u,v))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w)))) and
      all(u,all(x,'$E'(u,x)))).
theory(folmod,kd45,relational,_,varying,
	all(u,exists(v,'R$'(u,v))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w)))) and
      all(u,exists(x,'$E'(u,x)))).
theory(folmod,kd45,semi_functional,abbreviated,increasing,
	all(u,all(v,all(x,'R$'(u,v::x)))) and
      all(x,'$E'('o$','Ex$'('o$')^^x)) and
      all(u,all(v,all(x,all(y,'$E'(u::x,'Ex$'(v)^^y))))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,semi_functional,full,increasing,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w)))) and
      all(u,exists(x,'$E'(u,x))) and
      all(u,all(v,all(x,'R$'(u,v) and '$E'(u,x) => '$E'(v,x)))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,semi_functional,abbreviated,decreasing,
	all(u,all(v,all(x,'R$'(u,v::x)))) and
      all(x,'$E'('o$','Ex$'('o$')^^x)) and
      all(u,all(v,all(x,all(y,'$E'(u,'Ex$'(v::x)^^y))))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,semi_functional,full,decreasing,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w)))) and
      all(u,exists(x,'$E'(u,x))) and
      all(u,all(v,all(x,'R$'(u,v) and '$E'(v,x) => '$E'(u,x)))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,semi_functional,abbreviated,constant,
	all(u,all(v,all(x,'R$'(u,v::x)))) and
      all(u,all(x,'$E'(u,x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,semi_functional,full,constant,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w)))) and
      all(u,all(x,'$E'(u,x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,semi_functional,abbreviated,varying,
	all(u,all(v,all(x,'R$'(u,v::x)))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,semi_functional,full,varying,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w)))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,functional,_,increasing,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(x,y) => exists(z,(u::z) = y)))) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(u,y) => exists(z,(x::z) = y)))) and
      all(u,all(x,all(y,'R$'(u,x) and '$E'(u,y) => exists(z,('Ex$'(x)^^z) = y)))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,functional,_,decreasing,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(x,y) => exists(z,(u::z) = y)))) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(u,y) => exists(z,(x::z) = y)))) and
      all(u,all(x,all(y,'R$'(u,x) and '$E'(x,y) => exists(z,('Ex$'(u)^^z) = y)))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,functional,_,constant,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(x,y) => exists(z,(u::z) = y)))) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(u,y) => exists(z,(x::z) = y)))) and
      all(u,all(x,'$E'(u,x))) and
      all(u,'Normal$'(u))).
theory(folmod,kd45,functional,_,varying,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(x,y) => exists(z,(u::z) = y)))) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(u,y) => exists(z,(x::z) = y)))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,'Normal$'(u))).

theory(propmod,s5,relational,_,_,
	all(u,'R$'(u,u)) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w))))).
theory(propmod,s5,semi_functional,abbreviated,_,
	all(u,all(v,'R$'(u,v))) and
      all(u,'Normal$'(u))).
theory(propmod,s5,semi_functional,full,_,
	all(u,all(x,'R$'(u,u::x))) and
	all(u,all(v,'R$'(u,u))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w)))) and
      all(u,'Normal$'(u))).
theory(propmod,s5,functional,_,_,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,exists(x,u::x = u)) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(u,y) => exists(z,(x::z) = y)))) and
      all(u,'Normal$'(u))).
theory(folmod,s5,relational,_,_,
	all(u,'R$'(u,u)) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => 'R$'(v,w)))) and
      all(u,all(x,'$E'(u,x)))).
theory(folmod,s5,semi_functional,abbreviated,_,
	all(u,all(v,'R$'(u,v))) and
      all(u,all(x,'$E'(u,x))) and
      all(u,'Normal$'(u))).
theory(folmod,s5,semi_functional,full,_,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,x))) and
      all(u,exists(x,u::x = u)) and
      all(u,all(x,all(y,exists(z,(u::x::z) = (u::y))))) and
      all(u,'Normal$'(u))).
theory(folmod,s5,functional,_,_,
	all(u,all(x,'R$'(u,u::x))) and
      all(u,all(x,'$E'(u,'Ex$'(u)^^x))) and
      all(u,exists(x,u::x = u)) and
      all(u,all(x,all(y,'R$'(u,x) and 'R$'(u,y) => exists(z,(x::z) = y)))) and
      all(u,'Normal$'(u))).

theory(propmod,s42,relational,full,_,
	all(u,'R$'(u,u)) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => exists(x,'R$'(v,x) and 'R$'(w,x)))))).
theory(propmod,s42,relational,abbreviated,_,
	all(u,'R$'(u,u)) and
      all(u,'R$'('o$',u)) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      exists(x,all(y,'R$'(y,x)))).
theory(propmod,s42,semi_functional,full,_,
	all(u,'R$'(u,u)) and
      all(u,all(x,'R$'(u,u::x))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(v,w) => 'R$'(u,w)))) and
      all(u,all(v,all(w,'R$'(u,v) and 'R$'(u,w) => exists(x,'R$'(v,x) and 'R$'(w,x)))))).
theory(propmod,s42,semi_functional,abbreviated,_,
	all(u,'R$'(u,u)) and
      all(u,'R$'('o$',u)) and
      all(u,all(v,all(x,'R$'(u,v) => 'R$'(u,v::x)))) and
      exists(x,all(y,'R$'(y,x)))).
