% Assigns to a variable a unique integer
% used for naming Boolean variables
% Note that always a list of all known Boolean variables in the 
% call should be given when used in bvars2ints. Since only then
% all occurences of variables are maintained in its attribute
% if all are linked via some constraint.
% For retrieving then give at most one variable.
% Author: pb

:- module(booleanvars,[
	var_init/0,   % initialize counter of vars to start with 1
	bvars2ints/2, % if you are sure to have just BVars
	blits2ints/2, % or if you have BVars or Literals
	knownblit2int/2, % if BVar is surely already assigned
	bints2lits/3  % 3rd arg: either a BVar or a list of BVars
]).

:- use_module(library(atts)).

:- attribute intrep/2.

var_init :-
	bb_put(booleanvariablesnextfree,1).

bvars2ints(Vs,Is) :-
	do_bvars2ints(Vs,Is),
	merge_vars_list(Vs,[],AllVs),
	replace_vars_list(AllVs,AllVs).

blits2ints(Ls,Is) :-
	do_blits2ints(Ls,Vs,Is),
	merge_vars_list(Vs,[],AllVs),
	replace_vars_list(AllVs,AllVs).

knownblit2int(V,I) :- var(V),!,knownbvar2int(V,I).
knownblit2int('~'(V),I) :- var(V),knownbvar2int(V,NI),I is -NI.

knownbvar2int(V,I) :- 
	var(V),
	get_atts(V,intrep(I,_)).

do_bvars2ints([],[]).
do_bvars2ints([V|Vs],[I|Is]) :-
	do_bvar2int(V,I),
	do_bvars2ints(Vs,Is).
	
do_bvar2int(V,N) :-
	(get_atts(V,intrep(N,_)) ->
	    true
	;
	    bb_get(booleanvariablesnextfree,N),
	    put_atts(V,intrep(N,[V])), % list is updated later; self always present
	    N1 is N+1,
	    bb_put(booleanvariablesnextfree,N1)
	).

do_blits2ints([],[],[]).
do_blits2ints([L|Ls],[V|Vs],[I|Is]) :-
	do_blit2int(L,V,I),
	do_blits2ints(Ls,Vs,Is).

do_blit2int(L,V,N) :-
	(var(L) -> 
	    do_bvar2int(L,N),
	    V = L	
	;
	    L = '~'(V),
	    var(V),
	    do_bvar2int(V,NN),
	    N is -NN
	).

merge_vars_list([],AVs,AVs).
merge_vars_list([V|Vs],AVs0,AVs2) :-
	(var(V) ->
	    get_atts(V,intrep(_,L)),
	    merge_two_lists(L,AVs0,AVs1),
	    merge_vars_list(Vs,AVs1,AVs2)
	;
	    merge_vars_list(Vs,AVs0,AVs2)
	).

merge_two_lists([],AVs,AVs).
merge_two_lists([V|Vs],AVs0,AVs2) :-
	(var(V) ->
	    (var_member(V,AVs0) ->
		AVs1 = AVs0
	    ;
		AVs1 = [V|AVs0]
	    )
	;
	    AVs1 = AVs0
	),
	merge_two_lists(Vs,AVs1,AVs2).

var_member(V,[VV|_]) :- V == VV,!.
var_member(V,[_|VVs]) :- var_member(V,VVs).
	
replace_vars_list([],_).
replace_vars_list([V|Vs],AllVs) :-
	get_atts(V,intrep(N,_)),
	put_atts(V,intrep(N,AllVs)),
	replace_vars_list(Vs,AllVs).
	
	
do_blit2int(V,N) :-
	(var(V) -> 
	    bvar2int(V,N)
	;
	    V = '~'(NV),
	    var(NV),
	    do_bvar2int(NV,NN),
	    N is -NN
	).

bints2lits(Ns,Ls,BV) :-
	var(BV),!,
	get_atts(BV,intrep(_,AllBVars)),
	do_bints2lits(Ns,Ls,AllBVars).
bints2lits(Ns,Ls,[BV|BVs]) :-
	var(BV),!,
	del_constants([BV|BVs],[Var|VarList]),
	bvars2ints([Var|VarList],_), % update all bindings	
	get_atts(Var,intrep(_,AllBVars)),
	do_bints2lits(Ns,Ls,AllBVars).
bints2lits(Ns,Ls,[_BV|BVs]) :-
	bints2lits(Ns,Ls,BVs).

del_constants([],[]).
del_constants([BV|BVs],VarList) :-
	(atomic(BV) ->
	    del_constants(BVs,VarList)
	;
	    VarList = [BV|VarListTail],
	    del_constants(BVs,VarListTail)
	).


do_bints2lits([],[],_).
do_bints2lits([N|Ns],[L|Ls],AllBVars) :-
	get_blit(N,L,AllBVars),
	do_bints2lits(Ns,Ls,AllBVars).

get_blit(N,V,A) :-
	N > 0,!,
	get_bvar(N,V,A).
get_blit(N,'~'(V),A) :-
	N < 0,
	PN is -N,
	get_bvar(PN,V,A).

get_bvar(N,V,[E|R]) :-
	(get_atts(E,intrep(N,_)) ->
	    V = E
	;
	    get_bvar(N,V,R)
	).
	


% Do nothing special in this module for unification
% must always be called with a variable (or 0 and 1?)
verify_attributes(_V,Val,[]) :- 
	(var(Val) ; Val == 0; Val == 1).

% collect for one variable all constraints
attribute_goal(Var,Goal) :-
	(var(Var) ->
	    get_atts(Var,intrep(_,_)),
	    (bb_get(collected,no) -> % the first time succeeded 
		bb_put(collected,yes),
		interface:pb_getlpbis(LPBIS,Var),
		Goal = LPBIS  % pretty printing via portray	    
	    ;
		fail 
	    )
	;
	    fail
	).

% 
project_attributes(QueryVars,CreatedVars) :-
	% just assure that all variables are connected
	bb_put(collected,no),
	merge_vars_list(QueryVars,CreatedVars,AllVs),
	replace_vars_list(AllVs,AllVs).


	

% initialize black board
:- var_init.


/*
?- var_init,blits2ints([A,~B,C],L),blits2ints([D,E,F,B],LL),bints2lits([2,3],[X,Y],D).

L = [1,-2,3],
X = B,
Y = C,
LL = [4,5,6,2] ? 
*/

/*
test :-
	user:cputime(T0),
	List = [A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z],
	blits2ints(List,_),
	ntimes(test_single(List),100),
	user:cputime(T1),
	Time is T1-T0,
	format("Time used (~w)",[Time]).

ntimes(G,N) :-
	N == 0,!.
ntimes(G,N) :-
	call(G),
	N1 is N-1,
	ntimes(G,N1).

test_single([]).
test_single([A|L]) :-
	knownblit2int(A,_),
	test_single(L).
*/