% the interface to the pseudo-Boolean constraint solver
% Author: pb

:- module(interface,[
% tell the constraint store
	pb_caddlpbis/1,
% tell without backup
	pb_caddlpbis_nobackup/1,
% optimization
	pb_maximize/2,
	pb_minimize/2,
% initialization stuff
	pb_init/0,
	pb_set_option/2,
% special stuff
	pb_fixing/0,
	pb_fixingcheck/0,
	pb_eqcheck/0,
	pb_coeffred/0,
	pb_deletesimpimplied/0,
% get information from constraint store
	pb_getlpbis/2,
	pb_getnocovers/1,
	pb_getnonodes/1,
	pb_getnofixed/1
%	pb_getstate/4 % Ineqs, Equations, FixedLits, Consistent
]).

:- use_module(library(atts)).
:- use_module(booleanvars).

% pb_caddlpbis(LPBIS)
% The list of linear pseudo-Boolean inequaltiies is added to the
% internal solver.
% The actual state is backed up and restored on backtracking.
pb_caddlpbis(LPBIS) :-
%	write('1 '),pb_dump_state,
	p_backup,
%	write('2 '),pb_dump_state,
	pb_collect_vars(LPBIS,[],Vars),
	bvars2ints(Vars,_), % update the var to int structure
	pb_do_addlpbis(LPBIS),
	p_coversolve(1),      % assure consistency
	pb_fixing,
%	pb_fixingcheck, % fixingcheck is harder!
	( Vars = [BV|_] -> % Only if there are variables in the constraint
	    booleanvars:get_atts(BV,intrep(_,AllVars)),
	    pb_dofix(AllVars)
	;
	    true
	).
%	write('3 '),pb_dump_state.
pb_caddlpbis(_LPBIS) :-
	p_restore, % restore the state of the constraint solver
%	write('4 '),pb_dump_state,
        fail.      % and fail since add_lpbis is not backtrackable.



pb_caddlpbis_nobackup(LPBIS) :-
	pb_collect_vars(LPBIS,[],Vars),
	bvars2ints(Vars,_), % update the var to int structure
	pb_do_addlpbis(LPBIS),
	p_coversolve(1),      % assure consistency
	pb_fixing,
	( Vars = [BV|_] -> % Only if there are variables in the constraint
	    booleanvars:get_atts(BV,intrep(_,AllVars)),
	    pb_dofix(AllVars)
	;
	    true
	).

pb_collect_vars([],Vars,Vars).
pb_collect_vars([lpbi(CoeffLits,_)|LPBIs],Vars0,Vars2) :-
	pb_collect_vars_cl(CoeffLits,Vars0,Vars1),
	pb_collect_vars(LPBIs,Vars1,Vars2).

pb_collect_vars_cl([],Vars,Vars).
pb_collect_vars_cl([_-Lit|CoeffLits],Vars0,Vars2) :-
	(var(Lit) ->
	    Var = Lit
	;
	    Lit = '~'(Var)
	),
	(var_member(Var,Vars0) ->
	    Vars1 = Vars0
	;
	    Vars1 = [Var|Vars0]
	),
	pb_collect_vars_cl(CoeffLits,Vars1,Vars2).

var_member(V,[VV|_]) :- V == VV,!.
var_member(V,[_|VVs]) :- var_member(V,VVs).

pb_do_addlpbis([]).
pb_do_addlpbis([LPBI|LPBIs]) :-
	pb_addlpbi(LPBI),
	pb_do_addlpbis(LPBIs).

pb_addlpbi(lpbi(CoeffLits,Rhs)) :-
	make_list_coefflits(CoeffLits,[Rhs],List),
	p_term2intlist(List),
	p_addlpbi.

make_list_coefflits([],List,List).
make_list_coefflits([C-L|CLs],List0,[C,I|List1]) :-
%	blits2ints([L],[I]),
	knownblit2int(L,I),  % since already converted at the beginning
	make_list_coefflits(CLs,List0,List1).

pb_init :- 
	p_initialize,
	var_init,
	(bb_get(firstinitialized,no) ->
	    true
	;
	    bb_put(firstinitialized,no),
	    pb_set_option(maxcutnodes,500),
	    pb_set_option(verbosity,0),
	    pb_set_option(enumverbosity,0),
	    pb_set_option(heuristic,1),
	    pb_set_option(dellitheu,4)
	),
	p_resetcoverstatistics,
	p_resetenumstatistics.
	
pb_maximize(Objf,Max) :-
	pb_collect_vars_cl(Objf,[],Vars),
	bvars2ints(Vars,_), % update the var to int structure
	make_list_coefflits(Objf,[],List),
	p_term2intlist(List),
	p_maximize(Max).

pb_minimize(Objf,Min) :-
	pb_collect_vars_cl(Objf,[],Vars),
	bvars2ints(Vars,_), % update the var to int structure
	make_list_coefflits(Objf,[],List),
	p_term2intlist(List),
	p_minimize(Min).



pb_set_option(Option,Number) :-
	(integer(Number) ->
	    pb_set_option2(Option,Number)
	;
	    format("ps_set_option : Number expected, ~w found ~n",[Number])
	).

pb_set_option2(heuristic,Number) :- p_setheuristic(Number).
pb_set_option2(maxcutnodes,Number) :- p_setmaxcutnodes(Number).
pb_set_option2(dellitheu,Number) :- p_setdellitheu(Number).
pb_set_option2(verbosity,Number) :- p_setverbosity(Number).
pb_set_option2(enumverbosity,Number) :- p_setenumverbosity(Number).


pb_fixing :- 
	p_fixing(X), 
	(X > 0 -> 
	    format("~w Literals fixed ~n",[X]) 
	; 
	    true
	).
pb_fixingcheck :- 
	p_fixingcheck(X), 
	(X > 0 -> 
	    format("~w Literals fixed ~n",[X]) 
	; 
	    true
	).

pb_eqcheck :- 
	p_eqcheck(X), 
	(X > 0 -> format("~w Equations detected ~n",[X]) ; true).
pb_coeffred :- 
	p_coeffred(X), 
	(X > 0 -> format("~w Coefficients reduced ~n",[X]) ; true).

pb_delsimpimplied :- 
	p_deleteimplied(X), 
	(X > 0 -> format("~w simply implied ineqs deleted ~n",[X]) ; true).

pb_dofix(BV) :- 
	p_getfixed,
	p_intlist2term(List),	
	fix_new_fixed(List,BV).

fix_new_fixed([],_).
fix_new_fixed([I|List],BV) :-
	(bints2lits([I],[FL],BV) ->
	    (var(FL) ->
		FL = 1
	    ;
		FL = '~'(0)
	    )
	;
	    true % already fixed
	),
	fix_new_fixed(List,BV).

pb_dump_state :-
	p_first,
	format("dumping~n",[]),
	do_dump_state,
	p_getfixed,
	p_intlist2term(List),
	format("fixed ~w~n",[List]).
	

do_dump_state :- p_last(1),nl,!.
do_dump_state :- p_next,p_intlist2term(List),format("~w",[List]).

pb_getlpbis(LPBIs,BV) :-
	p_first,
	do_getlpbis(LPBIs,BV).

do_getlpbis(LPBIs,_BV) :-
	p_last(1),!,LPBIs = [].
do_getlpbis([lpbi(CoeffLits,Rhs)|LPBIs],BV) :-
	p_next,
	p_intlist2term(List),
	list_to_clr(List,CoeffLits,Rhs,BV),
	do_getlpbis(LPBIs,BV).

list_to_clr([Rhs],[],Rhs,_BV) :- !.
list_to_clr([Coeff,IntLit|List],[Coeff-Lit|CoeffLits],Rhs,BV) :-
	bints2lits([IntLit],[Lit],BV),
	list_to_clr(List,CoeffLits,Rhs,BV).


pb_getnocovers(X) :- p_getnocovers(X).
pb_getnonodes(X) :- p_getnonodes(X).
pb_getnofixed(X) :- p_getnofixed(X).

pb_enum([]).
pb_enum([V|R]) :-
	(var(V) ->
	    (
		pb_caddlpbis([lpbi([1-V],1)])
	    ;
		pb_caddlpbis([lpbi([1- '~'(V)],1)])
	    )
	;
	    true
	),
	pb_enum(R).





% do the external stuffs and loadings
:- dynamic foreign/2, foreign_file/2.

foreign(intlist2term,c,p_intlist2term([-term])).
foreign(term2intlist,c,p_term2intlist(+term)).

foreign(c_backup,c,p_backup).
foreign(c_restore,c,p_restore).

foreign(c_initialize,c,p_initialize).
foreign(c_addlpbi,c,p_addlpbi).

foreign(c_fixing,c,p_fixing([-integer])).
foreign(c_fixingcheck,c,p_fixingcheck([-integer])).
foreign(c_eqcheck,c,p_eqcheck([-integer])).
foreign(c_coeffred,c,p_coeffred([-integer])).
foreign(c_coeffredm,c,p_coeffredm([-integer])).
foreign(c_deleteimplied,c,p_deleteimplied([-integer])).
foreign(c_resetenumstatistics,c,p_resetenumstatistics).
foreign(c_resetcoverstatistics,c,p_resetcoverstatistics).
foreign(c_getnocovers,c,p_getnocovers([-integer])).
foreign(c_getnonodes,c,p_getnonodes([-integer])).
foreign(c_getnofixed,c,p_getnofixed([-integer])).
foreign(c_addcover,c,p_addcover(-integer,[-integer])).
foreign(c_coversolve,c,p_coversolve([-integer])).
foreign(c_solve,c,p_solve([-integer])).
foreign(c_maximize,c,p_maximize([-integer])).
foreign(c_minimize,c,p_minimize([-integer])).

foreign(c_getfixed,c,p_getfixed).
foreign(c_getsol,c,p_getsol).
foreign(c_first,c,p_first).
foreign(c_last,c,p_last([-integer])).
foreign(c_next,c,p_next).
foreign(c_setheuristic,c,p_setheuristic(+integer)).
foreign(c_setmaxcutnodes,c,p_setmaxcutnodes(+integer)).
foreign(c_setdellitheu,c,p_setdellitheu(+integer)).
foreign(c_setverbosity,c,p_setverbosity(+integer)).
foreign(c_setenumverbosity,c,p_setenumverbosity(+integer)).

foreign_file('interface.o',[c_backup,c_restore,c_initialize,c_addlpbi,c_fixing,c_fixingcheck,c_eqcheck,c_coeffred,c_coeffredm,c_deleteimplied,c_resetenumstatistics,c_resetcoverstatistics,c_getnocovers,c_getnonodes,c_getnofixed,c_addcover,c_coversolve,c_solve,c_maximize,c_minimize,c_getfixed,c_getsol,c_first,c_last,c_next,c_setheuristic,c_setmaxcutnodes,c_setdellitheu,c_setverbosity,c_setenumverbosity]).
foreign_file('intlist.o',[intlist2term,term2intlist]).

:- load_foreign_files(['intlist.o','interface.o'],['CLPPB/libpbcs.a','/opt/SUNWspro/SC3.0.1/lib/libC.a']).

do_load_once :-
	(bb_get(loaded,yes) ->
	    true
	;
	    bb_put(loaded,yes),
	    do_load
	).

% :- do_load_once.


