
:- module(cp,[constructor_abstraction/3,
	      constructor_inconsistency/3,
	      critical_pairs/3,
	      reflexivity_resolution/3]).

:- use_module(library(maplist),[maplist/2,maplist/3]).
:- use_module(formulas).
:- use_module(st_terms).
:- use_module(rewrite,[reducible/2]).

:- use_module(global_vars, [cont/2]).

:- use_module(theory,[delete_active/3,
		      type_check_active/3,
		      unify_active/3,
		      unify/3]).
:- use_module(library(basics),[member/2]).



critical_pairs(Ru,	% +Ru : Rule
	       Active,  % +Active : Theory
	       Eqs      % ?Eqs : [Equation]  minimally instantiated
	      ):-
	findall(Eq,cp_into(Ru,Active,Eq),CP1),
	numbers_simpified_active(CP1,Deleted),

	% rules to be deleted are excluded from cp-from computation
	findall(Eq,cp_from(Ru,Active,Deleted,Eq),CP2),

	% putting all together
	user:append(CP1,CP2,CP),
	maplist(mk_equation,CP,Eqs).



	



cp_into(Ru,Active,eq(ST,TT,TyST,PST)):-
	   % cp of Ru into a rule Ru1 in Active
	cont(ord,ord(OType,OParams)),
	
	   % consider all feasible orientations L -> R of Ru
	ord_rule(Ru,LL,RR),
	term2prolog((LL,RR),(L,R)),
	
	number(Ru,N),
%	(N=19 -> trace; true),
	orientation(Ru,Orient),
	
	   % any rule L1 -> T in Active
	unify_active(L1,Active,eq(T,Ru1)),
	number(Ru1,N1),
	N\==N1,
	type(Ru1,TyST),

	  % select a non-variable subterm Sub of L1
	  % root overlaps are dealt with by cp_from
	path_arg_(P,L1,Sub),
	nonvar_(Sub),

	  % type check and unify with L
	type_check_active(Active,Sub,Ru),
	unify(Active,L,Sub),

	  %  check whether rule instances are ordered
	(Orient=check ->
	    \+ OType:greater_eq(OParams,R,L)
	;
	    true
	),
	orientation(Ru1,OR),
	(OR=check ->
	    \+ OType:greater_eq(OParams,T,L1)
	;
	    true
	),

	  % left and below the superposition subterm the
	  % term should be irreducible
%	irreducibleLeftOf(P,Active,L1),
%	irreducibleSubterms(Active,Sub),

	  % compute L1[R]_P, that is, replace the subterm in L1 by R
	change_path_arg_(P,L1,S,R),
 

	  % convert cp from Prolog terms to st_terms
	prolog2term((S,T),(ST,TT)),

	  % add proof information
	(   left(Ru1,L1),   % the unifier of L and Sub applied to L1 matches L1
	    (   Orient\==check  ; OType:greater(OParams,L,R)  )
		            % the instance of Ru is reducing
	->
	    PST=inference(reduction_max,eq(Ru1),[N])
%	    format('~Ncp: ~p',[(ST=TT) : PST])
	;
				% filter out some trivial cases
	    S\==T,
	    PST=inference(cp,[N,N1])
	).

	
ord_rule(Ru1,L,R):-
	orientation(Ru1,Orient),
	(Orient=lr ->
	    left(Ru1,L),
	    right(Ru1,R)
	;Orient=rl ->
	    left(Ru1,R),
	    right(Ru1,L)
	;% check
	    (   left(Ru1,L),
		right(Ru1,R)
	    ;
		left(Ru1,R),
		right(Ru1,L)
	    )
	).



cp_from(Ru1,Active,Excluded,eq(ST,TT,TyST,PST)):-
	% cp of any rule in Active that is not excluded by Exclude into
	% any orientation of Ru1
	cont(ord,ord(OType,OParams)),
	ord_rule(Ru1,L11,T1),
	term2prolog((L11,T1),(L1,T)),
	type(Ru1,TyST),
	number(Ru1,NST),

	% select a position in L1
	path_arg_(P,L1,Sub),
	P\==[],
	nonvar_(Sub),
	
	% select any rule in Active
	unify_active(Sub,Active,eq(R,Ru)),  % use only unifiable index rules
	number(Ru,N),
	\+ member(N,Excluded),
	type_check_active(Active,Sub,Ru),

	orientation(Ru,Orient),

	(Orient=check ->
	    \+ OType:greater_eq(OParams,R,Sub)
	;
	    true
	),
	orientation(Ru1,OR),
	(OR=check ->
	    \+ OType:greater_eq(OParams,T,L1)
	;
	    true
	),
	
%	irreducibleLeftOf(P,Active,L1),
%	irreducibleSubterms(Active,Sub),

	
	change_path_arg_(P,L1,S,R),
	S\==T,
	prolog2term((S,T),(ST,TT)),
	
	PST=inference(cp,[N,NST]).






reflexivity_resolution(Thy,E,[In]):-
	type(E,prop),
	left(E,L),
	functor_(L,'$neq',2),
	term2prolog(L,LP),
	arg_(1,LP,S),
	arg_(2,LP,T),
	unify(Thy,S,T),
	number(E,N),
	!,
	mk_equation('$false','$true',prop,
		    inference(reflexivity,[N]),In).
reflexivity_resolution(_Thy,_E,[]).



/* ------------------------------------------------------------------------------

 Constructor stuff


------------------------------------------------------------------------------ */
/*
constructor_abstraction(sig(Sig,cons(Conss)),E,Eqs):-
	Conss\==[],
	left(E,L),
	right(E,R),
	(   constructor_context(Conss,Path,L,Sub),
	    S=L,
	    T=R
	;
	    constructor_context(Conss,Path,R,Sub),
	    S=R,
	    T=L
	),
	!,
	lpoc:newconstant(C),
	equational:term_type(Sig,Sub,Ty),
	abstraction(Path,S,T,E,C,Sub,Ty,Eqs).



abstraction(P,S,T,E,C,Sub,TySub,[E1,E2]):-
	change_path_arg_(P,S,S1,C),
	number(E,N),
	type(E,Ty),
	mk_equation(S1,T,Ty,inference(abstraction_c,[N]),E1),
	mk_equation(C,Sub,TySub,inference(abstraction,[N]),E2).


*/

constructor_abstraction(sig(Sig,cons(Conss)),E,Eqs):-
	Conss\==[],
	left(E,L),
	right(E,R),
	(   constructor_context(Conss,_Path,L,Sub)
	;
	    constructor_context(Conss,_Path,R,Sub)
	),
	!,
	lpoc:newconstant(C),
	equational:term_type(Sig,Sub,Ty),
	abstraction(E,C,Sub,Ty,Eqs).
constructor_abstraction(_,_,[]).



constructor_context(Conss,[I|Path],T,Sub):-
	nonvar_(T),
	\+ object_var_(T),
	functor_(T,F,_),
	member(F,Conss),
	genarg_(I,T,TI),
	nonproper_constructor_context(Conss,Path,TI,Sub),
	!.

	   
nonproper_constructor_context(Conss,Path,T,Sub):-
	nonvar_(T),
	\+object_var_(T),
	functor_(T,F,N),
	(member(F,Conss) ->
	    constructor_context(Conss,Path,T,Sub)
	;
	    N>0,    % ok for Sig constants too????
	    Path=[],
	    Sub=T
	).

abstraction(E,C,Sub,TySub,[E2]):-
	number(E,N),
	mk_equation(C,Sub,TySub,inference(abstraction,[N]),E2).




constructor_inconsistency(sig(_Sig,cons(Conss)),E,[In]):-
	Conss\==[],
	left(E,L),
	right(E,R),
	(constructor_equation(Conss,L,R) ->
	    \+orientation(E,rl)
	;constructor_equation(Conss,R,L) ->
	    \+orientation(E,lr)
	),
	number(E,N),
	mk_equation('$false','$true',prop,
		    inference(constructor_collapse,[N]),In).
constructor_inconsistency(_,_,[]).





constructor_term(Conss,T):-
	nonvar_(T),
	univ_(T,[F|Args]),
	member(F,Conss),
	maplist(constructor_subterm(Conss),Args).

constructor_subterm(_Conss,T):-
	constant(T),
	!.
constructor_subterm(Conss,T):-
	constructor_term(Conss,T).


constructor_equation(Conss,L,R):-
	constant(R),
	constructor_term(Conss,L).

constant(T):-
	nonvar_(T),
	\+object_var_(T),
	functor(T,_,0).









/* -----------------------------------------------------------------------------

Redundancy checking for critical pairs

------------------------------------------------------------------------------ */




irreducibleLeftOf([],_,_).
irreducibleLeftOf([I|Is],Rs,T):-
	irreducibleLeftOf(I,Rs,T),
	arg_(I,T,TI),
	irreducibleLeftOf(Is,Rs,TI).
irreducibleLeftOf(1,_,_):-
	!.
irreducibleLeftOf(I,Rs,T):-
	number(I),
	J is I-1,
	arg_(J,T,TJ),
	\+ (   numbervars(TJ,0,_),
	       reducible(TJ,Rs)  % ,write('*')
	   ),
	irreducibleLeftOf(J,Rs,T).


irreducibleSubterms(_,T):-
	var_(T),
	!.
irreducibleSubterms(Rs,T):-
	functor_(T,_F,N),
	N1 is N+1,
	irreducibleLeftOf(N1,Rs,T).



/* -----------------------------------------------------------------------------

Deletion

------------------------------------------------------------------------------ */

numbers_simpified_active([eq(_S,_T,_Ty,inference(reduction_max,eq(Ru),_))|Eqs],
			[N|Ns]):-
	!,
	number(Ru,N),
	numbers_simpified_active(Eqs,Ns).
numbers_simpified_active([_|Eqs],Ns):-
	numbers_simpified_active(Eqs,Ns).
numbers_simpified_active([],[]).
