% -*- Module: st_terms; -*-

:- module(st_terms,[
		    arg_/3,
		    change_path_arg_/4,
		    contains_var/2,
		    functor_/3,
		    genarg_/3,
		    is_function_symbol/1,
		    is_term/1,
		    nonvar_/1,
		    object_var_/1,
		    path_arg_/3,
		    permuted_term/2,
		    prolog2term/2,
		    sample_terms/1,
		    term2prolog/2,
		    univ_/2,
		    var_/1
		    ]).

:- use_module(satsrc(mylists),
	      [
	       is_list/1,
	       nth1/3
	      ]).

:- use_module(library(change_arg),[change_arg/5]).


/* ---------------------------------------------------------------------------

Terms as Prolog Terms

In this representation we assume that type information
will be computed as needed when given additional
information about the environment.
The rationale here is that for most operations
type information is irrelevant.

We might have to extend this to higher-order syntax as
some decision procedures might accept full first-order
constraints.


Term            ::= Prolog_term 
Object_Variable ::= '$VAR'(N)
N               ::= Prolog_nat
Function_Symbol ::= Prolog_atom


------------------------------------------------------------------------------ */


/* ----------------------------------------------------------------------------

Type checking

------------------------------------------------------------------------------ */

is_term(T):-      % incomplete test
	nonvar(T).
is_function_symbol(F):-
	atomic(F).



/* ----------------------------------------------------------------------------

Primitive operations on terms as we have them for Prolog terms

------------------------------------------------------------------------------ */


nonvar_(X):-
	nonvar(X).

% we assume that terms contain no Prolog variable
var_(X):-
	var(X).
   
object_var_(X):-
	nonvar(X),
	X='$VAR'(_).
   

functor_(T,F,N):-
	functor(T,F,N).


arg_(I,T,TI):-
	arg(I,T,TI).


:- use_module(library(arg),[genarg/3]).
genarg_(N, Term, Arg) :- genarg(N, Term, Arg).


univ_(T,FTs):- T=..FTs.


/* ----------------------------------------------------------------------------

                Nonprimitive operations

Some code is adapted from Quintus Prolog library occurs.pl


------------------------------------------------------------------------------ */



%   contains_var(+Variable, +Term)
%   is true when the given Term contains at least one sub-term which
%   is identical to the given Variable.  We use '=='/2 to check for
%   the variable so it can be used to check
%   for arbitrary terms, not just variables.

contains_var(Variable, Term) :-
	\+ free_of_var(Variable, Term).


%   free_of_var(+Variable, +Term)
%   is true when the given Term contains no sub-term identical to the
%   given Variable (which may actually be any term, not just a var).
%   For variables, this is precisely the "occurs check" which is
%   needed for sound unification.

free_of_var(Variable, Term) :-
	Term == Variable,
	!,
	fail.
free_of_var(Variable, Term) :-
	nonvar(Term),
	!,
	functor(Term, _, Arity),
	free_of_var(Arity, Term, Variable).
free_of_var(_, _).

free_of_var(1, Term, Variable) :- !,
	arg(1, Term, Argument),
	free_of_var(Variable, Argument).
free_of_var(N, Term, Variable) :-
	arg(N, Term, Argument),
	free_of_var(Variable, Argument),
	M is N-1, !,
	free_of_var(M, Term, Variable).




path_arg_( [], T, T ).
path_arg_([I|Path], S, Sub ):-
        nonvar_( S ),
	genarg_(I,S,SI),
	path_arg_(Path,SI,Sub).


change_path_arg_([],_T,New,New).
change_path_arg_([I|Path], S, New,NewArg ):-
	change_arg(I,S,SI,New,NewI),
	change_path_arg_(Path,SI,NewI,NewArg).





/* ----------------------------------------------------------------------------

                Conversions


------------------------------------------------------------------------------ */

prolog2term(TP,T):-
	copy_term(TP,T),
	numbervars(T,0,_).

term2prolog(T,TP):-
	term2prolog(T,TP,_EmptySubst),
	!.

term2prolog('$VAR'(N),X,Subst):-
	lookup(Subst,N,X).
term2prolog(T,TP,Subst):-
	functor_(T,F,N),
	functor_(TP,F,N),
	term2prolog(N,T,TP,Subst).


term2prolog(N,T,TP,Subst):-
	N>0,
	!,
	arg_(N,T,TN),
	arg_(N,TP,TNP),
	term2prolog(TN,TNP,Subst),
	N1 is N-1,
	term2prolog(N1,T,TP,Subst).
term2prolog(0,_,_,_Subst).

lookup([N-X|_Subst],N,X):-
	!.
lookup([_|Subst],N,X):-
	lookup(Subst,N,X).



	


permuted_term(T,TP):-
	term2prolog(T,TP,Subst),
	unzip(Subst,Indexes,PrologVars),
	!,
	permutation(Indexes,IndexesPermuted),
	zip(SubstP,IndexesPermuted,PrologVars),
	instantiate(SubstP).


unzip(X,_Y,_Z):-
	var(X),
	!.
unzip([K-I|X],[K|Y],[I|Z]):-
	unzip(X,Y,Z).

zip(_X,Y,_Z):-
	var(Y),
	!.
zip([K-I|X],[K|Y],[I|Z]):-
	zip(X,Y,Z).


instantiate(X):-
	var(X),
	!.
%instantiate([I-'$VAR'(I)|X]):-
% we want constants for the variables
instantiate([I-I|X]):-
	instantiate(X).

permutation(X,X):-
	var(X).
permutation(Xs,[X|Zs]):-
	select(X,Xs,Ys),
	permutation(Ys,Zs).

select(_,X,_):-
	var(X),
	!,
	fail.
select(X,[X|Xs],Xs).
select(X,[Y|Xs],[Y|Ys]):-
	select(X,Xs,Ys).

