:- module(theory,[add_actives/3,
		  empty_actives/1,
		  empty_actives/2,
		  extend_ordering/2,
		  delete_active/3,
		  constructor_equation/2,
		  inferences/3,
		  insert_active/3,
		  match_active/4,
		  redundant/2,
		  type_check_active/3,
		  type_check_type/3,
		  unify/3,
		  unify_active/3]).


:- use_module(library(change_arg),[change_arg/4]).
:- use_module(formulas,[left/2,
			left/3,
			mk_equation/5,
			number/2,
			orientation/2,
			orientation/3,
			right/2,
			right/3,
			type/2]).

:- use_module(satsrc(substitutiontree)).
:- use_module(cp,[constructor_abstraction/3,
		  constructor_inconsistency/3,
		  critical_pairs/3,
		  reflexivity_resolution/3]).
:- use_module(cong_closure,[]).
:- use_module(global_vars, [
        cont/2,
	store/2
   ]).
:- use_module(st_terms).
:- use_module(library(flatten),[flatten/2]).

/* ------------------------------------------------------------------------------

This module implements theory-specific inferences and
operations on their sets of active equations (also called rules).



Theory   ::= equational(Signature,STree)               % unit equations over
                                                       % the empty theory
          |  cong_closure(STree)


------------------------------------------------------------------------------ */




unify_active(L,equational(_,Rs),Ru):-
	equational:unify_rule(L,Rs,Ru).
unify_active(L,cong_closure(Rs),Ru):-
	cong_closure:unify_rule(L,Rs,Ru).

unify(equational(_Sig,_),S,S):-
	prolog:'$acyclic'(S).
unify(cong_closure(_),S,S).




/* ------------------------------------------------------------------------------

Find matching rewrite rule

------------------------------------------------------------------------------ */


match_active(equational(Sig,Rs),Term,R,Ru):-
	equational:match_rule(Sig,Rs,Term,R,Ru).
match_active(cong_closure(Rs),Term,R,Ru):-
	cong_closure:match_rule(Rs,Term,R,Ru).





/* ------------------------------------------------------------------------------

Insering and deleting active rules

------------------------------------------------------------------------------ */


add_actives(S,S,[]).
add_actives(S,S1,[E|Es]):-
	number(E,_), % give it some number
	insert_active(E,S,S2),
	add_actives(S2,S1,Es).




% Rules are assumed to be either oriented from left to right
% or else not uniformly orientable

insert_active(Ru,equational(Sig,Rs),equational(Sig,Rs1)):-
	equational:insert_rule(Ru,Rs,Rs1).
insert_active(Ru,cong_closure(Rs),cong_closure(Rs1)):-
	cong_closure:insert_rule(Ru,Rs,Rs1).




delete_active(Ru,equational(Sig,Rs),equational(Sig,Rs1)):-
	equational:delete_rule(Ru,Rs,Rs1).
delete_active(Ru,cong_closure(Rs),cong_closure(Rs1)):-
	cong_closure:delete_rule(Ru,Rs,Rs1).

empty_actives(equational(_Sig,Rs)):-
	equational:empty_rules(Rs).
empty_actives(cong_closure(Rs)):-
	cong_closure:empty_rules(Rs).

empty_actives(equational(Sig,_),equational(Sig,Rs)):-
	equational:empty_rules(Rs).
empty_actives(cong_closure(_),cong_closure(Rs)):-
	cong_closure:empty_rules(Rs).


type_check_active(equational(sig(Sig,_),_),Term,Ty):-
	equational:type_check_rule(Sig,Term,Ty).
type_check_active(cong_closure(_),_Term,_Ty).


type_check_type(equational(sig(Sig,_),_),Term,Ty):-
	equational:type_check_type(Sig,Term,Ty).
type_check_type(cong_closure(_),_Term,_Ty).



/* ------------------------------------------------------------------------------

Inferences

------------------------------------------------------------------------------ */



inferences(equational(Sig,ActiveEqs),E,NewEqs):-
	reflexivity_resolution(equational(Sig,ActiveEqs),E,NewEqs1),
	constructor_abstraction(Sig,E,NewEqs2),
	constructor_inconsistency(Sig,E,NewEqs3),
	critical_pairs(E,equational(Sig,ActiveEqs),NewEqs4),
	flatten([NewEqs1,NewEqs2,NewEqs3,NewEqs4],NewEqs).



inferences(cong_closure(ActiveEqs),E,NewEqs):-
	cong_closure:reflexivity_resolution(E,NewEqs1),
	cong_closure:critical_pairs(E,ActiveEqs,NewEqs2),
	user:append(NewEqs1,NewEqs2,NewEqs).








	


/* ------------------------------------------------------------------------------

Redundancy Detection

------------------------------------------------------------------------------ */

redundant(E,_):-
	equational:trivial(E),
	!.
redundant(E,equational(_,Active)):-
	% subsumption check; only needed for non-orientable equs of
	% which we have both directions in the Index
	equational:subsumed_by(E,Active).



