/*
:- module(emptyTheory, [
        currentTheory/1,
	equal/2,
	extension/4,
	match/2,
	match/4,
        mgu/2,
	unifiable/2,
	unify/2

     ]).
/*

/* this module provides the basic operations
   for solving equational constraints
   over finite trees (modulo the empty theory)
   where finite trees are
   at present identified with Prolog terms.
   Equational constraints are single equations
   or lists (representing conjunctions) of equations.
*/



:- use_module(myterms, [
	apply_subst/3,
	occurs_in/2
   ]).


:- use_module(auxiliary,[lookup_ml/2]).

:- use_module(sound_unify,[sound_unify/2]).


currentTheory(empty).





%------------------------------------------------------------------------------
/*
mgu(+E,?U)

is true when U is the mgu of the equations in E.
E may be a single equation or a list of equations.
The unifier is represented as an equational problem in solved form,
i.e. a conjunction of equations Xi=Ti with pairwise different Xi
not occurring in any of the Ti.

Uses mgu/3 in which the additional argument (arg-2) serves
to keep initial parts of the final solved form.

*/
%------------------------------------------------------------------------------

mgu( L, U ):-
	mgu( L, [], U ),
	!.

mgu( [], U, U ).				% the end
mgu( [T1=T2|R], S, U ):-			% delete
	T1==T2,
	!,
	mgu( R, S, U ).
mgu( [T1=T2|R], S, U ):-			
	(var(T1) ->
	     (\+ occurs_in( T1, T2 ) ->           % occur check
		  apply_subst( T1=T2, R, R1 ), 
		  apply_subst( T1=T2, S, S1 ), 
		  mgu(R1, [T1=T2|S1], U )
	     )
	;var(T2) ->
	     (\+ occurs_in( T2, T1 ) ->           % occur check
		  apply_subst( T2=T1, R, R1 ),
		  apply_subst( T2=T1, S, S1 ),
		  mgu(R1, [T2=T1|S1], U )
	     )
	; % decompose
	     functor(T1,F,N),
	     functor(T2,F,N),
	     equalArgs( N, T1, T2, L/R ),
	     mgu( L, S, U )
	).
mgu(T1=T2,S,U):-
	mgu([T1=T2],S,U).

		  


% equalArgs(+I,+N,+S,+T,?E) is true when E is the list of unification
% problems between corresponding subterms of S and T between 
% indexes N-I+1 and N. E is represented in form of a diffence list
% E1/E2 to allow easy appending with surrounding unification problems

equalArgs(0,_,_,L/L):-
	!.
equalArgs(I,T1,T2,[A1=A2|L]/R):-
	arg(I,T1,A1),
	arg(I,T2,A2),
	I1 is I-1,
	equalArgs(I1,T1,T2,L/R).


%------------------------------------------------------------------------------
/*
equal(+S,+T)
is true when S and T are equal (for all substitutions)
*/
%------------------------------------------------------------------------------


equal(S,T):-
	S==T.


%------------------------------------------------------------------------------
/*
match(+Pattern,+Term)
is true when Term is an instance of pattern before the call;
the call instantiates Pattern correspondingly.

*/
%------------------------------------------------------------------------------

match( P, T ):- 
	subsumes_chk( P, T ),
	P = T.




%------------------------------------------------------------------------------
/*
match(+P,+T,+L1,?L2) 

is true when T = P(\sigma\cup\tau),
where \sigma and \tau are the substitutions represented 
by L1 and L2, respectively.
The latter are equational problems in solved form.

*/
%------------------------------------------------------------------------------



match( Left, T, _, _ ):-
	\+ subsumes_chk( Left, T ),
	!,
	fail.
match(Left,T,L,L1):-
	(var(Left) ->
                (lookup_ml(Left=B,L) ->
		     B==T,
		     L=L1
		;    L1=[Left=T|L]
		)
	;% nonvar(Left) ->
	        functor(Left,F,N),
		functor(T,F,N),
		match_args(N,Left,T,L,L1)
	).


match_args(0,_,_,L,L):-
	!.
match_args(I,P,T,L1,L2):-
	arg(I,P,PI),
	arg(I,T,TI),
	match(PI,TI,L1,L),
	!,
	I1 is I-1,
	match_args(I1,P,T,L,L2).





unify(A,B):-
	sound_unify(A,B).

unifiable(A,B):-
	\+ \+ sound_unify(A,B).



% no extended rules
extension(L,R,L,R).