% mscg.pl
% most specific common generalization for substitution trees
% handles indicator (star) variables as constants!
% Does not! work for nonlinear terms, and hence doesn't create
% nonlinear terms (despite what is currently wrong defined)
% Author: pb

:- module(mscg,[
	mscg_term/8,
	mscg_subst/6,
	nonemptymscg_subst/2
]).

:- use_module(substitution).
:- use_module(vars).

% mscg_term(TX,TY,Sigma1,Sigma2,TZ,Tau1,Tau2,Star).
% specialized most specific common generalization for linear terms
mscg_term(TX,TY,Sigma1,Sigma2,TZ,Tau1,Tau2,_Star) :-
	TX == TY,!, % for identical terms nothing must be done
	TX = TZ,
	Tau1 = Sigma1,
	Tau2 = Sigma2.
mscg_term(TX,TY,Sigma1,Sigma2,TZ,Tau1,Tau2,Star) :-
	var(TX),!,
	(var_memberchk(TX,Star) -> % it is an indicator variable
	    % treat it as constant; hence different top symbols
	    Tau1 = [Xj-TX|Sigma1],
	    Tau2 = [Xj-TY|Sigma2],
	    TZ = Xj
	;	  
	    TX = TZ,
	    Sigma1 = Tau1,
	    [TX-TY|Sigma2] = Tau2
	).
% TY a non star variable can not happen (P.Graf)
mscg_term(TX,TY,Sigma1,Sigma2,TZ,Tau1,Tau2,Star) :-
	% nonvar(TX),
	nonvar(TY), % TY can't be a variable but a star variable! 
                    % but then there is no chance to succeed since TX star var is handled elsewhere
	functor(TX,F,A),
	functor(TY,F,A),!, % same top symbol
	functor(TZ,F,A),   % build new term TZ with this topsymbol
	mscg_arg(A,TX,TY,Sigma1,Sigma2,TZ,Tau1,Tau2,Star).
mscg_term(TX,TY,Sigma1,Sigma2,TZ,Tau1,Tau2,_Star) :- 
	% different top symbols for TX and TY
        Tau1 = [Xj-TX|Sigma1],
	Tau2 = [Xj-TY|Sigma2],
	TZ = Xj.


mscg_arg(0,_TX,_TY,Sigma1,Sigma2,_TZ,Sigma1,Sigma2,_Star) :- !.
mscg_arg(A,TX,TY,Sigma1,Sigma2,TZ,Tau1,Tau2,Star) :-
	arg(A,TX,STX),
	arg(A,TY,STY),
	arg(A,TZ,STZ),
	mscg_term(STX,STY,Sigma1,Sigma2,STZ,Sigma1Tmp,Sigma2Tmp,Star),
	A1 is A-1,
	mscg_arg(A1,TX,TY,Sigma1Tmp,Sigma2Tmp,TZ,Tau1,Tau2,Star).

% mscg_subst(Tau,Ro,Mu,Sigma1,Sigma2,StarVars).
% mscg for substitutions
mscg_subst(Tau,Ro,Mu,Sigma1,Sigma2,StarVars) :-
	mscg_subst(Tau,Ro,[],Mu,[],Sigma1,[],Sigma2,StarVars).
	
mscg_subst([],_Ro,Mu,Mu,Sigma1,Sigma1,Sigma2,Sigma2,_Star).
mscg_subst([V-T|Tau],Ro,MuIn,MuOut,Sigma1In,Sigma1Out,Sigma2In,Sigma2Out,Star) :- 
	% first, split on dom_member(Ro,V,TT) 
	(dom_member(Ro,V,TT) ->
	    % freeze mix or divide; pass through
	    mscg_subst(V,T,TT,Tau,Ro,MuIn,MuOut,Sigma1In,Sigma1Out,Sigma2In,Sigma2Out,Star)
	;   % bind if \+ dom_member(Ro,V) and continue directly
	    mscg_subst(Tau,Ro,MuIn,MuOut,[V-T|Sigma1In],Sigma1Out,Sigma2In,Sigma2Out,Star)
	).
	    
% freeze, mix, or divide
mscg_subst(V,T,TT,Tau,Ro,MuIn,MuOut,Sigma1In,Sigma1Out,Sigma2In,Sigma2Out,Star) :- % freeze
	T == TT,!,
	mscg_subst(Tau,Ro,[V-T|MuIn],MuOut,Sigma1In,Sigma1Out,Sigma2In,Sigma2Out,Star).
mscg_subst(V,T,TT,Tau,Ro,MuIn,MuOut,Sigma1In,Sigma1Out,Sigma2In,Sigma2Out,Star) :- % mix
	nonvar(T),
	nonvar(TT),
	functor(T,F,A),
	functor(TT,F,A),!, % same top symbol
	functor(TZ,F,A),
	mscg_arg(A,T,TT,Sigma1In,Sigma2In,TZ,Sigma1Tmp,Sigma2Tmp,Star),
	mscg_subst(Tau,Ro,[V-TZ|MuIn],MuOut,Sigma1Tmp,Sigma1Out,Sigma2Tmp,Sigma2Out,Star).
mscg_subst(V,T,TT,Tau,Ro,MuIn,MuOut,Sigma1In,Sigma1Out,Sigma2In,Sigma2Out,Star) :- % divide
	mscg_subst(Tau,Ro,MuIn,MuOut,[V-T|Sigma1In],Sigma1Out,[V-TT|Sigma2In],Sigma2Out,Star).


nonemptymscg_subst([V-T|_],Ro) :-
	dom_member(Ro,V,TT),
	equal_or_same_top(T,TT),!.
nonemptymscg_subst([_|Tau],Ro) :-
	nonemptymscg_subst(Tau,Ro).

equal_or_same_top(T,TT) :-
	T == TT,!.
equal_or_same_top(T,TT) :-
	nonvar(T),
	nonvar(TT),
	functor(T,F,A),
	functor(TT,F,A).
	

% debug stuff!

correct_mscg_term(T,Star) :- 
	(correct_mscg_term(T,Star,[],_) ->
	    true
	;
	    write('no correct mscg_term  : '),
	    write(T),nl
	).

correct_mscg_term(T,Star,V0,V1) :-
	var(T),!,
	(var_memberchk(T,Star) ->
	    true
	;
	    (var_memberchk(T,V0) ->
		fail
	    ;
		V1 = [T|V0]
	    )
	).
correct_mscg_term(T,Star,V0,V1) :-
	functor(T,_F,A),
	correct_mscg_term(A,T,Star,V0,V1).
	
correct_mscg_term(0,_T,_Star,V,V) :- !.
correct_mscg_term(A,T,Star,V0,V2) :-
	arg(A,T,ST),
	correct_mscg_term(ST,Star,V0,V1),
	A1 is A-1,
	correct_mscg_term(A1,T,Star,V1,V2).

	
% Examples
% mscg_term(h(a,A2,X1,A1,A1),h(a,A2,c,b,b),[],[],G,S1,S2,[A1,A2]).
% mscg_subst([Y-a,X1-a,X2-b,X3-g(c)],[X1-a,X2-c,X3-g(a)],[],Mu,[],Sigma1,[],Sigma2,[]).

% BUG
% mscg_term(e(e(X,e(Y,e(Z,e(e(U,e(V,Z))))))),e(e(e(e(X,e(Y,Z)),e(Y,X)),e(Z,U))),[],[],TZ,S1,S2,[X,U,Z,U,V]).

