%declarations%

:- module(myterms, [ 
	apply/3,
	apply_subst/3,
	apply_subst/4,
	apply_substitution/3,
	delete_var/3,
	deskolemize/2,
	function_symbol_in/2,
	general_function_call/1,
	identity_substitution/1,
	is_list_of_different_vars/1,
	is_var_set/1,
	intersection_vars/3,
	definition/2,
	difference_vars/3,
	linear/1,
	list_of_vars/2,
	mapargs/3,
	mapargs/4,
	multiset_of_vars/2,
	occurs_in/2,
	skolemized/3,
	skolemizedR/3,
	subset_vars/2
   ]).




/* ----------------------------------------------------------------------------


Operations on Prolog Terms




---------------------------------------------------------------------------- */




:- meta_predicate(mapargs(+,:,+,-)).
:- meta_predicate(mapargs(:,+,-)).
:- meta_predicate(mapargs1(+,+,:,+,-)).


:- use_module(library(arg), [
	genarg/3
   ]).

:- use_module(library(sets), [
	select/3
   ]).

:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(maplist), [
	maplist/2
   ]).

:- use_module(library(call), [
	call/3
   ]).

:- use_module(auxiliary, [ 
        get_prolog_data_operators/2,
	member212/2
   ]).

:- use_module(normalization,[normalization/5]).

:- use_module(equationalTheory, [equal/2]).

%end%



apply(F,T,FT):-
	FT=..[F,T].



%------------------------------------------------------------------------------
% apply_substitution(+S,+I,?O): 
% applies substitutions S of the form [X1=t1,...,Xn=tn], by where
% each Xi only appears once, and not inside the ti, 
% to I without side-effects, i.e.,
% by producing a copy of I with the Xi replaced by the ti.
% The Xi can also be atomic terms, in case we are dealing with
% skolemized variables.
%------------------------------------------------------------------------------
 
% new version


apply_subst(S,T,X,U):-
	apply_substitution([X=U],S,T).


apply_subst(E,S,T):-
	apply_substitution([E],S,T).


apply_substitution([],S,S):-
	!.
apply_substitution(E,S,T):-
	convertSubst(E,E1),
	(apply_subst1(E1,S,T) ->
	     true
	;    S=T).

convertSubst([E],[A=B]):-
	(E=(A=B);E=[A,B];E=A-B),
	!.
convertSubst([E|Es],[A=B|Es1]):-
	(E=(A=B);E=[A,B];E=A-B),
	convertSubst(Es,Es1),
	!.
convertSubst([],[]).	

apply_subst1( E, S, T ):-
	(lookup(E,S,T) ->
	     true
	;nonvar(S) ->
	     functor(S,F,N),
	     apply_subst_args1(E,S,T,N,F,N)
	).

apply_subst_args1(E,S,T,I,F,N):-
	I>0,
	arg(I,S,SI),
	I1 is I-1,
	(apply_subst1(E,SI,TI) ->
	     functor(T,F,N),
	     arg(I,T,TI),
	     sameSubterms(I,S,T,N),
	     apply_subst_args(E,S,T,I1)
	;    apply_subst_args1(E,S,T,I1,F,N)).


sameSubterms(N,_,_,N).
sameSubterms(I,S,T,N):-
	I<N,
	I1 is I+1,
	arg(I1,S,Sub),
	arg(I1,T,Sub),
	sameSubterms(I1,S,T,N).


apply_subst_args(_,_,_,0):-
	!.
apply_subst_args(E,S,T,I):-
	I1 is I-1,
	arg(I,S,SI),
	arg(I,T,TI),
	apply_substitution(E,SI,TI),
	apply_subst_args(E,S,T,I1).




lookup([X=U|_],Y,U):-
	equal(X,Y),
	!.
lookup([_,X=U|_],Y,U):-
	equal(X,Y),
	!.
lookup([_,_,X=U|_],Y,U):-
	equal(X,Y),
	!.
lookup([_,_,_|E],Y,U):-
	lookup(E,Y,U).




subset_vars(Sub,T):-
	list_of_vars(Sub,VS),
	list_of_vars(T,VT),
	\+ (member(X,VS), \+member212(X,VT)).


% end new version


% occurs  check

occurs_in( X, Y ):- X==Y, !.
occurs_in( _, Y ):- var(Y), !, fail.
occurs_in( X, T ):-
	genarg(_,T,Arg),
	occurs_in( X, Arg ),!.

/* better? 

occurs_in( X, Y ):- X==Y, !.
occurs_in( X, Y ):-
	\+subsumes_chk(X,Y).
*/ 

function_symbol_in(F,T):-
	get_prolog_data_operators( T, L ),
	member(F,L).



% mapping predicates over sub term lists


mapargs(N,P,S,T):-
	mapargs1(0,N,P,S,T).

mapargs1(N,N,_,_,_):-
	!.
mapargs1(I,N,P,S,T):-
	J is I+1,
	arg(J,S,SJ),
	arg(J,T,TJ),
	call(P,SJ,TJ),
	mapargs1(J,N,P,S,T).



mapargs(P,S,T):-
	nonvar(S),
	!,
	functor(S,F,N),
	functor(T,F,N),
	mapargs1(0,N,P,S,T).
mapargs(P,S,T):-
	functor(T,F,N),
	functor(S,F,N),
	mapargs1(0,N,P,S,T).






%-------------------------------------------------
% list (set) of variables of a prolog structure
%-------------------------------------------------

list_of_vars( T, L ):-
        list_vars( T, L1/[] ),
        sort( L1, L ),!.

multiset_of_vars(T,V):-
	list_vars( T, V/[] ).

list_vars( X, L ):- 
	(var(X) ->
	     L=[X|V]/V
	;atomic(X) ->
	     L=V/V
	;X=[X1|R] ->
	     list_vars( X1, V1/V2 ),
	     list_vars( R, V2/V3 ),
	     L=V1/V3
	;% go down arguments ->
             X =.. [_|Args],
	     list_vars( Args, L )
	).

is_list_of_different_vars(Args):-
	maplist(var,Args),
	sort(Args,ArgsS),
	length(Args,N),
	length(ArgsS,N).




intersection_vars([],_,[]).
intersection_vars([X|Xs],S,[X|Ys]):-
	member212(X,S),
	!,
	intersection_vars(Xs,S,Ys).
intersection_vars([_X|Xs],S,Ys):-
	intersection_vars(Xs,S,Ys).

difference_vars(X,[],X).
difference_vars(Xs,[Y|Ys],Zs):-
	delete_var(Xs,Y,Us),
	difference_vars(Us,Ys,Zs).


delete_var([],_,[]).
delete_var([X|Xs],Y,Xs):-
	X==Y,
	!.
delete_var([_X|Xs],Y,Xs):-
	delete_var(Xs,Y,Xs).




%------------------------------------------------------------------------------

% Skolemization

%------------------------------------------------------------------------------

skolemized(TV,TC,Subst):-
	var(Subst),
	!,
	normalization(TV,TC,Subst,[],_NewVars),
	numbervars(TC,0,_).
skolemized(TV,TC,Subst):-
	apply_substitution(Subst,TC,TV).




skolemizedR(TV,TC,SubstR):-
	normalization(TV,TC,Subst,[],_NewVars),
	normalization:reverse_var_subst(SubstR,Subst),
	numbervars(TC,0,_).




% deskolemize(+TS,?TV) is true when TV is like TS except that all
% skolem constants $VAR(N) are replaced by variables.
% TS must be a ground term.

deskolemize(TS,TV):-
	deskolemize(_,TS,TV).

deskolemize(E,'$VAR'(N),X):-
	!,
	assoc(E,N,X).
deskolemize(E,N,X):-
	atom(N),
	name(N,[88|_]),  % 88 = X  % should be sometime deleted
	!,
	assoc(E,N,X).
deskolemize(E,T,TV):-
	functor(T,F,N),
	functor(TV,F,N),
	deskolemizeArg(E,N,T,TV).


deskolemizeArg(E,N,T,TV):-
	(N=0 ->
	     true
	;
	     arg(N,T,TN),
	     arg(N,TV,TVN),
	     deskolemize(E,TN,TVN),
	     N1 is N-1,
	     deskolemizeArg(E,N1,T,TV)
	).


assoc([N-V|_Es],N,V):-
	!.
assoc([_-_|Es],N,V):-
	assoc(Es,N,V).




identity_substitution([]).
identity_substitution([A=B|S]):-
	A==B,
	identity_substitution(S).


linear(T):-
	multiset_of_vars(T,V),
	select(X,V,V1),
	member212(X,V1),
	!,
	fail.
linear(_).



is_var_set([]).
is_var_set([X|Xs]):-
	var(X),
	\+ member212(X,Xs),
	is_var_set(Xs).
	


general_function_call(T):-
	nonvar(T),
	T=..[_|Args],
	is_var_set(Args),
	linear(Args).
%	T=..[_,A|Args],
%	is_var_set([A|Args]),
%	linear([A|Args]).



% end module
% -----------------------------------------------------------------------------

