% -*- Module: saturation; -*-

%declarations%

:- module(saturation, [ 
	add_inference/3,
	sat/0,
	sat/1,
	sat/2,
	sat/6,
	update_minimalWeight/1
   ]).

:- use_module(props, [
        agoal/0,
	blocked_clause/1,
	congruence_relation_symbol/1,
	flag_on/2,
	flat_proofs/4,
	initial_var_depth/1,
	iterative_saturation/1,
	search_depth/1,
	scaled_inference_weight/11,
	scaled_inference_weight_D/11,
	scaled_clause_weight/7,
	scaled_clause_weight_D/7,
	tactic/1,
	theory_clause/1
   ]).

:- use_module(constraintSolver, [
	currentSolver/1
   ]).

:- use_module(termOrder, [
	currentOrdering/1
   ]).


:- use_module(adt_declarations, [
	message/2
   ]).

:- use_module(classes, [
	active_term/2,
	empty/1,
	blocked/1,
	literals/2,
	delete_rule/3,
	delete_rules_no_dependents/3,
	groundClause/1,
	moduleNumber/2,
	number/2,
	weight/2,
	weightBound/2,
	incr_weight_bound/2,
	is_brand_new/1,
	is_literal/1,
	is_new/1,
	is_old/1,
	kind/2,
	makeNew/2,
	makeOld/2,
	proof_depth/2,
	reductive/1,
	size/2,
	search_depth/2,
	theory_flag/2
   ]).

:- use_module(clause_processing, [
	process_clause/3
   ]).

:- use_module(clause_redundancy_prover, [
	backward_simplification/3,
	delete_redundant_old_rules/0,
	redundant_rule/1
   ]).

:- use_module(commands, [
	rp/0
   ]).

:- use_module(auxiliary,[get_prolog_data_operators/2]).


:- use_module(global, [
	cont/2,
	destroy/1,
	incr/1,
	initialize_vars/0,
	store/2
   ]).

:- use_module(inference_computation, [
	explicit_inferences/2,
	simplifying_inference/5,
	treat_inference/7
   ]).

:- use_module(inference_rules, [
	inference/7
   ]).

:- use_module(proofState, [
	clause/1,
	clause_counter/1,
	deleteAllRules/0,
	deleteClause/1,
	deleteClause/2,
	goal_symbol/1,
	inferenceComputed/2,
	recordInferenceComputed/2,
	rule/2,
	recordRule/2,
	deleteRule/2
   ]).

:- use_module(output, [
	display_rules/0
   ]).



:- use_module(adt_term, [
	f_substring/2,
	free_vars/2
   ]).

:- use_module(timing, [
	start_timing/0,
	push_timing/1,
	pop_timing/1,
	runtime/1,
	stime/0
   ]).

:- use_module(thisProlog,[compile_nqa/2,timeOut/3]).

:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(maplist), [
	maplist/2
   ]).

:-use_module(index,[initial_index/1]).

:- use_module(propLogic, [propLogicInit/0]).

%end%



% -----------------------------------------------------------------------------

	       


sat(Ord,Solver):-
	member((Ord,Solver),
	       [(lpo,lpo),(lpo,generic),(rpo,generic),
		(sto,sto),(sto,generic),(sto,generic)
		]),
	(currentOrdering(Ord) -> 
	     true
	;    name(Ord,ON),
	     user:append(ON,"Interface",ONI),
	     name(OI,ONI),
	     compile_nqa(saturation,satsrc(OI))),
	(currentSolver(Solver) -> 
	     true
	;
	     name(Solver,SN),
	     user:append(SN,"ConstraintSolver",FN),
	     name(F,FN),
	     compile_nqa(saturation,satsrc(F))),
	sat.


sat:-
	sat(1000000).



sat(ClBound,TimeBound,SatFlag,TimeUsed,NInf,NClGen):-
	on_exception(E,
		     sat1(ClBound,TimeBound,SatFlag,TimeUsed,NInf,NClGen),
		     (E\==time_out ->
			  raise_exception(E)
		     )
		    ),
	!.
sat(_ClBound,_TimeBound,exception,'?','?','?').

sat1(ClBound,TimeBound,SatFlag,TimeUsed,NInf,NClGen):-
	(timeOut(sat(ClBound),TimeBound,success) ->
	     result_of_saturation(TimeUsed,NInf,NClGen),
	     (rp ->
		  SatFlag=proved
	     ;\+sat_bound_not_reached ->
	          SatFlag='?'
	     ;
	          SatFlag=consistent
	     )
	;
	     SatFlag='?',
	     TimeUsed=exceeded,
	     result_of_saturation(_,NInf,NClGen)
	).

	     

sat(N):-	
	start_timing, 
	store(max_clauses,N),
	initialize_vars,

	initial_index(ST),

	on_exception(proof_found,
		     on_exception(saturate_error(S,P),
				  saturate_start(ST),
				  (  append("~n*** Saturate Error: ",S,Str),
				     format(Str,P)
				  )
				 ),
		     (  deleteAllRules,
			terminate
		     )
		    ).


saturate_start(ST):-
	(iterative_saturation(on) ->
	     store(proof_depth,0),
	     it_sat(ST,_STO)
         ;
	     saturate_start(ST,_STO)
	).


it_sat(ST,STE):-
	cont(proof_depth,D),
	format('~n~nSetting search depth to ~p~n',[D]),
	props:option(search_depth(D)),
	saturate_start(ST,STO),
        (rule(_,Ru), empty(Ru) ->
	        true
	;       make_rules_new,
	        incr(proof_depth),
	        it_sat(STO,STE)).





% we start out by processing all initially given clasues
saturate_start(ST,STE):-
	deleteClause(Cl),
	process_clause( Cl,ST,STO ),
	saturate_start(STO,STE), !.
saturate_start(ST,STO):-
	saturate(ST,STO).
						  % alternatively
%	minimal_bound(B),
%	saturate(orient,ST,B).



/*
Saturation performs the following steps:
1. Simplify and preprocess  all ``small clauses'' (facts clause/1).
   A small clause does not contain more variables than the current
   upper bound ``var_depth''.
   The ones that are not redundant become ``rules'' (facts rule/1).
2. If non-small clauses are left raise the var_depth bound if that
   is required for the fairness of the process (inc_var_depth_now=0)
   and restart saturation.
3. Otherwise, select a ``new'' rule, that is, one for which not all
   inferences with smaller rules have been computed yet,
   compute some inferences with it, and then restart saturation.
4. If no unprocessed clauses are left, terminate.
*/

saturate(ST,STE):-
	deleteClause(Cl,small_clause(Cl)),
	process_clause( Cl,ST,STO ),
	saturate(STO,STE), !.
saturate(ST,STE):-
	(clause(_) ->
		% this clause was ignored because of too many variables
		% which should not happen infinitely often
	     cont(inc_var_depth_now,N),
	     (N=0 ->
		     cont(var_depth,D),
		     D1 is D+1,
		     % format('~n[increasing var depth to ~p]~n',[D1]),
		     store(var_depth,D1),
		     store(inc_var_depth_now,50),
		     saturate(ST,STE)  % reconsider clauses
	     ;       N1 is N-1,
		     store(inc_var_depth_now,N1),
		     fail )    % compute some inferences first 
	;    initial_var_depth(D),
	     store(var_depth,D),
	     store(inc_var_depth_now,50),
	     fail     % compute some inferences now 
	),
	!.
saturate(ST,STE):-
	cont(new_rules,Rs),
	flag_on(ir,_),
	Rs\==[],
	sort(Rs,Ns),
	backward_simplification(Ns,ST,STO),
	!,
	saturate(STO,STE).
saturate(ST,STE):-
	flag_on(ir,_),
	push_timing(cinf),
	select_new_rule( Ru ),
	(sat_bound_not_reached ->
	      compute_inferences(ST,Ru,ST1 )
	),
	pop_timing(cinf),
	saturate(ST1,STE),
	!.
saturate(ST,ST):-
	\+ clause(_),
	!,
	terminate.
saturate(ST,STE):-
	cont(var_depth,D),
	D1 is D+1,
	store(var_depth,D1),
	store(inc_var_depth_now,50),
	saturate(ST,STE).

sat_bound_not_reached:-
	cont(max_clauses,Max),
	cont(nmb_rules,N),
	N=<Max.



terminate:- 
	(rp ->
	     true
	;(\+sat_bound_not_reached ->
	      format('~n~nMax clauses exceeded.~n',[])
	 ;    format('~n~nSaturation terminated.~n',[])
	 ),
	     display_rules
	),
	nl, 
	propLogicInit,  % to exit PB's top level
	stime,
	!.



result_of_saturation(Total,NInf,NCl):-
	runtime(Total),
%	cont(nmb_inferences,NInf),
	cont(nmb_inferences_skex,N1),
	cont(nmb_inferences_forward,N2),
	NInf is N1+N2,
	cont(nmb_rules,NCl).




saturate(orient,ST,B):-
	(deleteClause(Cl,small_clause(Cl,B)) ->
	     process_clause(Cl,ST,STO),
	     saturate(orient,STO,B)
	;
             saturate(inf,ST,B)
	).
saturate(inf,ST,B):-
	clause_counter(Ctr),
	(select_new_rule( Ru ) ->
	     (sat_bound_not_reached ->
		  push_timing(cinf),
		  compute_inferences( ST, Ru, ST1 ),
		  pop_timing(cinf)
	     ;
	          finish_up
	     ),
	     clause_counter(Ctr1),
	     (Ctr1=<Ctr+3 ->
						  % not many new clause
		  saturate(inf,ST1,B)
	     ;
	          minimal_bound(B1),
		  saturate(orient,ST1,B1)
	     )
	;
						  % no more inferences
	     (clause(_) ->
		  increase_bound(B,B1),
		  saturate(orient,ST,B1)
	     ;
						  % we're finished
	          finish_up
	     )
	).



finish_up:-
	(rp ->
	     true
	;(\+sat_bound_not_reached ->
	      format('~n~nMax clauses exceeded.~n',[])
	 ;    format('~n~nSaturation terminated.~n',[])
	 ),
	     display_rules
	),
	nl, 
	propLogicInit,  % to exit PB's top level
	stime, 
	!.
	         

small_clause(Cl,[VarDepth,ClauseSize|_]):-
	literals(Cl,L),
	free_vars(L,V),
	length(V,N),
	N=<VarDepth,
	size(Cl,S),
	S=<ClauseSize,
	!.



increase_bound([VarDepth,ClauseSize|B],[VarDepth1,ClauseSize1|B]):-
	clause_bound_increment(S),
	ClauseSize1 is ClauseSize+S,
	(ClauseSize1>2*S ->
	     VarDepth1 is VarDepth+1
	;
	     VarDepth=VarDepth1
	).




minimal_bound([VarDepth,ClauseSize|_]):-
	initial_var_depth(VarDepth),
	clause_bound_increment(ClauseSize).



clause_bound_increment(15).



small_clause(Cl):-
	literals(Cl,L),
	free_vars(L,V),
	length(V,N),
	cont(var_depth,D),
	N=<D,
	!.


/*             Rule Selection 
                    for
             Inference Computation


*/


% select_new_rule(?Ru)
% is true when Ru is a new rule with minimal weight.
% Also, Ru is not redundant.
% In that case we signal that saturation is not yet finished
% by setting notDone to false.

select_new_rule( Ru ):-
% We first retry to eliminate redundant ``old'' rules
	delete_redundant_old_rules,
	minimal_new_rule( Ru ),
	store(notDone,false).



% minimal_new_rule( ?Ru ) 
% is true when Ru is a ``new'' rule with minimal weight
% regardless of the weight we compute unary inferences with all
% brand new rules first.

minimal_new_rule( Ru ):-
        rule( _, Ru ),
	\+ blocked(Ru),
	is_brand_new( Ru ),
	!,
	increase_weight_bound(Ru).
minimal_new_rule( _Ru ):-
	next_selection,
	store(minimalWeight,(none,10000000)),   % empty set of new rules
        rule( _, Ru ),
	is_new( Ru ),
	\+ blocked(Ru),
	update_minimalWeight(Ru),
	fail.
minimal_new_rule( Ru ):-
	cont(minimalWeight,(Ru,_Z)),
	Ru\== none,
	message( comw, [Ru,Ru] ).




update_minimalWeight( Ru):-
	cont(minimalWeight,( _RuOldMin, MZ )),
	!,
%	trace,
	weight_with_bound_and_depth( Ru, Z ),

	(very_bad(Ru) ->
	     Z1 is Z*10
	;
	     Z1=Z
	),
/*
	(only_goal_operators(Ru) ->
	     Z2=Z1
	;
	     Z2 is 2*Z1
	),
*/
	(MZ =< Z1 ->
	     true
	;	
	     store(minimalWeight,(Ru,Z))
	).
update_minimalWeight(_).



only_goal_operators(_Ru):-
	\+goal_symbol(_),
	!.  % degenerate case
only_goal_operators(Ru):-
	active_term(Ru,T),
	get_prolog_data_operators(T,Ops),
	\+ (member(O,Ops),\+ like_a_goal_symbol(O)).



like_a_goal_symbol('$true$'/0).
like_a_goal_symbol(== /2).
like_a_goal_symbol(O/2):-
	congruence_relation_symbol(O).
like_a_goal_symbol(O):-
	goal_symbol(O).


very_bad(_):-
	!,
	fail.
very_bad(Ru):-
	literals(Ru,L),
	member(E,L),
%	f_substring(E,[F,F,F]),
	f_substring(E,[F,F]),
	F\=='.',
	!.




is_large(R):-
	literals(R,L),
	length(L,N),
	N>2.

depth_guided_selection:-
	flat_proofs(LB,UB,Freq,_),
%	agoal,
	cont(selCtr,N),
	(N =< LB ->
	     true
	;N=<UB ->
	     N mod Freq =0
	).

age_guided_selection:-
	cont(selCtr,N),
	N mod 7 = 0.

sos_selection:-
	cont(selCtr,N),
	N mod 2 = 0.



next_selection:-
	cont(selCtr,N),
	N1 is N+1,
	store(selCtr,N1).

weight_with_bound_and_depth( R, WB ):-
	weightBound(R,B),
        weight(R,W),
        proof_depth(R,D),
        theory_flag(R,T),
	kind(R,K),
        (reductive(R) ->
             RedFlag=reductive
        ;
             RedFlag=nonReductive
        ),
        (depth_guided_selection ->
	     scaled_clause_weight_D(W,D,B,T,RedFlag,K,WB)
        ;age_guided_selection ->
	     number(R,WB)
	;
             scaled_clause_weight(W,D,B,T,RedFlag,K,WB)
	).






scaled_inference_weight_with_depth(S,Sold,D,Dold,T,Told,K,Kold,Size):-
        (depth_guided_selection ->
	     scaled_inference_weight_D(S,Sold,D,Dold,T,Told,K,Kold,Size)
	;
	      scaled_inference_weight(S,Sold,D,Dold,T,Told,K,Kold,Size)
	).


%===============================================================
%
%	INFERENCE COMPUTATION
%
%===============================================================



compute_inferences( STI, R, STO ):-
	redundant_rule(R),
	delete_rule(R,STI,STO),
	message( ruca, [R] ), 
	!.
compute_inferences( ST, R, _STO ):-
	on_exception(ruleSimplified,
		     (	premises( R, Premises ),
%			format('~Nselecting ~p',[Premises]),
			add_inference(ST, R, Premises),
			fail
		     ),
		     fail).
compute_inferences(ST,_,STO):-
	cont(rules_to_be_deleted,Dels),
	delete_rules_no_dependents(Dels,ST,STO).








/*
add_inference(+ST,+R,+Premises)
is true when a nonredundant inference from Premises (where R is one
of the Premises) exists. In that case the conclusion of the
inference is added as a new clause. Upon backtracking, all inferences
from Premises are computed. 

The exception ``ruleSimplified'' is raised when R
has been deleted due to the fact that the inference has 
produced a simplification of R. In that case backtracking should
be stopped.

As a side effect the set of rules for which the inference has been simplifying
is computed in the global variable ``rules_to_be_deleted''.


*/

add_inference( ST, R, Premises) :-
	store(rules_to_be_deleted,[]),
	flag_on( ir, Key ),
	explicit_inferences(Premises,Key),
%	(Premises=[Ru], number(Ru,41) -> trace; true),
	inference( Key, Premises, ST, Unif, Conclusion, OrdConstraint, CBound ),
	(simplifying_inference(Key,Premises,Conclusion,Unif,Simplified) -> 
	     Simplifying=true
	;    true),

	treat_inference(Key, Premises, ST, Unif, Conclusion, OrdConstraint, CBound),

	(nonvar(Simplifying) ->

	     % if we arrive here the inference was non-redundant and its
	     % conclusion has been added so that the premise "Simplified"
	     % can be deleted
	     cont(rules_to_be_deleted,Dels),
	     store(rules_to_be_deleted,[Simplified|Dels]),
	     (Simplified==R -> 
		  raise_exception(ruleSimplified)
	     ;    true)

	;    true).




/*   
premises(+R,?Premises)
is true when Premises is a set of one or two clauses including R
from which no non-neglectable inferences have been computed before.
Inferences are neglectable if one of the premises is blocked
or if they are all theory clauses. What blocked clauses
or theory clauses are is specified with the input.


A modular form of inputs is supported where a module
is a saturated set of clauses. If the module number is the same
for two clauses, inferences from them have been considered before.

If clauses are too big they are not considered unless absolutely
needed for fairness. The latter is determined using the weight 
of a clause and the current weight bound for R. After all premises up to the
current weight bound for R have been produced upon backtracking,
the weight bound for R is increased. Once all premises including
R have been returned, R is marked as ``old'', and henceforth will
never again be selected by select_new_rule. 

Facts inferenceComputed/2 are
used to memoize which inferences have been computed.
The classification ``old''/``new'' also indicates what has
been considered already. We can restrict attention to binary inferences,
in which one of the premises is old.
*/



% Version 2


premises2( R, [R] ):-
	is_brand_new(R),     % rule to be considered for the first time
	!,
	number(R,N),
	\+ (theory_clause(T), member(N,T)),
	\+ (blocked_clause(T), member(N,T)).
premises2( R, Prem ):-
        firstPremise(R,N,NR,B,S,D,T,Red,K),
	rule(NRold, Rold ),
	(is_old( Rold );NR=NRold),
	moduleNumber( Rold, Nold ),
	N \== Nold,
	secondPremise(Rold,Sold,Dold,Told,Redold,Kold),
	(Told=theory, T=theory ->
	     \+sos
	;
	     true
	),


	scaled_inference_weight(S,Sold,D,Dold,T,Told,Red,Redold,K,Kold,Size),
	      
	(Size>B
	% ,clause_counter(Ctr),Ctr<800 
	->
	     store(notDone,true),
	     fail
	;
	     \+ inferenceComputed(NR,NRold),
	     recordInferenceComputed(NR,NRold)
	     
	),


	( Prem = [R,Rold] 
	; 
	  NR \== NRold, 
	  Prem = [Rold,R]
	).
premises2(R,_):-
	cont(notDone,false),
	% now all inferences with R have been computed, 
	% so make R become ``old''
	make_rule_old( R ), 
	!,
	fail.
premises2(R,_):-
	increase_weight_bound(R),
	!,
	fail.


firstPremise(R,N,NR,B,S,D,T,RF,K):-
	number( R, NR ), 
	moduleNumber( R, N ),
	weightBound(R,B),
	weight(R,S),
	proof_depth(R,D),
	theory_flag(R,T),
	reductivity_flag(R,RF),
	kind(R,K),
	!.

secondPremise(Rold,Sold,Dold,Told,RFold,Kold):-
	weight(Rold,Sold),
	proof_depth(Rold,Dold),
	theory_flag(Rold,Told),
	reductivity_flag(Rold,RFold),
	kind(Rold,Kold),
	!.

reductivity_flag(R,reductive):-
	classes:reductive(R),
	!.
reductivity_flag(_R,nonReductive).

increase_weight_bound(R):-
	incr_weight_bound(R,Ru),
	number(R,N),
	deleteRule(N,R),
	recordRule(N,Ru).

make_rule_old( R ):-
	makeOld( R, Ru ), 
	number(R,N),
	deleteRule(N,R),
	recordRule(N,Ru ).




make_rules_new:-
	rule(_,R),
	make_rule_new(R),
	fail.
make_rules_new.




make_rule_new( R ):-
	makeNew( R, Ru ), 
	number(R,N),
	deleteRule(N,R),
	recordRule(N,Ru ),
	!.







% the less complicted version in which weigth bounds are not considered

premises1( R, [R] ):-
	is_brand_new(R),     % rule to be considered for the first time
	!,
	number(R,N),
	\+ (theory_clause(T), member(N,T)),
	\+ (blocked_clause(T), member(N,T)).
premises1( R, Prem ):-
	moduleNumber( R, N ),
        search_depth(R,DR),
	rule( NRold, Rold ),
	search_depth(Rold,DRold),

	\+ \+ (   search_depth(D),
		  DR<D,
		  DRold<D
	      ;	  groundClause(R)
	      ;   groundClause(Rold)
	      ),

	number( R, NR ), 
	(is_old( Rold );NR=NRold),
	moduleNumber( Rold, Nold ),
	N \== Nold,

	( Prem = [R,Rold] 
	; 
	  NR \== NRold, 
	  Prem = [Rold,R]
	).
premises1(R,_):-
	make_rule_old( R ), 
	!,
	fail.
    


premises(R,Rs):-
	(flag_on(rp,ruca) ->
	     premises2(R,Rs)
	;
	     premises1(R,Rs)
	).

/*
premises(R,Rs):-
	premises2(R,Rs).
*/



forall_goal_tactic(0):-
	tactic(T),
	member(forall_goal,T),
	!.
forall_goal_tactic(D):-
	tactic(T),
	member(forall_goal(D),T),
	!.


:- dynamic sos/0.
