:-module(sequents, [
	atom_in_expansion/2,
	negative/1,
	positive/1,
	sk/5,
	simple_sk/3,
	simplify_sequent/3,
	tableau_expansion_on/3
   ]).


/* ----------------------------------------------------------------------------

This module provides a partial implementation of the sequent calculus
for use with saturation-based methods.

The general idea is to not considerer CNF transformation as a separate
step of preprocessing but integrate it into resolution.

In many respects we follow Avron JAR 10(2), 1993.
A main modification is that we envisage to apply sequent calculus rules
only for maximal (or selected) formulas in sequents.

--------------------------------------------------------------------------- */


:- use_module(adt_term, [
	abstract_from_terms/3,
        free_vars/2,
        binder/4,
	instantiate/1,
	is_closed/1,
	select_skolem_term/2,
	standard_atom/1,
	subst/3,
	subst/4
   ]).


:- use_module(adt_precedence, [
        abbreviation_for_literal/4,
        max_skolem_function/1
   ]).


:- use_module(props, [
        finite_domain/1,
	structural_expansion/1,
        full_cnf/1
		     ]).


:- use_module(adt_atom, [
	atom_as_term/2,
	create_atom/2,
	get_atom_pre_rel_left_right/5,
	is_standard_atom/1,
	orient_atom/2
   ]).

:- use_module(hgstuff, [
	split/4
   ]).

:- use_module(classes, [
	antecedent/2,
	succedent/2,
	createSimpleClause/2
   ]).

:- use_module(library(maplist), [
	maplist/2,
	maplist/3
   ]).

:- use_module(library(sets), [
	select/3
   ]).





/* ----------------------------------------------------------------------------

sk(Am,Ap,Sm,Sp,Seqs)

is true when the sequents in Seqs imply


        S = Am,Ap   -> Sm,Sp

and where and  N,S is satisfiable if and only if N,Seqs
is satifiable, for any matrix N.

Moreover, the sequents in Seqs are smaller than S
in some well-founded ordering.

The predicate fails if all formulas in Am and Sm are either atoms
or else equivalences of the form A==B where the maximal formula among
the A and B is an atom.

sk will be applied in situations where Am holds the maximal negative
and Sm the maximal positive formulas of a sequent. In other words,
for any A in Ap cup Sp there exists a B in Am cup Sm such that B > A.

---------------------------------------------------------------------------- */


%sk([],Ap,[],Sp,[(Ap,Sp)]).
sk(Am,Ap,[GAt|Sm],Sp,Sequents):-
	expand_general_literal([+|GAt],Clauses),
	!,
	maplist(clause_to_sequent,Clauses,Seqts),
	append(Am,Ap,A),
	append(Sm,Sp,S),
	maplist(weaken_sequent((A,S)),Seqts,Sequents).
sk([GAt|Am],Ap,Sm,Sp,Sequents):-
	expand_general_literal([-|GAt],Clauses),
	!,
	maplist(clause_to_sequent,Clauses,Seqts),
	append(Am,Ap,A),
	append(Sm,Sp,S),
	maplist(weaken_sequent((A,S)),Seqts,Sequents).
sk(Am,Ap,[GAt|Sm],Sp,Seqs):-
	% GAt need not be expanded
	!,
	sk(Am,Ap,Sm,Sp,SSeqs),
	maplist(weaken_sequent(([],[GAt])),SSeqs,Seqs).
sk([GAt|Am],Ap,Sm,Sp,Seqs):-
	!,
	sk(Am,Ap,Sm,Sp,SSeqs),
	maplist(weaken_sequent(([GAt],[])),SSeqs,Seqs).





tableau_expansion_on([Sign|At],(A,S),Sequents):-
	(   structural_expansion(off)
	;
	    maplist(is_standard_atom,A),
	    maplist(is_standard_atom,S)
	),
	!,
	expand_general_literal([Sign|At],Clauses),
	maplist(clause_to_sequent,Clauses,Seqts),
	maplist(weaken_sequent((A,S)),Seqts,Seqs),
	maplist(simplify_sequent,Seqs,Sequents).
tableau_expansion_on([Sign|At],([],[]),Sequents):-
	!,
	expand_general_literal([Sign|At],Clauses),
	!,
	maplist(clause_to_sequent,Clauses,Seqs),
	maplist(simplify_sequent,Seqs,Sequents).
tableau_expansion_on([Sign|At],(A,S),Sequents):-
	expand_general_literal([Sign|At],Clauses),
	length(Clauses,1),
	!,
	maplist(clause_to_sequent,Clauses,Seqts),
	maplist(weaken_sequent((A,S)),Seqts,Seqs),
	maplist(simplify_sequent,Seqs,Sequents).
tableau_expansion_on([Sign|At],(A,S),ResultSequents):-
	expand_general_literal([Sign|At],_Clauses), % something can be done
	!,
%	abstract_from_terms(At,AA,Subst),  % to allow for general defs
	At=AA,Subst=[],
	expand_general_literal([Sign|AA],Clauses),  % abstracted clause
	(   (A=[E],S=[],Sign1= -;A=[],S=[E],Sign1= +),
	    maplist(is_standard_unit_clause,Clauses) 
	->
	    % no abbreviation needed; would become binary clauses
	     maplist(clause_to_sequent,Clauses,Seqts),
	    ( Sign1= - ->
		maplist(weaken_sequent(([E],[])),Seqts,Sequents1)
	    ;
		maplist(weaken_sequent(([],[E])),Seqts,Sequents1)
	    ),
	    copy_term(Sequents1,Seqs),
	    maplist(simplify_sequent,Seqs,Sequents),
                           % to keep the def. of the new pred general
	    instantiate(Subst),
	    ResultSequents=Sequents
	;
	    ResultSequents=[Cl|Sequents],
	    abbreviation_for_literal([Sign|At],AA,[Sign1|NewPredAtom],Status),
	    ( Status=new ->
		maplist(clause_to_sequent,Clauses,Seqts),
		( Sign1= - ->
		    maplist(weaken_sequent(([],[NewPredAtom])),Seqts,Sequents1),
		    Cl=([NewPredAtom|A],S)
		;   
		    maplist(weaken_sequent(([NewPredAtom],[])),Seqts,Sequents1),
		    Cl=(A,[NewPredAtom|S])
		),
		
		copy_term(Sequents1,Seqs),
		maplist(simplify_sequent,Seqs,Sequents),
                           % to keep the def. of the new pred general
		instantiate(Subst)
                           % to specialize NewPredAtom 
	    ;	
		( Sign1= - ->
		    Cl=([NewPredAtom|A],S)
		;   
		    Cl=(A,[NewPredAtom|S])
		),
		Sequents=[]
	    )
	).
%	\+ \+ (numbervars([[Sign|At],(A,S),[Cl|Sequents]],0,_),
%	       format('~2nExpanded: ~p in context ~p~n   to ~p',[[Sign|At],(A,S),[Cl|Sequents]])).


is_standard_unit_clause([[_|E]]):-
	is_standard_atom(E).




simplify_sequent((A,S),Seq):-
	simplify_sequent((A,S),_,Seq).


% strictly size decreasing simplifications:
% (the negation rules are in the index!)
simplify_sequent((A,S),_Yes,Seq):-
	select(E,S,S2),
	atom_as_term(E,ET),
	functor(ET,=>,2),
	!,
	arg(1,ET,ET1),
	arg(2,ET,ET2),
	create_atom(ET1,E1),
	create_atom(ET2,E2),
	simplify_sequent(([E1|A],[E2|S2]),_,Seq).
simplify_sequent((A,S),_Yes,Seq):-
	select(E,S,S2),
	atom_as_term(E,ET),
	functor(ET,or,2),
	!,
	arg(1,ET,ET1),
	arg(2,ET,ET2),
	create_atom(ET1,E1),
	create_atom(ET2,E2),
	simplify_sequent((A,[E1,E2|S2]),_,Seq).
simplify_sequent((A,S),_Yes,Seq):-   % no! also need to shift
	select(E,S,S2),
	atom_as_term(E,ET),
	binder(ET,_Sk,Body,all),
	!,
	create_atom(Body,BodyE),
	subst(BodyE,_New,BodyES),
	simplify_sequent((A,[BodyES|S2]),_,Seq).
simplify_sequent((A,S),_Yes,Seq):-
	select(E,A,A2),
	atom_as_term(E,ET),
	binder(ET,_Sk,Body,exists),
	!,
	create_atom(Body,BodyE),
	subst(BodyE,_New,BodyES),
	simplify_sequent(([BodyES|A2],S),_,Seq).
simplify_sequent((A,S),_Yes,Seq):-
	select(E,A,A2),
	atom_as_term(E,ET),
	functor(ET,and,2),
	!,
	arg(1,ET,ET1),
	arg(2,ET,ET2),
	create_atom(ET1,E1),
	create_atom(ET2,E2),
	simplify_sequent(([E1,E2|A2],S),_,Seq).
simplify_sequent(S,no,S).




simple_sk([At|A],S,Sequent):-
	weakly_expand([-|At],Clause),
	!,
	clause_to_sequent(Clause,Seqt),
	weaken_sequent((A,S),Seqt,Sequent).
simple_sk(A,[At|S],Sequent):-
	weakly_expand([+|At],Clause),
	!,
	clause_to_sequent(Clause,Seqt),
	weaken_sequent((A,S),Seqt,Sequent).
simple_sk(A,[At|S],Sequent):-
	!,
	simple_sk(A,S,Seqt),
	weaken_sequent(([],[At]),Seqt,Sequent).
simple_sk([At|A],S,Sequent):-
	!,
	simple_sk(A,S,Seqt),
	weaken_sequent(([At],[]),Seqt,Sequent).





expand_general_literal([-|GAt],Clauses):-
	% get the generalized atom GAt; if the relation symbol is not ==
	% we have a standard atom and nothing is to be done for it
	% negative equivalences need to be expanded, otherwise incomplete:
	% p==q -> [].  p->[]. q->[].  is inconsistent!
	% Alternatively transform p->[] into  p==min. This restricts
	% selection, though.

	get_atom_pre_rel_left_right(GAt,_,==,F,G),
	G\=='$true$',G\=='$min',
	F\=='$true$',F\=='$min',
	!,
	orient_atom([_,==,F,G],[PreO,==,F2,G2]),
	(PreO=l ->
	    (standard_atom(F2),full_cnf(no) ->
		FClauses=[[[+|F2== ~G2]]]
	    ;   expand([-|F2==G2],FClauses)
	    )
	;PreO=r ->
	    (standard_atom(G2),full_cnf(no) ->
		FClauses=[[[+|G2== ~F2]]]
	    ;   expand([-|F2==G2],FClauses)
	    )
	;
	    expand([+|G2== ~F2],FClauses)
	),
	maplist(fClause_to_clause,FClauses,Clauses).
expand_general_literal([-|GAt],Clauses):-
	get_atom_pre_rel_left_right(GAt,_,==,F,G),
	(G\=='$true$' ->
		expand([-|G==F],FClauses)
	;F\=='$true$' ->
		expand([-|F==G],FClauses)
	),
	maplist(fClause_to_clause,FClauses,Clauses).
expand_general_literal([+|GAt],Clauses):-
	% get the generalized atom GAt; if the relation symbol is not ==
	% we have a standard atom and nothing is to be done for it

	get_atom_pre_rel_left_right(GAt,PreO,==,F,G),

	% expand one larger side of the equivalence
	
	Sign= +,
	(PreO\==r,\+ standard_atom(F) ->
	     !,
	     expand([Sign|F==G],FClauses)
	;PreO\==l,\+ standard_atom(G) ->
	     !,
	     expand([Sign|G==F],FClauses)
	% else the maximal side(s) are atomic; nothing to be done for this GAt
	),
	maplist(fClause_to_clause,FClauses,Clauses).





weakly_expand([Sign|GAt],Clause):-
	% get the generalized atom GAt; if the relation symbol is not ==
	% we have a standard atom and nothing is to be done for it

	get_atom_pre_rel_left_right(GAt,_PreO,==,F,G),
	simplifyE([Sign|F==G],H),
	expand1(H,[SG]),
	fClause_to_clause(SG,Clause).



/*
it seems better to expand negative equivalences in standard atoms

weakly_expand([Sign|GAt],Clause):-
	% get the generalized atom GAt; if the relation symbol is not ==
	% we have a standard atom and nothing is to be done for it

	get_atom_pre_rel_left_right(GAt,_PreO,==,F,G),
	(PreO\==r,\+ standard_atom(F) ->
	     !,
	     simplifyE([Sign|F==G],H),
	     expand1(H,[SG])
	;PreO\==l,\+ standard_atom(G) ->
	     !,
	     simplifyE([Sign|F==G],H),
	     expand1(H,[SG])
	% else the maximal side(s) are atomic; nothing to be done for this GAt
	),
	fClause_to_clause(SG,Clause).

*/






/* ----------------------------------------------------------------------------


expand(+SF,?Cls)

   is true whenever Cls
   is the set of premises premise (as general clauses) of an
   invertible (modulo provability) sequent calculus expansion for 
   the signed formula SF.


---------------------------------------------------------------------------- */

expand(F,G):-
	simplifyE(F,H),
	(  expand1(H,G)
	;
	   expand2(H,G)
	).



% expand1: simplifying expansions or simplification with constants


expand1(F,G):-
	simplify(F,G).
expand1(F,G):-
	expand3(F,G).


	
% expand3: simplifying expansions

	
expand3(SF,[[F,G]]):-
	beta_formula(SF,F,G),
	!.
expand3(SF,[[F1]]):-
	gamma_formula(SF,Body),
	subst(Body,_New,F1).
expand3(SF,[[F]]):-
	negation(SF,F).



% expand2: structural expansions

expand2(SF,[[F],[G]]):-
	alpha_formula(SF,F,G),
	!.
expand2(SF,[[F1]|E]):-
	delta_formula(SF,SkolemTerm,Body),
	select_skolem_term(SkolemTerm,Sk),  % crucial for some examples,
	  % disastreous for others; we could be failing because of this
%	Sk=SkolemTerm,
	subst(Body,Sk,F1),
	(SkolemTerm==Sk -> E=[]
	;   E=[[EqS]],
%	    create_atom(Sk=SkolemTerm,Eq),
%	    EqS=[+|Eq],
	    EqS=[def|Sk=SkolemTerm]
%	    , print([[F1]|E])
	).

	

% need unique bound variables variables 
% will always expand nonmaximal formulas if expansion yields
% just one clause.

/* ----------------------------------------------------------------------------


      atoms in formulas

we need to know which atoms can result from tableau expansion


---------------------------------------------------------------------------- */
atom_in_expansion(X,X):-
	var(X),
	!.
atom_in_expansion([Sign|At],[Sign1|A]):-
	get_atom_pre_rel_left_right(At,_PreO,==,F,G),
	!,
	((var(F);F='$true$') ->
	     atom_in_expansion([Sign|G],[Sign1|A1])
	;(var(G);G='$true$') ->
	     atom_in_expansion([Sign|F],[Sign1|A1])
	;     
	     atom_in_expansion([Sign|F==G],[Sign1|A1])
	),
	\+ top(A1),
	\+ bottom(A1),
	create_atom(A1,A).  
atom_in_expansion(F,A):-
	alpha_formula(F,A1,A2),
	!,
	( atom_in_expansion(A1,A)
	; atom_in_expansion(A2,A)
	).
atom_in_expansion(F,A):-
	beta_formula(F,A1,A2),
	!,
	( atom_in_expansion(A1,A)
	; atom_in_expansion(A2,A)
	).
atom_in_expansion(F,B):-
	gamma_formula(F,A),
	!,
	atom_in_expansion(A,A1),
	subst(A1,_New,B).
atom_in_expansion(F,B):-
	delta_formula(F,SkolemTerm,A),
	!,
	atom_in_expansion(A,A1),
%	simplify_skolem_term(SkolemTerm,STS),
	subst(A1,SkolemTerm,B).
atom_in_expansion(F,F).




/* ----------------------------------------------------------------------------


      formula types


---------------------------------------------------------------------------- */

alpha_formula([+|A and B],[+|A],[+|B]).
alpha_formula([+|A == B],[+|A=>B],[+|B=>A]).
alpha_formula([+|A <=> B],[+|A=>B],[+|B=>A]).
alpha_formula([-|A or B],[-|A],[-|B]).
alpha_formula([-|A => B],[+|A],[-|B]).


beta_formula([+|A or B],[+|A],[+|B]).
beta_formula([-|A and B],[-|A],[-|B]).
beta_formula([+|A => B],[-|A],[+|B]).
beta_formula([-|A == B],[-|A=>B],[-|B=>A]).
%beta_formula([-|A == B],[+|A and ~B],[+|~A and B]).
beta_formula([-|A <=> B],[-|A=>B],[-|B=>A]).
beta_formula([+|'$exists'(SkT,B)],[+|B1],[+|B2]):-
	finite_domain([I,J]),
	free_vars(SkT,[]),
	subst(B,I,B1),
	subst(B,J,B2).
beta_formula([-|'$all'(SkT,B)],[-|B1],[-|B2]):-
	finite_domain([I,J]),
	free_vars(SkT,[]),
	subst(B,I,B1),
	subst(B,J,B2).





gamma_formula([+|'$all'(_,B)],[+|B]).
gamma_formula([-|'$exists'(_,B)],[-|B]).


/* maybe wrong because in different contexts different instantiations
   might be present
   
delta_formula([-|'$all'(SkT,B)],SkT1,[-|B]):-
	SkT=..[F|SkArgs],
	split(is_closed,SkArgs,_,VarArgs),
	SkT1=..[F|VarArgs].
delta_formula([+|'$exists'(SkT,B)],SkT1,[+|B]):-
	SkT=..[F|SkArgs],
	split(is_closed,SkArgs,_,VarArgs),
	SkT1=..[F|VarArgs].
*/

delta_formula([-|'$all'(SkT,B)],SkT,[-|B]).
delta_formula([+|'$exists'(SkT,B)],SkT,[+|B]).


negation([+| ~A],[-|A]).
negation([-| ~A],[+|A]).

top([+|'$true$']).
top([-|'$min']).

bottom([-|'$true$']).
bottom([+|'$min']).


/* ----------------------------------------------------------------------------


      simplification with constants


---------------------------------------------------------------------------- */

% simplifyE removes redundancies from our representation of atoms
% it shouldn't fail if irredundant

simplifyE([P|F==G],[P|F]):-
	top([+|G]),
	!.
simplifyE(F,F).



% simplify does real simplification, and fails, if not applicable
% should now be included in rewriting


simplify([P|F==G],[P|F]):-
	top([+|G]),
	!.
simplify([P|G==F],[P|F]):-
	top([+|G]),
	!.
simplify([P|F==G],F1):-
	bottom([+|G]),
	negate([P|F],F1),
	!.
simplify([P|G==F],F1):-
	bottom([+|G]),
	negate([P|F],F1),
	!.
simplify(F,A):-
	alpha_formula(F,A,B),
	top(B),
	!.
simplify(F,B):-
	alpha_formula(F,A,B),
	top(A),
	!.
simplify(F,B):-
	alpha_formula(F,_,B),
	bottom(B),
	!.
simplify(F,A):-
	alpha_formula(F,A,_),
	bottom(A),
	!.
simplify(F,B):-
	beta_formula(F,_,B),
	top(B),
	!.
simplify(F,A):-
	beta_formula(F,A,_),
	top(A),
	!.
simplify(F,A):-
	beta_formula(F,A,B),
	bottom(B),
	!.
simplify(F,B):-
	beta_formula(F,A,B),
	bottom(A),
	!.



negate([+|A],[-|A]).
negate([-|A],[+|A]).


simplify_skolem_term(T,ST):-
	functor(T,SkF,_N),
	free_vars(T,Vars),
	ST=..[SkF|Vars].


/* ----------------------------------------------------------------------------


      polarities


---------------------------------------------------------------------------- */



positive([Pol|At]):-
	get_atom_pre_rel_left_right(At,_PreO,==,F,G),
	simplifyE([Pol|F==G],H),
	positive1(H).


positive1(F):-
	alpha_formula(F,F1,F2),
	positive1(F1),
	positive1(F2).
positive1(F):-
	beta_formula(F,F1,F2),
	positive1(F1),
	positive1(F2).
positive1(F):-
	gamma_formula(F,F2),
	positive1(F2).
positive1(F):-
	delta_formula(F,_,F2),
	positive1(F2).
positive1(F):-
	negation(F,F2),
	negative1(F2).
positive1(F):-
	top(F).
positive1(F):-
	bottom(F).
positive1([+|F]):-
	standard_atom(F).

negative([Pol|At]):-
	get_atom_pre_rel_left_right(At,_PreO,==,F,G),
	simplifyE([Pol|F==G],H),
	negative1(H).

negative1(F):-
	alpha_formula(F,F1,F2),
	negative1(F1),
	negative1(F2).
negative1(F):-
	beta_formula(F,F1,F2),
	negative1(F1),
	negative1(F2).
negative1(F):-
	gamma_formula(F,F2),
	negative1(F2).
negative1(F):-
	delta_formula(F,_,F2),
	negative1(F2).
negative1(F):-
	negation(F,F2),
	positive1(F2).
negative1(F):-
	top(F).
negative1(F):-
	bottom(F).
negative1([-|F]):-
	standard_atom(F).





polar_subformula([Pol|At],T,H,S):-
	get_atom_pre_rel_left_right(At,_PreO,==,F,G),
	polar_sub([Pol|F==G],T,H,S).


% polar_sub(+F,?T,+H,?S) 
%   is true when F[H/S] = T, with T,S in [$true$,$min].
%   H need not necessarily be a subformula of F

polar_sub(F,T,F,T).
polar_sub(F,'$true$',H,S):-
	alpha_formula(F,F1,F2),
	polar_sub(F1,'$true$',H,S),
	polar_sub(F2,'$true$',H,S).
polar_sub(F,'$min',H,S):-
	alpha_formula(F,F1,F2),
	(  polar_sub(F1,'$min',H,S)
	;  
	   polar_sub(F2,'$min',H,S)
	).
polar_sub(F,'$min',H,S):-
	beta_formula(F,F1,F2),
	polar_sub(F1,'$min',H,S),
	polar_sub(F2,'$min',H,S).
polar_sub(F,'$min',H,S):-
	alpha_formula(F,F1,F2),
	(  polar_sub(F1,'$min',H,S)
	;  
	   polar_sub(F2,'$min',H,S)
	).
polar_sub(F,T,H,S):-
	gamma_formula(F,F1),
	polar_sub(F1,T,H,S).
polar_sub(F,T,H,S):-
	delta_formula(F,_,F1),
	polar_sub(F1,T,H,S).
polar_sub(F,T,H,S):-
	negation(F,F1),
	neg(T,T1),
	polar_sub(F1,T1,H,S).
polar_sub(F,'$true$',_,_):-
	top(F).
polar_sub(F,'$min',_,_):-
	bottom(F).


neg('$true$','$min').
neg('$min','$true').



/* ----------------------------------------------------------------------------


      representation conversions


---------------------------------------------------------------------------- */



clause_to_sequent([],([],[])).
clause_to_sequent([[-|E]|L],(A1,S)):-
	F=[-|E],
	(top(F) ->
	     A1=A
	;
	     A1=[E|A]
	),
	clause_to_sequent(L,(A,S)).
clause_to_sequent([[+|E]|L],(A,S1)):-
	F=[+|E],
	(top(F) ->
	     S1=S
	;
	     S1=[E|S]
	),
	clause_to_sequent(L,(A,S)).
clause_to_sequent([[def,_|E]|L],(A,[[def|E]|S])):-
	clause_to_sequent(L,(A,S)).



sequent_to_clause((A,S),L):-
	maplist(formula_as_neg_formula,A,AL),
	maplist(formula_as_pos_formula,S,SL),
	append(AL,SL,L),
	!.




formula_as_neg_formula(A,[-|A]).
formula_as_pos_formula(A,[+|A]).



fClause_to_clause(C,C1):-
	maplist(create_literal,C,C1).


create_literal([P|F],[P|A]):-
	create_atom(F,A).
	

weaken_sequent(_,([],[[def|E]]),([],[[n|E]])):-
	!.
weaken_sequent((AW,SW),(A1,S1),(A2,S2)):-
	append(AW,A1,A2),
	append(SW,S1,S2).



createSimpleClauseP((A,S),C):-
	createSimpleClause([A,S],C).


% --------------------------------- test ----------------------------------



cl(([],[A])):-
%	E=(eq(X,Y)==all(member('i$0',X)==member('i$0',Y))),
%	E = (~all(all(sub(u('i$1','i$0'),u('i$0','i$1'))and sub(u('i$0','i$1'),u('i$1','i$0'))))),
%	E =(~all(all(all(in('i$0','i$1')or in('i$0','i$0')=>in('i$0','i$0')or in('i$0','i$1'))and all(in('i$0','i$0')or in('i$0','i$1')=>in('i$0','i$1')or in('i$0','i$0'))))),

	E=(~all((f('i$0')=>h('i$0'))and all(g('i$0')=>j('i$0')))==all(all(f('i$1')and g('i$0')=>h('i$1')and j('i$0')))),
	create_atom(E,A).


	
test((A,S),C1):-
	(sk(A,[],S,[],Cls) ->
	     basics:member(C2,Cls),
%	     format('~Nselecting ~p',[C2]),
	     test(C2,C1)
	;
	     C1=(A,S)
	).
	
test1(C):-
	setof(CC,test(C,CC),C1),
	maplist:maplist(printc,C1).


printc((A,S)):-
	nl,
	adt_atom:print_list_atoms(A), write(' -> '),adt_atom:print_list_atoms(S).

