% -*- Module: softtypes; -*-

:-module(softtypes,[abstraction/3,
		    no_abstraction/3,
		    abstract_program/2,
		    delete_goals/2,
		    read_program/2,
		    magic/2,
		    magicS/2,
		    original_predicate/1,
		    prolog_to_sequent/2,
		    write_program/3
		   ]).


:-use_module(myterms,[list_of_vars/2,
		      skolemized/3]).


:-use_module(commands,[out_as_dfg/2
		       ]).

:-use_module(hgstuff,[split/4]).


:-use_module(thisProlog,[portray_clause2/2]).

:-use_module(auxiliary,[appendAtoms/3,
		  list_as_conjunction/2,
		  list_as_disjunction/2,
		  lookup_ml/2,
		  member212/2]).

:-use_module(global,[store/2]).


:-use_module(library(maplist),[maplist/3]).
:-use_module(library(basics),[member/2]).
:-use_module(library(flatten),[flatten/2]).
:-use_module(library(sets),[select/3]).
:-use_module(library(change_arg), [change_path_arg/4]).
:-use_module(library(arg), [path_arg/3]).


:-dynamic original_predicate/1.

/* ----------------------------------------------------------------------------

Types
=====

Cls  = list[Clause]
Clause = (Head:-Body)
Body = [ true  | Atom | (Atom , Body) ]
Head = [ false | Atom | (Atom ; Body) ]
Atom = Prolog atom

Goals are represented in the form false:-Body.


Mode = [ dfg | saturate | prolog ]

Sequent = (list[Atom] -> list[Atom])   Prolog clauses in sequent notation

Filename = Prolog atom 
          (don't forget the single quotes if the filename is not an identifier)



Predicates
==========

abstract_program(+P:Cls,?A:Cls)

  abstracts a Prolog program  P
  via Tau_P into program A.




abstraction(+In:filename,?Out:filename,+M:Mode)

  abstracts a Prolog program from input file In
  via Tau_P into an abstract program on output file Out.

  Mode specifies the output format:

  M=saturate: Saturate format
  M=dfg:      dfg format (suited for SPASS)
  M=prolog:   Prolog format



read_program(+F:filename,-P:Cls)

  read Prolog program from file F into P


write_program(+Out:filename,+P:Cls,+M:Mode)

  write program P to file Out in format M.


cd(+D:filename)

  cd to directory D (you need write permission on D)


delete_goals(+P:Cls, ?P1:Cls)
 
  P1 is P minus goal clauses


magicC(+C:Clause,  ?MC:Cls)
magicS(+C:Sequent,?MC:list[Sequent])
magic(+C:Cls,?MC:Cls)

  magic set transformation for a single clause,
  a single clause in sequent notation, and
  a list of clauses




Typical call sequence
=====================



:- 
     read_program('dnf.pl',P),
     magic(P,PM),
     delete_goals(PM,PM1),
     abstract_program(PM1,PA),
     write_program('dnfma.dfg',PA,dfg).




----------------------------------------------------------------------------*/


abstraction(In,Out,Mode):-
	read_program(In,P),
	abstract_program(P,A),
	write_program(Out,A,Mode).

no_abstraction(In,Out,Mode):-
	read_program(In,P),
	write_program(Out,P,Mode).



write_program(Out,A,saturate):-
	!,
	maplist(clause_prolog_saturate,A,C),
	write_program(Out,C),
	write_sat_epilogue(Out).
write_program(Out,A,dfg):-
	!,
	(setof(O,original_predicate(O),Os) ->
	     maplist(dfg_operators,Os,Os_),
	     abolish(original_predicate,1),
	     maplist(assert_opred,Os_)
	;true),
	maplist(dfg_operators,A,A_),
	maplist(clause_prolog_saturate,A_,SatC),
	write_program('/tmp/sat_input',SatC),
	store(better_precedence,no),  % complexity problem
	user:in('/tmp/sat_input'),
	abolish(original_predicate,1),
	store(better_precedence,yes),
	out_as_dfg(Out,dfg).
write_program(Out,A,prolog):-
	maplist(normalize_Prolog_clause,A2,A),
	write_program(Out,A2).
write_program(_Out,_A,_Mode):-
	raise_exception(unknown_mode).



assert_opred(P):-
	assert(original_predicate(P)).


read_program(F,PN):-
	open(F,read,S),
	readTerms(S,P),
	close(S),
	maplist(normalize_Prolog_clause,P,PN),
	!.


readTerms(S,[T|Ts]):-
	read(S,T),
	\+atend(T),
	!,
	readTerms(S,Ts).
readTerms(_S,[]).



atend(X):-
	X==end_of_file.


normalize_Prolog_clause(X,Y):-
	var(X),
	var(Y),
	!,
	raise_exception(clause_is_variable).
normalize_Prolog_clause((:-B),(false:-B)):-
	!.
normalize_Prolog_clause((?-B),(false:-B)):-
	!.
normalize_Prolog_clause((H:-B),(H:-B)):-
	!.
normalize_Prolog_clause(H,(H:-true)).

delete_goals(P,P1):-
	split(is_goal,P,_Goals,P1).


is_goal((false:-_)).


abstract_program(PN,A):-
	abolish(original_predicate,1),
	% for precedence generation we store those original predicates
	% which become function symbols by transformation into monadic form
	number_list(PN,NumberedP),
	maplist(abstract_clause,NumberedP,As),
	flatten(As,A).

write_program(F,P):-
	open(F,write,S),
	maplist(portray_clause2(S),P),
	close(S).


write_sat_epilogue(F):-
	open(F,append,S),
	portray_clause2(S,option((select(Empty,[[At1,At2],[]]):-
				   (  Empty=At1,
				      Prop=At2
				   ;  Empty=At2,
				      Prop=At1
				   ),
				    atom(Prop),
				    call(adt_term:free_vars(Empty,[_])),
				    !
				   ))
		       ),
	close(S).





% bidirectional
clause_prolog_saturate((false:-H),(H->[])):-
	!.
clause_prolog_saturate((H:-B),(H<-B)):-
	B\==true,
	!.
clause_prolog_saturate((H:-true),H):-
	!.


clause_prolog_dfg((false:-B),C):-
	list_of_vars(B,Vs),
	list_as_conjunction(Atoms,B),
	maplist(negate,Atoms,Literals),
	Disj=..[or|Literals],
	(Vs=[] ->
	     C='$dfg'(clause(Disj))
	;
	     C='$dfg'(clause(forall(Vs,Disj)))
	).
clause_prolog_dfg((H:-true),C):-
	!,
	list_of_vars(H,Vs),
	(Vs=[] ->
	     C='$dfg'(clause(or(H)))
	;
	     C='$dfg'(clause(forall(Vs,or(H))))
	).
clause_prolog_dfg((H:-B),C):-
	list_of_vars((H,B),Vs),
	list_as_conjunction(Atoms,B),
	maplist(negate,Atoms,Literals),
	Disj=..[or|[H|Literals]],
	(Vs=[] ->
	     C='$dfg'(clause(Disj))
	;
	     C='$dfg'(clause(forall(Vs,Disj)))
	).





number_list(P,NP):-
	number_list(P,0,NP).



number_list([],_N,[]).
number_list([X|Xs],N,[(N1,X)|NXs]):-
	N1 is N+1,
	number_list(Xs,N1,NXs).



abstract_clause((_N,(H:-B1)),[(H:-B)]):-
%	(_N=37 -> trace; true),
	atom(H),
	!,
	monadic_clause((H:-B1),(H:-B)).
abstract_clause((N,(H1:-B1)),ACs):-
	monadic_clause((H1:-B1),(H:-B)),
	functor(H,P,_),
	head_functor_clauses(H,HFCs,P,N,E),
	delete_repetitions(E,E1),
	maplist(body_functor_clause(P,N,B),E1,BFCs),
	append(HFCs,BFCs,ACs).

body_functor_clause(P,N,Body,Eq,(Ty:-Body)):-
	vareq_as_type(P,N,Eq,Ty).

	



linearize(X,N,E, X1,N1,E1):-
	var(X),
	!,
	(lookup_ml(X=XN,E) ->
	     % non-linear Variable
	     % new X1 with old name XN
%	     X1=X,
	     N1=N
	;% linear variable
	     N1 is N+1,
	     X1=X,
	     XN=N1
	),
	     
	E1=[X1=XN|E].
linearize(X,N,E, X1,N1,E1):-
	X=..[F|Xs],
	linearize_list(Xs,N,E,Xs1,N1,E1),
	X1=..[F|Xs1].


linearize_list([],N,E,[],N,E).
linearize_list([X|Xs],N,E,[X1|Xs1],N1,E1):-
	linearize(X,N,E, X1,N2,E2),
	linearize_list(Xs,N2,E2,Xs1,N1,E1).



delete_repetitions([],[]).
delete_repetitions([Eq|E],E1):-
	(\+ \+ member(Eq,E)->
	     delete_repetitions(E,E1)
	;
	     delete_repetitions(E,E2),
	     E1=[Eq|E2]
	).


head_functor_clauses(H,HFCs,P,N,E):-
	linearize(H,0,[],HL,_,E),
	maplist(vareq_as_type(P,N),E,B),
	list_as_conjunction(B,BC),
	flattened_head(N,(HL:-BC),HFCs).


flattened_head(_,(H:-B),[(H:-B)]):-
	arg(1,H,H1),
	(var(H1);flat_term(H1)),
	!.
flattened_head(N,(H:-B),[(PFT:-B1)|HFCs]):-
	Pos=[_,_|_],
	adt_term:path_arg1(Pos,H,T),
	flat_term(T),
	!,
	functor(H,P,_),
	change_path_arg(Pos,H,H1,Y),
	(atom(T) ->
	     op_prolog_dfg(T,O),
	     appendAtoms([is,O],'_',PF),
	     B1=true
	;
	     functor(T,F,_),
	     op_prolog_dfg(F,O),
	     aux_predicate(P,O,N,Pos,PF),
	     non_var_disjoint_atoms(T,B,B1)
	),
	(B==true -> PFYB=PFY; PFYB=(PFY,B)),
	PFT=..[PF,T],
	PFY=..[PF,Y],
	flattened_head(N,(H1:-PFYB),HFCs).

flat_term(A):-
	nonvar(A),
	A=..[_|Xs],
	maplist(var,Xs).

aux_predicate(P,F,N,Pos,PF):-
	appendAtoms([P,F,N|Pos],'_',PF).

non_var_disjoint_atoms(T,B,B1):-
	list_as_conjunction(BL,B),
	split(adt_term:disjointVars(T),BL,_,BL1),
	list_as_conjunction(BL1,B1),
	!.
	

% ------------ magic set transformations --------

magic(C,D):-
	maplist(magicC,C,Ds),
	flatten(Ds,D).


magicC(C,Ds):-
	prolog_to_sequent(C,C1),
	magicS(C1,Ds1),
	maplist(prolog_to_sequent,Ds,Ds1).


magicS((B->H),[(B1->H)| GClauses ]):-
	maplist(goalform,H,GoalH),
	append(GoalH,B,B1),
	gclauses(B,[],GoalH,GClauses).


gclauses([],_,_,[]).
gclauses([A|As],B,GH,[(GB->[GA] ) | GClauses]):-
	goalform(A,GA),
	append(GH,B,GB),
	gclauses(As,[A|B],GH,GClauses).



	
	





% --------------- conversions -------------------


goalform(A,GA):-
	A=..[P|Ts],
	goalPredicate(P,GP),
	GA=..[GP|Ts].



goalPredicate(P,GP):-
	appendAtoms([goal,P],'__',GP).



vareq_as_type(P,N,V=Name,T):-
	appendAtoms([P,N,Name],'_',PA),
	T=..[PA,V].


monadic(A,tt_(A)):-
	functor(A,P,N),
	N>1,
	assert(original_predicate(P/N)),
	!.
monadic(A,A).


monadic_clause((H:-B),(HM:-BM)):-
	!,
	list_as_conjunction(BL,B),
	maplist(monadic,[H|BL],[HM|BML]),
	list_as_conjunction(BML,BM).
monadic_clause(A,AM):-
	monadic(A,AM).



prolog_to_sequent((H:-B),(BL->HL)):-
	list_as_disjunction(HL,H),
	list_as_conjunction(BL,B).






dfg_operators(S,S):-
	var(S),
	!.
dfg_operators(S,T):-
	S=..[F|Ss],
	op_prolog_dfg(F,G),
	!,
	maplist(dfg_operators,Ss,Ts),
	T=..[G|Ts].


op_prolog_dfg([],nil_).
op_prolog_dfg('.',cons_).
op_prolog_dfg(dnf,dnf_).
op_prolog_dfg(O,O).   % default


negate(L,not(L)).
