% st_delete.pl
% deletion in substitution trees
% In order to delete an identical term
%     st_delete_term(Term,ST0,ST1,identical)
% In order to delete all variants
%     st_delete_term(Term,ST0,ST1,true) or
%     st_delete_term(Term,ST0,ST1)
% In order to delete variants with special info
%     st_delete_term(Term,ST0,ST1,Module:Test)
% where Test is a predicate that starts with any number of arguments
% and gets Info appended if a variant is found.
% If Test succeds the entry is deleted (No bindings remain from applying test).
% Author: pb

:- module(st_delete,[
	st_delete_term/3,
	st_delete_term/4
]).

:- use_module(mylists).
:- use_module(substitution).
:- use_module(normalization).
:- use_module(variant).
:- use_module(sound_unify).

st_delete_term(Term,ST0,ST1) :-
	st_delete_term(Term,ST0,ST1,true).

st_delete_term(Term,substitutiontree(U0,ST0,Star0),substitutiontree(U0,ST1,Star0),Test) :-
	normalization(Term,NormalTerm,StarSubst,Star0,Star1),
	st_del(ST0,ST1,[U0-NormalTerm],StarSubst,Star1,Test).

st_del(t,t,_,_,_,_). % deletion in empty tree
st_del(st(Tau,leaf(Infos0)),ST1,Ro,StarSubst,StarVars,Test) :- 
        % a leaf node, check whether there is an identical term
	!, 
	(variant(Tau,Ro,StarVars,_Sigma) ->
%	    subst_concat(Ro,Sigma,RoSigma), % warum brauch ich das nicht?
	    (delete_if_test(Infos0,StarSubst,Test,Infos1) ->
		(Infos1 = [] -> % no more terms with this variant
		    ST1 = t
		;
		    ST1 = st(Tau,leaf(Infos1))
		)
	    ;
		ST1 = st(Tau,leaf(Infos0)) % not found; no change
	    )
	;
	    ST1 = st(Tau,leaf(Infos0)) % no variant; no change
	).

st_del(st(Tau,SubNodes0),ST1,Ro,StarSubst,StarVars,Test) :-
	variant(Tau,Ro,StarVars,Sigma),!,
	subst_concat(Ro,Sigma,RoSigma), 
	st_del_subnodes(SubNodes0,SubNodes1,RoSigma,StarSubst,StarVars,Test),
	st_del_repair(SubNodes1,NewSubNodes,Tau,NewTau,StarVars),
	(NewSubNodes = [] ->
	    ST1 = t
	;
	    ST1 = st(NewTau,NewSubNodes)
	).	    
st_del(ST,ST,_Ro,_StarSubst,_StarVars,_Test). % No variant; nothing to do


st_del_subnodes([],[],_Ro,_StarSubst,_StarVars,_Test).
st_del_subnodes([ST0|ST0s],ST1s,Ro,StarSubst,StarVars,Test) :-
	st_del(ST0,ST1,Ro,StarSubst,StarVars,Test),
	(ST1 == t ->
	    ST1s = ST2s
	;
	    ST1s = [ST1|ST2s]
	),
	st_del_subnodes(ST0s,ST2s,Ro,StarSubst,StarVars,Test).


st_del_repair([],[],_,_,_).
st_del_repair([st(TauPrime,SubNodes)],SubNodes,Tau,TauTauPrime,_StarVars) :-
	!, % just one subnode
	subst_join(Tau,TauPrime,TauTauPrime).
st_del_repair(SubNodes,SubNodes,Tau,Tau,_). % more than one SubNodes; just copy


% delete_identical_starsubst(Infos0,StarSubst,Infos1).
delete_if_test([],_,_,[]).
delete_if_test([StarSubst0-Info0|Infos0],StarSubst1,Test,Infos1) :-
	% if the initial terms are identical, then a simple check
	% of identical starsubst would do, i.e. StarSubst0 == StarSubst1
	% but one can also find variable renamed substitutions.
	% In that case, we just apply the starsubst while doing the test
	% by quoting it with findall
        findall(Res,do_the_test(StarSubst0,StarSubst1,Info0,Test,Res),LRes),       
	LRes = [t|_],
	!, % ok; delete it
	delete_if_test(Infos0,StarSubst1,Test,Infos1).		    
delete_if_test([S-I|Infos0],StarSubst1,Test,[S-I|Infos1]) :-
	delete_if_test(Infos0,StarSubst1,Test,Infos1).

do_the_test(StarSubst0,StarSubst1,Info,Test,Res) :-
	% the StarSubsts need to be applied;
	% both starsubsts are identical, the renaming are to
	% be unified
	(functor(Test,identical,_) ->
	    StarSubst0 == StarSubst1
	;
	    do_the_test_unify_starsubsts(StarSubst0,StarSubst1),
	    (functor(Test,true,_) ->
		Res = t
	    ;
		call(Test,[Info]),!,
		Res = t
	    )
	).


do_the_test_unify_starsubsts([],[]).
do_the_test_unify_starsubsts([V0-T0|StarSubst0],[V1-T1|StarSubst1]) :-
	(V0 == V1 ->
	    sound_unify(T0,T1)
	;
	    write('Damned; non identical starsubst domain in deletion-test'),nl
	),
	do_the_test_unify_starsubsts(StarSubst0,StarSubst1).

% the different deletions wrt. the merging Substitution function

st_delete_variant(Term,ST0,ST1) :-
	st_delete_gen(Term,ST0,ST1,variant).


% generic deletion procedure
% MergeSubst/2; a predicate that merges substitutions
% deletes all terms that are 
% st_delete_gen(Term,ST0,ST1,MergeSubst).
st_delete_gen(Term,substitutiontree(U0,ST0,Star0),substitutiontree(U0,ST1,Star0),MergeSubst) :-
	normalization(Term,NormalTerm,StarSubst,Star0,Star1),
	st_del_gen(ST0,ST1,[U0-NormalTerm],StarSubst,Star1,MergeSubst).

st_del_gen(t,t,_,_,_,_). % deletion in empty tree
st_del_gen(st(Tau,leaf(Infos0)),ST1,Ro,_StarSubst,StarVars,MergeSubst) :- 
        % a leaf node, do admissible deletions 
	!, 
	(call(MergeSubst,[Tau,Ro,StarVars,_Sigma]) -> % ok; now delete all
	    ST1 = t
	;
	    ST1 = st(Tau,leaf(Infos0)) % not found; no change
	).

st_del_gen(st(Tau,SubNodes0),ST1,Ro,StarSubst,StarVars,MergeSubst) :-
	call(MergeSubst,[Tau,Ro,StarVars,Sigma]),!,
	subst_concat(Ro,Sigma,RoSigma), 
	st_del_gen_subnodes(SubNodes0,SubNodes1,RoSigma,StarSubst,StarVars,MergeSubst),
	st_del_repair(SubNodes1,NewSubNodes,Tau,NewTau,StarVars),
	(NewSubNodes = [] ->
	    ST1 = t
	;
	    ST1 = st(NewTau,NewSubNodes)
	).	    
st_del_gen(ST,ST,_Ro,_StarSubst,_StarVars,_MergeSubst). % No variant; nothing to do


st_del_gen_subnodes([],[],_Ro,_StarSubst,_StarVars,_MergeSubst).
st_del_gen_subnodes([ST0|ST0s],ST1s,Ro,StarSubst,StarVars,MergeSubst) :-
	st_del_gen(ST0,ST1,Ro,StarSubst,StarVars),
	(ST1 == t ->
	    ST1s = ST2s
	;
	    ST1s = [ST1|ST2s]
	),
	st_del_gen_subnodes(ST0s,ST2s,Ro,StarSubst,StarVars,MergeSubst).


		



% subst_join(Tau,Ro,Join).
% applies Ro to all assignments in Tau and deletes the 
% used substitutions in Ro.
subst_join(Tau,Ro,Join) :-
	subst_join_init(Ro,VRo),
	subst_join2(Tau,VRo,TauRo),
	subst_join_delete(VRo,DRo),
	user:append(TauRo,DRo,Join).

subst_join_init([],[]).
subst_join_init([V-T|Ro],[V-T-_|VRo]) :-
	subst_join_init(Ro,VRo).

subst_join_delete([],[]).
subst_join_delete([V-T-Ind|VRo],Ro) :-
	(Ind == t ->
	    subst_join_delete(VRo,Ro)
	;
	    Ro = [V-T|Ro1],
	    subst_join_delete(VRo,Ro1)
	).
	
subst_join2([],_,[]).
subst_join2([V-T|Tau],Ro,[V-TT|TauRo]) :-
	subst_apply_mark(T,Ro,TT),
	subst_join2(Tau,Ro,TauRo).


subst_apply_mark(T0,Subst,T1) :-
	var(T0),!,
	(dom_member_mark(Subst,T0,T1) ->
	    true
	;
	    T0 = T1
	).
subst_apply_mark(T0,Subst,T1) :-
	functor(T0,F,A),
	functor(T1,F,A),
	subst_apply_mark(A,T0,Subst,T1).

subst_apply_mark(0,_,_,_) :- !.
subst_apply_mark(A,T0,Subst,T1) :-
	arg(A,T0,ST0),
	arg(A,T1,ST1),
	subst_apply_mark(ST0,Subst,ST1),
	A1 is A-1,
	subst_apply_mark(A1,T0,Subst,T1).

	    
dom_member_mark([V0-T0-Ind|Subst],V1,T1) :-
	(V0 == V1 ->
	    T0 = T1,
	    Ind = t
	;
	    dom_member_mark(Subst,V1,T1)
	).



% from O'Keefe; 
% calls a term with a list of extra arguments

call(Term,Arguments) :-
	strip_module_prefix(Term, user, Form, Module),
	append_term(Form,Arguments,Call),
	call(Module:Call).

append_term(Term, MoreArgs, FullTerm) :-
        functor(Term, Symbol, Arity),
        atom(Symbol),           % must test as MoreArgs=[] is ok.
        len(MoreArgs, Arity, FullArity),
        functor(FullTerm, Symbol, FullArity),
        append_term1(MoreArgs, Arity, FullTerm),
        append_term2(Arity, Term, FullTerm).

len(0, _, _) :- !, fail.        % catch partial lists
len([], N, N).
len([_|L], N0, N) :-
        N1 is N0+1,
        len(L, N1, N).

append_term1([A1,A2|Args], M, FullTerm) :- !,
        N1 is M+1, arg(N1, FullTerm, A1),
        N2 is M+2, arg(N2, FullTerm, A2),
        append_term1(Args, N2, FullTerm).
append_term1([A1], M, FullTerm) :- !,
        N1 is M+1, arg(N1, FullTerm, A1).
append_term1([], _, _).

append_term2(N, OldTerm, NewTerm) :-
        (   N > 1 ->
            arg(N, OldTerm, ArgN),
            arg(N, NewTerm, ArgN),
            M is N-1,
            arg(M, OldTerm, ArgM),
            arg(M, NewTerm, ArgM),
            L is M-1,
            append_term2(L, OldTerm, NewTerm)
        ;   N > 0 ->
            arg(1, OldTerm, Arg1),
            arg(1, NewTerm, Arg1)
        ;   true
        ).


strip_module_prefix(0, _, _, _) :- !,
        fail.
strip_module_prefix(Prefix:Term, _, Goal, Module) :- !,
        strip_module_prefix(Term, Prefix, Goal, Module).
strip_module_prefix(Goal, Module, Goal, Module) :-
        nonvar(Goal),           % should be callable(Goal)
        atom(Module).

