% st_insert.pl
% insert a substitution in a substitution tree
% usage: st_insert_term(Term,Info,ST0,ST1).
% inserts Term with associated Info into substitution tree ST0 yielding ST1
% Info may be ommitted and defaults to t.
% st_empty_tree(ST) unifies ST with th empty substitution tree
% Author: pb

:- module(st_insert,[
	st_insert_term/3,
	st_insert_term/4,
	st_empty_tree/1,
	st_is_empty_tree/1
]).

:- use_module(mylists).
:- use_module(normalization).
:- use_module(sound_unify).
:- use_module(substitution).
:- use_module(variant).
:- use_module(mscg).
:- use_module(vars).

st_empty_tree(substitutiontree(_U,t,[])).
st_is_empty_tree(substitutiontree(_U,ST,_)) :-
	ST == t.


st_insert_term(Term,ST0,ST1) :-
	st_insert_term(Term,t,ST0,ST1).

st_insert_term(Term,Info,substitutiontree(U0,ST0,Star0),substitutiontree(U0,ST1,Star1)) :-
	normalization(Term,NormalTerm,StarSubst,Star0,Star1),
	st_ins(ST0,ST1,[U0-NormalTerm],[U0],Star1,StarSubst,Info).

st_ins(t,st(RoRestricted,leaf([StarSubst-Info])),Ro,OV,_,StarSubst,Info) :-
	% the empty tree
	subst_restricted_to(Ro,OV,RoRestricted).
st_ins(st(Tau,leaf(Infos)),ST1,Ro,OV,StarVars,StarSubst,Info) :-
	!, % its a leaf
	(variant(Tau,Ro,StarVars,_Sigma) ->
	    % it is a variant of a leafnode; fine!
	    ST1 = st(Tau,leaf([StarSubst-Info|Infos]))
	;
	    % calculate mscg and split; as in last case
	    mscg_subst(Tau,Ro,Mu,Sigma1,Sigma2,StarVars),
	    st_ins_mscg(Tau,leaf(Infos),ST1,Ro,OV,StarVars,StarSubst,Info,Mu,Sigma1,Sigma2)
	).
st_ins(st(Tau,SubNodes0),ST1,Ro,OV,StarVars,StarSubst,Info) :-
	variant(Tau,Ro,StarVars,Sigma),!,
	st_ins_variant(st(Tau,SubNodes0),ST1,Ro,OV,StarVars,StarSubst,Info,Sigma).
st_ins(st(Tau,SubNodes0),ST1,Ro,OV,StarVars,StarSubst,Info) :-
	mscg_subst(Tau,Ro,Mu,Sigma1,Sigma2,StarVars),
	st_ins_mscg(Tau,SubNodes0,ST1,Ro,OV,StarVars,StarSubst,Info,Mu,Sigma1,Sigma2).

% insert if mscg is known
st_ins_mscg(Tau,SubNodes0,ST1,Ro,OV,_StarVars,StarSubst,Info,Mu,Sigma1,Sigma2) :-
	ST1 = st(Mu,[SubNode1,SubNode2]),
	SubNode1 = st(Sigma1,SubNodes0),
	domain(Tau,TauDomain),
	var_setminus(OV,TauDomain,Restriction),
	subst_restricted_to(Ro,Restriction,RoRestricted),
	user:append(Sigma2,RoRestricted,NewSigma2), % union for substitutions
	SubNode2 = st(NewSigma2,leaf([StarSubst-Info])).	

% insert if variant is known!
st_ins_variant(st(Tau,leaf(Infos)),ST1,_Ro,_OV,_StarVars,StarSubst,Info,_Sigma) :-
	!, % its a leaf
	% variant(Tau,Ro,StarVars,Sigma), already computed
        ST1 = st(Tau,leaf([StarSubst-Info|Infos])).
st_ins_variant(st(Tau,SubNodes0),ST1,Ro,OV,StarVars,StarSubst,Info,Sigma) :-
	% variant(Tau,Ro,StarVars,Sigma),!, %already computed variant (Sigma)
	subst_concat(Ro,Sigma,RoSigma),
	get_new_ov(OV,NewOV,Tau),
	(select_variant_node(SubNodes0,Node,SubNodes1,RoSigma,StarVars,Variant) ->
	    ST1 = st(Tau,[NewNode|SubNodes1]),
	    st_ins_variant(Node,NewNode,RoSigma,NewOV,StarVars,StarSubst,Info,Variant)
	;
	    (select_nonemptymscg_node(SubNodes0,Node,SubNodes1,RoSigma,StarVars) ->
		ST1 = st(Tau,[NewNode|SubNodes1]),
		st_ins(Node,NewNode,RoSigma,NewOV,StarVars,StarSubst,Info)
	    ;
		subst_restricted_to(RoSigma,NewOV,NewTau),
		NewNode = st(NewTau,leaf([StarSubst-Info])),	
		ST1 = st(Tau,[NewNode|SubNodes0])
	    )
	).

select_variant_node([Node0|SubNodes0],Node0,SubNodes0,RoSigma,StarVars,Variant) :-
	Node0 = st(Tau,_),
	variant(Tau,RoSigma,StarVars,Variant),!.
select_variant_node([Node0|SubNodes0],Node1,[Node0|SubNodes1],RoSigma,StarVars,Variant) :-
	select_variant_node(SubNodes0,Node1,SubNodes1,RoSigma,StarVars,Variant).

select_nonemptymscg_node([Node0|SubNodes0],Node0,SubNodes0,RoSigma,_StarVars) :-
	Node0 = st(Tau,_),
	nonemptymscg_subst(Tau,RoSigma),
	!. 
select_nonemptymscg_node([Node0|SubNodes0],Node1,[Node0|SubNodes1],RoSigma,StarVars) :-
	select_nonemptymscg_node(SubNodes0,Node1,SubNodes1,RoSigma,StarVars).






% debugging

star_var_bound([]).
star_var_bound([V|StarVars]) :-
	(var(V) ->
	    star_var_bound(StarVars)
	;
	    write('SCHEISSE!'),nl,
	    spy_it
	).
	    
spy_it :- write('spy').

get_new_ov(OV,NewOV,Tau) :-
	domain(Tau,TauDomain),
	var_codomain(Tau,TauVCOD),
	var_setminus(OV,TauDomain,Part1),
	var_union(TauVCOD,Part1,NewOV).

	
% Examples
% P. Graf Diss
% st_insert_term(f(Z,g(b)),substitutiontree(U,t,[S1,S2,S3]),ST1).
% st_insert_term(f(Z,g(b)),substitutiontree(U,t,[S1,S2,S3]),ST1),st_insert_term(f(Y,g(b)),ST1,ST2).
% st_insert_term(f(Z,g(b)),substitutiontree(U,t,[S1,S2,S3]),ST1),st_insert_term(f(Y,g(b)),ST1,ST2),st_insert_term(f(a,b),ST2,ST3).
% st_insert_term(f(Z,g(b)),substitutiontree(U,t,[S1,S2,S3]),ST1),st_insert_term(f(Y,g(b)),ST1,ST2),st_insert_term(f(a,b),ST2,ST3),st_insert_term(f(c,g(d)),ST3,ST4).
% st_insert_term(f(Z,g(b)),substitutiontree(U,t,[S1,S2,S3]),ST1),st_insert_term(f(Y,g(b)),ST1,ST2),st_insert_term(f(a,b),ST2,ST3),st_insert_term(f(c,g(d)),ST3,ST4),st_insert_term(f(b,g(a)),ST4,ST5).

% Braid4Rules
% st_insert_term(rule(f305(X),X),substitutiontree(U,t,[S1,S2,S3]),ST1).
% st_insert_term(rule(f305(X),X),substitutiontree(U,t,[S1,S2,S3]),ST1),st_insert_term(rule(fP(fp(X)),X),ST1,ST2).
% st_insert_term(rule(f305(X),X),substitutiontree(U,t,[S1,S2,S3]),ST1),st_insert_term(rule(fP(fp(X)),X),ST1,ST2),st_insert_term(rule(fp(fP(X)),X),ST2,ST3).

% ec_pos_500
% st_insert_term(p(e(e(X,e(Y,e(Z,e(e(U,e(V,Z)),e(V,U))))),e(Y,X))),substitutiontree(U,t,[S1,S2,S3,S4,S5]),ST1).
% st_insert_term(p(e(e(X,e(Y,e(Z,e(e(U,e(V,Z)),e(V,U))))),e(Y,X))),substitutiontree(UU,t,[S1,S2,S3,S4,S5]),ST1),st_insert_term(p(e(e(e(e(X,e(Y,Z)),e(Y,X)),e(Z,U)),u)),ST1,ST2).



print_subst_tree(substitutiontree(U,ST,StarVars)) :-
	print_subst_tree(substitutiontree(U,ST,StarVars),[]).
print_subst_tree(substitutiontree(U,ST,StarVars),WriteBinds) :-
	print_st(ST,0,U,StarVars,WriteBinds).

print_st(st(Subst,Nodes),Tab,U,StarVars,WriteBinds) :-
	tab_write(Tab),
	write_subst(Subst,StarVars,WriteBinds),
	Tab1 is Tab + 1,
	print_st_list(Nodes,Tab1,U,StarVars,WriteBinds).

print_st_list(leaf(Info),Tab,_U,StarVars,WriteBinds) :-
	!,
	tab_write(Tab),
	write_leaf(Info,StarVars,WriteBinds).
print_st_list([],_,_U,_StarVars,_WriteBinds).
print_st_list([Node|Nodes],Tab,U,StarVars,WriteBinds) :-
	print_st(Node,Tab,U,StarVars,WriteBinds),
	print_st_list(Nodes,Tab,U,StarVars,WriteBinds).
	
write_subst(Subst,StarVars,WriteBinds) :- 
	bind_star_vars(StarVars,['*1','*2','*3','*4','*5','*6','*7']),
	make_write_binds(WriteBinds),
	write(Subst),nl,
	fail.
write_subst(_Subst,_StarVars,_WriteBinds).

write_leaf(Info,StarVars,WriteBinds) :-
	make_write_binds(WriteBinds),
	bind_star_vars(StarVars,['*1','*2','*3','*4','*5','*6','*7']),
	write(leaf(Info)),nl,
	fail.
write_leaf(_Info,_StarVars,_WriteBinds).

make_write_binds([]).
make_write_binds([V-V|WriteBinds]) :-
	make_write_binds(WriteBinds).

bind_star_vars([],_).
bind_star_vars([V|Vs],[V|StarVars]) :-
	bind_star_vars(Vs,StarVars).

tab_write(0) :- !.
tab_write(N) :- write(' '),N1 is N-1,tab_write(N1).

