:- module(stoC, [ 
	simplify/2,
	simplify_light/2,
	simplify_med/2,
	simplify_constraint_vars/3
   ]).

:- use_module(sto, [
	greater/2,
	greater_op_sto/2,
	greaterEqual/2
   ]).

:- use_module(lpoC,[]).

:- use_module(timing, [
	timing/2
   ]).

:- use_module(proofState, [
        s_con/2,
	record_s_con/2
   ]).


:- use_module(props, [
        memo_constraints/1
   ]).

:- use_module(adt_precedence, [
	is_predicate/1
   ]).

:- use_module(auxiliary, [
	member212/2
   ]).

:- use_module(constraintSolverBasics, [
	implies/2,
	satisfiable_ineq/1
   ]).

:- use_module(equationalTheory, [
        mgu/2
     ]).

:- use_module(orderConstraints, [
	true/1,
	delete_implying_ands/2

   ]).

:- use_module(myterms, [
	apply_substitution/3,
	skolemized/3
   ]).

:- use_module(adt_term, [
        path_arg1/3
   ]).


:- use_module(library(maplist), [
	maplist/3
   ]).

:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(sets), [
	select/3
   ]).

:- use_module(library(arg), [
	genarg/3
   ]).

:- use_module(sound_unify, [
	sound_unify/2
   ]).






%------------------------------------------------------------------------------
% CONSTRAINT SIMPLIFICATION
%------------------------------------------------------------------------------

simplify_constraint_vars(A,B,C):-
	constraintSolverBasics:simplify_constraint_vars(A,B,C).

%------------------------------------------------------------------------------
% simplify_constraint_light( C1, C2 ): 
% 	efficient method that simplifies constraints, although not totally, i.e.
% 	if C1 is insatisfiable, then C2 will not always be syntactically bottom
%------------------------------------------------------------------------------

simplify_light( C,C1 ):-
	lpoC:simplify_light(C,C1).

%------------------------------------------------------------------------------
% simplify_constraint_med( C1, C2 ): 
% 	simplifies a constraint at medium level
% 	if C1 is insatisfiable, then C2 will not necessarily 
%	be syntactically bottom
%------------------------------------------------------------------------------

simplify_med( C, C ):- true( C ), !.
simplify_med( [], [] ):- !.
simplify_med( C, [conj([],[],[])] ):-
	member( conj([],[],[]), C ), !.	
simplify_med( C, C2 ):- 
	simplify_med1( C, C1 ),
	(member( conj([],[],[]), C1 ) ->
	     C2 = [conj([],[],[])]
	;    delete_implying_ands( C1, C2 ) ), !.

simplify_med1( [], [] ).
simplify_med1( [A|C], C1 ):-
	simplified_and( A, A1 ), !,
	( A1 = conj([],[],[]) ->
	      C1 = [conj([],[],[])] 
	;     simplify_med1( C, C2 ), 
	      C1 = [A1|C2]  ), !. 
simplify_med1( [_|C], C1 ):-
	simplifymed1( C, C1 ), !.

simplified_and( conj(I,E,O), conj(I3,U,O1) ):-
	mgu( E, U ),
	apply_substitution( U, I, I1 ),
	partially_solved_form( I1, I1, I2 ),
	apply_substitution( U, O, O1 ),
	eliminate_implied_ineqs( I2, I3 ), !.





%------------------------------------------------------------------------------
% simplify( C1, C2 ): 
% 	simplifies a constraint totally
% 	e.g. if C1 is insatisfiable, then C2 will be syntactically bottom
%------------------------------------------------------------------------------

simplify( C, C ):- true( C ), !.
simplify( [], [] ):- !.
simplify( C, [conj([],[],[])] ):-
	member( conj([],[],[]), C ), !.	
simplify( C, C2):-
	(memo_constraints(on) ->
	     lpoC:normalize_constraint(C,CN),  % to normalize a little before caching
	     skolemized(CN,CS,Subst),
	     hash_term(CS,Key),
	     (s_con(Key,(CS, CS1)) ->
%	          write('|'),ttyflush,
		  skolemized(C2,CS1,Subst)
	     ;
	          timing(cons,simplify_n(CN,C2)),
		  record_s_con(Key,(CN,C2))
	     )
	;
	     timing(cons,simplify_n(C,C2))
	).



% test case 
% constraintSolver:simplify([[[(f(B)>C)> (C>D),C>=D,(B>D)>(f(B)>C)],[]]],S).
% should produce S=[] (for f > '>')
% stoC:partially_solved_form([_233781:f(_233781,_233779)>diam(_233781):_233779],_234677)
% should be satisfiable for predicates : 



simplify_n(CN,C2):-
%	format('~N{working on constraint ~p',[CN]),ttyflush,
%	write('['),ttyflush,
%	(timeOut(simplify1(CN,C1),5000,success) ->
%	     true
%	; 
%	     format('~Ntime-out on constraint ~p',[CN]),
%	     C1=CN
%	),

	simplify1(CN,C1),

%	write('}'),ttyflush,

	(member( conj([],[],[]), C1 ) ->
	     C2 = [conj([],[],[])]
	;	 
	     delete_implying_ands( C1, C2 )
	), 
	!.



simplify1([],[]).
simplify1( [A|C], C1 ):-
	simplified_satisfiable_and( A, A1 ),
	( A1 = conj([],[],[]) ->
	      C1 = [conj([],[],[])] 
	; 
	      simplify1( C, C2 ), 
	      C1 = [A1|C2] ),
	!. 
simplify1( [_|C], C1 ):-
	simplify1( C, C1 ).


simplified_satisfiable_and( conj(I,E,O), conj(I3,U,O1) ):-
	mgu( E, U ),
	apply_substitution( U, I, I1 ),
	partially_solved_form( I1, I1, I2 ),
	lpoC: eliminate_implied_ineqs( I2, I3 ),
	flattenInEqs(I3,I3,[],[],_,I4),
	!,
	lpoC: partially_solved_form(I4,I4,I41),
	lpoC: eliminate_implied_ineqs( I41, I42 ),

	\+ (\+ lpoC:satisfiable_and( I42 )  ), 
	apply_substitution( U, O, O1 ),
	!.



/*
Inequations are flattened by replacing each nonvariable subterm
by a new variable, and by adding a constraint X>Y whenever
the term denoted by Y is a direct subterm of the term denoted
by X. We optimize
*/


flattenInEqs([],_,E,I,E,I).
flattenInEqs([A>B|Xs],OrigI,E,I,E1,[FA>FB|I1]):-
	flattenTerm(A,OrigI,FA,E,I,E2,I2),
	flattenTerm(B,OrigI,FB,E2,I2,E3,I3),
	flattenInEqs(Xs,OrigI,E3,I3,E1,I1).
flattenInEqs([A>=B|Xs],OrigI,E,I,E1,[FA>=FB|I1]):-
	flattenTerm(A,OrigI,FA,E,I,E2,I2),
	flattenTerm(B,OrigI,FB,E2,I2,E3,I3),
	flattenInEqs(Xs,OrigI,E3,I3,E1,I1).



flattenTerm(A,_OrigI,F,E,I,E,I):-
	isFlat(A,E,F).
flattenTerm(A,OrigI,F,E,I,E1,I1):-
	A=..[P|As],
	(is_predicate(P/_) ->
	     flattenList(As,OrigI,Fs,E,I,E1,I1),
	     F=..[P|Fs]
	;
	     criticalSubterms(A,OrigI,ASub),
	     flattenList(ASub,OrigI,Fs,E,I,E2,I2),
	     E1=[(A,F)|E2],
	     F='$c'(_),
	     ineqList(F,Fs,I2,I1)
	).


criticalSubterms(A,I,Sub):-
	setof(S,N^M^(path_arg1([N|M],A,S),constraintTerm(S,I)),Sub),
	!.
criticalSubterms(_A,_I,[]).

constraintTerm(T,I):-
	member(U>V,I),
	(argument(U,T);argument(V,T)),
	!.
constraintTerm(T,I):-
	member(U>=V,I),
	(argument(U,T);argument(V,T)),
	!.

argument(U,T):-
	U==T.
argument(U,T):-
	nonvar(U),
	functor(U,P,N),
	is_predicate(P/N),
	genarg(_,U,V),
	V==T.




flattenList([],_OrigI,[],E,I,E,I).
flattenList([X|Xs],OrigI,[F|Fs],E,I,E1,I1):-
	flattenTerm(X,OrigI,F,E,I,E2,I2),
	flattenList(Xs,OrigI,Fs,E2,I2,E1,I1).

isFlat(X,_E,'$c'(X)):-
	var(X),
	!.
isFlat(X,E,F):-
	member((Y,F),E),
	X==Y,
	!.

ineqList(_,[],I,I).
ineqList('$c'(A),['$c'(B)|Bs],I,[A>B|I1]):-
	ineqList('$c'(A),Bs,I,I1).




%------------------------------------------------------------------------------
% partially_solved_form:
%------------------------------------------------------------------------------
partially_solved_form( _, [], [] ):-!.
partially_solved_form( I, [S>T|R], R1 ):- 
	greater( S, T ), !, 
	partially_solved_form( I, R, R1 ), !.
partially_solved_form( I, [S>=T|R], R1 ):- 
	greaterEqual( S, T ), !, 
	partially_solved_form( I, R, R1 ), !.
partially_solved_form( I, [S>T|_], _ ):-
	implies(I,T>=S),
	!,
%	greaterEqual( T, S ), !, 
	fail.
partially_solved_form( I, [S>=T|_], _ ):-
	implies(I,T>S),
	!,
%	greater( T, S ), !, 
	fail.
partially_solved_form( I, [T1>T2|R], R1 ):-
        nonvar(T1),
	nonvar(T2),
        T1 =.. [F1|Args1],
        T2 =.. [F2|Args2],
	is_predicate(F1/_),
	is_predicate(F2/_),
        project_to_elem( >, I, T1, F1, Args1, T2, F2, Args2, L1),
	!,
	partially_solved_form( I, L1, L ),
        partially_solved_form( I, R, R2 ),
        user:append( L, R2, R1 ), !.
partially_solved_form( I, [T1>=T2|R], R1 ):-
        nonvar(T1),
        nonvar(T2),
        T1 =.. [F1|Args1],
        T2 =.. [F2|Args2],
	is_predicate(F1/_),
	is_predicate(F2/_),
        project_to_elem( >=, I, T1, F1, Args1,    T2, F2, Args2, L1),
	!,
	partially_solved_form( I, L1, L ),
        partially_solved_form( I, R, R2 ), 
        user:append( L, R2, R1 ), !.
partially_solved_form( I, [D|R], [D|R1] ):-
        partially_solved_form( I, R, R1 ), !.


project_to_elem( _, _I, T1,F1,_,    _,F2,Args2,  L ):-
        greater_op_sto( F1, F2 ),
        greater_term_list( T1, Args2, L ),
	!.
project_to_elem( _, _I, _,F1,Args1,    T2,F2,Args2, L ):-
        greater_op_sto( F2, F1 ), 
	delete_smaller_ones( T2, Args1, [A] ), 
	!,
	% we have one Arg1 term A left that might possibly be maximal
	greater_term_list(A,Args2,L).
project_to_elem( >, I, T1, F, L1,    _, F, L2,   L ):-
	is_predicate(F/_),
	create_lex_list( I, T1, L1, L2, L ),
        !.
project_to_elem( >=, I, T1, F, L1,    T2, F, L2,   L ):-
	is_predicate(F/_),
	\+ sound_unify( T1, T2 ), !,
	create_lex_list( I, T1, L1, L2, L ), !.
project_to_elem( >=, _I, T1, F, L1,    _, F, L2,   L ):-
	is_predicate(F/_),
	create_lex_list_geq( T1, L1, L2, L ), !.


								% Let a > b >= c
create_lex_list( I ,_, [S1,S2], [T1,T2], [S1>=T1] ):-   	
        % f(x,a) > f(y,b) => x>=y
	implies(I,S2>T2),
	!.
%	greater( S2, T2 ), !.
create_lex_list( _I, _, [S1,S2], [T1,T2], [S1>T1] ):-   	
	% f(x,z) > f(y,z) => x>y
	S2 == T2, !.
create_lex_list( _, _, [S1,S2], [T1,T2], [S1>T1] ):-   	
        % f(x,y) > f(y,x) => x>y
	(S1>S2) == (T2>T1), !.
create_lex_list( I, S, [S1,S2], [T1,T2], [S1>T1, S>T2] ):-	
        % f(x,c)>f(y,b) => x>y & f(x,c)>b
	implies(I,T2>=S2),
	!.
%	greaterEqual( T2, S2 ), !.
create_lex_list( I, _S, [S1,S2], [T1,T2], [S2>T1, S2>T2] ):-	
	implies(I,T1>S1),
        % f(b,x)>f(a,y) -> x>=f(a,y) -> x>a & x>y
	!.
create_lex_list( _, _, [S], [T], [S>T] ):- !.		% f(x)>f(y) => x>y 
create_lex_list( I, T1, [S|RS], [T|RT], L ):-		% f(x,Y)>f(x,Z) => 
	S == T, !,						% f(Y)>f(Z)
	create_lex_list( I, T1, RS, RT, L ), !.
create_lex_list( _I, T1, [S|_], [T|RT], [S>T|L] ):-	
	% f(s,X)>f(t,Y) => ... if s and t not unif.
	\+ sound_unify( S, T ),				   
        greater_term_list( T1, RT, L ), !.


create_lex_list_geq( _, [], [], [] ):- !.
create_lex_list_geq( _, AS, AT, [S>=T] ):-
	delete_synt_equals( AS, AT, [S], [T] ), !.

delete_synt_equals( [], [], [], [] ):- !.
delete_synt_equals( [S|RS], [T|RT], S1, T1 ):-
	S == T, !,
	delete_synt_equals( RS, RT, S1, T1 ), !.
delete_synt_equals( [S|RS], [T|RT], [S|S1], [T|T1] ):-
	delete_synt_equals( RS, RT, S1, T1 ), !.

greater_term_list( _, [], [] ).
greater_term_list( T, [S|L], [T>S|L1] ):-
	greater_term_list( T, L, L1 ).

delete_smaller_ones(  _, [], [] ).
delete_smaller_ones( T, [S|R], R1  ):-
	greater( T, S ), !,
	delete_smaller_ones(  T, R, R1 ), !.
delete_smaller_ones( T, [S|R], [S|R1] ):-
	delete_smaller_ones( T, R, R1 ), !.


