% -*- Module: theory; -*-

:- module(theory, [
        compile_theory_properties/0,
	display_theory/0,
	summary/1
		  ]).


:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(sets), [
	select/3
   ]).

:- use_module(library(maplist), [
	maplist/2
   ]).

:- use_module(myterms, [
        apply_substitution/3,
	function_symbol_in/2,
	is_list_of_different_vars/1,
	list_of_vars/2
   ]).


:- use_module(auxiliary, [
        appendAtoms/3,
	delete_all/3,
	get_prolog_data_operators/2,
	less_or_equal/2
   ]).

:- use_module(classes, [
	antecedent/2,
	as_new_axiom/2,
	as_new_goal/2,
	createClause/2,
	createSimpleClauseFromLiterals/2,
        kind/2,
        kind/3,
	literal_list_to_sequent/3,
	literals/2,
	not_cnf/2,
	succedent/2,
	sequent_to_literals/3,
	simplify_cnf/2,
	theory_flag/2
   ]).

:- use_module(global, [
        cont/2
   ]).

:- use_module(rewrite, [
	subsumes_literals/2
   ]).

:- use_module(clause_redundancy_prover, [
        tautology/1
   ]).

:- use_module(proofState, [
        rule/2,
	clause/1,
	create_new_axiom/2,
	create_new_axiom/3,
	deleteClause/2,
	addTransformedClauses/4,
	operators/1,
	predicates/1,
	goal_symbol/1,
	recordOperator/2,
	retractTempClause/1,
	tempClause/1,
	transformClauses/4
   ]).

:- use_module(thisProlog, [
        compile_nqa/2,
	portray_clause2/2
   ]).




:- use_module(props, [
        expand_definition_for/1,
        horn/1,
	theory/1,
	option/1,
	ordering/2,
	ac/1
   ]).

:- use_module(input,[
        create_clauses/0
   ]).

:- use_module(adt_precedence,[definition/2,oriented_definition/3,predicate/1]).

:- use_module(adt_atom,[
        create_atom/2,
	get_atom/4,
	get_atom_members/3,
	is_equational_atom/1,
	is_equivalence/1,
	is_standard_atom/1,
	orient_atom_fresh/2,
	print_list_atoms/1
   ]).


:- use_module(adt_term,[
        variant/2,logical_operator/1
   ]).

:- use_module(termOrder,[currentOrderingAC/0]).



/* ----------------------------------------------------------------------------

This module collects theory information from clauses, rules and
termporary clauses and provides a procedure 

      compile_theory_properties

which compiles these properties as facts into module prop.
It is assumed that clause parsing has been done so that
the kind attribute of a clause describes what is special
about the clause.

Here we also generate auxiliary predicates such as the reflexive
closure of a strict total ordering or the strict part
of a non-strict total ordering.

Most properties should be clear from their name.

A partial equivalence is a transitive, symmetric relation
on terms that is compatible with all predicates
(except equality =). For such relations we chain the arguments
of predicates. Chaining into variables is not performed.
That may result in incompleteness whenever the variable
is shielded by a function symbol. (Equivalences need not be
compatible withfunction application.) For positive chaining
that is not a problem. For negative chaining it might be incomplete
for certain selection strategies.

Relations of this kind include ``equal'' in many tptp-examples
where skolem functions are sometimes not compatible with ``equal''.

---------------------------------------------------------------------------- */




property(associative/1).
property(commutative/1).
property(anti_symmetric_relation_symbol/1).
property(composition_of_relations/3).
property(congruence_relation_symbol/1).
property(dense_ordering_relation_symbol/1).
property(dense_ordering_wo_end_relation_symbol/1).
property(equality_relation_symbol/1).
property(explicit_nonstrict_total_ordering_relation_symbol/1).
property(explicit_total_ordering_relation_symbol/1).
property(finite_domain/1).
property(function_as_predicate/5).
property(irreflexive_relation_symbol/1).
property(nonstrict_total_ordering_relation_symbol/1).
property(noendL_relation_symbol/1).
property(noendR_relation_symbol/1).
property(partial_congruence_relation_symbol/1).
property(partial_equivalence_relation_symbol/1).
property(reflexive_closure/2).
property(reflexive_relation_symbol/1).
property(strict_part/2).
property(special_relation/1).
property(special_user_relation/1).
property(symbol_admits_var_elim/1).
property(symmetric_relation_symbol/1).
property(total_ordering_relation_symbol/1).
property(total_relation_symbol/1).
property(transitive_relation_symbol/1).




dynamic_property(associative/1).
dynamic_property(commutative/1).
dynamic_property(reflexive_relation_symbol/1).





    % Strategies are compiled after properties have been
    % compiled.
strategy(horn/0).
strategy(resolution/0).
strategy(agoal/0).
strategy(firstOrder/0).
strategy(predicate_with_extension/3).  
    % that analysis requires that we know the logical properties of 
    % relation symbols.

		     

compile_theory_properties:-
	compile_theory(property),
		   
	create_clauses,
	compile_theory(strategy),
	(props:resolution ->
	     commands:sair([3,4])
	;
	     true
	),
	(props:firstOrder ->
	     commands:sair(+[7])
	;
	     true
	),
	props:set_weights,

	(props:finite_domain(_) ->
%	     commands:(saci([2]),sama([2])),
	     commands:saci([2]),
	     true
	;
	     true
	),
	(  ac(on), 
	   associative(O), 
	   commutative(O) ->
		theory(ac),
		(currentOrderingAC ->
		     true
		;
		     ordering(aclpo,generic)
		)
	;  ac(on), 
	   commutative(_) ->
		theory(c),
		(currentOrderingAC ->
		     true
		;
		     ordering(rpo,generic)
		)
		
	;
	        true
	),
	delete_congruence_axioms,
	(dense_ordering_relation_symbol(_) ->
	    option(structural_expansion(off))
	    % seems to counteract with variable elimination
	;
	    true).
	     



compile_theory(T):-
	compile_special_axioms,
	cont(props_tmp_file,F),
	open(F,write,Str),
	write_props(Str,T),
	close(Str),
	inspect,
	compile_nqa(props,F).




delete_congruence_axioms:-
	congruence_relation_symbol(R),
	R\== =,
	deleteClause(C,congruence_prop(C,R)),
	fail.
delete_congruence_axioms.

	
congruence_prop(C,R):-
	kind(C,K),
	congruence_prop(C,R,K).


congruence_prop(_,R,reflexivity(R)).
congruence_prop(_,R,symmetry(R)).
congruence_prop(_,R,composition(R,R,R)).
congruence_prop(_,R,monotonicity(_F,_I,R,_P,_R)).



inspect.       % can be spyed

%display_theory.  % to be written

display_theory:-
	nl,
	(property(P/N);strategy(P/N)),
	length(A,N),
	PP=..[P|A],
	
	(\+ \+ props:PP ->
	     props:PP,
	     (N=<3 ->
		  \+ member(=,A),
		  \+ member(==,A)
	     ;
	          true
	     ),

	     print('$prop'(PP)),
	     fail
	).
display_theory:-
	nl.

:- mymultifile user:portray/1.

user:portray('$prop'(predicate_with_extension(P/_N,_SF,[Args,_Pos,Cls]))):-
	!,
	PHead=..[P|Args],
	create_atom(PHead,PA),
	format("Extensional:~n~p.~n",['$ext'(PA,Cls)]).
user:portray('$prop'(CL)):-
	portray_clause(CL).
user:portray('$ext'(PA,[Cl])):-
	!,
	Cl1=[[+|PA]|Cl],
	\+ \+ (numbervars(Cl1,0,_),write('   '),print_list_atoms(Cl1)).
user:portray('$ext'(PA,[Cl|Cls])):-
	print('$ext'(PA,[Cl])),
	nl,
	print('$ext'(PA,Cls)).

	      

	

	
	
      


shouldBeTransitive(less).
shouldBeTransitive(less_equal).
shouldBeTransitive(less_than_or_equal).
shouldBeTransitive(subset).
shouldBeTransitive(subclass).


shouldBeCongruence(equal).


shouldBeEquivalence(equal).
shouldBeEquivalence(equivalent).




summary(S):-
	setof(F,interesting(F),S),
	!.
summary([]).



interesting(no_equiv(F)):-
	shouldBeEquivalence(F),
	symbol_occurs(F),
	\+ props:partial_equivalence_relation_symbol(F).
interesting(no_cong(F)):-
	shouldBeCongruence(F),
	symbol_occurs(F),
	props:partial_equivalence_relation_symbol(F),
	\+ props:congruence_relation_symbol(F).
interesting(not_trans(F)):-
	shouldBeTransitive(F),
	symbol_occurs(F),
	\+ props:transitive_relation_symbol(F).
interesting(func_as_rel(Rel,O)):-
	props:function_as_predicate(Rel,_,_,O,_).




symbol_occurs(F):-
	operators(Fs),
	member(F/_,Fs),
	!.
symbol_occurs(F):-
	predicates(Fs),
	member(F/_,Fs),
	!.






:- dynamic axiom_exists/1.

compile_special_axioms:-
	retractall(axiom_exists(_)),
	special_axiom(K),
	assert(axiom_exists(K)),
	fail.
compile_special_axioms.



special_axiom(K):-
	retractTempClause((K,K)).
special_axiom(K):-
	tempClause((_K,K)),
	K\==regular.
special_axiom(K):-
	clause(C),
	kind(C,K),
	K\==regular.
special_axiom(K):-
	rule(_,C),
	kind(C,K),
	K\==regular.

user_sequent(A,S,F):-
	user_sequent(A,S,F,_K).

user_clause(Cl,F,K):-
	user_sequent(A,S,F,K),
	sequent_to_literals(A,S,Cl).

user_sequent(A,S,F,K):-
	tempClause(([F,A,S],K)).
user_sequent(A,S,F,K):-
	clause(C),
	antecedent(C,A),
	succedent(C,S),
	kind(C,K),
	theory_flag(C,F).
user_sequent(A,S,F,K):-
	rule(_,C),
	antecedent(C,A),
	succedent(C,S),
	kind(C,K),
	theory_flag(C,F).


		
	


% :-dynamic theory/1.


write_props(Str,T):-
	TP=..[T,P/N],
	TP,
	prop_clause(P/N,PP),
	portray_clause2(Str,PP),
%	assert(theory(PP)),
	fail.
write_props(_,_).



prop_clause(P/N,(:-dynamic P/N)):-
	dynamic_property(P/N).
prop_clause(P/N,Clause):-
	length(A,N),
	PP=..[P|A],
	(setof(PP,PP,Facts) ->
	     member(Clause,Facts)

	;\+dynamic_property(P/N) ->
	     Clause=((PP:-fail))
	).


/* ----------------------------------------------------------------------------

Properties that we support

-----------------------------------------------------------------------------*/



associative(O):-
	axiom_exists(associativity(O,R)),
	\+ \+congruence_relation_symbol(R).

commutative(O):-
	axiom_exists(commutativity(O,R)),
	\+ \+congruence_relation_symbol(R).
commutative(O):-
	axiom_exists(symmetry(Rel,3,1,2)),
	function_as_predicate(Rel,3,3,O,R),
	\+ \+congruence_relation_symbol(R).


anti_symmetric_relation_symbol(R):-
	explicit_total_ordering_relation_symbol(S),
	reflexive_closure(S,R),
	S\==R.
%	nonstrict_total_ordering_relation_symbol(R).
anti_symmetric_relation_symbol(R):-
%	explicitly_transitive(R),
	axiom_exists(anti_symmetry(R,ER)),
	\+ \+ congruence_relation_symbol(ER).  % only testing!

composition_of_relations(C,R,R):-
	axiom_exists(partial_congruence(C)),
	special_relation(R),
	C\== ==,
	R\== == .
composition_of_relations(R,C,R) :-
	axiom_exists(partial_congruence(C)),
	special_relation(R),
	C\== ==,
	R\== ==,
	R\== C .
composition_of_relations(==,==,==).
composition_of_relations(=,R,R):-
	special_relation(R),
	R\== == .
composition_of_relations(R,=,R):-
	special_relation(R),
	R\== == .
composition_of_relations(R,RR,R):-
	reflexive_closure(R,RR).
composition_of_relations(RR,R,R):-
	reflexive_closure(R,RR),
	R\==RR.
composition_of_relations(A,B,C):-
	axiom_exists(composition(A,B,C)).

congruence_relation_symbol( R ):-
	partial_congruence_relation_symbol(R),
	\+ \+ reflexive_relation_symbol(R).

dense_ordering_relation_symbol(R):-
	total_ordering_relation_symbol(R),
	axiom_exists(density(R,left,strict)),
	axiom_exists(density(R,right,strict)).
dense_ordering_relation_symbol(R):-
	total_ordering_relation_symbol(R),
	reflexive_closure(R,RR),
	axiom_exists(density(RR,left,nonstrict)),
	axiom_exists(density(RR,right,nonstrict)).
	
dense_ordering_wo_end_relation_symbol(R):-
	dense_ordering_relation_symbol(R),
	axiom_exists(noendL(R)),
	axiom_exists(noendR(R)).
	
equality_relation_symbol(R):-
	congruence_relation_symbol(R).

equivalence_relation_symbol(R):-
	congruence_relation_symbol(R).
equivalence_relation_symbol(R):-
	partial_equivalence_relation_symbol(R),
	\+ \+ reflexive_relation_symbol(R).

explicit_nonstrict_total_ordering_relation_symbol(R):-
	axiom_exists(composition(R,R,R)),
	axiom_exists(nonstrict_totality(R)).

explicit_total_ordering_relation_symbol(R):-
	axiom_exists(composition(R,R,R)),
	axiom_exists(irreflexivity(R)),
	axiom_exists(totality(R,ER)),
	\+ \+ congruence_relation_symbol(ER).


finite_domain(Dom):-
	axiom_exists(finite_domain(Dom,R)),
	congruence_relation_symbol(R),
	!.  % we want just one domain



function_as_predicate(P,N,I,F,R):-
	axiom_exists(total_predicate(P,I,N,F)),
	axiom_exists(det_predicate(P,N,I,R)),
	%% compatibility with predicates is enough
	\+ \+ partial_equivalence_relation_symbol(R),
	(goal_symbol(P/N) ->
	     N1 is N-1,
	     recordOperator(F/N1,(goal))
	;
	     true
	).


irreflexive_relation_symbol(R):-
	axiom_exists(irreflexivity(R)).
irreflexive_relation_symbol(R):-
	total_ordering_relation_symbol(R).


nonstrict_total_ordering_relation_symbol(R):-
	explicit_nonstrict_total_ordering_relation_symbol(R).
nonstrict_total_ordering_relation_symbol(R):-
	explicit_total_ordering_relation_symbol(S),
	reflexive_closure(S,R),
	S\==R.

noendL_relation_symbol(R):-
	axiom_exists(noendL(R)).

noendR_relation_symbol(R):-
	axiom_exists(noendR(R)).

partial_congruence_relation_symbol( = ).
partial_congruence_relation_symbol( R ):-
	axiom_exists(partial_congruence(R)).
partial_congruence_relation_symbol( R ):-
	explicitly_transitive(R),
	symmetric_relation_symbol(R),
	allContextsReplacableWith(R),
	allRelationsCompatibleWith(R),
	\+ (  user_sequent(A,S,_),
	      ( member(E,A); member(E,S) ),
	      ( E=(_=_)    ; (E=(_== T), T\== '$true$'))
	   ).


partial_equivalence_relation_symbol(==).
partial_equivalence_relation_symbol( = ).
%partial_equivalence_relation_symbol(R):-
%	partial_congruence_relation_symbol( R ).
partial_equivalence_relation_symbol(R):-
	explicitly_transitive(R),
	symmetric_relation_symbol(R),
	allRelationsCompatibleWith(R).



reflexive_closure(R,RR):-
	explicit_total_ordering_relation_symbol(O),
	(axiom_exists(reflexivity(S)),
	 axiom_exists(reflexive_closure1(O,S)),
	 axiom_exists(reflexive_closure2(O,S)) ->
		S=RR
	;
	        appendAtoms(['=',O],'$',RR)
	),
	(   R=O
	;
	    R=RR
	).
reflexive_closure(R,RR):-
	explicit_nonstrict_total_ordering_relation_symbol(RR),
	(axiom_exists(reflexive_closure1(S,RR)),
	 axiom_exists(reflexive_closure2(S,RR)) ->
		true
	;
	        appendAtoms([RR,strict],'$',S)
	),
	(   R=S
	;
	    R=RR
	).


reflexive_relation_symbol(Rs,R):-
	member(R,Rs),
	!,
	fail.
reflexive_relation_symbol(_,=).
reflexive_relation_symbol(_,'==').
reflexive_relation_symbol(_,R):-
	axiom_exists(reflexivity(R)).
reflexive_relation_symbol(Rs,R):-
	axiom_exists(reflexive_closure1(R1,R)),
	reflexive_relation_symbol([R|Rs],R1).
reflexive_relation_symbol(_,R):-
	axiom_exists(nonstrict_totality(R)).
reflexive_relation_symbol(_,R):-
	reflexive_closure(_,R).


reflexive_relation_symbol(R):-
	reflexive_relation_symbol([],R).


special_relation(R):-
	% only special relations are not represented as the terms
	% they represent
	special_relation1(R),
	not_to_be_expanded(R).

not_to_be_expanded(R):-
	\+ (relation_has_definition(R), expand_definition_for(R/2)).


special_relation1(R):-
	axiom_exists(partial_congruence(R)).
special_relation1(R):-
	special_user_relation(R).
special_relation1(RR):-
	special_user_relation(R),
	reflexive_closure(R,RR),
	R\==RR.
special_relation1(R):-
	special_user_relation(RR),
	reflexive_closure(R,RR),
	R\==RR.
special_relation1(=).
special_relation1(==).

special_user_relation(R):-
	special_user_relation(R,[R]).

special_user_relation(R,_Rs):-
	axiom_exists(composition(R,R,R)).


/* this gives too many
	(    axiom_exists(composition(R,S,R))
	;    
	     axiom_exists(composition(S,R,R))
	),
	(member(S,Rs) ->
	     true
	;
	     special_user_relation(S,[S|Rs])
	).
*/



strict_part(R,RS):-
	reflexive_closure(RS,R),
	irreflexive_relation_symbol(RS).
strict_part(R,R):-
	irreflexive_relation_symbol(R).


symbol_admits_var_elim(RS):-
	explicitly_transitive(RS),
	( noendL_relation_symbol(RS)
	; noendR_relation_symbol(RS)
	; total_ordering_relation_symbol(RS)
	; reflexive_relation_symbol(RS)
	).

symmetric_relation_symbol(==).
symmetric_relation_symbol(=).
symmetric_relation_symbol(R):-
	axiom_exists(symmetry(R)),
	\+ adt_precedence: dual_predicate_pair(_,R).
symmetric_relation_symbol(R):-
	axiom_exists(partial_congruence(R)).


total_ordering_relation_symbol(R):-
	explicit_total_ordering_relation_symbol(R).
total_ordering_relation_symbol(R):-
	explicit_nonstrict_total_ordering_relation_symbol(RR),
	reflexive_closure(R,RR),
	R\==RR.

total_relation_symbol(R):-
	axiom_exists(totality(R,Eq)),
	congruence_relation_symbol(Eq).
total_relation_symbol(R):-
	total_ordering_relation_symbol(R).


explicitly_transitive(R):-
	axiom_exists(partial_congruence(R)).
explicitly_transitive(R):-
	axiom_exists(composition(R,R,R)).
explicitly_transitive(=).
explicitly_transitive(==).


transitive_relation_symbol(R):-
	explicitly_transitive(R).
transitive_relation_symbol(R1):-
	total_ordering_relation_symbol(R),
	reflexive_closure(R,R1).
transitive_relation_symbol(R1):-
	nonstrict_total_ordering_relation_symbol(R),
	strict_part(R,R1).


relation_has_definition(RS):-
	% this ought to work at compile time when all atoms are terms
	user_sequent([],[L==R],theory,_),
	(  definition(L,R),
	   functor(L,RS,2),
	   special_relation1(RS)
	;  definition(R,L),
	   functor(L,RS,2),
	   special_relation1(RS)
	).

% ----------------------------------- treating equivalences-----------------

/* We attempt to recognize equivalences of the form
  
          ~p(x) == exists y . exp(x,y)

   that is to say
           
           p(x) == all y. ~exp(x,y)

   where x and y are vectors of variables. (At present we only deal with
   y of length =< 1.) The atomic expresions in exp is restricted to 
   be of the form q(x,y). y occurs in the form of a skolem function call
   y(x) in clauses derived from the existential quantifier.

   Extensionality axioms are equivalences of this kind.

   If such an equivalence has been found, all negative occurrences
   ~p(t) of the predicate p are replaced by the set of clauses in
   equivalent to exp(t,y(t)).

   In case we have only the ``->'' direction of the first equivalence,
   some clauses of the new form are added.
   This may in particular be the case when we are unable to detect that the
   function symbol y is in fact a Skolem function.

*/


predicate_with_extension((=)/2,_SFs,[[X,Y],_,[[[-|Def]]]]):-
	% this case works only at runtime
	user_sequent([],[E],theory,_),
	is_equivalence(E),
	get_atom_members(E,L,R),
	(  definition(L,R),
	   L=(X=Y),
	   create_atom(R,Def)
	;  definition(R,L),
	   R=(X=Y),
	   create_atom(L,Def)
	).
predicate_with_extension(P/N,SFs,Definition):-
	predicate_with_extension([],Ps),
	member((P/N,SFs,Definition),Ps),
	N>1.

predicate_with_extension(Old,New):-
	extensionality_axioms(P/N,PArgs,SFs,NegatedDef),
	    % ~P(PArgs) -> exists SF. NegatedDef
	\+ member((P/N,_,_),Old),
	clauses_for_negated_predicate(P/N,Sufficient),

	(SFs=[SF] ->
	     SFT=..[SF|PArgs],
	     apply_substitution([SFT=_],NegatedDef,NegatedDefMatrix)
	;SFs=[] ->
	     NegatedDefMatrix=NegatedDef
	),
	not_cnf(NegatedDefMatrix,Def),
	simplify_cnf(Def,DefS),
	implies(Sufficient,DefS),
	!,
%	apply_equivalence_transformation(P/N,PArgs,DefS,NegatedDef),
	predicate_with_extension([(P/N,SFs,[PArgs,DefS,NegatedDef])|Old],New).
predicate_with_extension(Old,Old).

extensionality_axioms(P/N,Args,SFs,Cls):-
	extensionality_axiom_exists(P/N,SFs),
	findall((Args,Cl),extensionality_axiom(Cl,P/N,Args,SFs),ACls),
	length(Args,N),
	maplist(project(Args),ACls,Cls1),
	maplist(select_general_p_literal(P/N,+),Cls1,Cls).

extensionality_axiom_exists(P/N,SFs):-
	setof((P/N,SFs),Cl^Args^extensionality_axiom(Cl,P/N,Args,SFs),Axs),
	member((P/N,SFs),Axs).



extensionality_axiom(Cl,P/N,Args,SFs):-
	user_clause(Cl,theory,extensionality(P/N,Args,SFs)).
extensionality_axiom(Cl,P/2,Args,[]):-
	user_clause(Cl,theory,anti_symmetry(_R,P)),
	select_p_literal(P/2,+,Args,Cl,_).


project(F,(F,S),S).

clauses_for_negated_predicate(P/N,NegativePart):-
	setof(Cl,clause_for_negated_predicate(P/N,Cl),NegativePart).



clause_for_negated_predicate(P/N,Cl1):-
	user_clause(Cl,theory,_),
	select_general_p_literal(P/N,-,Cl,Cl1),
	maplist(admittedDep,Cl1).

select_general_p_literal(P/N,Pol,Cl,Cl1):-
	select([Pol|A],Cl,Cl1),
	(   get_atom(A,==,E,'$true$'),
	    E=..[P|Args],
	    length(Args,N)
	;
	    N=2,
	    get_atom(A,P,L,R),
	    Args=[L,R]
	),
	is_list_of_different_vars(Args).

select_p_literal(P/N,Pol,Args,Cl,Cl1):-
	select([Pol|A],Cl,Cl1),
	(   get_atom(A,==,E,'$true$'),
	    functor(E,P,N),
	    E=..[P|Args]
	;
	    N=2,
	    get_atom(A,P,L,R),
	    Args=[L,R]
	).


	
admittedDep([_|A]):-
	get_atom(A,_,L,R),
	(nonvar(L) ->
	     L=..[_|Args],
	     maplist(var,Args)
	;
	     true
	),
	(nonvar(R)->
	     R=..[_|Args1],
	     maplist(var,Args1)
	;
	     true
	).


no_occurrence_of_outside(SF,Cls):-
	user_clause(Cl,_,K),
	\+ (member(D,Cls),variant(D,Cl)),
	(function_symbol_in(SF,Cl) ->
	     \+ (K=monotonicity(SF,R,_,1,R),partial_equivalence_relation_symbol(R))
	),
	!,
	fail.
no_occurrence_of_outside(_SF,_Cls).

implies(Cls,Cls1):-
	maplist(createSimpleClauseFromLiterals,Cls,CLS),
	maplist(createSimpleClauseFromLiterals,Cls1,CLS1),
	implies1(CLS,CLS1).

	
		


implies1(_Cls,[]).
implies1(Cls,[C|Cls1]):-
	tautology(C),
	!,
	implies1(Cls,Cls1).
implies1(Cls,[C|Cls1]):-
	literals(C,Ls),
	\+ \+ ( numbervars(Ls,0,_),
		member(D,Cls),
		literals(D,DLs),
		subsumes_literals(DLs,Ls)
	      ),
	!,
	implies1(Cls,Cls1).



% ----------------------------------------------------------------



horn:-
	special_relation(R),
	\+congruence_relation_symbol(R),  % R\== =,
	R\== ==,
	!,
	fail.
horn:-
	horn(on),
	\+ firstOrder,
	\+ ( 
	      user_sequent(_A,S,_TFlag),
	      length(S,N),
	      (N>1 ->
		   true
	      ;S=[E] ->
		   is_equational_atom(E)
	      )

	   ),
	\+ user_sequent(_,[],theory),
	user_sequent(_,[],(goal)),
	!.

firstOrder:-
	      user_sequent(A,S,_TFlag),
	      (  member(E,S)
	      ;  member(E,A)
	      ),
	      \+ is_standard_atom(E),
	      !.


	

resolution:-
	firstOrder,
	!,
	fail.
resolution:-
	props:special_relation(R),
	R\== =,
	R\== ==,
	!,
	fail.
resolution:-
	props:symmetric_relation_symbol(R),
	R\== =,
	R\== ==,
	!,
	fail.
resolution:-
	\+ ( 
	      user_sequent(_A,S,_TFlag),
	      member(E,S),
	      is_equational_atom(E)

	   ),
	\+ ( 
	      user_sequent(A,_S,_TFlag),
	      member(E,A),
	      is_equivalence(E)

	   ).


	
	


agoal:-
	user_sequent(_,_,(goal)),
	!,
	\+ ( 
	      user_sequent(A,S,(goal)),
	      list_of_vars((A,S),[_|_])

	   ),
	\+ firstOrder.

	
	

% ----------------------------------- auxiliary -----------------------------



allRelationsCompatibleWith(R):-
	\+ notCompatibleWith(_,R).

allContextsReplacableWith(R):-
	\+ notMonotone(_F,R).



% we assume for now that new relations relate terms, not propositions
% so that == is the only relation on propositions.
% Hence compatiblity with == is not required for a partial equivalence.
% Also, if = does not occur in the input clauses, we assume
% compatibility with =.
notCompatibleWith(R1,R):-
	predicate(R1/N),
	(N=2,special_relation(R1) ->
	     R1\== =,
	     R1\== ==,
	     \+ reflexive_closure(_,R1)
	;
	     true
	),
	less_or_equal(I,N),
	\+ compatibleWith(R1/N,I,R).


compatibleWith(R1/2,1,R):-
	composition_of_relations(R,R1,R1).
compatibleWith(R1/2,2,R):-
	composition_of_relations(R1,R,R1).
compatibleWith(R1/N,I,R):-
	axiom_exists(monotonicity(R1/N,Is,R,1,==)),
	member(I,Is).


notMonotone(F,R):-
	operators(Os),
	member(F/N,Os),
%	\+ special_user_relation(F),
	less_or_equal(I,N),
	\+ monotone_operator(F,I,R,1,R),
	\+ (  
	       N1 is N+1,
	       axiom_exists(symmetry(Rel,N1,I,J)),
	       function_as_predicate(Rel,N1,N1,F,R),
	       monotone_operator(F,J,R,1,R)
	   ),
	!.
	



monotone_operator(F,I,R,P,R1):-
	axiom_exists(monotonicity(F/_,Is,R,P,R1)),
	member(I,Is).



	
