:- module(cong_closure,[empty_rules/1,
		  delete_rule/3,
		  insert_rule/3,
		  match_rule/4,
		  type_check_rule/3,
		  type_check_type/3,
		  unify/2,
		  unify_rule/3]).


:- use_module(library(change_arg),[change_arg/4]).
:- use_module(library(basics),[member/2]).
:- use_module(library(maplist),[maplist/2,maplist/3]).

:- use_module(formulas,[left/2,
			mk_equation/2,
			mk_equation/5,
			number/2,
			orientation/2,
			orientation/3,
			right/2,
			type/2]).

:- use_module(satsrc(substitutiontree)).
:- use_module(global_vars, [cont/2,store/2]).

:- use_module(st_terms).






unify_rule(L,STree,Ru):-
	st_us_unify(L,STree,Ru).

unify(S,S).



/* ----------------------------------------------------------------------------

Matching rewrite rule

---------------------------------------------------------------------------- */



match_rule(STree,Term,R,Ru):-
	% all rules oriented, no type checking required
	% although STree has only variable-free keys, 
	% undoing of internal bindings is required
	\+ \+ (st_us_unify(Term,STree,eq(R,Ru)) ->
		  store('$found',(R,Ru))
	      ),
	(cont(irreducibility_check,true) ->
	    raise_exception(reducible)
	;
	    true
	),
	cont('$found',(R,Ru)),
				%			 proof(Ru,P),
	!.


/* ----------------------------------------------------------------------------

Insertion and Deletion

We exploit that equations are all uniformly orientable

---------------------------------------------------------------------------- */

empty_rules(STree):-
	st_empty_tree(STree).


insert_rule(Ru,STree,STree1):-
	orientation(Ru,O),
	insert_rule(Ru,O,STree,STree1).

insert_rule(Ru,lr,STree,STree1):-
	left(Ru,L),
	right(Ru,R),
	term2prolog((L,R),(LP,RP)),  % needed for negative equations which may have
	                             % variables
	st_insert_term(LP,eq(RP,Ru),STree,STree1).
insert_rule(Ru,rl,STree,STree1):-
	right(Ru,L),
	left(Ru,R),
	left(Ru,Ru1,L),
	right(Ru1,RuInv,R),
	term2prolog((L,R),(LP,RP)),
	st_insert_term(LP,eq(RP,RuInv),STree,STree1).



delete_rule(Ru,STree,STree1):-
	left(Ru,L),
	number(Ru,N),           % need to disambiguate by number
	st_delete_term(L,STree,STree1,equational:same_number(N)).




/* ----------------------------------------------------------------------------

Inferences

---------------------------------------------------------------------------- */




reflexivity_resolution(E,[In]):-
	left(E,L),
	type(E,prop),
	functor_(L,'$neq',2),
	arg_(1,L,S),
	arg_(2,L,T),
	unify(S,T),
	number(E,N),
	!,
	mk_equation('$false','$true',prop,inference(reflexivity,[N]),In).
reflexivity_resolution(_E,[]).



critical_pairs(Ru,	% +Ru : Rule
	       Active,  % +Active : STree
	       Eqs      % ?Eqs : [Equation]  minimally instantiated
	      ):-
	findall(Eq,cp_into(Ru,Active,Eq),CPs),
	% the cp_from cps are implicitly computed during reduction of
	% passive equations with active ones
	maplist(mk_equation,CPs,Eqs).


cp_into(Ru,Active,eq(ST,TT,TyST,PST)):-
	   % cp of Ru into a rule Ru1 in Active; Ru1 is oriented
	   % from L -> R.
	   % Every rule is variable-free so that each cp is a reduction
	   % or another rule.
	   % Type checking is not needed.
	left(Ru,L),
	right(Ru,R),
	number(Ru,N),
	type(Ru,Ty),
	Ty\== prop,
	
	   % any rule L1 -> T in Active
	   % this could be a negative equation with variables
	unify_rule(L1,Active,eq(T,Ru1)),
	number(Ru1,N1),
	N\==N1,
	type(Ru1,TyST),

	  % select a non-variable subterm Sub of L1
	path_arg_(P,L1,Sub),
	nonvar_(Sub),

	  % type check and unify with L
	unify(L,Sub),

	  % compute L1[R]_P, that is, replace the subterm in L1 by R
	change_path_arg_(P,L1,S,R),
 

	  % convert cp from Prolog terms to st_terms
	prolog2term((S,T),(ST,TT)),

	  % add proof information
	((TyST\==prop ; left(Ru1,L1))
			% the unifier of L and Sub applied to L1 matches L1
	->
	    PST=inference(reduction_max,eq(Ru1),[N])
%	    format('~Ncp: ~p',[(ST=TT) : PST])
	;
				% filter out some trivial cases
	    S\==T,
	    PST=inference(cp,[N,N1])
	).
